#define _DS_INCLUDE_XML_HEADERS

#include <stdio.h>
#include <stdint.h>
#include <unistd.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/types.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <errno.h>

#include <opclib.h>
#include <opccommands.h>
#include <opcdefines.h>
#include <dstime.h>
#include <dsxml.h>
#include "dime.h"

#define SERVER_IP "127.0.0.1"
#define SERVER_PORT 9333

//#define SERVER_IP "141.52.64.15"
#define REQUEST "<Read xmlns=\"http://opcfoundation.org/webservices/XMLDA/1.0/\"><ItemList ItemPath=\"/Examples/random\"><Item ItemName=\"complex\"/></ItemList></Read>"

static int opc_socket = -1;
static char opc_buffer[OPC_BUFFER_SIZE+16];
static int opc_error = 0;

char opc_header[16];
unsigned long opc_header_size;
char opc_request[256];
unsigned long opc_request_size;


static void opc_close_socket() {
    close(opc_socket);
    opc_socket=-1;
}

static int opc_init_socket() {
    struct sockaddr_in mysock;
    int err=0;
    
    opc_socket=socket(PF_INET,SOCK_STREAM,IPPROTO_TCP);
    if (opc_socket<=0) return DS_ERR_SOCKET_CREATE;

	// Optimize string operations
    mysock.sin_family=AF_INET;
    mysock.sin_port=htons(SERVER_PORT);
    mysock.sin_addr.s_addr=inet_addr(SERVER_IP);
    if (connect(opc_socket,(struct sockaddr*)&mysock,sizeof(struct sockaddr_in))<0) {
	opc_close_socket();
	return DS_ERR_SOCKET_CONNECT;
    }

    return err;
}


static int opc_new_request(char *req_string, unsigned long req_size, unsigned long req_response) {
    int readed;
    if (send(opc_socket,req_string,req_size,0)<0) return DS_ERR_SOCKET_WRITE;
    readed=recv(opc_socket,opc_buffer,OPC_BUFFER_SIZE,0);
    if (readed<=0) return DS_ERR_SOCKET_READ;
    if (readed!=sizeof(unsigned long int)) return DS_ERR_INVALID_RESPONSE;
    if (ntohl(*((unsigned long int*)opc_buffer))!=req_response) {
	switch (*(unsigned long*)opc_buffer) {
	    case OPC_FECMD_REPEAT:
		return opc_new_request(req_string,req_size,req_response);
	    case OPC_FECMD_INVALIDREQ:
	        return DS_ERR_INVALID_REQUEST;
	    case OPC_FECMD_BUSY:
		return DS_ERR_BUSY;
	    break;
	    case OPC_FECMD_OPC_INVALIDREQ:
	        return DS_ERR_INVALID_REQUEST;
	    default:
	        return DS_ERR_INVALID_RESPONSE;
	}
    }
    return 0;
}



double smin=1E+10, smax=0, smid=0;
double rmin=1E+10, rmax=0, rmid=0;
unsigned long iters=0;


int mainloop() {
    FILE *thefile;
    xmlDocPtr doc;
    xmlNodePtr node;
    struct timeval tm1, tm2, tm3;
    double time1, time2;
    double duration;
    const char *rcvtime, *replytime;
    struct DSTimeT otime1, otime2;
    struct DSDurationT oduration;
    int err;
    unsigned long length;
    unsigned long readed, pos = 0, len;
    
    
    thefile = fopen("output.xml", "w");
    dime = 0;



    gettimeofday(&tm3,NULL);
    err=opc_new_request(opc_header,opc_header_size,OPC_FECMD_NEWREQ);
    if (err) return err;

    gettimeofday(&tm1,NULL);
    if (send(opc_socket,opc_request,opc_request_size,0)<0) return err;
    while (1) {
	if (!pos) {
	    readed=recv(opc_socket,opc_buffer,OPC_BUFFER_SIZE,0);
	    if (readed<sizeof(unsigned long)) return 1;
	}
	else if (readed<sizeof(unsigned long)) {
	    if (readed) memmove(opc_buffer, opc_buffer+pos, readed);
	    pos=recv(opc_socket,opc_buffer,OPC_BUFFER_SIZE,0);
	    if (pos<=0) return 1;
	    readed+=pos;
	    if (readed<sizeof(unsigned long)) return 1;
	    pos = 0;
	}
    
	len=ntohl(*(unsigned long*)(opc_buffer+pos));
	if (len>OPC_FECMD) {
	    if (len==OPC_FECMD_XMLEND) break;
	    if (len==OPC_FECMD_DIME_RESPONSE) {
		    readed-=sizeof(uint32_t);
		    pos+=sizeof(uint32_t);
		    writefunc(opc_buffer+pos, 1, readed, thefile);
		    moredata = 0;
		    while (moredata>=0) {
			readed=recv(opc_socket,opc_buffer,OPC_BUFFER_SIZE,0);
			if (readed<=0) return 1;
		    	writefunc(opc_buffer, 1, readed, thefile);
		    }
		    break;
	    }		    		    
	    return 1;
	}	
	readed-=sizeof(unsigned long);
	pos+=sizeof(unsigned long);

	if (readed<=len) {
	    if (readed) {
	    	fwrite(opc_buffer+pos,1,readed,thefile);
		len-=readed;
	    }
	    pos = 0;
	    
	    if (len) {
		readed=recv(opc_socket,opc_buffer,OPC_BUFFER_SIZE,0);
		if (readed<=0) 1;
	    
		while (readed<=len) {
	    	    fwrite(opc_buffer,1,readed,thefile);
		    len-=readed;
		    if (len) {
			readed=recv(opc_socket,opc_buffer,OPC_BUFFER_SIZE,0);
			if (readed<=0) return 1;
		    }
		}
	    }
	}
	
	if (len) {
	    fwrite(opc_buffer+pos,1,len,thefile);
	    pos+=len;
	    readed-=len;
	}
    }

    gettimeofday(&tm2,NULL);
    time1 = 1.*((tm2.tv_sec - tm1.tv_sec)*1000000 + tm2.tv_usec - tm1.tv_usec)/1000;
    time2 = 1.*((tm1.tv_sec - tm3.tv_sec)*1000000 + tm1.tv_usec - tm3.tv_usec)/1000;
    fclose(thefile);
    if (dime<-1) return 1;
    
    printf("\n");
    printf("Connect: %.3lfms, Request: %.3lfms, ", time2, time1);

    
    doc = xmlReadFile("output.xml", NULL, 0);
    node=xmlDocGetRootElement(doc)->children;
    rcvtime=dsXmlGetPropString(node,"RcvTime",NULL);
    replytime=dsXmlGetPropString(node,"ReplyTime",NULL);
    dsTimeSetDateTime(&otime1, rcvtime);
    dsTimeSetDateTime(&otime2, replytime);
    dsTimeCalcDuration(&oduration, &otime1, &otime2);
    xmlFreeDoc(doc);
    duration = 1000.*oduration.s + 1.*oduration.us/1000;
    printf("Internal: %.3lfms\n", duration);
    
    if (duration<smin) smin=duration;
    if (duration>smax) smax=duration;
    smid = (smid*iters + duration)/(iters+1);

    if (time1<rmin) rmin=time1;
    if (time1>rmax) rmax=time1;
    rmid = (rmid*iters + time1)/(iters+1);
    
    printf("Internal: %3.3lf(%.3lf,%.3lf)ms, iter: %lu\n",smid,smin,smax,iters);
    printf("Request : %3.3lf(%.3lf,%.3lf)ms, iter: %lu\n",rmid,rmin,rmax,iters);
    iters++;
    printf("\n");
    return 0;
}

int main(int argc, char *argv[]) {
    int err;
    unsigned long iters = 1;
    
    
    if (argc>=2) {
	iters = atol(argv[1]);
	if (!iters) iters=-1;
    }

    *((unsigned long int*)opc_header)=htonl(OPC_FECMD_NEWREQ);
    opc_header[sizeof(unsigned long int)]=0;
    opc_header[sizeof(unsigned long int)+1]=0;
    opc_header_size = sizeof(unsigned long)+2;
    
    opc_request_size = strlen(REQUEST);
    *((unsigned long int*)(opc_request))=htonl(opc_request_size);
    *((unsigned long int*)(opc_request+opc_request_size+sizeof(unsigned long)))=htonl(OPC_FECMD_XMLEND);
    memcpy(opc_request+sizeof(unsigned long),REQUEST,opc_request_size);
    opc_request_size += 2*sizeof(unsigned long);

    while (opc_init_socket()) sleep(1);
    
    for (;iters>0;iters--) {
	err=mainloop();
	if (err) {
	    opc_close_socket();	
	    while (opc_init_socket()) sleep(1);
	} // else sleep(1);
    }
	
    opc_close_socket();	
}
