#include <stdio.h>

#include <opclib.h>
#include <dsassert.h>
#include <dsplugin.h>
#include <dsdebug.h>
#include <dsbinaryconfiguration.h>


#include <opc.h>
#include <opcinterface.h>
#include <opcserverinterface.h>
#include <opcquality.h>

#include <mrandom.h>

RandomContext randomCreateContext() {
    RandomContext ctx;
    dsDebugServer(DS_DEBUG_NOTICE, "opcRandomCreateContext");    
    ctx = (RandomContext)malloc(sizeof(struct RandomContextT));
    return ctx;
}

static inline_function int PushDataComplex(OPCItemContext ictx) {
    int err;
    double *arr;
    unsigned long i;

#ifdef FAST_DATA_MANAGER
    DSAllocator allocator;
    DSAllocation allocation;
    struct _random_complex_t *item;

    allocator = dsDataManagerGetAllocator(ictx->dm, ictx->dm->native_representation, NULL);
    allocation = dsAllocatorSimpleAllocate(allocator, sizeof(struct _random_complex_t));
    item = opcAllocatorGet(allocator,allocation,NULL);
#else
    struct _random_complex_t sitem;
    struct _random_complex_t *item;
    
    item = &sitem;
#endif

    arr = item->array;
    for (i=0;i<SIZE;i++) 
	arr[i] = (1.*i/100);

#ifdef FAST_DATA_MANAGER
    err = opcItemPushNativeAllocation(ictx, NULL, allocator, allocation);
    dsDataManagerReturnAllocator(ictx->dm, ictx->dm->native_representation, allocator);
    return err;
#else
    return opcItemPushNativePointer(ictx, NULL, item, 0);
#endif
}

static inline_function int PushDataInt(OPCItemContext ictx) {
    int err;
    double *arr;
    unsigned long i;

#ifdef FAST_DATA_MANAGER
    DSAllocator allocator;
    DSAllocation allocation;
    uint32_t *item;

    allocator = dsDataManagerRepGetAllocator(ictx->dm, ictx->dm->native_representation, NULL);
    allocation = dsAllocatorSimpleAllocate(allocator, sizeof(uint32_t));
    item = opcAllocatorGet(allocator,allocation,NULL);
#else
    uint32_t sitem;
    uint32_t *item;
    
    item = &sitem;
#endif

    *item = rand();

#ifdef FAST_DATA_MANAGER
    err = opcItemPushNativeAllocation(ictx, NULL, allocator, allocation);
    dsDataManagerReturnAllocator(ictx->dm, ictx->dm->native_representation, allocator);
    return err;
#else
    return opcItemPushNativePointer(ictx, NULL, item, 0);
#endif
}

static inline_function int PushDataDouble(OPCItemContext ictx) {
    int err;
    double *arr;
    unsigned long i;

#ifdef FAST_DATA_MANAGER
    DSAllocator allocator;
    DSAllocation allocation;
    double *item;

    allocator = dsDataManagerRepGetAllocator(ictx->dm, ictx->native_representation, NULL);
    allocation = dsAllocatorSimpleAllocate(allocator, sizeof(double));
    item = opcAllocatorGet(allocator,allocation,NULL);
#else
    double sitem;
    double *item;
    
    item = &sitem;
#endif

    *item = 1. * rand() / RAND_MAX;

#ifdef FAST_DATA_MANAGER
    err = opcItemPushNativeAllocation(ictx, NULL, allocator, allocation);
    dsDataManagerReturnAllocator(ictx->dm, ictx->dm->native_representation, allocator);
    return err;
#else
    return opcItemPushNativePointer(ictx, NULL, item, 0);
#endif
}

static int InitDM(OPCItemContext ictx, DSBinaryType type) {
    DSDataManagerRepContext dm;
    
#ifdef FAST_DATA_MANAGER
    dm=opcItemRegisterFastDataManager(ictx, type, NULL, 0);
# ifdef FAST_DATA_MANAGER_XML
    dsDataManagerFastSetUserizeFunction(dm, dsDataComplexFastGetOPCXML);
# endif
#endif

#ifdef LEGACY_DATA_MANAGER
# ifdef LEGACY_DATA_MANAGER_DATABUFFER
    dm = opcItemRegisterLegacyDataManager(ictx, type, NULL, DS_DATAMANAGERREP_FLAG_SEQUENTIAL);
# else
    dm = opcItemRegisterLegacyDataManager(ictx, type, NULL, 0);
# endif 
# ifdef LEGACY_DATA_MANAGER_BINARY
    ictx->representation = 0;
# endif
#endif
}

int randomInitContext(RandomContext ctx, DSPluginContext pctx, OPCContext sctx) {
    int err;
    int i;
    OPCItemContext ictx;
    char stmp[64];
    
    DSOPCBinaryContext fctx;
    
//    dsDebugSetLevel(DS_DEBUG_DEBUG);
    dsDebugServer(DS_DEBUG_NOTICE, "opcRandomInitContext");    
    assert(sctx);
    fctx = sctx->fd_opcbinary;
    assert(fctx);
    
    ctx->dict = dsOPCBinaryCreateDictionary(fctx, "random.xml");
    if (!ctx->dict) {
	dsDebugServer(DS_DEBUG_ERROR, "MRandom(%p): Error Creating Dictionary", ctx);
	return dsErrorModule(DS_ERR_CREATE, DS_MODULE_OPCBinaryDictionary);
    }

    err=dsOPCBinaryDictionaryLocalise(ctx->dict);
    if (err) {
	dsDebugServer(DS_DEBUG_ERROR, "MRandom(%p): Error Localsing Dictionary", ctx);
	return dsErrorModule(err, DS_MODULE_OPCBinaryDictionary);
    }
	

    ctx->complextype = dsOPCBinaryDictionaryCreateType(ctx->dict, "complex");
    if (!ctx->complextype) {
	dsDebugServer(DS_DEBUG_ERROR, "MRandom(%p): ComplexType creation is failed");
	return dsErrorModule(DS_ERR_CREATE, DS_MODULE_BinaryType);
    }
    ctx->inttype = dsOPCBinaryDictionaryCreateType(ctx->dict, "integer");
    if (!ctx->inttype) {
	dsDebugServer(DS_DEBUG_ERROR, "MRandom(%p): IntType creation is failed");
	return dsErrorModule(DS_ERR_CREATE, DS_MODULE_BinaryType);
    }
    ctx->doubletype = dsOPCBinaryDictionaryCreateType(ctx->dict, "double");
    if (!ctx->doubletype) {
	dsDebugServer(DS_DEBUG_ERROR, "MRandom(%p): DoubleType creation is failed");
	return dsErrorModule(DS_ERR_CREATE, DS_MODULE_BinaryType);
    }

    for (i=0;((i<INUM)||(i<DNUM)||(i<CNUM));i++) {
	sprintf(stmp,"Item%i",i+1);
	if (i<INUM) {
	    err = opcServerRegisterItem(sctx,&ictx,"/Examples/MultiRandom/Integer/",stmp);
	    ctx->ictx_int[i] = ictx;
	    if (!err) {
		InitDM(ictx,ctx->inttype);
		opcItemEnableDataManager(ictx);
		err = PushDataInt(ictx);
		if (err) {
		    dsDebugServerError(DS_DEBUG_ERROR, err, "MRandom(%p): Error pushing initial Int value, ", ctx);
		    return dsErrorModule(DS_ERR_CREATE, DS_MODULE_BinaryType);
		}
		opcItemResetManualMode(ictx);
	    }
	}
	if (i<DNUM) {
	    err = opcServerRegisterItem(sctx,&ictx,"/Examples/MultiRandom/Double/",stmp);
	    ctx->ictx_double[i] = ictx;
	    if (!err) {
		InitDM(ictx,ctx->doubletype);
		opcItemEnableDataManager(ictx);
		err = PushDataDouble(ictx);
		if (err) {
		    dsDebugServerError(DS_DEBUG_ERROR, err, "MRandom(%p): Error pushing initial Double value, ", ctx);
		    return dsErrorModule(DS_ERR_CREATE, DS_MODULE_BinaryType);
		}
		opcItemResetManualMode(ictx);
	    }
	}
	if (i<CNUM) {
	    err = opcServerRegisterItem(sctx,&ictx,"/Examples/MultiRandom/Complex/",stmp);
	    ctx->ictx_complex[i] = ictx;
	    if (!err) {
		InitDM(ictx,ctx->complextype);
		opcItemEnableDataManager(ictx);
		err = PushDataComplex(ictx);
		if (err) {
		    dsDebugServerError(DS_DEBUG_ERROR, err, "MRandom(%p): Error pushing initial Double value, ", ctx);
		    return dsErrorModule(DS_ERR_CREATE, DS_MODULE_BinaryType);
		}
		opcItemResetManualMode(ictx);
	    }
	}
    }
    
    ctx->pctx=pctx;
    dsDebugServer(DS_DEBUG_INFO, "opcRandomInitContext Ok");    
    return err;
}

int randomRun(RandomContext ctx) {
    unsigned long j;
    unsigned long i = 0;
    assert(ctx);
    while (dsPluginContinue(ctx->pctx)) {
	if ((i++)%CLIENTS==0) {
	    dsDebugServer(DS_DEBUG_NOTICE, "=====> Pushing Data");    
	    for (j=0;j<INUM;j++)
		PushDataInt(ctx->ictx_int[j]);
	    for (j=0;j<DNUM;j++)
		PushDataDouble(ctx->ictx_double[j]);
	    for (j=0;j<CNUM;j++)
		PushDataComplex(ctx->ictx_complex[j]);
	}
#ifdef _WIN32
	Sleep(1000);
#else /* _WIN32 */
	sleep(1);
#endif /* _WIN32 */
    }
    return 0;
}

void randomFreeContext(RandomContext ctx) {
    dsDebugServer(DS_DEBUG_NOTICE, "opcRandomFreeContext");    
    assert(ctx);
}

void randomDestroyContext(RandomContext ctx) {
    dsDebugServer(DS_DEBUG_NOTICE, "opcRandomDestroyContext");    
    randomFreeContext(ctx);
    free(ctx);
}

dsPluginInfoFunctionImplementation {
    return &info;
}
