#include <string.h>

#include <dslib.h>
#include <dsdebug.h>
#include <dsassert.h>

#include <dsallocator.h>

DSAllocatorContext dsAllocatorCreateContext(DSAllocator allocator, DSMallocContext mctx) {
    DSAllocatorContext res;
    ds_assert_null(allocator);

    if (allocator->CreateContext) {
	res=allocator->CreateContext(mctx);
	memcpy(&res->allocator, allocator, sizeof(struct DSAllocatorT));
/*	res->allocator = allocator;*/
	res->malloc = mctx;
    } else 
	res=NULL;
	
    dsDebugServer(DS_DEBUG_DEBUG, "dsAllocator(%p): CreateContext: %p", res, allocator);    

    return res;
}


int dsAllocatorInitContext(DSAllocatorContext ctx, DSAllocatorOptions options) {
/*    DSAllocator allocator; */
    
    dsDebugServer(DS_DEBUG_DEBUG, "dsAllocator(%p): InitContext", ctx);    

    ds_assert_param(ctx);
/*    allocator = ctx->allocator;
    ds_assert_param(allocator);*/
    
    if ((!ctx->allocator.allocate)||(!ctx->allocator.free)||
	(!ctx->allocator.accept)||(!ctx->allocator.accept_chunk)||
	(!ctx->allocator.cancel)||(!ctx->allocator.cancel_chunk)||
	(!ctx->allocator.call)||(!ctx->allocator.sz)||
	(!ctx->allocator.get)||(!ctx->allocator.get_next))
	return DS_ERR_ALLOCATOR_INCOMPLETE;

    if (ctx->allocator.InitContext) return ctx->allocator.InitContext(ctx,options);
    return 0;
}

void dsAllocatorFreeContext(DSAllocatorContext ctx) {
/*    DSAllocator allocator;*/

    dsDebugServer(DS_DEBUG_DEBUG, "dsAllocator(%p): FreeContext", ctx);    

    ds_assert_return(ctx);

/*   
    allocator = ctx->allocator;
    ds_assert_return(allocator);
*/

    if (ctx->allocator.FreeContext) ctx->allocator.FreeContext(ctx);
}

void dsAllocatorDestroyContext(DSAllocatorContext ctx) {
/*    DSAllocator allocator;*/

    dsDebugServer(DS_DEBUG_DEBUG, "dsAllocator(%p): DestroyContext", ctx);    

    ds_assert_return(ctx);
/*    
    allocator = ctx->allocator;
    ds_assert_return(allocator);
*/
    
    if (ctx->allocator.DestroyContext) ctx->allocator.DestroyContext(ctx);
    else {
	if (ctx->allocator.FreeContext) ctx->allocator.FreeContext(ctx);
	dsFree(ctx->malloc, ctx);
    }
}


DSAllocatorOptions dsAllocatorGetDefaultOptions(DSAllocator ctx) {
    int err;
    DSAllocatorOptions options;
    
    if (ctx->CreateOptions) return ctx->CreateOptions();
    
    options = (DSAllocatorOptions)malloc(sizeof(struct DSAllocatorOptionsT));
    if (!options) return options;

    err = dsAllocatorInitDefaultOptions(ctx, options);
    if (err) {
	dsAllocatorDestroyOptions(ctx, options);
	return NULL;
    }
    return options;
}

int dsAllocatorInitDefaultOptions(DSAllocator ctx, DSAllocatorOptions options) {
    if (ctx->InitOptions) return ctx->InitOptions(options);
    memset(options, 0, sizeof(struct DSAllocatorOptionsT));
    return 0;
}

void dsAllocatorDestroyOptions(DSAllocator ctx, DSAllocatorOptions options) {
    if (ctx->DestroyOptions) ctx->DestroyOptions(options);
    else free(options);
}

DSAllocation dsAllocatorSimpleAllocate(DSAllocatorContext allocator, unsigned long size) {
    DSAllocation allocation;
    if (!allocator) return NULL;
    if (allocator->allocator.simple) {
	return allocator->allocator.simple(allocator,size);
    } else {
	if (allocator->allocator.start) allocation = allocator->allocator.start(allocator,size);
	else allocation = NULL;
	if (allocator->allocator.simple_chunk) allocator->allocator.simple_chunk(allocator,allocation,size);
	else {
	    allocator->allocator.allocate(allocator,allocation,size,0);
	    allocator->allocator.accept_chunk(allocator,allocation,NULL);
	}
	return allocator->allocator.accept(allocator,allocation);	
    }    
}
