#ifndef _DS_ALLOCATOR_H
#define _DS_ALLOCATOR_H

#include <dsmalloc.h>
#include <dslib.h>

typedef struct DSAllocatorOptionsT *DSAllocatorOptions;
typedef struct DSAllocatorContextT *DSAllocatorContext;
typedef const struct DSAllocatorT *DSAllocator;
typedef void *DSAllocation;

typedef void *(*DSAllocatorCreateFunction)(DSMallocContext mctx);
typedef int (*DSAllocatorInitFunction)(DSAllocatorContext ctx, DSAllocatorOptions options);
typedef void (*DSAllocatorFreeFunction)(DSAllocatorContext ctx);
typedef void (*DSAllocatorDestroyFunction)(DSAllocatorContext ctx);

typedef DSAllocatorOptions (*DSAllocatorGetDefaultOptions)();
typedef int (*DSAllocatorInitDefaultOptions)(DSAllocatorOptions options);
typedef void (*DSAllocatorDestroyOptions)(DSAllocatorOptions options);

typedef DSAllocation (*DSAllocatorStartAllocationFunction)(DSAllocatorContext ctx, unsigned long preallocate);
typedef void *(*DSAllocatorAllocateMemFunction)(DSAllocatorContext ctx, DSAllocation allocation, unsigned long size, unsigned char alignment);
typedef void (*DSAllocatorFreeMemFunction)(DSAllocatorContext ctx, DSAllocation allocation);
typedef void *(*DSAllocatorAcceptChunkFunction)(DSAllocatorContext ctx, DSAllocation allocation, unsigned long *size);
typedef DSAllocation (*DSAllocatorAcceptFunction)(DSAllocatorContext ctx, DSAllocation allocation);
typedef void (*DSAllocatorCancelFunction)(DSAllocatorContext ctx, DSAllocation allocation);
typedef void (*DSAllocatorCancelChunkFunction)(DSAllocatorContext ctx, DSAllocation allocation);
typedef void (*DSAllocatorTrimFunction)(DSAllocatorContext ctx, DSAllocation allocation, unsigned long size);
typedef unsigned long (*DSAllocatorReserveFunction)(DSAllocatorContext ctx, DSAllocation allocation, unsigned long size);

typedef void *(*DSAllocatorGetFunction)(DSAllocatorContext ctx, DSAllocation allocation, unsigned long *size);
typedef void *(*DSAllocatorGetNextFunction)(DSAllocatorContext ctx, DSAllocation allocation, void *current, unsigned long *size);
typedef int (*DSAllocatorCallFunction)(DSAllocatorContext ctx, DSAllocation allocation, DSSendFunction func, void *attr, unsigned char align);
typedef unsigned long (*DSAllocatorSizeofFunction)(DSAllocatorContext ctx, DSAllocation allocation, unsigned char align);

typedef void *(*DSAllocatorAllocateSimpleChunkFunction)(DSAllocatorContext ctx, DSAllocation allocation, unsigned long size);
typedef DSAllocation (*DSAllocatorAllocateSimpleFunction)(DSAllocatorContext ctx, unsigned long size);

typedef void (*DSAllocatorMantainFunction)(DSAllocatorContext ctx);

struct DSAllocatorT {
    char *name;
    
    DSAllocatorCreateFunction CreateContext;
    DSAllocatorInitFunction InitContext;
    DSAllocatorFreeFunction FreeContext;
    DSAllocatorDestroyFunction DestroyContext;

    DSAllocatorGetDefaultOptions CreateOptions;
    DSAllocatorInitDefaultOptions InitOptions;
    DSAllocatorDestroyOptions DestroyOptions;
    
    DSAllocatorMantainFunction mantain;
    
    DSAllocatorStartAllocationFunction start;
    DSAllocatorAllocateMemFunction allocate;
    DSAllocatorFreeMemFunction free;
    DSAllocatorAcceptChunkFunction accept_chunk;
    DSAllocatorCancelChunkFunction cancel_chunk;
    DSAllocatorAcceptFunction accept;
    DSAllocatorCancelFunction cancel;

    DSAllocatorAllocateSimpleChunkFunction simple_chunk;
    DSAllocatorAllocateSimpleFunction simple;

    DSAllocatorReserveFunction reserve;
    DSAllocatorTrimFunction trim;

    DSAllocatorGetFunction get;
    DSAllocatorGetNextFunction get_next;
    DSAllocatorCallFunction call;
    DSAllocatorSizeofFunction sz;
};

struct DSAllocatorContextT {
    struct DSAllocatorT allocator;
    DSMallocContext malloc;
};

struct DSAllocatorOptionsT {
    void *pool;
    unsigned long minimal_size;
    unsigned long maximal_size;
    unsigned long fixed_allocation_size;
};

DSAllocatorOptions dsAllocatorGetDefaultOptions(DSAllocator ctx);
int dsAllocatorInitDefaultOptions(DSAllocator ctx, DSAllocatorOptions options);
void dsAllocatorDestroyOptions(DSAllocator ctx, DSAllocatorOptions options);


DSAllocatorContext dsAllocatorCreateContext(DSAllocator allocator, DSMallocContext mctx);
int dsAllocatorInitContext(DSAllocatorContext ctx, DSAllocatorOptions options);
void dsAllocatorFreeContext(DSAllocatorContext ctx);
void dsAllocatorDestroyContext(DSAllocatorContext ctx);


#define dsAllocatorStart(ctx, size) \
    (((ctx)&&(ctx)->allocator.start)?(ctx)->allocator.start(ctx,size):NULL)
#define dsAllocatorAllocate(ctx, allocation, size, alignment) \
    ((ctx)?(ctx)->allocator.allocate(ctx,allocation,size,alignment):NULL)
#define dsAllocatorFree(ctx, ptr) \
    ((ctx)?(ctx)->allocator.free(ctx,ptr):0)
#define dsAllocatorAcceptChunk(ctx,allocation,size) \
    ((ctx)?(ctx)->allocator.accept_chunk(ctx,allocation,size):NULL)
#define dsAllocatorAccept(ctx,allocation) \
    ((ctx)?(ctx)->allocator.accept(ctx,allocation):NULL)
#define dsAllocatorCancelChunk(ctx,allocation) \
    ((ctx)?(ctx)->allocator.cancel_chunk(ctx,allocation):0)
#define dsAllocatorCancel(ctx,allocation) \
    ((ctx)?(ctx)->allocator.cancel(ctx,allocation):0)
#define dsAllocatorCall(ctx, ptr, func, attr, align) \
    ((ctx)?(ctx)->allocator.call(ctx, ptr, func, attr, align):DS_ERR_INVALID_ARGUMENT)
#define dsAllocatorSizeof(ctx, ptr, align) \
    ((ctx)?(ctx)->allocator.sz(ctx, ptr, align):0)
#define dsAllocatorTrim(ctx, allocation, size) \
    (((ctx)&&((ctx)->allocator.trim))?(ctx)->allocator.trim(ctx, allocation, size):0)
#define dsAllocatorReserve(ctx, allocation, size) \
    (((ctx)&&((ctx)->allocator.reserve))?(ctx)->allocator.reserve(ctx, allocation, size):0)
#define dsAllocatorMantain(ctx) \
    (((ctx)&&((ctx)->allocator.mantain))?(ctx)->allocator.mantain(ctx):0)
#define dsAllocatorSimpleChunk(ctx,allocation,size) \
    ((ctx)?(((ctx)->allocator.simple_chunk)?ctx->allocator.simple_chunk(ctx,allocation,size):((ctx->allocator.allocate(ctx,allocation,size,0))?ctx->allocator.accept_chunk(ctx,allocation,NULL):NULL)):NULL)
#define dsAllocatorGet(ctx,allocation,size) \
    ((ctx)?(ctx)->allocator.get(ctx,allocation,size):NULL)
#define dsAllocatorGetNext(ctx,allocation,current,size) \
    ((ctx)?(ctx)->allocator.get_next(ctx,allocation,current,size)

DSAllocation dsAllocatorSimpleAllocate(DSAllocatorContext allocator, unsigned long size);
    
#endif /* _DS_ALLOCATOR_H */
