#define _DS_ALLOCATOR_SIMPLE_C
#include <string.h>

#include <dslib.h>
#include <dsdebug.h>
#include <dsassert.h>

#include <dsbinaryconfiguration.h>

#include <dsallocatorsimple.h>

#define HEADER_SIZE (sizeof(void*) + sizeof(unsigned long))
#define OUTBUF_SIZE 4096
#define ALIGN(pos, align) \
    if ((align)&&(pos&(align-1))) pos = (pos&~(align-1))+align;

static unsigned long header_size = 0;


DSAllocator dsAllocatorSimple() {
    return &ds_allocator_simple;
}


DSAllocatorSimpleContext dsAllocatorSimpleCreateContext(DSMallocContext mctx) {
    DSAllocatorSimpleContext ctx;

    ctx = (DSAllocatorSimpleContext)dsMalloc(mctx, sizeof(struct DSAllocatorSimpleContextT));
    if (ctx) {
	memcpy(&ctx->base.allocator, &ds_allocator_simple, sizeof(struct DSAllocatorT));
	ctx->base.malloc = mctx;
	
	if (!header_size) {
	    header_size = HEADER_SIZE;
	    ALIGN(header_size, ds_binary_configuration->align_allocation);
	}
    }
    return ctx;
}

int dsAllocatorSimpleInitContext(DSAllocatorSimpleContext ctx, const void *options) {
    return 0;
}

void dsAllocatorSimpleFreeContext(DSAllocatorSimpleContext ctx) {
    ds_assert_return(ctx);
}

void dsAllocatorSimpleDestroyContext(DSAllocatorSimpleContext ctx) {
    dsAllocatorSimpleFreeContext(ctx);
    dsFree(ctx->base.malloc, ctx);
}

DSAllocation dsAllocatorSimpleStart(DSAllocatorSimpleContext ctx, unsigned long preallocate) {
    DSAllocationSimple allocation;

    allocation = (DSAllocationSimple)dsMalloc(ctx->base.malloc, sizeof(struct DSAllocationSimpleT));
    if (!allocation) return NULL;
    
    allocation->buffer = NULL;
    allocation->prebuffer = NULL;
    allocation->allocation = NULL;
    return (DSAllocation)allocation;
    
}

void *dsAllocatorSimpleAllocateMemory(DSAllocatorSimpleContext ctx, DSAllocation allocation, unsigned long size, unsigned char alignment) {
    char *buffer;
    unsigned long new_size, buff_size;

    ds_assert_null(ctx);
    ds_assert_null(allocation);
    ds_assert_null(size>0);

    if (((DSAllocationSimple)allocation)->buffer) {
	new_size = ((DSAllocationSimple)allocation)->current_size;
	ALIGN(new_size,alignment);
	new_size+=size + header_size;

	if (new_size>((DSAllocationSimple)allocation)->buffer_size) {
	    for (buff_size=((DSAllocationSimple)allocation)->buffer_size*2;buff_size<new_size;buff_size*=2);
	    buffer=dsRealloc(ctx->base.malloc,((DSAllocationSimple)allocation)->buffer,buff_size);
	    if (buffer) {
		((DSAllocationSimple)allocation)->buffer = buffer;
		((DSAllocationSimple)allocation)->buffer_size = buff_size;
	    } else return NULL;
	}
	buff_size = ((DSAllocationSimple)allocation)->current_size;
	((DSAllocationSimple)allocation)->current_size = new_size - header_size;
	return buffer+buff_size;
    } else {
	new_size = size + header_size;
	ALIGN(new_size, ds_binary_configuration->align_allocation);
	((DSAllocationSimple)allocation)->buffer=dsMalloc(ctx->base.malloc, new_size);
	if (((DSAllocationSimple)allocation)->buffer) {
	    ((DSAllocationSimple)allocation)->current_size = size;
	    ((DSAllocationSimple)allocation)->buffer_size = new_size;
	}
	return ((DSAllocationSimple)allocation)->buffer + header_size;
    }
}

void *dsAllocatorSimpleAcceptChunk(DSAllocatorSimpleContext ctx, DSAllocation allocation, unsigned long *size) {
    char *buffer;

    ds_assert_null(ctx);
    ds_assert_null(allocation);
    
    buffer=((DSAllocationSimple)allocation)->buffer;

    if (buffer) {
	if (((DSAllocationSimple)allocation)->prebuffer) *((void**)(((DSAllocationSimple)allocation)->prebuffer+sizeof(unsigned long)))=buffer;
	else ((DSAllocationSimple)allocation)->allocation = (DSAllocation)buffer;
	*((unsigned long*)(buffer))=((DSAllocationSimple)allocation)->current_size;
	*((void**)(buffer+sizeof(unsigned long)))=NULL;
	((DSAllocationSimple)allocation)->prebuffer = buffer;
	((DSAllocationSimple)allocation)->buffer = NULL;
	if (size) *size = ((DSAllocationSimple)allocation)->current_size;
	return buffer + header_size;
    }
    return NULL;
}

DSAllocation dsAllocatorSimpleAccept(DSAllocatorSimpleContext ctx, DSAllocation allocation) {
    DSAllocation res;

    ds_assert_null(ctx);
    ds_assert_null(allocation);

    if (((DSAllocationSimple)allocation)->buffer) dsAllocatorSimpleAcceptChunk(ctx, allocation, NULL);
    res = (DSAllocation)((DSAllocationSimple)allocation)->allocation;
    dsFree(ctx->base.malloc,allocation);
    return res;
}


void dsAllocatorSimpleCancelChunk(DSAllocatorSimpleContext ctx, DSAllocation allocation) {
    ds_assert_return(ctx);
    ds_assert_return(allocation);
    
    if (((DSAllocationSimple)allocation)->buffer) {
	dsFree(ctx->base.malloc, ((DSAllocationSimple)allocation)->buffer);
	((DSAllocationSimple)allocation)->buffer = NULL;
    }
}

void dsAllocatorSimpleCancel(DSAllocatorSimpleContext ctx, DSAllocation allocation) {
    char *ptr, *next;

    ds_assert_return(ctx);
    ds_assert_return(allocation);
    
    dsAllocatorSimpleCancelChunk(ctx, allocation);
    if (((DSAllocationSimple)allocation)->allocation) {
	for (ptr = (char*)((DSAllocationSimple)allocation)->allocation;ptr;ptr=next) {
	    next = *((void**)(ptr+sizeof(unsigned long)));
	    dsFree(ctx->base.malloc, ptr);
	}
    }
    dsFree(ctx->base.malloc, allocation);
}

void dsAllocatorSimpleFree(DSAllocatorSimpleContext ctx, DSAllocation allocation) {
    char *ptr, *next;
    
    ds_assert_return(ctx);
    ds_assert_return(allocation);

    for (ptr = (char*)allocation;ptr;ptr=next) {
        next = *((void**)(ptr+sizeof(unsigned long)));
        dsFree(ctx->base.malloc, ptr);
    }
}

void *dsAllocatorSimpleAllocateChunk(DSAllocatorSimpleContext ctx, DSAllocation allocation, unsigned long size) {
    unsigned long new_size;
    char *buffer;

    ds_assert_null(ctx);
    ds_assert_null(!((DSAllocationSimple)allocation)->buffer);

    new_size = size + header_size;
    ALIGN(new_size, ds_binary_configuration->align_allocation);
    buffer=dsMalloc(ctx->base.malloc, new_size);
    if (!buffer) return NULL;
    
    if (((DSAllocationSimple)allocation)->prebuffer) 
	*((void**)(((DSAllocationSimple)allocation)->prebuffer+sizeof(unsigned long)))=buffer;
    else 
	((DSAllocationSimple)allocation)->allocation = (DSAllocation)buffer;
    *((unsigned long*)(buffer))=size;
    *((void**)(buffer+sizeof(unsigned long)))=NULL;
    ((DSAllocationSimple)allocation)->prebuffer = buffer;
    
    return buffer + header_size;
}

DSAllocation dsAllocatorSimpleAllocateSimple(DSAllocatorSimpleContext ctx, unsigned long size) {
    unsigned long new_size;
    char *buffer;
    
    ds_assert_null(ctx);

    new_size = size + header_size;
    ALIGN(new_size, ds_binary_configuration->align_allocation);
    buffer=dsMalloc(ctx->base.malloc, new_size);
    if (!buffer) return NULL;
    *((unsigned long*)(buffer))=size;
    *((void**)(buffer+sizeof(unsigned long)))=NULL;
    return (DSAllocation)buffer;
}

void* dsAllocatorSimpleGet(DSAllocatorSimpleContext ctx, DSAllocation allocation, unsigned long *size) {
    char *buffer;
    
    ds_assert_null(ctx);
    ds_assert_null(allocation);
    
    buffer = (char*)allocation;
    if (size) *size = *((unsigned long*)buffer);
    return buffer+header_size;
}

void* dsAllocatorSimpleGetNext(DSAllocatorSimpleContext ctx, DSAllocation allocation, void *current, unsigned long *size) {
    char *buffer;

    ds_assert_null(ctx);
    ds_assert_null(allocation);
    ds_assert_null(current);
    
    buffer = *((char**)((char*)current-header_size+sizeof(unsigned long)));
    if (!buffer) return buffer;
    
    if (size) *size = *((unsigned long*)buffer);
    return buffer+header_size;
}

int dsAllocatorSimpleCall(DSAllocatorSimpleContext ctx, DSAllocation allocation, DSSendFunction func, void *attr, unsigned char align) {
    int err = 0;
    unsigned long size;
    unsigned long pos = 0;
    char *ptr, *next;
    char *buffer;
    
    ds_assert_param(ctx);
    ds_assert_param(allocation);
    
    buffer = dsMalloc(ctx->base.malloc, OUTBUF_SIZE);
    if (!buffer) return DS_ERR_MEM;

    for (ptr = (char*)allocation;ptr;ptr=next) {
	size = *((unsigned long*)ptr);
        next = *((void**)(ptr+sizeof(unsigned long)));
	if ((next)&&(align))
	    ALIGN(size,ds_binary_configuration->align_allocation);

	if (pos+size>OUTBUF_SIZE) {
	    if (pos) {
	        err=func(attr,buffer,pos);
		if (err) break;
	    }

	    if (size>OUTBUF_SIZE/2) {
		err=func(attr,ptr+header_size,size);
		if (err) break;
		pos = 0;
	    } else {
		memcpy(buffer,ptr+header_size,size);
		pos = size;
	    }
	} else {
	    memcpy(buffer+pos,ptr+header_size,size);
	    pos += size;
	}
    }
    if ((!err)&&(pos)) err=func(attr,buffer,pos);
    
    free(buffer);
    return err;
}

unsigned long dsAllocatorSimpleSizeof(DSAllocatorSimpleContext ctx, DSAllocation allocation, unsigned char align) {
    unsigned long res=0, size;
    char *ptr, *next;
    
    ds_assert_zero(ctx);
    ds_assert_zero(allocation);

    for (ptr = (char*)allocation;ptr;ptr=next) {
	size = *((unsigned long*)ptr);
        next = *((void**)(ptr+sizeof(unsigned long)));
	if ((next)&&(align))
	    ALIGN(size,ds_binary_configuration->align_allocation);
	res += size;
    }
    return res;
}
