#ifndef _DS_ALLOCATOR_SIMPLE_H
#define _DS_ALLOCATOR_SIMPLE_H

#include <dsallocator.h>

typedef struct DSAllocatorSimpleContextT *DSAllocatorSimpleContext;
typedef struct DSAllocationSimpleT *DSAllocationSimple;

struct DSAllocationSimpleT {
    DSAllocation allocation;
    char *buffer;
    char *prebuffer;
    unsigned long current_size;
    unsigned long buffer_size;
};

struct DSAllocatorSimpleContextT {
    struct DSAllocatorContextT base;
};


DSAllocator dsAllocatorSimple();


DSAllocatorSimpleContext dsAllocatorSimpleCreateContext(DSMallocContext mctx);
int dsAllocatorSimpleInitContext(DSAllocatorSimpleContext ctx, const void *options);
void dsAllocatorSimpleFreeContext(DSAllocatorSimpleContext ctx);
void dsAllocatorSimpleDestroyContext(DSAllocatorSimpleContext ctx);

DSAllocation dsAllocatorSimpleStart(DSAllocatorSimpleContext ctx, unsigned long preallocate);
void *dsAllocatorSimpleAllocateMemory(DSAllocatorSimpleContext ctx, DSAllocation allocation, unsigned long size, unsigned char alignment);
void *dsAllocatorSimpleAcceptChunk(DSAllocatorSimpleContext ctx, DSAllocation allocation, unsigned long *size);
void dsAllocatorSimpleCancelChunk(DSAllocatorSimpleContext ctx, DSAllocation allocation);
DSAllocation dsAllocatorSimpleAccept(DSAllocatorSimpleContext ctx, DSAllocation allocation);
void dsAllocatorSimpleCancel(DSAllocatorSimpleContext ctx, DSAllocation allocation);
void dsAllocatorSimpleFree(DSAllocatorSimpleContext ctx, DSAllocation allocation);
void *dsAllocatorSimpleAllocateChunk(DSAllocatorSimpleContext ctx, DSAllocation allocation, unsigned long size);
DSAllocation dsAllocatorSimpleAllocateSimple(DSAllocatorSimpleContext ctx, unsigned long size);
void* dsAllocatorSimpleGet(DSAllocatorSimpleContext ctx, DSAllocation allocation, unsigned long *size);
void* dsAllocatorSimpleGetNext(DSAllocatorSimpleContext ctx, DSAllocation allocation, void *current, unsigned long *size);
int dsAllocatorSimpleCall(DSAllocatorSimpleContext ctx, DSAllocation allocation, DSSendFunction func, void *attr, unsigned char align);
unsigned long dsAllocatorSimpleSizeof(DSAllocatorSimpleContext ctx, DSAllocation allocation, unsigned char align);

#ifdef _DS_ALLOCATOR_SIMPLE_C
static const struct DSAllocatorT ds_allocator_simple = {
    "Simple",
    (DSAllocatorCreateFunction)dsAllocatorSimpleCreateContext,
    (DSAllocatorInitFunction)dsAllocatorSimpleInitContext,
    (DSAllocatorFreeFunction)dsAllocatorSimpleFreeContext,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    (DSAllocatorStartAllocationFunction)dsAllocatorSimpleStart,
    (DSAllocatorAllocateMemFunction)dsAllocatorSimpleAllocateMemory,
    (DSAllocatorFreeMemFunction)dsAllocatorSimpleFree,
    (DSAllocatorAcceptChunkFunction)dsAllocatorSimpleAcceptChunk,
    (DSAllocatorCancelChunkFunction)dsAllocatorSimpleCancelChunk,
    (DSAllocatorAcceptFunction)dsAllocatorSimpleAccept,
    (DSAllocatorCancelFunction)dsAllocatorSimpleCancel,
    (DSAllocatorAllocateSimpleChunkFunction)dsAllocatorSimpleAllocateChunk,
    (DSAllocatorAllocateSimpleFunction)dsAllocatorSimpleAllocateSimple,
    NULL,
    NULL,
    (DSAllocatorGetFunction)dsAllocatorSimpleGet,
    (DSAllocatorGetNextFunction)dsAllocatorSimpleGetNext,
    (DSAllocatorCallFunction)dsAllocatorSimpleCall,
    (DSAllocatorSizeofFunction)dsAllocatorSimpleSizeof
};

#endif /* _DS_ALLOCATOR_SIMPLE_C */

#endif /* _DS_ALLOCATOR_SIMPLE_H */

