#ifndef _DS_BUFFER_POOL_H
#define _DS_BUFFER_POOL_H

#include <dsmalloc.h>
#include <dslock.h>

typedef struct DSBufferPoolContextT *DSBufferPoolContext;
typedef struct DSBufferPoolBufferT *DSBufferPoolBuffer;

typedef DSBufferPoolBuffer (*DSBufferPoolCreateBufferFunction)(DSBufferPoolContext ctx, unsigned long buffer_size);
typedef int (*DSBufferPoolRestoreBufferFunction)(DSBufferPoolContext ctx, DSBufferPoolBuffer buffer);
typedef int (*DSBufferPoolResizeBufferFunction)(DSBufferPoolContext ctx, DSBufferPoolBuffer *buffer, unsigned long buffer_size);
typedef int (*DSBufferPoolReplaceBufferFunction)(DSBufferPoolContext ctx, DSBufferPoolBuffer *buffer, unsigned long buffer_size);
typedef void (*DSBufferPoolDestroyBufferFunction)(DSBufferPoolContext ctx, DSBufferPoolBuffer buffer);

/*
    Buffer size should be multiple of align_allocation!
*/

struct DSBufferPoolBufferT {
    DSBufferPoolBuffer next;
    unsigned long buffer_size;
    unsigned char buffer[];
};


struct DSBufferPoolContextT {
    DSMallocContext malloc;
    
    DSLockContext lock;
    
    unsigned char bins;
    DSBufferPoolBuffer *first;
    DSBufferPoolBuffer *last;
    unsigned long *allocated;
    
    unsigned char lsr;
    
    unsigned long default_size;
    unsigned char default_bin;
    
    DSBufferPoolCreateBufferFunction create;
    DSBufferPoolRestoreBufferFunction restore;
    DSBufferPoolResizeBufferFunction resize;
    DSBufferPoolReplaceBufferFunction replace;
    DSBufferPoolDestroyBufferFunction destroy;
};

DSBufferPoolContext dsBufferPoolCreateContext(DSMallocContext mctx);
int dsBufferPoolInitContext(DSBufferPoolContext ctx, unsigned long first_bin, unsigned char bins);
void dsBufferPoolFreeContext(DSBufferPoolContext ctx);
void dsBufferPoolDestroyContext(DSBufferPoolContext ctx);

int dsBufferPoolSetDefaultSize(DSBufferPoolContext ctx, unsigned long default_size);
int dsBufferPoolSetFunctions(DSBufferPoolContext ctx, DSBufferPoolCreateBufferFunction create, DSBufferPoolRestoreBufferFunction restore, DSBufferPoolResizeBufferFunction resize, DSBufferPoolReplaceBufferFunction replace, DSBufferPoolDestroyBufferFunction destroy);

DSBufferPoolBuffer dsBufferPoolGetBuffer(DSBufferPoolContext ctx, unsigned long size);
void dsBufferPoolReturnBuffer(DSBufferPoolContext ctx, DSBufferPoolBuffer buffer);
int dsBufferPoolResizeBuffer(DSBufferPoolContext ctx, DSBufferPoolBuffer *buffer, unsigned long size);
int dsBufferPoolReplaceBuffer(DSBufferPoolContext ctx, DSBufferPoolBuffer *buffer, unsigned long size);
void dsBufferPoolMantain(DSBufferPoolContext ctx);

#endif /* _DS_BUFFER_POOL_H */
