#include <dsassert.h>
#include <dslib.h>

#include <dsbufferpoolsimplebuffer.h>

static DSBufferContext dsBufferPoolCreateSimpleBuffer(DSBufferPoolContext ctx, unsigned long buffer_size) {
    DSBufferContext buffer;
    
    ds_assert_null(ctx);
    ds_assert_null(buffer_size);
    
    buffer = (DSBufferContext)dsMalloc(ctx->malloc, sizeof(struct DSBufferContextT));
    if (!buffer) return buffer;
    
    buffer->buffer = (unsigned char*)dsMalloc(ctx->malloc, buffer_size);
    if (!buffer->buffer) {
	dsFree(ctx->malloc, buffer);
	return NULL;
    }
    buffer->buffer_size = buffer_size;

    
    if (dsBufferInitContext(buffer)) {
	dsFree(ctx->malloc, buffer);
	dsFree(ctx->malloc, buffer->buffer);
	return NULL;
    }
    
    buffer->pool = ctx;
    
    return buffer;
}

static int dsBufferPoolResizeSimpleBuffer(DSBufferPoolContext ctx, DSBufferContext *bufferptr, unsigned long buffer_size) {
    void *buff;
    DSBufferContext buffer = *bufferptr;
    
    buff = (char*)dsRealloc(ctx->malloc, buffer->buffer, buffer_size);
    if (!buff) return DS_ERR_MEM;
    buffer->buffer = buff;
    buffer->buffer_size = buffer_size;
    
    return 0;
}

static int dsBufferPoolRestoreSimpleBuffer(DSBufferPoolContext ctx, DSBufferContext buffer) {
    dsBufferRestoreContext(buffer);
    return 0;
}

static int dsBufferPoolReplaceSimpleBuffer(DSBufferPoolContext ctx, DSBufferContext *bufferptr, unsigned long buffer_size) {
    void *buff;
    DSBufferContext buffer = *bufferptr;
    
    buff = (char*)dsMalloc(ctx->malloc, buffer_size);
    if (!buff) return DS_ERR_MEM;
    dsFree(ctx->malloc, buffer->buffer);
    buffer->buffer = buff;
    buffer->buffer_size = buffer_size;
    
    return 0;
}

static void dsBufferPoolDestroySimpleBuffer(DSBufferPoolContext ctx, DSBufferContext buffer) {
    dsBufferFreeContext(buffer);
    if (buffer->buffer) dsFree(ctx->malloc, buffer->buffer);
    dsFree(ctx->malloc, buffer);
}


int dsBufferPoolSetSimpleBufferFunctions(DSBufferPoolContext ctx) {
    unsigned char i;
    
    ds_assert_param(ctx);
    ds_assert_init(ctx->first);
    
    for (i=0;i<ctx->bins;i++)
	if ((ctx->first[i])||(ctx->allocated[i])) return DS_ERR_INITIALIZED;
    if (ctx->allocated[ctx->bins]) return DS_ERR_INITIALIZED;

    ctx->create = (DSBufferPoolCreateBufferFunction)&dsBufferPoolCreateSimpleBuffer;
    ctx->restore = (DSBufferPoolRestoreBufferFunction)&dsBufferPoolRestoreSimpleBuffer;
    ctx->resize = (DSBufferPoolResizeBufferFunction)&dsBufferPoolResizeSimpleBuffer;
    ctx->replace = (DSBufferPoolReplaceBufferFunction)&dsBufferPoolReplaceSimpleBuffer;
    ctx->destroy = (DSBufferPoolDestroyBufferFunction)&dsBufferPoolDestroySimpleBuffer;
    return 0;
}

int dsBufferPoolCheckSimpleBufferFunctions(DSBufferPoolContext ctx) {
    ds_assert_param(ctx);
    
    if (ctx->create != (DSBufferPoolCreateBufferFunction)&dsBufferPoolCreateSimpleBuffer) return DS_ERR;
    if (ctx->restore != (DSBufferPoolRestoreBufferFunction)&dsBufferPoolRestoreSimpleBuffer) return DS_ERR;
    if (ctx->resize != (DSBufferPoolResizeBufferFunction)&dsBufferPoolResizeSimpleBuffer) return DS_ERR;
    if (ctx->replace != (DSBufferPoolReplaceBufferFunction)&dsBufferPoolReplaceSimpleBuffer) return DS_ERR;
    if (ctx->destroy != (DSBufferPoolDestroyBufferFunction)&dsBufferPoolDestroySimpleBuffer) return DS_ERR;
    
    return 0;
}

