#ifndef _DS_DATA_BUFFER_H
#define _DS_DATA_BUFFER_H

#include <dsmalloc.h>

typedef struct DSDataBufferContextT *DSDataBufferContext;

#include <dsallocator.h>
#include <dsbuffer.h>
#include <dsbufferpool.h>

typedef struct DSDataBufferOptionsT *DSDataBufferOptions;

struct DSDataBufferContextT {
    struct DSAllocatorContextT base;

    DSBufferPoolContext pool;
    DSBufferContext buffers;
    DSBufferContext current;
    DSBufferContext first_chunk;
    DSAllocation allocation;
    
    unsigned long preallocated;
    unsigned long block_size;
    
    unsigned long allocation_size;

    unsigned long n;
    unsigned long max1, max2;
    unsigned long save;
    
    unsigned long maxfill;
    
    unsigned long minimal_size;
    unsigned long maximal_size;
    unsigned long fixed_allocation_size;
};

struct DSDataBufferOptionsT {
    DSBufferPoolContext pool;
    unsigned long minimal_size;
    unsigned long maximal_size;
    unsigned long block_size;
    unsigned long fixed_allocation_size;
};

DSAllocator dsAllocatorDataBuffer();

DSDataBufferContext dsDataBufferCreateContext(DSMallocContext mctx);
int dsDataBufferInitContext(DSDataBufferContext ctx, DSDataBufferOptions options);
void dsDataBufferFreeContext(DSDataBufferContext ctx);
void dsDataBufferDestroyContext(DSDataBufferContext ctx);

DSDataBufferOptions dsDataBufferGetDefaultOptions();
int dsDataBufferInitDefaultOptions(DSDataBufferOptions options);
void dsDataBufferDestroyOptions(DSDataBufferOptions options);

int dsDataBufferOptionsSetSize(DSDataBufferOptions options, unsigned long minimal_size, unsigned long maximal_size);
int dsDataBufferOptionsSetBlockSize(DSDataBufferOptions options, unsigned long block_size);
int dsDataBufferOptionsSetAllocationSize(DSDataBufferOptions options, unsigned long fixed_allocation_size);
int dsDataBufferOptionsSetBufferPool(DSDataBufferOptions options, DSBufferPoolContext pool);


unsigned long dsDataBufferReserve(DSDataBufferContext ctx, DSAllocation notrequired, unsigned long size);
void dsDataBufferTrim(DSDataBufferContext ctx, DSAllocation notrequired, unsigned long size);
void *dsDataBufferAllocate(DSDataBufferContext ctx, DSAllocation notrequired, unsigned long size, unsigned char alignment);
void dsDataBufferFree(DSDataBufferContext ctx, DSAllocation buffer);

void *dsDataBufferAcceptChunk(DSDataBufferContext ctx, DSAllocation notrequired, unsigned long *req_size);
void dsDataBufferCancelChunk(DSDataBufferContext ctx, DSAllocation notrequired);
DSAllocation dsDataBufferAccept(DSDataBufferContext ctx, DSAllocation notrequired);
void dsDataBufferCancel(DSDataBufferContext ctx, DSAllocation notrequired);

void *dsDataBufferSimpleChunk(DSDataBufferContext ctx, DSAllocation notrequired, unsigned long size);
/* DSAllocation for DataBuffer is a pointer on the first buffer and therefore the pointer
returned by SimpleAllocate can be directly used to store data, - but it can be wrong for
other allocators! And normaly dsDataBufferGet call should be used */
DSAllocation dsDataBufferSimpleAllocate(DSDataBufferContext ctx, unsigned long size);

int dsDataBufferSend(DSAllocation allocation, DSSendFunction func, void *attr);
int dsDataBufferCall(DSDataBufferContext ctx, DSAllocation allocation, DSSendFunction func, void *attr, unsigned char align);
unsigned long dsDataBufferSizeof(DSDataBufferContext ctx, DSAllocation allocation, unsigned char align);
void* dsDataBufferGet(DSDataBufferContext ctx, DSAllocation allocation, unsigned long *size);
void* dsDataBufferGetNext(DSDataBufferContext ctx, DSAllocation allocation, void *current, unsigned long *size);

void dsDataBufferMantain(DSDataBufferContext ctx);


#ifdef _DS_DATA_BUFFER_C
static const struct DSAllocatorT ds_allocator_databuffer = {
    "DataBuffer",
    (DSAllocatorCreateFunction)dsDataBufferCreateContext,
    (DSAllocatorInitFunction)dsDataBufferInitContext,
    (DSAllocatorFreeFunction)dsDataBufferFreeContext,
    (DSAllocatorDestroyFunction)dsDataBufferDestroyContext,
    (DSAllocatorGetDefaultOptions)dsDataBufferGetDefaultOptions,
    (DSAllocatorInitDefaultOptions)dsDataBufferInitDefaultOptions,
    (DSAllocatorDestroyOptions)dsDataBufferDestroyOptions,
    (DSAllocatorMantainFunction)dsDataBufferMantain,
    NULL,
    (DSAllocatorAllocateMemFunction)dsDataBufferAllocate,
    (DSAllocatorFreeMemFunction)dsDataBufferFree,
    (DSAllocatorAcceptChunkFunction)dsDataBufferAcceptChunk,
    (DSAllocatorCancelChunkFunction)dsDataBufferCancelChunk,
    (DSAllocatorAcceptFunction)dsDataBufferAccept,
    (DSAllocatorCancelFunction)dsDataBufferCancel,
    (DSAllocatorAllocateSimpleChunkFunction)dsDataBufferSimpleChunk,
    (DSAllocatorAllocateSimpleFunction)dsDataBufferSimpleAllocate,
    (DSAllocatorReserveFunction)dsDataBufferReserve,
    (DSAllocatorTrimFunction)dsDataBufferTrim,
    (DSAllocatorGetFunction)dsDataBufferGet,
    (DSAllocatorGetNextFunction)dsDataBufferGetNext,
    (DSAllocatorCallFunction)dsDataBufferCall,
    (DSAllocatorSizeofFunction)dsDataBufferSizeof
};

static const struct DSDataBufferOptionsT ds_databuffer_options = {
    NULL,
    0,
    0,
    0,
    0
};
#endif /* _DS_DATA_BUFFER_C */

#endif /* _DS_DATA_BUFFER_H */
