#include <dslib.h>
#include <dsassert.h>
#include <dsdebug.h>

#include <dsdatahandler.h>
#include <dsrepresentation.h>

DSRepresentationOptions dsRepresentationGetDefaultOptions() {
    int err;
    DSRepresentationOptions options;

    options = (DSRepresentationOptions)malloc(sizeof(struct DSRepresentationOptionsT));
    if (!options) return options;
    
    err = dsRepresentationInitDefaultOptions(options);
    if (err) {
	dsRepresentationDestroyOptions(options);
	return NULL;
    }
    return options;
}

int dsRepresentationInitDefaultOptions(DSRepresentationOptions options) {
    ds_assert_param(options);
    memset(options, 0, sizeof(struct DSRepresentationOptionsT));
    return 0;
}

void dsRepresentationDestroyOptions(DSRepresentationOptions options) {
    if (options) free(options);
}

int dsRepresentationOptionsSetDataHandler(DSRepresentationOptions options, DSDataHandler handler) {
    ds_assert_param(options);
    
    options->handler = handler;
    
    return 0;
}


DSRepresentationInstance dsRepresentationCreateInstance(DSRepresentation ctx, DSMallocContext mctx, DSRepresentationOptions options) {
    int err;
    
    DSRepresentationInstance instance;
    
    instance = dsNewClass(mctx, DSRepresentationInstance);
    if (instance) {
	instance->malloc = mctx;
	instance->representation = ctx;
	instance->base = NULL;
	
	if (ctx->CreateInstance) {
	    instance->data = ctx->CreateInstance(mctx, options);
	    if (!instance->data) {
		dsFree(mctx, instance);
		instance = NULL;		
	    }
	} else {
	    instance->data = NULL;
	}
	
	if ((options)&&(options->handler)) {
	    instance->handler = options->handler;
	    dsRepresentationSetupHandler(instance, instance->handler);
	} else {
	    instance->handler = dsDataHandlerCreate(mctx);
	    if (instance->handler) {
		dsRepresentationSetupHandler(instance, instance->handler);
		instance->destroyhandler = 1;
	    } else {
		dsFree(mctx, instance);
		instance = NULL;
	    }
	}
    }
    
    return instance;
}

void dsRepresentationDestroyInstance(DSRepresentationInstance instance) {
    ds_assert_return(instance);

    if ((instance->handler)&&(instance->destroyhandler)) {
	dsDataHandlerDestroy(instance->malloc, instance->handler);
	instance->handler = NULL;
    }
    if (instance->data) {
	if (instance->representation->DestroyInstance) instance->representation->DestroyInstance(instance->malloc, instance->data);
	else dsFree(instance->malloc, instance->data);
    }
}

DSDataConst dsRepresentationGet(DSRepresentationInstance ctx, DSDataChunkContext chunk) {
    int err;
    DSData data;
    
    ds_assert_null(ctx);
    ds_assert_null(chunk);
    
    
    data = (DSData)dsDataChunkGet(chunk, ctx->id);
    if ((data)&&(!data->handler)) data->handler = ctx->handler;

    return (DSDataConst)data;
}

int dsRepresentationPushPointer(DSRepresentationInstance ctx, DSDataChunkContext chunk, void *buffer, size_t size) {
    int err;
    DSDataS data;
    
    err = dsDataSetPointer(&data, NULL, buffer, size);
    if (err) return err;
    
    return dsRepresentationPushData(ctx, chunk, &data);
}
