#ifndef _DS_REPRESENTATION_H
#define _DS_REPRESENTATION_H

#include <dsallocator.h>
#include <dsdatachunk.h>
#include <dsdatahandler.h>

typedef struct DSRepresentationT *DSRepresentation;
typedef struct DSRepresentationInstanceT *DSRepresentationInstance;
typedef void *DSRepresentationInstanceData;
typedef struct DSRepresentationOptionsT *DSRepresentationOptions;

struct DSRepresentationOptionsT {
    DSDataHandler handler;
};
typedef struct DSRepresentationOptionsT DSRepresentationOptionsS;

DSRepresentationOptions dsRepresentationGetDefaultOptions();
int dsRepresentationInitDefaultOptions(DSRepresentationOptions options);
void dsRepresentationDestroyOptions(DSRepresentationOptions options);
int dsRepresentationOptionsSetDataHandler(DSRepresentationOptions options, DSDataHandler handler);


typedef DSRepresentationInstanceData (*DSRepresentationCreateFunction)(DSMallocContext mctx, DSRepresentationOptions options);
typedef void (*DSRepresentationDestroyFunction)(DSMallocContext mctx, DSRepresentationInstance ctx);
typedef int (*DSRepresentationSetupHandlerFunction)(DSRepresentationInstance ctx, DSDataHandler handler);
typedef int (*DSRepresentationMantainFunction)(DSRepresentationInstance ctx);

typedef int (*DSRepresentationPushDataFunction)(DSRepresentationInstance ctx, DSDataChunkContext chunk, DSDataConst data);

typedef int (*DSRepresentationReadBaseFunction)(DSRepresentationInstance ctx, DSDataChunkContext chunk);
typedef int (*DSRepresentationWriteBaseFunction)(DSRepresentationInstance ctx, DSDataChunkContext chunk);

struct DSRepresentationT {
    DSRepresentationCreateFunction CreateInstance;
    DSRepresentationDestroyFunction DestroyInstance;
    DSRepresentationSetupHandlerFunction SetupHandler;
    
    DSRepresentationMantainFunction Mantain;

    DSRepresentationPushDataFunction PushData;

    DSRepresentationReadBaseFunction ReadBase;
    DSRepresentationWriteBaseFunction WriteBase;
};
typedef struct DSRepresentationT DSRepresentationS;

struct DSRepresentationInstanceT {
    char *title;

    DSDataChunkForm id;    
    DSRepresentation representation;
    DSMallocContext malloc;
    
    unsigned char destroyhandler;
    DSDataHandler handler;

    DSRepresentationInstanceData data;
    
    DSRepresentationInstance base;
};
typedef struct DSRepresentationInstanceT DSRepresentationInstanceS;

DSRepresentationInstance dsRepresentationCreateInstance(DSRepresentation ctx, DSMallocContext mctx, DSRepresentationOptions options);
void dsRepresentationDestroyInstance(DSRepresentationInstance instance);


#define dsRepresentationSetupHandler(ctx, handler) \
    ((ctx&&ctx->representation->SetupHandler)?ctx->representation->SetupHandler(ctx, handler):0)

#define dsRepresentationMantain(ctx) \
    ((ctx&&ctx->representation->Mantain)?ctx->representation->Mantain(ctx):DS_ERR_NOTIMPLEMENTED)
#define dsRepresentationPushData(ctx,chunk,data) \
    ((ctx&&ctx->representation->PushData)?ctx->representation->PushData(ctx,chunk,data):DS_ERR_NOTIMPLEMENTED)
#define dsRepresentationReadBase(ctx,chunk) \
    ((ctx&&ctx->representation->ReadBase)?ctx->representation->ReadBase(ctx,chunk):DS_ERR_NOTIMPLEMENTED)
#define dsRepresentationWriteBase(ctx,chunk) \
    ((ctx&&ctx->representation->WriteBase)?ctx->representation->WriteBase(ctx,chunk):DS_ERR_NOTIMPLEMENTED)

DSDataConst dsRepresentationGet(DSRepresentationInstance ctx, DSDataChunkContext chunk);
int dsRepresentationPushPointer(DSRepresentationInstance ctx, DSDataChunkContext chunk, void *buffer, size_t size);

#endif /* _DS_REPRESENTATION_H */
