#include <string.h>

#include <dslib.h>
#include <dsassert.h>
#include <dsdebug.h>

#include <dsdata.h>

void dsDataZero(DSData data) {
    memset(data, 0, sizeof(DSDataS));
}

int dsDataIsZero(DSData data) {
    return (data&&data->data)?0:1;
}

DSData dsDataCreate(DSMallocContext mctx) {
    DSData data;
    
    data = (DSData)dsMalloc(mctx, sizeof(DSDataS));
    if (data) dsDataZero(data);
    return data;
}

void dsDataDestroy(DSMallocContext mctx, DSData data) {
    ds_assert_return(data);
    
    dsFree(mctx, data);
}

void dsDataDestroyWithContent(DSMallocContext mctx, DSData data) {
    dsDataDestroyContent(data);
    dsDataDestroy(mctx, data);
}


int dsDataSet(DSData data, DSDataType type, void *owner, void *ptr, size_t size) {
    ds_assert_param(data);
    
    memset(data, 0, sizeof(DSDataS));

    data->type = type;
    data->owner = owner;
    data->data = ptr;
    data->size = size;
    
    return 0;
}

int dsDataSetPointer(DSData data, DSMallocContext mctx, void *ptr, size_t size) {
    return dsDataSet(data, DS_DATA_TYPE_POINTER, mctx, ptr, size);
}

int dsDataSetAllocation(DSData data, DSAllocatorContext allocator, DSAllocation allocation, size_t size) {
    if (!size) {
	size = dsDataBufferSizeof(allocator, allocation, 1);
	if (size == (size_t)-1) size = 0;
    }
    return dsDataSet(data, DS_DATA_TYPE_ALLOCATION, allocator, allocation, size);
}

int dsDataSetReference(DSData data, DSData orig) {
    return dsDataSet(data, DS_DATA_TYPE_REFERENCE, NULL, orig, 0);
}

int dsDataSetIncomplete(DSData data, void *ptr, size_t size) {
    return dsDataSet(data, DS_DATA_TYPE_INCOMPLETE, NULL, ptr, size);    
}


int dsDataSetOwner(DSData data, const void *owner) {
    ds_assert_param(data);
    data->owner = owner;
    return 0;
}

int dsDataSetHandler(DSData data, DSDataHandler handler) {
    ds_assert_param(data);

    data->handler = handler;    
    return 0;
}

int dsDataCopy(DSData dst, DSDataConst src) {
    ds_assert_param(dst);
    ds_assert_param(src);
    
    memcpy(dst, src, sizeof(DSDataS));
    return 0;
}

DSDataType dsDataGetType(DSDataConst data) {
    ds_assert_default(data, DS_DATA_TYPE_INCOMPLETE);

    if (data->type == DS_DATA_TYPE_REFERENCE) return dsDataGetType((DSDataConst)data->data);
    return data->type;
}

const void *dsDataGet(DSDataConst data) {
    ds_assert_default(data, NULL);
    
    if (data->type == DS_DATA_TYPE_REFERENCE) return dsDataGet((DSDataConst)data->data);
    return data->data;
}

int dsDataSetUserType(DSData data, DSDataUserType type) {
    ds_assert_param(data);

    data->user_type = type;
    return 0;
}

int dsDataCheckUserType(DSData data, DSDataUserType type) {
    ds_assert_default(data, 0);

    if (type == data->user_type) return 1;
    return 0;
}
