#ifndef _DS_DATA_H
#define _DS_DATA_H

#include <dsmalloc.h>
#include <dsallocator.h>

typedef struct DSDataT *DSData;
typedef const struct DSDataT *DSDataConst;
typedef int DSDataUserType; /* For verification purposes */

#include <dsdatahandler.h>

#define DS_DATA_SIZE_UNKNOWN	0

enum DSDataTypeT {
    DS_DATA_TYPE_INCOMPLETE = 0, /* Require configuration from parent class */
    DS_DATA_TYPE_POINTER,
    DS_DATA_TYPE_DATABUFFER,
    DS_DATA_TYPE_ALLOCATION,
    DS_DATA_TYPE_XMLNODEPTR,
    DS_DATA_TYPE_REFERENCE,
    DS_DATA_TYPE_CUSTOM		/* Processing is performed using handler */
};
typedef enum DSDataTypeT DSDataType;


/* 
   Main data structure. It is allowed to extend this structure with additional
   properties which could be set with dsDataSet functions. The destructor
   and send functions could serve an example. 

   The context of data should not be sequential and fully depends on its type
   and owner. All functions just taking copying the pointer, not the data.
   For copying the data dsData[Copy|Set|Change|Destroy]Content function should 
   be created/used.
   
   Malloc context is not used since the data is controlled by some object,
   for example dsMallocContext
*/
 
struct DSDataT {
    DSDataType type;
    const void *owner;

    const void *data;
    size_t size;
    
    DSDataHandler handler;
    DSDataUserType user_type;
};
typedef struct DSDataT DSDataS;

DSData dsDataCreate(DSMallocContext mctx);
void dsDataDestroy(DSMallocContext mctx, DSData data);
void dsDataDestroyWithContent(DSMallocContext mctx, DSData data);
void dsDataZero(DSData data);
int dsDataIsZero(DSData data);
int dsDataCopy(DSData dst, DSDataConst src);

int dsDataSet(DSData data, DSDataType type, void *owner, void *ptr, size_t size);
int dsDataSetIncomplete(DSData data, void *ptr, size_t size);
int dsDataSetPointer(DSData data, DSMallocContext mctx, void *ptr, size_t size);
int dsDataSetAllocation(DSData data, DSAllocatorContext allocator, DSAllocation allocation, size_t size);
int dsDataSetReference(DSData data, DSData orig);

int dsDataSetOwner(DSData data, const void *owner);
int dsDataSetHandler(DSData data, DSDataHandler handler);
int dsDataSetUserType(DSData data, DSDataUserType type);
int dsDataCheckUserType(DSData data, DSDataUserType type);

DSDataType dsDataGetType(DSDataConst data);
const void *dsDataGet(DSDataConst data);

#endif /* _DS_DATA_H */
