#ifndef _DS_DATA_HANDLER_H
#define _DS_DATA_HANDLER_H

typedef struct DSDataHandlerT *DSDataHandler;

#include <dsdata.h>

typedef struct DSDataCopierT *DSDataCopier;

typedef int (*DSDataCopyFunction)(DSDataCopier copier, DSData dst, DSDataConst src);
typedef int (*DSDataPipeFunction)(DSDataConst data, DSPipeFunction func, void *ctx);
typedef int (*DSDataDestructorFunction)(DSData data);


enum DSDataCopierTypeT {
    DS_DATA_COPIER_TYPE_ANY,		/* Decided by copy routine, for example autocreate allocator and destruct it with DSDataDestroy handler */
    DS_DATA_COPIER_TYPE_ALLOCATOR,
    DS_DATA_COPIER_TYPE_MALLOC
};
typedef enum DSDataCopierTypeT DSDataCopierType;

struct DSDataCopierT {
    DSDataCopierType type;
    void *ctx;
};
typedef struct DSDataCopierT DSDataCopierS;

int dsDataSetupDefaultCopier(DSDataCopier copier);
int dsDataSetupAllocatorCopier(DSDataCopier copier, DSAllocatorContext allocator);
int dsDataSetupMallocCopier(DSDataCopier copier, DSMallocContext malloc);


#define DS_DATA_HANDLER(ctx) ((DSDataHandler)ctx)

typedef int DSDataHandlerUserType;
enum DSDataHandlerTypeT {
    DS_DATA_HANDLER_TYPE_STANDARD = 0,
    DS_DATA_HANDLER_TYPE_DATAMANAGER
};
typedef enum DSDataHandlerTypeT DSDataHandlerType;

struct DSDataHandlerT {
    void *ctx; /* Have precedence other owner */
    DSDataHandlerType type;
    DSDataHandlerUserType user_type;
    
    DSDataCopyFunction copy;
    DSDataPipeFunction pipe;
    DSDataDestructorFunction destructor;    
};
typedef struct DSDataHandlerT DSDataHandlerS;

DSDataHandler dsDataHandlerCreate(DSMallocContext mctx);
void dsDataHandlerDestroy(DSMallocContext mctx, DSDataHandler handler);
int dsDataHandlerConfigureDefaultMethods(DSDataHandler handler);
int dsDataHandlerCopy(DSDataHandler dst, DSDataHandler src);

int dsDataHandlerSetType(DSDataHandler handler, DSDataHandlerType type);
int dsDataHandlerSetUserType(DSDataHandler handler, DSDataHandlerUserType type);

int dsDataHandlerSetContext(DSDataHandler handler, void *ctx);
int dsDataHandlerSetCopyFunction(DSDataHandler handler, DSDataCopyFunction func);
int dsDataHandlerSetPipeFunction(DSDataHandler handler, DSDataPipeFunction func);
int dsDataHandlerSetDestructorFunction(DSDataHandler handler, DSDataDestructorFunction func);

/* The following functions are executed only if the type is supported or
the appropriate handler is provided */
int dsDataDestroyContent(DSData data);
int dsDataCopyContent(DSDataCopier copier, DSData dst, DSDataConst src);
int dsDataPipeContent(DSDataConst data, DSPipeFunction func, void *attr);

int dsDataDestroyContentDummy(DSData data);

extern DSDataHandler DS_DATA_HANDLER_STATIC_DATA;

#endif /* _DS_DATA_HANDLER_H */
