#include <dslib.h>
#include <dsassert.h>
#include <dsdebug.h>

#include <dsbinaryformat.h>

DSBinaryFormatInstance dsBinaryFormatCreateInstance(DSBinaryFormat ctx, DSMallocContext mctx, DSBinaryFormatInstanceOptions options) {
    int err;
    
    DSBinaryFormatInstance instance;
    instance = dsNewClass(mctx, DSBinaryFormatInstance);
    if (instance) {
	instance->malloc = mctx;
	instance->format = ctx;
	instance->options = options;

	if (ctx->CreateInstance) {
	    instance->data = ctx->CreateInstance(mctx, options);
	    if (!instance->data) {
		dsFree(mctx, instance);
		instance = NULL;
	    }
	} else {
	    instance->data = NULL;
	}
    }
    
    return instance;
}

void dsBinaryFormatDestroyInstance(DSBinaryFormatInstance instance) {
    ds_assert_return(instance);

    if (instance->data) {
	if (instance->format->DestroyInstance) instance->format->DestroyInstance(instance->malloc, instance->data);
	else dsFree(instance->malloc, instance->data);
    }
}

int dsBinaryFormatReadType(DSBinaryFormatInstance instance, DSBinaryType type, const char *source, size_t *readed) {
    int err;

    DSBinaryFormat format;
    DSTreeNodeContext node;
    char *source_ptr;

    unsigned long bytes;
    unsigned long pos = 0;
    unsigned char align;
        
    format = instance->format;
    source_ptr = (char*)source;

    switch (type->type) {
	case DS_BINARY_TYPE_CLASS_COMPLEX:
	break;
	case DS_BINARY_TYPE_CLASS_INTEGER:
	    if (format->ReadInteger)
	        return format->ReadInteger(instance, type, (void*)source, readed);
	    else 
	        return DS_ERR_NOTIMPLEMENTED;
	case DS_BINARY_TYPE_CLASS_FLOAT:
	    if (format->ReadFloat)
	        return format->ReadFloat(instance, type, (void*)source, readed);
	    else
	        return DS_ERR_NOTIMPLEMENTED;
	default:
	    return DS_ERR_NOTIMPLEMENTED;
    }
    
    for (node=((DSTreeComplexNodeContext)type)->first_child;node;node=node->next) {
	align = ((DSBinaryType)node)->align;
	if ((align>1)&&(pos&(align-1))) pos=(pos&~(align-1))+align;

	switch (((DSBinaryType)node)->type) {
	    case DS_BINARY_TYPE_CLASS_COMPLEX:
		// source_ptr viravnivaem
//1.		format->WriteStructureStart(format, ctx, (void*)source_ptr, conversion);
//		dsBinaryTypeWrite((DSBinaryType)node, ctx, source_ptr, conversion)
//		format->WriteStructureEnd(format, ctx, (void*)..., conversion);
//2.		Complex vizvat a iz nego vernutsya (ili net)
		return DS_ERR_NOTIMPLEMENTED;
	    break;
	    case DS_BINARY_TYPE_CLASS_INTEGER:
		if (format->ReadInteger)
		    err = format->ReadInteger(instance, (DSBinaryType)node, (void*)(source_ptr+pos), &bytes);
		else 
		    return DS_ERR_NOTIMPLEMENTED;

		if (err) return err;
		pos+=bytes;
	    break;
	    case DS_BINARY_TYPE_CLASS_FLOAT:
		if (format->ReadFloat)
	    	    err=format->ReadFloat(instance, (DSBinaryType)node, (void*)(source_ptr+pos), &bytes);
		else
		    return DS_ERR_NOTIMPLEMENTED;
		
		if (err) return err;
		pos+=bytes;
	    break;
	    default:
		return DS_ERR_NOTIMPLEMENTED;
	    break;
	}
    }
    
    *readed = pos;

    return 0;
}

int dsBinaryFormatWriteType(DSBinaryFormatInstance instance, DSBinaryType type, const char *source, size_t *writed) {
    return DS_ERR_NOTIMPLEMENTED;
}
