#ifndef _DS_BINARY_FORMAT_H
#define _DS_BINARY_FORMAT_H

#include <dslib.h>
#include <dsdata.h>
#include <dsallocator.h>
#include <dsbinarytype.h>

typedef struct DSBinaryFormatT *DSBinaryFormat;
typedef struct DSBinaryFormatInstanceT *DSBinaryFormatInstance;
typedef void *DSBinaryFormatInstanceData;
typedef const void *DSBinaryFormatInstanceOptions;

enum DSBinaryFormatTypeT {
    DS_BINARY_FORMAT_NATIVE = 0,
    DS_BINARY_FORMAT_BINARY,
    DS_BINARY_FORMAT_OPCXML,
    DS_BINARY_FORMAT_XML,
    DS_BINARY_FORMAT_UNKNOWN
};
typedef enum DSBinaryFormatTypeT DSBinaryFormatType;

enum DSBinaryFormatAllocatorTypeT {
    DS_BINARY_FORMAT_ALLOCATOR_TYPE_ABSENT = 0,
    DS_BINARY_FORMAT_ALLOCATOR_TYPE_MINIMAL,
    DS_BINARY_FORMAT_ALLOCATOR_TYPE_FULL
};
typedef enum DSBinaryFormatAllocatorTypeT DSBinaryFormatAllocatorType;

typedef DSBinaryFormatInstanceData (*DSBinaryFormatCreateFunction)(DSMallocContext mctx, DSBinaryFormatInstanceOptions options);
typedef void (*DSBinaryFormatDestroyFunction)(DSMallocContext mctx, DSBinaryFormatInstanceData ctx);

typedef int (*DSBinaryFormatImportDataFunction)(DSBinaryFormatInstance ctx, DSBinaryType type, DSAllocatorContext allocator, DSData data, DSDataConst source);
typedef int (*DSBinaryFormatDestructDataFunction)(DSBinaryFormatInstance ctx, DSBinaryType type, DSData data);
typedef int (*DSBinaryFormatCopyFunction)(DSBinaryFormatInstance ctx, DSBinaryType type, DSAllocatorContext allocator, DSData dst, DSDataConst src);
typedef int (*DSBinaryFormatPipeFunction)(DSBinaryFormatInstance ctx, DSBinaryType type, DSDataConst data, DSPipeFunction func, void *attr);

typedef int (*DSBinaryFormatReadFromBaseFunction)(DSBinaryFormatInstance ctx, DSBinaryType type, DSAllocatorContext allocator, DSData data, DSDataConst native);
typedef int (*DSBinaryFormatWriteToBaseFunction)(DSBinaryFormatInstance ctx, DSBinaryType type, DSAllocatorContext allocator, DSDataConst data, DSData native);

typedef int (*DSBinaryFormatWriteTypeFunction)(DSBinaryFormatInstance ctx, DSBinaryType type, const char *source, size_t *bytes);
typedef int (*DSBinaryFormatReadTypeFunction)(DSBinaryFormatInstance ctx, DSBinaryType type, const char *source, size_t *bytes);

struct DSBinaryFormatT {
    char *title;
    
    DSBinaryFormatType type;
    DSBinaryFormatAllocatorType allocator;

    DSBinaryFormatCreateFunction CreateInstance;
    DSBinaryFormatDestroyFunction DestroyInstance;

    DSBinaryFormatImportDataFunction ImportData;
    DSBinaryFormatDestructDataFunction DestructData;
    DSBinaryFormatCopyFunction CopyData;
    DSBinaryFormatPipeFunction PipeData;

    DSBinaryFormatReadFromBaseFunction ReadFromBase;
    DSBinaryFormatWriteToBaseFunction WriteToBase;
    
    DSBinaryFormatWriteTypeFunction ReadInteger;
    DSBinaryFormatWriteTypeFunction ReadFloat;
};
typedef struct DSBinaryFormatT DSBinaryFormatS;

/*
    For Convert operation the appropriate allocator should be used. Of OPCXML
    for ReadFromBase calls and of Native for WriteTobase calls.
*/

struct DSBinaryFormatInstanceT {
    DSMallocContext malloc;

    DSBinaryFormat format;
    DSBinaryFormatInstanceOptions options;
    
    DSBinaryFormatInstanceData data;
};
typedef struct DSBinaryFormatInstanceT DSBinaryFormatInstanceS;


DSBinaryFormatInstance dsBinaryFormatCreateInstance(DSBinaryFormat ctx, DSMallocContext mctx, DSBinaryFormatInstanceOptions options);
void dsBinaryFormatDestroyInstance(DSBinaryFormatInstance instance);


#define dsBinaryFormatImport(ctx,type,allocator,data,source) \
    (((ctx)&&(ctx->format)&&(ctx->format->ImportData))?ctx->format->ImportData(ctx,type,allocator,data,source):DS_ERR_NOTSUPPORTED)

#define dsBinaryFormatDestruct(ctx,type,data) \
    (((ctx)&&(ctx->format)&&(ctx->format->DestructData))?ctx->format->DestructData(ctx, type, data):DS_ERR_NOTSUPPORTED)

#define dsBinaryFormatCopy(ctx,type,allocator,dst,src) \
    (((ctx)&&(ctx->format)&&(ctx->format->CopyData))?ctx->format->CopyData(ctx,type,allocator,dst,src):DS_ERR_NOTSUPPORTED)

#define dsBinaryFormatPipe(ctx,type,data,func,attr) \
    (((ctx)&&(ctx->format)&&(func)&&(ctx->format->PipeData))?ctx->format->PipeData(ctx,type,data,func,attr):DS_ERR_NOTSUPPORTED)

#define dsBinaryFormatReadFromBase(ctx,type,allocator,data,native) \
    (((ctx)&&(ctx->format)&&(ctx->format->ReadFromBase))?ctx->format->ReadFromBase(ctx, type, allocator, data, native):DS_ERR_NOTSUPPORTED)

#define dsBinaryFormatWriteToBase(ctx,type,allocator,data,native) \
    (((ctx)&&(ctx->format)&&(ctx->format->WriteToBase))?ctx->format->WriteToBase(ctx, type, allocator, data, native):DS_ERR_NOTSUPPORTED)

int dsBinaryFormatReadType(DSBinaryFormatInstance instance, DSBinaryType type, const char *source, size_t *readed);
int dsBinaryFormatWriteType(DSBinaryFormatInstance instance, DSBinaryType type, const char *source, size_t *writed);

#endif /* _DS_BINARY_FORMAT_H */
