#include <dslib.h>
#include <dsassert.h>
#include <dsdebug.h>

#include <dsbinaryformatnative.h>

static int dsBinaryFormatNativeImportData(DSBinaryFormatInstance instance, DSBinaryType type, DSAllocatorContext allocator, DSData data, DSDataConst source) {
    DSAllocation allocation;
    size_t size = 0;
    void *ptr;

    ds_assert_param(instance);
    ds_assert_param(type);
    ds_assert_param(allocator);
    ds_assert_param(data);
    ds_assert_param(source);
    

    switch (source->type) {
	case DS_DATA_TYPE_ALLOCATION:
	case DS_DATA_TYPE_DATABUFFER:
	    if (source->size) size = source->size;
	    else size = dsAllocatorSizeof((DSAllocatorContext)source->owner, (DSAllocation)source->data, 1);
	    
	    if (source->owner == allocator) {
		dsDataSetAllocation(data, allocator, (DSAllocation)source->data, size);
		break;
	    }
	break;
	case DS_DATA_TYPE_POINTER:
	    if ((!size)&&(type->size)) size=type->size;
	    
	    if (size) {
		allocation = dsAllocatorSimpleAllocate(allocator, size);
		if (allocation) ptr = dsAllocatorGet(allocator, allocation, NULL);
		else return DS_ERR_ALLOCATOR_ALLOCATE;
		memcpy(ptr, source->data, size);
		dsDataSetAllocation(data, allocator, allocation, dsAllocatorSizeof(allocator, allocation, 1));
	    }  else {
		// ToDo: go by items
		return DS_ERR_NOTIMPLEMENTED;
	    }
	break;
	default:
	    return DS_ERR_NOTSUPPORTED;
    } 
    
    return 0;
}

static DSBinaryFormatS ds_binary_format_native = {
	"Native",
	DS_BINARY_FORMAT_NATIVE,
	DS_BINARY_FORMAT_ALLOCATOR_TYPE_FULL,
	NULL,							/* CreateInstance */
	NULL,							/* DestroyInstance */
	dsBinaryFormatNativeImportData,
	NULL,							/* DestructData */
	NULL,							/* CopyData */
	NULL,							/* PipeData */
	NULL,							/* ReadFrom */
	NULL,							/* WriteTo */
	NULL,							/* ReadInt */
	NULL							/* ReadFloat */
};

DSBinaryFormat dsBinaryFormatNative() {
    return &ds_binary_format_native;
}
