#define _DS_INCLUDE_XML_HEADERS

#include <dslib.h>
#include <dsassert.h>
#include <dsdebug.h>

#include <dsxml.h>

#include <dsbinarytype.h>
#include <dsbinaryformat.h>


struct DSBinaryFormatOPCXMLInstanceDataT {
    xmlNodePtr node;
    xmlNsPtr xsins;
    xmlNsPtr xsdns;
};
typedef struct DSBinaryFormatOPCXMLInstanceDataT *DSBinaryFormatOPCXMLInstanceData;
typedef struct DSBinaryFormatOPCXMLInstanceDataT DSBinaryFormatOPCXMLInstanceDataS;

static DSBinaryFormatInstanceData dsBinaryFormatOPCXMLCreateInstanceData(DSMallocContext mctx, DSBinaryFormatInstanceOptions options) {
    DSBinaryFormatInstanceData data;
    return dsNewClass(mctx, DSBinaryFormatOPCXMLInstanceData);
}

static void dsBinaryFormatOPCXMLDestroyInstanceData(DSMallocContext mctx, DSBinaryFormatInstanceData data) {
    if (data) {
	if (((DSBinaryFormatOPCXMLInstanceData)data)->node) xmlFreeNode((xmlNodePtr)(((DSBinaryFormatOPCXMLInstanceData)data)->node));
	dsFree(mctx, data);
    }
}

static const char *opcxml_subtypes[] = {
    "byte", "short", "int", "long",
    "unsignedByte", "unsignedShort", "unsignedInt", "unsignedLong",
    "float", "double", "decimal", "boolean",
    "string", "dateTime"
};

const char *dsBinaryFormatOPCXMLSetType(xmlNsPtr xsins, DSBinaryType type, xmlNodePtr node) {
    unsigned char flag = 0;
    
    if ((type->element_count>1)||(type->element_count_ref)||(type->terminator)) flag = 1;

    switch(type->type) {
	case DS_BINARY_TYPE_CLASS_COMPLEX:
	    if (node) xmlSetNsProp(node, xsins, "type", "ArrayOfAnyType");
	    return NULL;
	case DS_BINARY_TYPE_CLASS_INTEGER:
	    switch (((DSBinaryTypeInteger)type)->bytes) {
		case 1:
		    if (((DSBinaryTypeInteger)type)->is_signed) {
			if (flag) {
			    if (node) xmlSetNsProp(node, xsins, "type", "ArrayOfByte");
			    return opcxml_subtypes[0];
		        } else {
			    if (node) xmlSetNsProp(node, xsins, "type", "xsd:byte");
			    return NULL;
			}
		    } else {
			if (flag) {
			    if (node) xmlSetNsProp(node, xsins, "type", "ArrayOfUnsignedByte");
			    return opcxml_subtypes[4];
		        } else {
			    if (node) xmlSetNsProp(node, xsins, "type", "xsd:unsignedByte");
			    return NULL;
			}
		    }
		case 2:
		    if (((DSBinaryTypeInteger)type)->is_signed) {
			if (flag) {
			    if (node) xmlSetNsProp(node, xsins, "type", "ArrayOfShort");
			    return opcxml_subtypes[1];
		        } else {
			    if (node) xmlSetNsProp(node, xsins, "type", "xsd:short");
			    return NULL;
			}
		    } else {
			if (flag) {
			    if (node) xmlSetNsProp(node, xsins, "type", "ArrayOfUnsignedShort");
			    return opcxml_subtypes[5];
		        } else {
			    if (node) xmlSetNsProp(node, xsins, "type", "xsd:unsignedShort");
			    return NULL;
			}
		    }
		case 4:
		    if (((DSBinaryTypeInteger)type)->is_signed) {
			if (flag) {
			    if (node) xmlSetNsProp(node, xsins, "type", "ArrayOfInt");
			    return opcxml_subtypes[2];
		        } else {
			    if (node) xmlSetNsProp(node, xsins, "type", "xsd:int");
			    return NULL;
			}
		    } else {
			if (flag) {
			    if (node) xmlSetNsProp(node, xsins, "type", "ArrayOfUnsignedInt");
			    return opcxml_subtypes[6];
		        } else {
			    if (node) xmlSetNsProp(node, xsins, "type", "xsd:unsignedInt");
			    return NULL;
			}
		    }
		case 8:
		    if (((DSBinaryTypeInteger)type)->is_signed) {
			if (flag) {
			    if (node) xmlSetNsProp(node, xsins, "type", "ArrayOfLong");
			    return opcxml_subtypes[3];
		        } else {
			    if (node) xmlSetNsProp(node, xsins, "type", "xsd:long");
			    return NULL;
			}
		    } else {
			if (flag) {
			    if (node) xmlSetNsProp(node, xsins, "type", "ArrayOfUnsignedLong");
			    return opcxml_subtypes[7];
		        } else {
			    if (node) xmlSetNsProp(node, xsins, "type", "xsd:unsignedLong");
			    return NULL;
			}
		    }
		default:
		    if (flag) {
			if (node) xmlSetNsProp(node, xsins, "type", "ArrayOfDecimal");
			return opcxml_subtypes[10];
		    } else {
			if (node) xmlSetNsProp(node, xsins, "type", "xsd:decimal");
			return NULL;
		    }
	    }
	case DS_BINARY_TYPE_CLASS_FLOAT:
	    switch (((DSBinaryTypeInteger)type)->bytes) {
		case 4:
		    if (flag) {
			if (node) xmlSetNsProp(node, xsins, "type", "ArrayOfFloat");
			return opcxml_subtypes[8];
		    } else {
			if (node) xmlSetNsProp(node, xsins, "type", "xsd:float");
			return NULL;
		    }
		case 8:
		    if (flag) {
			if (node) xmlSetNsProp(node, xsins, "type", "ArrayOfDouble");
			return opcxml_subtypes[9];
		    } else {
			if (node) xmlSetNsProp(node, xsins, "type", "xsd:double");
			return NULL;
		    }
		default:
		    if (flag) {
			if (node) xmlSetNsProp(node, xsins, "type", "ArrayOfDecimal");
			return opcxml_subtypes[10];
		    } else {
			if (node) xmlSetNsProp(node, xsins, "type", "xsd:decimal");
			return NULL;
		    }
	    }
    }

    return NULL;
}

static inline_function int dsBinaryFormatOPCXMLinteger(DSBinaryTypeInteger type, const void *source, char *stmp) {
    if (type->is_signed) {
        switch(type->bytes) {
	    case sizeof(char):
	        sprintf(stmp, "%hhi", *(char*)source);
	    break;
	    case sizeof(short):
		sprintf(stmp, "%hi", *(short*)source);
	    break;
	    case sizeof(int):
	        sprintf(stmp, "%i", *(int*)source);
	    break;
#if __WORDSIZE == 64 
	    case sizeof(long):
	        sprintf(stmp, "%li", *(long*)source);
	    break;
#else
	    case sizeof(long long):
	        sprintf(stmp, "%lli", *(long long*)source);
	    break;
#endif 
	    default:
	        return DS_ERR_DATA_INVALID_TYPE_SIZE;
	}
    } else {
	switch(type->bytes) {
	    case sizeof(char):
	        sprintf(stmp, "%hhu", *(unsigned char*)source);
	    break;
	    case sizeof(short):
	        sprintf(stmp, "%hu", *(unsigned short*)source);
	    break;
	    case sizeof(int):
	        sprintf(stmp, "%u", *(unsigned long*)source);
	    break;
#if __WORDSIZE == 64 
	    case sizeof(long):
	        sprintf(stmp, "%lu", *(unsigned long*)source);
	    break;
#else
	    case sizeof(long long):
	        sprintf(stmp, "%llu", *(unsigned long long*)source);
	    break;
#endif
	    default:
	        return DS_ERR_DATA_INVALID_TYPE_SIZE;
	}
    }
    return 0;
}

static inline_function int dsBinaryFormatOPCXMLfloat(DSBinaryTypeFloat type, const void *source, char *stmp) {
    switch(type->bytes) {
	case sizeof(float):
	    sprintf(stmp, "%e", *(float*)source);
	break;
	case sizeof(double):
	    sprintf(stmp, "%le", *(double*)source);
	break;
#ifndef _WIN32
	case sizeof(long double):
	    sprintf(stmp, "%Le", *(long double*)source);
	break;
#endif /* _WIN32 */
	default:
	    return DS_ERR_DATA_INVALID_TYPE_SIZE;
    }
    return 0;
}


static int dsBinaryFormatOPCXMLImportData(DSBinaryFormatInstance instance, DSBinaryType type, DSAllocatorContext allocator, DSData data, DSDataConst source) {
    ds_assert_param(instance);
    ds_assert_param(type);
    ds_assert_param(data);

	// ToDo: check if xml in text form and parse
    return DS_ERR_NOTIMPLEMENTED;
}

static int dsBinaryFormatOPCXMLDestructData(DSBinaryFormatInstance instance, DSBinaryType type, DSData data) {
    ds_assert_param(instance);
    ds_assert_param(type);
    ds_assert_param(data);
    
    if (data->data) {
	ds_assert_param(data->type == DS_DATA_TYPE_XMLNODEPTR);
	xmlFreeNode((xmlNodePtr)data->data);
    }
    return 0;
}



static int dsBinaryFormatOPCXMLReadFromBase(DSBinaryFormatInstance instance, DSBinaryType type, DSAllocatorContext allocator, DSData data, DSDataConst native) {
    int err;

    xmlNodePtr node;
    xmlNsPtr xsins, xsdns;

    size_t bytes = 0;
    const void *source;

    ds_assert_param(instance);
    ds_assert_param(type);
    ds_assert_param(data);
    ds_assert_param(native);

    switch (native->type) {
	case DS_DATA_TYPE_POINTER:
	case DS_DATA_TYPE_DATABUFFER:
	    source = native->data;
	break;
	case DS_DATA_TYPE_ALLOCATION:
	    source = dsAllocatorGet(((DSAllocatorContext)native->owner), ((DSAllocation)native->data), NULL);
	break;
	default:
	    return DS_ERR_NOTIMPLEMENTED;	    
    }

    node = xmlNewNode(NULL,"Value");
    if (node) {
	xsins = xmlNewNs(node, DS_XML_XSI_NAMESPACE, "xsi");
	xsdns = xmlNewNs(node, DS_XML_XSD_NAMESPACE, "xsd");
    } else return DS_ERR_XML;
	
    ((DSBinaryFormatOPCXMLInstanceData)instance->data)->node = node;
    ((DSBinaryFormatOPCXMLInstanceData)instance->data)->xsins = xsins;
    ((DSBinaryFormatOPCXMLInstanceData)instance->data)->xsdns = xsdns;

    dsBinaryFormatOPCXMLSetType(xsins, type, node);
    
    err = dsBinaryFormatReadType(instance, type, source, &bytes);
    if (err) return err;
    
    err = dsDataSet(data, DS_DATA_TYPE_XMLNODEPTR, NULL, node, 0);
    if (err) return err;

    ((DSBinaryFormatOPCXMLInstanceData)instance->data)->node = NULL;
    ((DSBinaryFormatOPCXMLInstanceData)instance->data)->xsins = NULL;
    ((DSBinaryFormatOPCXMLInstanceData)instance->data)->xsdns = NULL;
    
    return 0;
}

static int dsBinaryFormatOPCXMLWriteToBase(DSBinaryFormatInstance instance, DSBinaryType type, DSAllocatorContext allocator, DSDataConst data, DSData native) {
    ds_assert_param(instance);
    ds_assert_param(type);
    ds_assert_param(data);
    ds_assert_param(native);

	// Implement parsing of OPCXML
    return DS_ERR_NOTIMPLEMENTED;
}



static int dsBinaryFormatOPCXMLReadInteger(DSBinaryFormatInstance instance, DSBinaryType type, const char *source, size_t *bytes) {
    int err;
    DSBinaryFormat ctx;
    xmlNodePtr node, cnode;
    xmlNsPtr xsins;
    unsigned long elements, i;    
    char stmp[64];
    const char *subtype;

    node = ((DSBinaryFormatOPCXMLInstanceData)instance->data)->node;
    xsins = ((DSBinaryFormatOPCXMLInstanceData)instance->data)->xsins;
    
    elements=type->element_count;
    if (elements) {
	if (elements>1) {
	    if (type->options&DS_BINARY_TYPE_CLASS_OPTION_MASTER) {
		cnode = node;
		subtype = dsBinaryFormatOPCXMLSetType(NULL,type,NULL);
	    } else {
		cnode = xmlNewChild(node, NULL, "anyType", NULL);
		subtype = dsBinaryFormatOPCXMLSetType(xsins,type,cnode);
	    }
    	    for (i=type->element_count;i>0;i--) {
		err = dsBinaryFormatOPCXMLinteger((DSBinaryTypeInteger)type, source, stmp);
		dsDebugServer(DS_DEBUG_CRITICAL, "Write Float: (%lx)\n", err);
		if (err) return err;
		node=xmlNewChild(cnode,NULL,subtype,stmp);
		source+=((DSBinaryTypeInteger)type)->bytes;
	    }
	    if (bytes) *bytes = ((DSBinaryTypeInteger)type)->bytes * type->element_count;
	} else {
	    err = dsBinaryFormatOPCXMLinteger((DSBinaryTypeInteger)type, source, stmp);
	    if (err) return err;

	    if (type->options&DS_BINARY_TYPE_CLASS_OPTION_MASTER) {
		puts("add");
		xmlNodeAddContentLen(node, stmp, (int)strlen(stmp));
	    } else {
		cnode=xmlNewChild(node, NULL, "anyType", stmp);
		dsBinaryFormatOPCXMLSetType(xsins,type,cnode);
	    }
	    
	    if (bytes) *bytes = ((DSBinaryTypeInteger)type)->bytes;
	}
    } else if (type->element_count_ref) {
	return DS_ERR_NOTIMPLEMENTED;
    } else if (type->terminator) {
	return DS_ERR_NOTIMPLEMENTED;
    } else return DS_ERR_NOTIMPLEMENTED;
    return 0;
}



static int dsBinaryFormatOPCXMLReadFloat(DSBinaryFormatInstance instance, DSBinaryType type, const char *source, size_t *bytes) {
    int err;
    xmlNodePtr node,cnode;
    xmlNsPtr xsins;
    unsigned long elements, i;    
    char stmp[64];
    const char *subtype;

    node = ((DSBinaryFormatOPCXMLInstanceData)instance->data)->node;
    xsins = ((DSBinaryFormatOPCXMLInstanceData)instance->data)->xsins;
    
    elements=type->element_count;
    if (elements) {
	if (elements>1) {
	    if (type->options&DS_BINARY_TYPE_CLASS_OPTION_MASTER) {
		cnode = node;
		subtype=dsBinaryFormatOPCXMLSetType(NULL,type,NULL);
	    } else {
		cnode=xmlNewChild(node, NULL, "anyType", NULL);
		subtype=dsBinaryFormatOPCXMLSetType(xsins,type,cnode);
	    }
    	    for (i=type->element_count;i>0;i--) {
		err = dsBinaryFormatOPCXMLfloat((DSBinaryTypeFloat)type, source, stmp);
		if (err) return err;
		node=xmlNewChild(cnode,NULL,subtype,stmp);
		source+=((DSBinaryTypeFloat)type)->bytes;
	    }
	    if (bytes) *bytes = ((DSBinaryTypeFloat)type)->bytes * type->element_count;
	} else {
	    err = dsBinaryFormatOPCXMLfloat((DSBinaryTypeFloat)type, source, stmp);
	    if (err) return err;
	    
	    if (type->options&DS_BINARY_TYPE_CLASS_OPTION_MASTER) {
		puts("add");
		xmlNodeAddContentLen(node, stmp, (int)strlen(stmp));
	    } else {
		cnode=xmlNewChild(node, NULL, "anyType", stmp);
		dsBinaryFormatOPCXMLSetType(xsins,type,cnode);
	    }
	    
	    if (bytes) *bytes = ((DSBinaryTypeFloat)type)->bytes;
	}
    } else if (type->element_count_ref) {
	return DS_ERR_NOTIMPLEMENTED;
    } else if (type->terminator) {
	return DS_ERR_NOTIMPLEMENTED;
    } else return DS_ERR_NOTIMPLEMENTED;

    return 0;
}


static DSBinaryFormatS ds_binary_format_opcxml = {
	"OPCXML",
	DS_BINARY_FORMAT_OPCXML,
	DS_BINARY_FORMAT_ALLOCATOR_TYPE_ABSENT,
	dsBinaryFormatOPCXMLCreateInstanceData,
	dsBinaryFormatOPCXMLDestroyInstanceData,
	dsBinaryFormatOPCXMLImportData,
	dsBinaryFormatOPCXMLDestructData,
	NULL,							/* CopyData */
	NULL,							/* PipeData */
	dsBinaryFormatOPCXMLReadFromBase,
	dsBinaryFormatOPCXMLWriteToBase,
	dsBinaryFormatOPCXMLReadInteger,
	dsBinaryFormatOPCXMLReadFloat
};

DSBinaryFormat dsBinaryFormatOPCXML() {
    return &ds_binary_format_opcxml;
}
