#include <dslib.h>
#include <dsassert.h>
#include <dsdebug.h>

#include <dsbinaryformatpool.h>
#include <dsbinaryformatnative.h>
#include <dsbinaryformatopcxml.h>

void dsBinaryFormatPoolZeroContext(DSBinaryFormatPoolContext ctx) {
    ctx->formats = NULL;
    ctx->binary_format = NULL;
    ctx->native_format = NULL;
    ctx->opcxml_format = NULL;
}

DSBinaryFormatPoolContext dsBinaryFormatPoolCreateContext(DSMallocContext mctx) {
    DSBinaryFormatPoolContext ctx;

    dsDebugServer(DS_DEBUG_NOTICE, "dsBinaryFormatPoolCreateContext");

    ctx = (DSBinaryFormatPoolContext)dsMalloc(mctx,sizeof(struct DSBinaryFormatPoolContextT));
    if (ctx) {
	ctx->malloc=mctx;
	dsBinaryFormatPoolZeroContext(ctx);
    }
    return ctx;
}

int dsBinaryFormatPoolInitContext(DSBinaryFormatPoolContext ctx, size_t scaling) {
    int err;
    unsigned long formats_hash_size;

    ds_assert_param(ctx);
    
    formats_hash_size = 2 * scaling;
    if (formats_hash_size<4) formats_hash_size=4;
    
    dsDebugServer(DS_DEBUG_NOTICE, "dsBinaryFormatPoolInitContext");
    dsBinaryFormatPoolZeroContext(ctx);

    ctx->formats = dsHashTableInit(ctx->malloc, formats_hash_size, DS_HASH_TYPE_POINTER, NULL);
    if (!ctx->formats) {
	dsBinaryFormatPoolFreeContext(ctx);
	return DS_ERR_MEM;
    }

    ctx->native_format = dsBinaryFormatNative();
    err = dsBinaryFormatPoolRegisterFormat(ctx, ctx->native_format);
    if (err) {
	dsBinaryFormatPoolFreeContext(ctx);
	return err;
    }

/*
    ctx->opcxml_format = dsBinaryFormatOPCXML();

    err = dsBinaryFormatPoolRegisterFormat(ctx, ctx->opcxml_format);
    if (err) {
	dsBinaryFormatPoolFreeContext(ctx);
	return err;
    }
*/

    return 0;
}

void dsBinaryFormatPoolFreeContext(DSBinaryFormatPoolContext ctx) {
    ds_assert_return(ctx);
    dsDebugServer(DS_DEBUG_NOTICE, "dsBinaryFormatPoolFreeContext");
    
    if (ctx->formats) dsHashTableDestroy(ctx->formats);
    dsBinaryFormatPoolZeroContext(ctx);
}

void dsBinaryFormatPoolDestroyContext(DSBinaryFormatPoolContext ctx) {
    dsDebugServer(DS_DEBUG_NOTICE, "dsBinaryFormatPoolDestroyContext");
    dsBinaryFormatPoolFreeContext(ctx);
    dsFree(ctx->malloc,ctx);
}


DSBinaryFormat dsBinaryFormatPoolFindFormat(DSBinaryFormatPoolContext ctx, const char *format) {
    ds_assert_null(ctx);
    ds_assert_null(format);
    
    return (DSBinaryFormat)dsHashTableFind(ctx->formats, (void*)format);
}


int dsBinaryFormatPoolRegisterFormat(DSBinaryFormatPoolContext ctx, DSBinaryFormat fctx) {
    ds_assert_param(ctx);
    ds_assert_param(fctx);
    
    return dsHashTableAdd(ctx->formats, fctx, 1);
}
