#ifndef _DS_DATA_CHUNK_H
#define _DS_DATA_CHUNK_H

#include <dsmalloc.h>

#include <dslock.h>
#include <dscondition.h>
#include <dstime.h>

#include <dsdata.h>
#include <dsdatahandler.h>

typedef struct DSDataChunkContextT *DSDataChunkContext;

typedef void *DSDataChunkInfo;
typedef unsigned int DSDataChunkForm;

#define DS_DATA_CHUNK_FORM_INVALID	((DSDataChunkForm)-1)
#define DS_DATA_CHUNK_FORM_MISSING	((DSDataChunkForm)-1)
#define DS_DATA_CHUNK_FORM_1		0


struct DSDataChunkContextT {
    DSDataS info; /* Should be first element */

    DSData data;
    
    DSMallocContext malloc;
    DSLockContext lock;
    DSConditionContext cond;
    DSDataChunkContext next;

    unsigned long referers;
    unsigned long delayed_referers;

    DSDataHandler defhandler;
    unsigned long supported_forms;
};

DSDataChunkContext dsDataChunkCreateContext(DSMallocContext mctx);
int dsDataChunkInitContext(DSDataChunkContext ctx, DSDataHandler defhandler, unsigned long supported_forms);
int dsDataChunkRestoreContext(DSDataChunkContext ctx, DSDataHandler defhandler);
void dsDataChunkFreeContext(DSDataChunkContext ctx);
void dsDataChunkDestroyContext(DSDataChunkContext ctx);

int dsDataChunkSet(DSDataChunkContext ctx, DSDataChunkForm format, DSDataConst data);
DSDataConst dsDataChunkGet(DSDataChunkContext ctx, DSDataChunkForm format);
DSData dsDataChunkGetData(DSDataChunkContext ctx, DSDataChunkForm format);
DSDataChunkForm dsDataChunkFindData(DSDataChunkContext ctx);

int dsDataChunkSetInfo(DSDataChunkContext ctx, DSDataConst info);
DSDataConst dsDataChunkGetInfo(DSDataChunkContext ctx);
DSDataChunkContext dsDataChunkGetChunkByInfo(DSDataConst ctx);

int dsDataChunkWaitUnref(DSDataChunkContext ctx, unsigned char delayed);

unsigned long dsDataChunkRef(DSDataChunkContext ctx);
unsigned long dsDataChunkUnref(DSDataChunkContext ctx);
unsigned long dsDataChunkGetRefs(DSDataChunkContext ctx);
unsigned long dsDataChunkDelayedRef(DSDataChunkContext ctx);
unsigned long dsDataChunkRefDelayed(DSDataChunkContext ctx);
unsigned long dsDataChunkDelayedUnref(DSDataChunkContext ctx);
unsigned long dsDataChunkGetDelayedRefs(DSDataChunkContext ctx);

/* Pointer on the Info returned by GetInfo method should be passed, not
a copy of this structure */
unsigned long dsDataChunkRefByInfo(DSDataConst ctx);
unsigned long dsDataChunkUnrefByInfo(DSDataConst ctx);
unsigned long dsDataChunkGetRefsByInfo(DSDataConst ctx);
unsigned long dsDataChunkDelayedRefByInfo(DSDataConst ctx);
unsigned long dsDataChunkRefDelayedByInfo(DSDataConst ctx);
unsigned long dsDataChunkDelayedUnrefByInfo(DSDataConst ctx);
unsigned long dsDataChunkGetDelayedRefsByInfo(DSDataConst ctx);

void dsDataChunkLock(DSDataChunkContext ctx);
void dsDataChunkUnlock(DSDataChunkContext ctx);

typedef void (*DSDataChunkRunFunction)(void *ctx, DSData data);
void dsDataChunkRun(DSDataChunkContext ctx, DSDataChunkRunFunction func, void *attr);

#endif /* _DS_DATA_CHUNK_H */
