#include <string.h>
#include <time.h>
#include <dsassert.h>

#include <dslib.h>

#include <dsmalloc.h>
#include <dsdebug.h>
#include <dslock.h>

#include <dsholderbucket.h>

DSHolderBucketContext dsHolderBucketInitContext(DSMallocContext mctx, void *ptr, DSDuration lifetime, DSDuration steptime, DSDestroyContextFunction destructor) {
    DSHolderBucketContext ctx;
    
    ctx = (DSHolderBucketContext)dsMalloc(mctx, sizeof(struct DSHolderBucketContextT));
    dsDebugServer(DS_DEBUG_NOTICE, "dsHolderBucketInitContext: %p", ctx);    
    if (!ctx) return NULL;

    ctx->malloc = mctx;

    ctx->lock = dsLockCreateContext(mctx);
    if (!ctx->lock) {
	dsFree(mctx,ctx);
	return NULL;
    }
    if (dsLockInitContext(ctx->lock)) {
	dsLockDestroyContext(ctx->lock);
	dsFree(mctx,ctx);
	return NULL;
    }

    if (lifetime) {
	dsTimeCurrentTime(&ctx->destruction_time);
	dsTimeAdd(&ctx->destruction_time, lifetime);
	dsDurationCopy(&ctx->step_time, steptime);
    } else dsTimeZeroTime(&ctx->destruction_time);
    ctx->destructor=destructor;
    ctx->data=ptr;
    return ctx;
}

void dsHolderBucketDestroyContext(DSHolderBucketContext ctx) {
    dsDebugServer(DS_DEBUG_NOTICE, "dsHolderBucketDestroyContext: %p", ctx);    

    ds_assert_return(ctx);
    
    /* should be unlinked from opcHolder at this point */
    dsLockWrite(ctx->lock);
    dsLockReturn(ctx->lock);
    dsLockDestroyContext(ctx->lock);
    dsFree(ctx->malloc, ctx);
}

void dsHolderBucketLock(DSHolderBucketContext ctx) {
    ds_assert_return(ctx);
    
    dsLockWrite(ctx->lock);
}

void dsHolderBucketUnlock(DSHolderBucketContext ctx) {
    ds_assert_return(ctx);
    
    dsLockReturn(ctx->lock);
}

void dsHolderBucketRestartTimer(DSHolderBucketContext ctx) {
    DSTimeS destruction_time;

    ds_assert_return(ctx);
    
    if (!dsTimeIsZero(&ctx->destruction_time)) {
	dsTimeCurrentTime(&destruction_time);
	dsTimeAdd(&destruction_time, &ctx->step_time);
	if (dsTimeCmp(&destruction_time, &ctx->destruction_time)>0) 
	    dsTimeCopy(&ctx->destruction_time, &destruction_time);
    }
}


int dsHolderBucketMantain(DSHolderBucketContext ctx, void *attr) {
    ds_assert_param(ctx);
    
    if (dsLockTryWrite(ctx->lock)) return 0;
    dsLockReturn(ctx->lock); /* Since opcHolderContext is locked nobody can accquire new lock */
    
    if (dsTimeCurrentCmp(&ctx->destruction_time)>0) return 0;
    dsHolderBucketDestroyContext(ctx);
    return 1;
}

int dsHolderBucketSetNumbers(DSHolderBucketContext ctx, unsigned long magic) {
    ds_assert_param(ctx);
    
    ctx->magic = magic;
    ctx->secret = PTR2INT(ctx->data);
    return 0;
}

int dsHolderBucketCheckMagicNumber(DSHolderBucketContext ctx, unsigned long magic) {
    ds_assert_param(ctx);
    
    return (ctx->magic==magic);
}

int dsHolderBucketCheckSecretNumber(DSHolderBucketContext ctx, unsigned long secret) {
    ds_assert_param(ctx);
    
    return (ctx->secret==secret);
}
