#ifndef _DS_DATA_MANAGER_H
#define _DS_DATA_MANAGER_H

#define DS_DATA_MANAGER_FORM_NATIVE "Native"
#define DS_DATA_MANAGER_FORM_OPCXML "OpcXml"

#include <dslib.h>
#include <dsmodule.h>

#include <dsdata.h>
#include <dschunkpool.h>
#include <dsdatachunk.h>
#include <dsallocator.h>
#include <dsdatamanageruserize.h>

#define DS_DATA_MANAGER(ctx) ((DSDataManagerContext)ctx)
#define DS_DATA_MANAGER_OPTIONS(options) ((DSDataManagerOptions)options)

typedef struct DSDataManagerContextT *DSDataManagerContext;

typedef int (*DSDataManagerAcquireFunction)(DSDataManagerContext ctx);
typedef int (*DSDataManagerPushDataFunction)(DSDataManagerContext ctx, DSDataChunkForm repid, DSDataConst info, DSData data);
typedef DSAllocatorContext (*DSDataManagerGetAllocatorFunction)(DSDataManagerContext ctx, DSDataChunkForm repid);
typedef int (*DSDataManagerReturnAllocatorFunction)(DSDataManagerContext ctx, DSDataChunkForm repid, DSAllocatorContext allocator);
typedef int (*DSDataManagerGetDataFunction)(DSDataManagerContext ctx, DSDataChunkForm repid, DSData *info, DSData *data);
typedef int (*DSDataManagerReturnDataFunction)(DSDataManagerContext ctx, DSDataConst info);
typedef int (*DSDataManagerMantainFunction)(DSDataManagerContext ctx);
typedef DSDataChunkForm (*DSDataManagerGetFormByNameFunction)(DSDataManagerContext ctx, const char *repname);

enum DSDataManagerFlagsT {
    DS_DATA_MANAGER_FLAG_SEQUENTIAL_SOURCE = 1
};
typedef enum DSDataManagerFlagsT DSDataManagerFlags;

struct DSDataManagerOptionsT {
    DSDataManagerFlags flags;

    DSChunkPoolContext chunk_pool;
    unsigned long cache_min;
    unsigned long cache_max;
};
typedef struct DSDataManagerOptionsT DSDataManagerOptionsS;
typedef struct DSDataManagerOptionsT *DSDataManagerOptions;

DSDataManagerOptions dsDataManagerGetDefaultOptions();
int dsDataManagerInitDefaultOptions(DSDataManagerOptions options);
void dsDataManagerDestroyOptions(DSDataManagerOptions options);
int dsDataManagerOptionsSetCacheSize(DSDataManagerOptions options, unsigned long min, unsigned long max);
int dsDataManagerOptionsSetFlags(DSDataManagerOptions options, unsigned long flags);
int dsDataManagerOptionsSetChunkPool(DSDataManagerOptions options, DSChunkPoolContext chunk_pool);
int dsDataManagerOptionsSetMissingChunkPool(DSDataManagerOptions options, DSChunkPoolContext chunk_pool);

struct DSDataManagerT {
    DSDestroyContextFunction Destructor;
    DSDataManagerAcquireFunction Acquire;
    DSDataManagerPushDataFunction PushData;
    DSDataManagerGetAllocatorFunction GetAllocator;
    DSDataManagerReturnAllocatorFunction ReturnAllocator;
    DSDataManagerGetDataFunction GetData;
    DSDataManagerReturnDataFunction ReturnData;
    DSDataManagerMantainFunction Mantain;
    DSDataManagerGetFormByNameFunction GetFormByName;
};
typedef struct DSDataManagerT DSDataManagerS;
typedef struct DSDataManagerT *DSDataManager;

struct DSDataManagerContextT {
    DSModuleS module;
    DSDataManagerS manager;

    DSDataManagerFlags flags;

	// Referers Managment
    DSLockContext lock;
    unsigned long refs;
    
	// Chunk Managment
    DSLockContext chunk_lock;
    DSChunkPoolContext chunk_pool;
    DSDataChunkContext chunks;
    DSDataChunkContext current;
    unsigned long nchunks;
    unsigned long cache_min;
    unsigned long cache_max;
};
typedef struct DSDataManagerContextT DSDataManagerContextS;

int dsDataManagerInitContext(DSDataManagerContext ctx, DSDataManagerOptions options);
void dsDataManagerFreeContext(DSDataManagerContext ctx);

int dsDataManagerRef(DSDataManagerContext ctx);
int dsDataManagerUnref(DSDataManagerContext ctx);

int dsDataManagerPushPointer(DSDataManagerContext ctx, DSDataChunkForm repid, DSDataConst info, void *data, size_t size);
int dsDataManagerPushAllocation(DSDataManagerContext ctx, DSDataChunkForm repid,  DSDataConst info, DSAllocatorContext allocator, DSAllocation allocation);

int dsDataManagerRegisterChunk(DSDataManagerContext ctx, DSDataChunkContext chunk);
int dsDataManagerFastClear(DSDataManagerContext ctx);

#define dsDataManagerAcquire(ctx) \
    ((ctx&&ctx->manager.Acquire)?(ctx->manager.Acquire(ctx)):DS_ERR_NOTSUPPORTED)
#define dsDataManagerPushData(ctx,repid,info,data) \
    ((ctx&&ctx->manager.PushData)?(ctx->manager.PushData(ctx,repid,info,data)):DS_ERR_NOTIMPLEMENTED)
#define dsDataManagerGetAllocator(ctx,repid,info) \
    ((ctx&&ctx->manager.GetAllocator)?(ctx->manager.GetAllocator(ctx,repid,info)):DS_ERR_NOTIMPLEMENTED)
#define dsDataManagerReturnAllocator(ctx,repid,allocator) \
    ((ctx&&ctx->manager.ReturnAllocator)?(ctx->manager.ReturnAllocator(ctx,repid,allocator)):0)
#define dsDataManagerGetData(ctx,repid,info,data) \
    ((ctx&&ctx->manager.GetData)?(ctx->manager.GetData(ctx,repid,info,data)):DS_ERR_NOTIMPLEMENTED)

#define dsDataManagerMantain(ctx) \
    ((ctx&&ctx->manager.Mantain)?ctx->manager.Mantain(ctx):0)
#define dsDataManagerGetFormByName(ctx, name) \
    ((ctx&&ctx->manager.GetFormByName)?ctx->manager.GetFormByName(ctx,name):0)
    


int dsDataManagerReturnData(DSDataManagerContext ctx, DSDataConst info);
int dsDataManagerSetAcquireFunction(DSDataManagerContext ctx, DSDataManagerAcquireFunction acquire);


/*
int dsDataManagerGetDataFirst(DSDataManagerContext ctx, unsigned long repid, DSData *info, DSData *data);
int dsDataManagerGetDataNextFunction)(DSDataManagerContext ctx, unsigned long repid, DSData *info, DSData *data);
*/
#endif /* _DS_DATA_MANAGER_H */
