#ifndef _DS_DATA_MANAGER_FAST_H
#define _DS_DATA_MANAGER_FAST_H

#include <dslock.h>
#include <dschunkpool.h>
#include <dsdatamanager.h>
#include <dsdatabuffer.h>

typedef struct DSDataManagerFastContextT *DSDataManagerFastContext;
typedef struct DSDataManagerFastOptionsT *DSDataManagerFastOptions;

struct DSDataManagerFastOptionsT {
    DSDataManagerOptionsS manager;
    struct DSDataBufferOptionsT allocator;
};
typedef struct DSDataManagerFastOptionsT DSDataManagerFastOptionsS;

DSDataManagerFastOptions dsDataManagerFastGetDefaultOptions();
int dsDataManagerFastInitDefaultOptions(DSDataManagerFastOptions options);
void dsDataManagerFastDestroyOptions(DSDataManagerFastOptions options);
int dsDataManagerFastOptionsSetBufferSize(DSDataManagerFastOptions options, unsigned long min, unsigned long max);
int dsDataManagerFastOptionsSetBufferPool(DSDataManagerFastOptions options, void *pool);
int dsDataManagerFastOptionsSetMissingBufferPool(DSDataManagerFastOptions options, void *pool);
int dsDataManagerFastOptionsSetFixedAllocationSize(DSDataManagerFastOptions options, size_t size);

struct DSDataManagerFastContextT {
    DSDataManagerContextS manager;
    DSDataManagerDataHandlerS handler;
    
    DSBinaryFormatInstance format;
    DSBinaryType type;

    DSLockContext allocator_lock;
    unsigned char grabed_allocator;
    DSDataBufferContext allocator;
};
typedef struct DSDataManagerFastContextT DSDataManagerFastContextS;

DSDataManagerFastContext dsDataManagerFastCreateContext(DSMallocContext mctx);
int dsDataManagerFastInitContext(DSDataManagerFastContext ctx, DSBinaryType type, DSDataManagerFastOptions options);

int dsDataManagerFastPushData(DSDataManagerFastContext ctx, DSDataChunkForm repid, DSDataConst info, DSDataConst data);
int dsDataManagerFastGetData(DSDataManagerFastContext ctx, DSDataChunkForm repid, DSDataConst *info, DSDataConst *data);
DSAllocatorContext dsDataManagerFastGetAllocator(DSDataManagerFastContext ctx, DSDataChunkForm repid);
int dsDataManagerFastReturnAllocator(DSDataManagerFastContext ctx, DSDataChunkForm repid, DSAllocatorContext allocator);

int dsDataManagerFastMantain(DSDataManagerFastContext ctx);
DSDataChunkForm dsDataManagerFastGetFormByName(DSDataManagerContext ctx, const char *title);

int dsDataManagerFastSetUserizeFunction(DSDataManagerFastContext ctx, DSDataManagerGetUserDataFunction get, void *userizer);

int dsDataManagerFastDataDestroy(DSData data);
int dsDataManagerFastDataCopy(DSDataCopier copier, DSData dst, DSDataConst src);
int dsDataManagerFastDataPipe(DSDataConst data, DSPipeFunction func, void *attr);

#endif /* _DS_DATA_MANAGER_FAST_H */
