#ifndef _DS_DATA_MANAGER_REP_H
#define _DS_DATA_MANAGER_REP_H

#include <dslock.h>
#include <dstime.h>
#include <dsmalloc.h>
#include <dshashtable.h>

#include <dschunkpool.h>
#include <dsbinaryformat.h>

#include <dsdataconverter.h>
#include <dsrepresentation.h>

#include <dsallocator.h>
#include <dsdatabuffer.h>

#include <dsdatamanager.h>

typedef struct DSDataManagerRepContextT *DSDataManagerRepContext;
typedef struct DSDataManagerRepOptionsT *DSDataManagerRepOptions;

struct DSDataManagerRepOptionsT {
    DSDataManagerOptionsS manager;
    struct DSDataBufferOptionsT allocator;
};
typedef struct DSDataManagerRepOptionsT DSDataManagerRepOptionsS;

DSDataManagerRepOptions dsDataManagerRepGetDefaultOptions();
int dsDataManagerRepInitDefaultOptions(DSDataManagerRepOptions options);
void dsDataManagerRepDestroyOptions(DSDataManagerRepOptions options);

int dsDataManagerRepOptionsSetBufferSize(DSDataManagerRepOptions options, unsigned long min, unsigned long max);
int dsDataManagerRepOptionsSetBufferPool(DSDataManagerRepOptions options, DSBufferPoolContext buffer_pool);
int dsDataManagerRepOptionsSetMissingBufferPool(DSDataManagerRepOptions options, DSBufferPoolContext buffer_pool);
int dsDataManagerRepOptionsSetFixedAllocationSize(DSDataManagerRepOptions options, size_t size);

enum DSDataManagerRepRepresentationFlagsT {
    DS_DATA_MANAGER_REP_REPRESENTATION_NATIVE,
    DS_DATA_MANAGER_REP_REPRESENTATION_BINARY
};
typedef enum DSDataManagerRepRepresentationFlagsT DSDataManagerRepRepresentationFlags;

/* Our numbering scheme is completely independent from the Converter */
struct DSDataManagerRepRepresentationDataT {
    const char *title; /* Expects static pointer */
    DSDataChunkForm formid;
    DSRepresentationInstance reference;
    DSDataManagerRepRepresentationFlags flags;

    DSDataManagerDataHandlerS handler;
};
typedef struct DSDataManagerRepRepresentationDataT DSDataManagerRepRepresentationDataS;
typedef struct DSDataManagerRepRepresentationDataT *DSDataManagerRepRepresentationData;

struct DSDataManagerRepContextT {
    DSDataManagerContextS manager;
    struct DSDataBufferOptionsT databuffer_options;

    DSDataConverterContext converter;

    DSLockContext writer_lock;    
    DSLockContext representation_lock;

    DSDataChunkForm allocated_forms;
    DSDataChunkForm supported_forms;
    DSDataManagerRepRepresentationData representation;
    DSHashTable rephash;
};
typedef struct DSDataManagerRepContextT DSDataManagerRepContextS;

DSDataManagerRepContext dsDataManagerRepCreateContext(DSMallocContext mctx);
int dsDataManagerRepInitContext(DSDataManagerRepContext ctx, DSDataManagerRepOptions options);
int dsDataManagerRepRef(DSDataManagerRepContext ctx);
void dsDataManagerRepUnref(DSDataManagerRepContext ctx);

DSDataManagerRepContext dsDataManagerRepCreateOPCXML(DSMallocContext mctx, DSBinaryType type, DSDataManagerRepOptions options);

DSDataChunkForm dsDataManagerRepRegisterRepresentation(DSDataManagerRepContext ctx, DSDataChunkForm base, const char *title, DSRepresentation representation, DSRepresentationOptions options);
DSDataChunkForm dsDataManagerRepRegisterBinaryRepresentation(DSDataManagerRepContext ctx, DSDataChunkForm base, const char *title, DSBinaryFormat format, DSBinaryType type, DSAllocatorContext allocator);
DSDataChunkForm dsDataManagerRepRegisterNativeRepresentation(DSDataManagerRepContext ctx, DSBinaryType type, DSAllocatorContext allocator);
DSDataChunkForm dsDataManagerRepRegisterOPCXMLRepresentation(DSDataManagerRepContext ctx, DSBinaryType type);

DSDataChunkForm dsDataManagerRepGetFormByName(DSDataManagerRepContext ctx, const char *repname);
int dsDataManagerRepPushData(DSDataManagerRepContext ctx, DSDataChunkForm repid, DSDataConst info, DSDataConst data);
int dsDataManagerRepGetData(DSDataManagerRepContext ctx, DSDataChunkForm repid, DSDataConst *info, DSDataConst *data);
int dsDataManagerRepMantain(DSDataManagerRepContext ctx);

DSAllocatorContext dsDataManagerRepGetAllocator(DSDataManagerRepContext ctx, DSDataChunkForm repid);
int dsDataManagerRepReturnAllocator(DSDataManagerRepContext ctx, DSDataChunkForm repid, DSAllocatorContext allocator);

int dsDataManagerRepSetUserizeFunction(DSDataManagerRepContext ctx, DSDataChunkForm repid, DSDataManagerGetUserDataFunction get, void *userizer);

#endif /* _DS_DATA_MANAGER_REP_H */
