#ifndef _DS_OPCBINARY_DICTIONARY_H
#define _DS_OPCBINARY_DICTIONARY_H

#include <dsxml.h>
#include <dsmalloc.h>
#include <dsenumeration.h>
#include <dslock.h>

#include <dsbinarytype.h>


typedef struct DSOPCBinaryDictionaryContextT *DSOPCBinaryDictionaryContext;

#include <dsopcbinary.h>

struct DSOPCBinaryDictionaryContextT {
    DSMallocContext malloc;
    DSEnumerationContext types;
    
    DSLockContext lock;

    xmlXPathContextPtr xpath_ctx;    
    xmlDocPtr xml_doc;
    xmlNsPtr xsins;
    
    unsigned char default_big_endian;
    unsigned char default_char_width;
    char *default_string_encoding;
    char *default_float_format;
};

DSOPCBinaryDictionaryContext dsOPCBinaryDictionaryCreateContext(DSMallocContext mctx);
int dsOPCBinaryDictionaryInitContext(DSOPCBinaryDictionaryContext ctx, xmlDocPtr doc, const char *opcbinary_namespace);
void dsOPCBinaryDictionaryFreeContext(DSOPCBinaryDictionaryContext ctx);
void dsOPCBinaryDictionaryDestroyContext(DSOPCBinaryDictionaryContext ctx);

int dsOPCBinaryDictionaryLocalise(DSOPCBinaryDictionaryContext ctx);

DSBinaryType dsOPCBinaryDictionaryCreateType(DSOPCBinaryDictionaryContext ctx, const char *type);

/* ToDo:
int dsOPCBinaryDictionaryAddType(DSOPCBinaryDictionaryContext ctx, xmlDocPtr doc);
int dsOPCBinaryDictionaryLocalize(DSOPCBinaryDictionaryContext ctx);
int dsOPCBinaryDictionaryPublish(DSOPCBinaryDictionaryContext ctx);
*/


/* ToDo: in case of problems rewrite truth set of defines */
enum DSOPCBinaryTypesT {
    DS_OPCBINARY_INTEGER,
    DS_OPCBINARY_FLOATINGPOINT,
    DS_OPCBINARY_CHARSTRING,
    DS_OPCBINARY_TYPEREFERENCE,
    DS_OPCBINARY_INT8,
    DS_OPCBINARY_INT16,
    DS_OPCBINARY_INT32,
    DS_OPCBINARY_INT64,
    DS_OPCBINARY_UINT8,
    DS_OPCBINARY_UINT16,
    DS_OPCBINARY_UINT32,
    DS_OPCBINARY_UINT64,
    DS_OPCBINARY_SINGLE,
    DS_OPCBINARY_DOUBLE,
    DS_OPCBINARY_ASCII,
    DS_OPCBINARY_UNICODE
};

typedef enum DSOPCBinaryTypesT DSOPCBinaryTypes;


struct DSOPCBinaryTypesEnumerationT {
    char *label;
    unsigned long type;
};

#ifdef _DS_OPCBINARY_DICTIONARY_C
struct DSOPCBinaryTypesEnumerationT ds_opcbinary_types_enumeration[] = {
    {"Integer", DS_OPCBINARY_INTEGER},
    {"FloatingPoint", DS_OPCBINARY_FLOATINGPOINT},
    {"CharString", DS_OPCBINARY_CHARSTRING},
    {"TypeReference", DS_OPCBINARY_TYPEREFERENCE},
    {"Int8", DS_OPCBINARY_INT8},
    {"Int16", DS_OPCBINARY_INT16},
    {"Int32", DS_OPCBINARY_INT32},
    {"Int64", DS_OPCBINARY_INT64},
    {"UInt8", DS_OPCBINARY_UINT8},
    {"UInt16", DS_OPCBINARY_UINT16},
    {"UInt32", DS_OPCBINARY_UINT32},
    {"UInt64", DS_OPCBINARY_UINT64},
    {"Single", DS_OPCBINARY_SINGLE},
    {"Double", DS_OPCBINARY_DOUBLE},
    {"Ascii", DS_OPCBINARY_ASCII},
    {"Unicode", DS_OPCBINARY_UNICODE},
    {NULL, 0}
};
#endif /* _DS_OPCBINARY_DICTIONARY_C */

#endif /* _DS_OPCBINARY_DICTIONARY_H */
