#include <dslib.h>
#include <dsassert.h>
#include <dstreenode.h>

#include <dsbinarytype.h>

void dsBinaryTypeZero(DSBinaryType ctx) {
    ctx->options = 0;
    ctx->terminator = NULL;
    ctx->element_count_ref = NULL;
    ctx->element_count = 1;
    ctx->is_pointer = 0;
    ctx->size = 0;
    ctx->align = 0;
}

DSBinaryType dsBinaryTypeCreate(DSMallocContext mctx) {
    DSBinaryType ctx;
    
    ctx = dsMalloc(mctx, sizeof(struct DSBinaryTypeT));
    if (ctx) {
	((DSTreeNodeContext)ctx)->malloc = mctx;
	dsBinaryTypeZero(ctx);
    }
    return ctx;
}

int dsBinaryTypeInit(DSBinaryType ctx, DSBinaryTypeClass type, DSDestroyContextFunction destructor) {
    ds_assert_param(ctx);
    dsBinaryTypeZero(ctx);
    if (!destructor) destructor=(DSDestroyContextFunction)dsBinaryTypeDestroy;
    dsTreeComplexNodeInitContext((DSTreeComplexNodeContext)ctx, destructor);
    ctx->type = type;
    return 0;
}

void dsBinaryTypeFree(DSBinaryType ctx) {
    ds_assert_return(ctx);
    dsTreeComplexNodeFreeContext((DSTreeComplexNodeContext)ctx);
    if (ctx->terminator) dsExchangeString(((DSTreeNodeContext)ctx)->malloc, &ctx->terminator, NULL);
    dsBinaryTypeZero(ctx);
}

void dsBinaryTypeDestroy(DSBinaryType ctx) {
    DSDestroyContextFunction destructor;

    destructor = ctx->node.node.destructor;
    if ((destructor)&&(destructor != (DSDestroyContextFunction)dsBinaryTypeDestroy)) {
	ctx->node.node.destructor = NULL;
	destructor(ctx);
	return;
    }
    dsBinaryTypeFree(ctx);
    dsFree(((DSTreeNodeContext)ctx)->malloc,ctx);
}



void dsBinaryTypeSetElementCount(DSBinaryType ctx, unsigned long element_count) {
    ds_assert_return(ctx);
    ds_assert_return(!ctx->terminator);
    ds_assert_return(!ctx->element_count_ref);
    
    ctx->element_count = element_count;
    
    if ((element_count>1)&&(ctx->size)) ctx->size*=element_count;
}

void dsBinaryTypeSetElementCountRef(DSBinaryType ctx, DSBinaryType ref) {
    ds_assert_return(ctx);
    ds_assert_return(!ctx->terminator);
    ds_assert_return(!ctx->element_count);
    
    ctx->element_count = 0;
    ctx->element_count_ref = ref;
}

void dsBinaryTypeSetTerminator(DSBinaryType ctx, const char *terminator) {
    ds_assert_return(ctx);
    ds_assert_return(!ctx->element_count);
    ds_assert_return(!ctx->element_count_ref);

    ctx->element_count = 0;
    dsExchangeString(((DSTreeNodeContext)ctx)->malloc, &ctx->terminator, terminator);
}

void dsBinaryTypeSetExternal(DSBinaryType ctx) {
    ds_assert_return(ctx);
    
    ctx->is_pointer = 1;
}

void dsBinaryTypeSetMaster(DSBinaryType ctx) {
    ds_assert_return(ctx);
    
    ctx->options|=DS_BINARY_TYPE_CLASS_OPTION_MASTER;
}

void dsBinaryTypeResetMaster(DSBinaryType ctx) {
    ds_assert_return(ctx);
    
    ctx->options&=~DS_BINARY_TYPE_CLASS_OPTION_MASTER;
}

void dsBinaryTypeSetAlignment(DSBinaryType ctx, unsigned char align) {
    ds_assert_return(ctx);
    
    ctx->align = align;
}

int dsBinaryTypePush(DSBinaryType ctx, DSBinaryType child) {
    ds_assert_param(ctx);
    ds_assert_param(child);
    ds_assert_param(ctx->type == DS_BINARY_TYPE_CLASS_COMPLEX);
    return dsTreeComplexNodeAddChild((DSTreeComplexNodeContext)ctx,(DSTreeNodeContext)child);
}


