#include <dsassert.h>

#include <dsbinarytypesimple.h>

static int dsBinaryTypeIntegerInit(DSBinaryTypeInteger ctx, unsigned char is_signed, unsigned char bytes, DSByteOrder byte_order) {
    dsBinaryTypeInit((DSBinaryType)ctx, DS_BINARY_TYPE_CLASS_INTEGER, (DSDestroyContextFunction)dsBinaryTypeIntegerDestroy);

    ctx->is_signed = is_signed;
    ctx->bytes = bytes;
    ctx->byte_order = byte_order;
    ctx->base.size = bytes;

    ctx->enumeration = NULL;
    return 0;
}

static int dsBinaryTypeFloatInit(DSBinaryTypeFloat ctx, DSFloatFormat format, unsigned char bytes, DSByteOrder byte_order) {
    dsBinaryTypeInit((DSBinaryType)ctx, DS_BINARY_TYPE_CLASS_FLOAT, NULL);

    ctx->bytes = bytes;
    ctx->byte_order = byte_order;
    ctx->float_format = format;
    ctx->base.size = bytes;
    return 0;
}

DSBinaryTypeInteger dsBinaryTypeIntegerCreate(DSMallocContext mctx, unsigned char is_signed, unsigned char bytes, DSByteOrder byte_order) {
    DSBinaryTypeInteger ctx;
    
    ctx = (DSBinaryTypeInteger)dsMalloc(mctx, sizeof(struct DSBinaryTypeIntegerT));
    if (!ctx) return ctx;
    
    ((DSTreeNodeContext)ctx)->malloc=mctx;
    dsBinaryTypeIntegerInit(ctx, is_signed, bytes, byte_order);

    return ctx;
}

void dsBinaryTypeIntegerDestroy(DSBinaryTypeInteger ctx) {
    if (ctx->enumeration) dsEnumerationDestroyContext(ctx->enumeration);    
    dsBinaryTypeDestroy((DSBinaryType)ctx);
}

int dsBinaryTypeIntegerSetEnumeration(DSBinaryTypeInteger ctx, DSEnumerationContext enumeration) {
    ds_assert_param(ctx);

    if (ctx->enumeration) dsEnumerationDestroyContext(ctx->enumeration);
    ctx->enumeration = enumeration;
    return 0;
}

int dsBinaryTypeIntegerSetEnumerationFromArray(DSBinaryTypeInteger ctx, const char **array, unsigned char copy_strings) {
    int err;
    DSEnumerationContext enumeration;
    
    enumeration = dsEnumerationCreateContext(((DSTreeNodeContext)ctx)->malloc);
    if (!enumeration) return DS_ERR_MEM;
    
    err = dsEnumerationInitContextFromArray(enumeration, array, copy_strings);
    if (err) dsEnumerationDestroyContext(enumeration);
    else {
	err = dsBinaryTypeIntegerSetEnumeration(ctx, enumeration);
	if (err) dsEnumerationDestroyContext(enumeration);
    }

    return err;
}


DSBinaryTypeFloat dsBinaryTypeFloatCreate(DSMallocContext mctx, DSFloatFormat format, unsigned char bytes, DSByteOrder byte_order) {
    DSBinaryTypeFloat ctx;
    
    ctx = (DSBinaryTypeFloat)dsMalloc(mctx, sizeof(struct DSBinaryTypeFloatT));
    if (!ctx) return ctx;
    
    ((DSTreeNodeContext)ctx)->malloc=mctx;
    dsBinaryTypeFloatInit(ctx, format, bytes, byte_order);

    return ctx;
}

DSBinaryTypeString dsBinaryTypeStringCreate(DSMallocContext mctx, const char *encoding, DSByteOrder byte_order) {
    DSBinaryTypeString ctx;
    
    ctx = (DSBinaryTypeString)dsMalloc(mctx, sizeof(struct DSBinaryTypeStringT));
    if (!ctx) return ctx;
    
    ((DSTreeNodeContext)ctx)->malloc=mctx;
    dsBinaryTypeInit((DSBinaryType)ctx, DS_BINARY_TYPE_CLASS_STRING, (DSDestroyContextFunction)dsBinaryTypeStringDestroy);
    
    ctx->string_encoding = NULL;
    dsExchangeString(mctx, &ctx->string_encoding, encoding);

	// DS: Detect (unicode strings)
    ctx->byte_order = byte_order;
    ctx->char_width = 1;
    ctx->char_count_ref = NULL;

    return ctx;

}

void dsBinaryTypeStringDestroy(DSBinaryTypeString ctx) {
    dsExchangeString(((DSTreeNodeContext)ctx)->malloc, &ctx->string_encoding, NULL);
    dsBinaryTypeDestroy((DSBinaryType)ctx);
}

