#ifndef _DS_BINARY_TYPE_SIMPLE_H
#define _DS_BINARY_TYPE_SIMPLE_H

#include <dsbinarytype.h>
#include <dsenumeration.h>

typedef enum DSByteOrderT DSByteOrder;
typedef enum DSFloatFormatT DSFloatFormat;
typedef struct DSBinaryTypeIntegerT *DSBinaryTypeInteger;
typedef struct DSBinaryTypeFloatT *DSBinaryTypeFloat;
typedef struct DSBinaryTypeStringT *DSBinaryTypeString;
typedef struct DSBinaryTypeEnumerationT *DSBinaryTypeEnumeration;

enum DSByteOrderT {
    DS_LITTLE_ENDIAN = 0,
    DS_BIG_ENDIAN,
    DS_MIXED_ENDIAN
};

enum DSFloatFormatT {
    DS_FLOAT_IEEE = 0
};

struct DSBinaryTypeIntegerT {
    struct DSBinaryTypeT base;
    DSEnumerationContext enumeration;

    unsigned char is_signed;
    DSByteOrder byte_order;
    unsigned char bytes;
};
typedef struct DSBinaryTypeIntegerT DSBinaryTypeIntegerS;

struct DSBinaryTypeFloatT {
    struct DSBinaryTypeT base;
    DSFloatFormat float_format;
    DSByteOrder byte_order;
    unsigned char bytes;
};
typedef struct DSBinaryTypeFloatT DSBinaryTypeFloatS;

struct DSBinaryTypeStringT {
    struct DSBinaryTypeT base;

    DSByteOrder byte_order;
    unsigned char char_width;
    char *string_encoding;
    DSBinaryType char_count_ref;
};
typedef struct DSBinaryTypeStringT DSBinaryTypeStringS;

DSBinaryTypeInteger dsBinaryTypeIntegerCreate(DSMallocContext mctx, unsigned char is_signed, unsigned char bytes, DSByteOrder byte_order);
void dsBinaryTypeIntegerDestroy(DSBinaryTypeInteger ctx);
int dsBinaryTypeIntegerSetEnumeration(DSBinaryTypeInteger ctx, DSEnumerationContext enumeration); /* eating */
int dsBinaryTypeIntegerSetEnumerationFromArray(DSBinaryTypeInteger ctx, const char **array, unsigned char copy_strings);

DSBinaryTypeFloat dsBinaryTypeFloatCreate(DSMallocContext mctx, DSFloatFormat format, unsigned char bytes, DSByteOrder byte_order);
#define dsBinaryTypeFloatDestroy(ctx) dsBinaryTypeDestroy(DS_BINARY_TYPE(ctx))

DSBinaryTypeString dsBinaryTypeStringCreate(DSMallocContext mctx, const char *encoding, DSByteOrder byte_order);
void dsBinaryTypeStringDestroy(DSBinaryTypeString ctx);



#endif /* _DS_BINARY_TYPE_SIMPLE_H */
