#include <dsassert.h>
#include <dsbinaryconfiguration.h>

#include <dsbinarytypestandard.h>

DSBinaryTypeInteger ds_binary_type_int_8t = NULL;
DSBinaryTypeInteger ds_binary_type_int_16t = NULL;
DSBinaryTypeInteger ds_binary_type_int_32t = NULL;
DSBinaryTypeInteger ds_binary_type_int_64t = NULL;

DSBinaryTypeInteger ds_binary_type_uint_8t = NULL;
DSBinaryTypeInteger ds_binary_type_uint_16t = NULL;
DSBinaryTypeInteger ds_binary_type_uint_32t = NULL;
DSBinaryTypeInteger ds_binary_type_uint_64t = NULL;

DSBinaryTypeFloat ds_binary_type_float = NULL;
DSBinaryTypeFloat ds_binary_type_double = NULL;

DSBinaryTypeString ds_binary_type_utf8_string = NULL;

int dsBinaryTypeStandardInit() {
    ds_binary_type_int_8t = dsBinaryTypeIntegerCreate(NULL, 1, 1, ds_binary_configuration->endianess);
    ds_binary_type_int_16t = dsBinaryTypeIntegerCreate(NULL, 1, 2, ds_binary_configuration->endianess);
    ds_binary_type_int_32t = dsBinaryTypeIntegerCreate(NULL, 1, 4, ds_binary_configuration->endianess);
    ds_binary_type_int_64t = dsBinaryTypeIntegerCreate(NULL, 1, 8, ds_binary_configuration->endianess);

    ds_binary_type_uint_8t = dsBinaryTypeIntegerCreate(NULL, 0, 1, ds_binary_configuration->endianess);
    ds_binary_type_uint_16t = dsBinaryTypeIntegerCreate(NULL, 0, 2, ds_binary_configuration->endianess);
    ds_binary_type_uint_32t = dsBinaryTypeIntegerCreate(NULL, 0, 4, ds_binary_configuration->endianess);
    ds_binary_type_uint_64t = dsBinaryTypeIntegerCreate(NULL, 0, 8, ds_binary_configuration->endianess);

    ds_binary_type_float = dsBinaryTypeFloatCreate(NULL, ds_binary_configuration->float_format, ds_binary_configuration->float_size, ds_binary_configuration->endianess);
    ds_binary_type_double = dsBinaryTypeFloatCreate(NULL, ds_binary_configuration->float_format, ds_binary_configuration->double_size, ds_binary_configuration->endianess);

    ds_binary_type_utf8_string = dsBinaryTypeStringCreate(NULL, "UTF-8", ds_binary_configuration->endianess);
    
    return 0;    
}


void dsBinaryTypeStandardDestroy() {
    if (ds_binary_type_int_8t) dsBinaryTypeIntegerDestroy(ds_binary_type_int_8t);
    if (ds_binary_type_int_16t) dsBinaryTypeIntegerDestroy(ds_binary_type_int_16t);
    if (ds_binary_type_int_32t) dsBinaryTypeIntegerDestroy(ds_binary_type_int_32t);
    if (ds_binary_type_int_64t) dsBinaryTypeIntegerDestroy(ds_binary_type_int_64t);

    if (ds_binary_type_uint_8t) dsBinaryTypeIntegerDestroy(ds_binary_type_uint_8t);
    if (ds_binary_type_uint_16t) dsBinaryTypeIntegerDestroy(ds_binary_type_uint_16t);
    if (ds_binary_type_uint_32t) dsBinaryTypeIntegerDestroy(ds_binary_type_uint_32t);
    if (ds_binary_type_uint_64t) dsBinaryTypeIntegerDestroy(ds_binary_type_uint_64t);

    if (ds_binary_type_float) dsBinaryTypeFloatDestroy(ds_binary_type_float);
    if (ds_binary_type_double) dsBinaryTypeFloatDestroy(ds_binary_type_double);

    if (ds_binary_type_utf8_string) dsBinaryTypeStringDestroy(ds_binary_type_utf8_string);

    ds_binary_type_int_8t = NULL;
    ds_binary_type_int_16t = NULL;
    ds_binary_type_int_32t = NULL;
    ds_binary_type_int_64t = NULL;
    ds_binary_type_uint_8t = NULL;
    ds_binary_type_uint_16t = NULL;
    ds_binary_type_uint_32t = NULL;
    ds_binary_type_uint_64t = NULL;

    ds_binary_type_float = NULL;
    ds_binary_type_double = NULL;
    
    ds_binary_type_utf8_string = NULL;
}

DSBinaryTypeInteger dsBinaryTypeStandardCreateEnumerationFromArray(DSMallocContext mctx, const char **array, unsigned char copy_strings) {
    DSBinaryTypeInteger type;
    
    type = dsBinaryTypeIntegerCreate(mctx, 0, sizeof(int), ds_binary_configuration->endianess);
    if (type) dsBinaryTypeIntegerSetEnumerationFromArray(type, array, copy_strings);
    return type;
}


DSBinaryType dsBinaryTypeStandardCreate(DSMallocContext mctx, DSBinaryTypeStandard datatype, const void *attr) {
    switch (datatype) {
	case DS_BINARY_TYPE_STANDARD_UNSIGNED_CHAR:
	    return (DSBinaryType)dsBinaryTypeIntegerCreate(mctx, 0, sizeof(char), ds_binary_configuration->endianess);
	case DS_BINARY_TYPE_STANDARD_UNSIGNED_SHORT:
	    return (DSBinaryType)dsBinaryTypeIntegerCreate(mctx, 0, sizeof(short), ds_binary_configuration->endianess);
	case DS_BINARY_TYPE_STANDARD_UNSIGNED:
	    return (DSBinaryType)dsBinaryTypeIntegerCreate(mctx, 0, sizeof(int), ds_binary_configuration->endianess);
	case DS_BINARY_TYPE_STANDARD_UNSIGNED_LONG:
	    return (DSBinaryType)dsBinaryTypeIntegerCreate(mctx, 0, sizeof(long), ds_binary_configuration->endianess);
	case DS_BINARY_TYPE_STANDARD_CHAR:
	    return (DSBinaryType)dsBinaryTypeIntegerCreate(mctx, 1, sizeof(char), ds_binary_configuration->endianess);
	case DS_BINARY_TYPE_STANDARD_SHORT:
	    return (DSBinaryType)dsBinaryTypeIntegerCreate(mctx, 1, sizeof(short), ds_binary_configuration->endianess);
	case DS_BINARY_TYPE_STANDARD_INT:
	    return (DSBinaryType)dsBinaryTypeIntegerCreate(mctx, 1, sizeof(int), ds_binary_configuration->endianess);
	case DS_BINARY_TYPE_STANDARD_LONG:
	    return (DSBinaryType)dsBinaryTypeIntegerCreate(mctx, 1, sizeof(long), ds_binary_configuration->endianess);
	case DS_BINARY_TYPE_STANDARD_FLOAT:
	    return (DSBinaryType)dsBinaryTypeFloatCreate(mctx, ds_binary_configuration->float_format, ds_binary_configuration->float_size, ds_binary_configuration->endianess);
	case DS_BINARY_TYPE_STANDARD_DOUBLE:
	    return (DSBinaryType)dsBinaryTypeFloatCreate(mctx, ds_binary_configuration->float_format, ds_binary_configuration->double_size, ds_binary_configuration->endianess);
	case DS_BINARY_TYPE_STANDARD_STRING:
	    return (DSBinaryType)dsBinaryTypeStringCreate(mctx, attr?((const char *)attr):"UTF-8", ds_binary_configuration->endianess);
	case DS_BINARY_TYPE_STANDARD_QNAME:
	    return (DSBinaryType)dsBinaryTypeStringCreate(mctx, attr?((const char *)attr):"ASCII", ds_binary_configuration->endianess);
	case DS_BINARY_TYPE_STANDARD_ENUMERATION:
	    return (DSBinaryType)dsBinaryTypeStandardCreateEnumerationFromArray(mctx, (const char**)attr, 0);
	case DS_BINARY_TYPE_STANDARD_COMPLEX:
	case DS_BINARY_TYPE_STANDARD_STRINGLIST:
	case DS_BINARY_TYPE_STANDARD_XML:
	case DS_BINARY_TYPE_STANDARD_DATETIME:
	case DS_BINARY_TYPE_STANDARD_DATE:
	case DS_BINARY_TYPE_STANDARD_TIME:
	    return NULL;
	default:
	    return NULL;
    }
}


