#ifndef _DS_ASSERT_H
#define _DS_ASSERT_H

#include <stdlib.h>
#include <dslib_error.h>

#if defined (__GNUC__) && !defined (__STRICT_ANSI__) && !defined (__cplusplus)
# define DS_STMT_START		(void) __extension__ (
# define DS_STMT_END		)
#else
# if (defined (sun) || defined (__sun__))
#  define DS_STMT_START		if (1)
#  define DS_STMT_END		else (void)0
# else
#  define DS_STMT_START		do
#  define DS_STMT_END		while (0)
# endif
#endif

#define assert(expr)			DS_STMT_START{		\
    if (expr) { } else						\
    dsAssert("File: %s, Line: %d, Assertion Failed: (%s)\n",	\
	      __FILE__,						\
	      __LINE__,						\
	      #expr);			}DS_STMT_END

#define fatal_assert(expr)		DS_STMT_START{		\
    if (expr) { } else						\
    dsAssertFatal("File: %s, Line: %d, Assertion Failed: (%s)\n",	\
	      __FILE__,						\
	      __LINE__,						\
	      #expr);			}DS_STMT_END

#define DS_INITIALIZED 0x7E
#define ds_assert(a) DS_STMT_START{ fatal_assert(a); }DS_STMT_END
#define ds_assert_info(a) assert(a)
#define ds_assert_return(a) DS_STMT_START{ assert(a); if (!(a)) return; }DS_STMT_END
#define ds_assert_err(a,b) DS_STMT_START{ assert(a); if (!(a)) return b; }DS_STMT_END
#define ds_assert_default(a,b) DS_STMT_START{ assert(a); if (!(a)) return b; }DS_STMT_END
#define ds_assert_null(a)  DS_STMT_START{ assert(a); if (!(a)) return NULL; }DS_STMT_END
#define ds_assert_zero(a)  DS_STMT_START{ assert(a); if (!(a)) return 0; }DS_STMT_END

#define ds_assert_param(a) ds_assert_err(a, DS_ERR_INVALID_ARGUMENT)
#define ds_assert_options(a) ds_assert_err(a, DS_ERR_INVALID_OPTIONS)
#define ds_assert_init(a) ds_assert_err(a, DS_ERR_NOTINITIALIZED)

#define ds_assert_ctx(a) DS_STMT_START{ ds_assert_param(a); ds_assert_init(a->initialized == DS_INITIALIZED); }DS_STMT_END
#define ds_assert_complexctx(a,b) DS_STMT_START{ ds_assert_param(a); ds_assert_init(b); }DS_STMT_END
#define ds_assert_returnctx(a) DS_STMT_START{ ds_assert_return(a); ds_assert_return(a->initialized == DS_INITIALIZED); }DS_STMT_END
#define ds_assert_nullctx(a) DS_STMT_START{ ds_assert_null(a); ds_assert_null(a->initialized == DS_INITIALIZED); }DS_STMT_END
#define ds_assert_zeroctx(a) DS_STMT_START{ ds_assert_zero(a); ds_assert_zero(a->initialized == DS_INITIALIZED); }DS_STMT_END

#define ds_assert_module(a,b) DS_STMT_START{ ds_assert_param(a); ds_assert_init(b); }DS_STMT_END
#define ds_assert_module_return(a,b) ds_assert_return((a)&&(b))
#define ds_assert_module_default(a,b,c) ds_assert_default(((a)&&(b)),c)


void dsAssert(const char *format, ...);
void dsAssertFatal(const char *format, ...);

#endif /* _DS_ASSERT_H */
