#include <stdio.h>

#include <dslib.h>
#include <dsdebug.h>
#include <dsassert.h>

#include <dsbinaryconfiguration.h>

static struct DSBinaryConfigurationContextT ds_binary_configurationT;
DSBinaryConfigurationContext ds_binary_configuration = &ds_binary_configurationT;

static unsigned char IEEE_754_4_bigendian[] = 
  {0x3c, 0x00, 0x00, 0x00};
static unsigned char IEEE_754_4_littleendian[] = 
  {0x00, 0x00, 0x00, 0x3c};
static unsigned char IEEE_754_8_bigendian[] = 
  {0x3f, 0x80, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
static unsigned char IEEE_754_8_littleendian[] = 
  {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x80, 0x3f};
static unsigned char IEEE_754_8_mixedendian[] = 
  {0x00, 0x00, 0x80, 0x3f, 0x00, 0x00, 0x00, 0x00};
static unsigned char IEEE_754_16_bigendian[] = 
  {0x3f, 0xf8, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
static unsigned char IEEE_754_16_littleendian[] = 
  {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xf8, 0x3f};
static unsigned char IEEE_754_16_mixedendian[] = 
  {0x00, 0x00, 0xf8, 0x3f, 0x00, 0x00, 0x00, 0x00, 
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};

DSBinaryConfigurationContext dsBinaryConfigurationCreateContext(DSMallocContext mctx) {
    DSBinaryConfigurationContext ctx;

    ctx = (DSBinaryConfigurationContext)dsMalloc(mctx,sizeof(struct DSBinaryConfigurationContextT));
    if (ctx) {
	ctx->malloc=mctx;
//	dsBinaryConfigurationZeroContext(ctx);
    }
    return ctx;
}

int dsBinaryConfigurationInitContext(DSBinaryConfigurationContext ctx) {
    int i;
    uint16_t test=0xFF;
    double float_magic=0.0078125;
    char *alloc1, *alloc2;
    enum ds_binary_configuration_test_enumeration {
	DS_BINARY_CONFIGURATION_TEST_ENUMERATION 
    };
    struct {
	int8_t first;
	void *offset;
    } *struct_pointer = NULL;
    struct {
	int8_t first;
	int8_t offset;
    } *struct_int8 = NULL;
    struct {
	int8_t first;
	int16_t offset;
    } *struct_int16 = NULL;
    struct {
	int8_t first;
	int32_t offset;
    } *struct_int32 = NULL;
    struct {
	int8_t first;
	int64_t offset;
    } *struct_int64 = NULL;
    struct {
	int8_t first;
	enum ds_binary_configuration_test_enumeration  offset;
    } *struct_enum = NULL;
    struct {
	int8_t first;
	float offset;
    } *struct_float = NULL;
    struct {
	int8_t first;
	double offset;
    } *struct_double = NULL;
    struct {
	int8_t first;
	long double offset;
    } *struct_long_double = NULL;

    struct {
	int8_t first;
	struct {
	    int8_t first;
	} offset;
    } *struct_struct_min = NULL;
    struct {
	int8_t first;
	struct {
	    int8_t first;
	    long double second;
	} offset;
    } *struct_struct_combo = NULL;

    struct {
	int8_t first;
	struct {
	    long double first;
	} offset;
    } *struct_struct_max = NULL;
    

    ds_assert_param(ctx);
    
    ctx->pointer_size = sizeof(void*);
    ctx->float_size = sizeof(float);
    ctx->double_size = sizeof(double);
    ctx->long_double_size = sizeof(long double);
    ctx->enum_size = sizeof(enum ds_binary_configuration_test_enumeration);
    if ((uint8_t)test==0xFF) ctx->endianess = ENDIANESS_LITTLE_ENDIAN;
    else ctx->endianess = ENDIANESS_BIG_ENDIAN;
    
    ctx->float_endianess = 0xFF;
    switch (ctx->double_size) {
    case 4:
	if (memcmp(&float_magic, &IEEE_754_4_bigendian[0], 4) == 0) {
	    ctx->float_endianess = ENDIANESS_BIG_ENDIAN;
	} else if (memcmp(&float_magic, &IEEE_754_4_littleendian[0], 4) == 0) {
	    ctx->float_endianess = ENDIANESS_LITTLE_ENDIAN;
	}
	break;
    case 8:
	if (memcmp(&float_magic, &IEEE_754_8_bigendian[0], 8) == 0) {
	    ctx->float_endianess = ENDIANESS_BIG_ENDIAN;
	} else if (memcmp(&float_magic, &IEEE_754_8_littleendian[0], 8) == 0) {
	    ctx->float_endianess = ENDIANESS_LITTLE_ENDIAN;
	} else if (memcmp(&float_magic, &IEEE_754_8_mixedendian[0], 8) == 0) {
	    ctx->float_endianess = ENDIANESS_MIXED_ENDIAN;
	}
	break;
    case 16:
	if (memcmp(&float_magic, &IEEE_754_16_bigendian[0], 16) == 0) {
	    ctx->float_endianess = ENDIANESS_BIG_ENDIAN;
	} else if (memcmp(&float_magic, &IEEE_754_16_littleendian[0], 16) ==0){
	    ctx->float_endianess = ENDIANESS_LITTLE_ENDIAN;
	} else if (memcmp(&float_magic, &IEEE_754_16_mixedendian[0], 16) == 0){
	    ctx->float_endianess = ENDIANESS_MIXED_ENDIAN;
	}
	break;
    } 
    if (ctx->float_endianess==0xFF) return DS_ERR_BINARYCONFIGURATION_UNKNOWN_FLOAT_FORMAT;
    else ctx->float_format = FLOAT_FORMAT_IEEE;

/*  ToDo: Really I think zero is OK */
    alloc1 = (char*)malloc(1);
    alloc2 = (char*)malloc(1);
    if ((!alloc1)||(!alloc2)) return DS_ERR_BINARYCONFIGURATION_UNKNOWN_ALLOCATION_ALIGMENT;
    ctx->align_allocation = (unsigned short)(alloc2 - alloc1);
    free(alloc1);
    free(alloc2);

    for (i=0;i<6;i++) 
	if (ctx->align_allocation==(1<<i)) break;
    if (i==6) ctx->align_allocation = 0; //return ERR_OPC_DETERMINE_ALLOCATION_ALIGNMENT;
    
    ctx->align_pointer = (unsigned short)PTR2INT(&struct_pointer->offset);
    ctx->align_int8 = (unsigned short)PTR2INT(&struct_int8->offset);
    ctx->align_int16 = (unsigned short)PTR2INT(&struct_int16->offset);
    ctx->align_int32 = (unsigned short)PTR2INT(&struct_int32->offset);
    ctx->align_int64 = (unsigned short)PTR2INT(&struct_int64->offset);
    ctx->align_enum = (unsigned short)PTR2INT(&struct_enum->offset);
    ctx->align_float = (unsigned short)PTR2INT(&struct_float->offset);
    ctx->align_double = (unsigned short)PTR2INT(&struct_double->offset);
    ctx->align_long_double = (unsigned short)PTR2INT(&struct_long_double->offset);
    ctx->align_struct_min = (unsigned short)PTR2INT(&struct_struct_min->offset);
    ctx->align_struct_max = (unsigned short)PTR2INT(&struct_struct_max->offset);

    while (ctx->align_allocation%ctx->align_pointer) ctx->align_allocation*=2;
    while (ctx->align_allocation%ctx->align_int64) ctx->align_allocation*=2;
    while (ctx->align_allocation%ctx->align_double) ctx->align_allocation*=2;
    while (ctx->align_allocation%ctx->align_long_double) ctx->align_allocation*=2;
    while (ctx->align_allocation%ctx->align_struct_max) ctx->align_allocation*=2;

    if (ctx->align_struct_min!=ctx->align_struct_max) {
        if (((unsigned short)PTR2INT(&struct_struct_max->offset))==ctx->align_struct_min) {
	    ctx->align_struct_algo = ALIGN_STRUCT_FIRST;
        } else if (((unsigned short)PTR2INT(&struct_struct_max->offset))==ctx->align_struct_max) {
	    ctx->align_struct_algo = ALIGN_STRUCT_MAX;
	} else {
	    return DS_ERR_BINARYCONFIGURATION_UNKNOWN_STRUCT_ALIGNMENT;
	}
    } else ctx->align_struct_algo = ALIGN_STRUCT_FIRST;

    dsDebugServer(DS_DEBUG_NOTICE, "BinaryConfiguration: Sizes. Pointer: %u, Enumeration %u",ctx->pointer_size, ctx->enum_size);
    dsDebugServer(DS_DEBUG_NOTICE, "BinaryConfiguration: Sizes. Float: %u, Double: %u, Long Double: %u", ctx->float_size, ctx->double_size, ctx->long_double_size);
    dsDebugServer(DS_DEBUG_NOTICE, "BinaryConfiguration: Endianess. Integer: %u, Floats: %u",ctx->endianess,ctx->float_endianess);
    dsDebugServer(DS_DEBUG_NOTICE, "BinaryConfiguration: Align. Allocation: %u, Pointer: %u",ctx->align_allocation, ctx->align_pointer);
    dsDebugServer(DS_DEBUG_NOTICE, "BinaryConfiguration: Align. Int8: %u, Int16: %u, Int32: %u, Int64: %u, Enum: %u",ctx->align_int8,ctx->align_int16,ctx->align_int32,ctx->align_int64,ctx->align_enum);
    dsDebugServer(DS_DEBUG_NOTICE, "BinaryConfiguration: Align. Float: %u, Double: %u, Long Double: %u",ctx->align_float,ctx->align_double,ctx->align_long_double);
    dsDebugServer(DS_DEBUG_NOTICE, "BinaryConfiguration: Align. Structure Algo: %u, Structure Min: %u, Structure Max: %u",ctx->align_struct_algo,ctx->align_struct_min,ctx->align_struct_max);
    return 0;
}



void dsBinaryConfigurationFreeContext(DSBinaryConfigurationContext ctx) {
    ds_assert_return(ctx);
}

void dsBinaryConfigurationDestroyContext(DSBinaryConfigurationContext ctx) {
    dsBinaryConfigurationFreeContext(ctx);
    dsFree(ctx->malloc,ctx);
}

int dsBinaryConfigurationInit() {
    return dsBinaryConfigurationInitContext(&ds_binary_configurationT);
}

void dsBinaryConfigurationDestroy() {
    dsBinaryConfigurationFreeContext(&ds_binary_configurationT);
}

/*
main() {
    DSBinaryConfigurationContext ctx;
    
    ctx = dsBinaryConfigurationCreateContext(NULL);
    dsBinaryConfigurationInitContext(ctx);
}
*/
