#ifndef _DS_DEBUG_H
#define _DS_DEBUG_H

# include <stdarg.h>

# ifdef _DS_DEBUG_C
static char *ds_debug_priority[] = {
    "ASSERTION",
    "FATAL    ",
    "EMERGENCY",
    "ALERT    ",
    "CRITICAL ",
    "ERROR    ",
    "WARNING  ",
    "NOTICE   ",
    "INFO     ",
    "DEBUG    ",
    NULL
};
# endif /* _DS_DEBUG_C */

enum DSDebugPriorityT {
    DS_DEBUG_NOTHING = 0,
    DS_DEBUG_FATAL,
    DS_DEBUG_EMERGENCY,
    DS_DEBUG_ALERT,
    DS_DEBUG_CRITICAL,
    DS_DEBUG_ERROR,
    DS_DEBUG_WARNING,
    DS_DEBUG_NOTICE,
    DS_DEBUG_INFO,
    DS_DEBUG_DEBUG
};
typedef enum DSDebugPriorityT DSDebugPriority;
    
enum DSDebugMethodT {
    DS_DEBUG_METHOD_STDERR = 0,
    DS_DEBUG_METHOD_FILE,
    DS_DEBUG_METHOD_SOCKET
};
typedef enum DSDebugMethodT DSDebugMethod;

#define DS_DEBUG_LOG_TIME 1
#define DS_DEBUG_LOG_LEVEL 2

int dsDebugInit();
int dsDebugSetLevel(DSDebugPriority priority);
int dsDebugCheckLevel(DSDebugPriority priority);
int dsDebugGetLevel();
int dsDebugSetMethod(DSDebugMethod method, const void *attr);
int dsDebugSetOptions(unsigned int options);
int dsDebugSetLogTimeOption(unsigned char flag);
int dsDebugSetLogLevelOption(unsigned char flag);
void dsDebugDestroy();

void dsDebugServerAssert(const char *format, va_list ap);
void dsDebugModule(DSDebugPriority priority, const char *module, const void *ptr, const char *format, ...);
void dsDebugModuleError(DSDebugPriority priority, const char *module, const void *ptr, int err, const char *format, ...);
void dsDebugServer(DSDebugPriority priority, const char *format, ...);
void dsDebugRequest(DSDebugPriority priority, const char *format, ...);
void dsDebugServerError(DSDebugPriority priority, int err, const char *format, ...);
void dsDebugRequestError(DSDebugPriority priority, int err, const char *format, ...);
void dsDebugServerSystemError(DSDebugPriority priority, int err, const char *prefix, ...);
void dsDebugRequestSystemError(DSDebugPriority priority, int err, const char *prefix, ...);

DSDebugPriority dsDebugGetLevelByName(const char *level);
const char *dsDebugGetName(DSDebugPriority level);

#define ds_debug_memcpy(a,b,c) \
	dsDebugServer(DS_DEBUG_DEBUG, "MEMCPY File: %s, Line: %d. %p - %p (%lu)",__FILE__,__LINE__, a, (char*)a+c, c); \
	memcpy(a,b,c)

#endif /* _DS_DEBUG_H */
