#ifndef _DS_ERROR_H
#define _DS_ERROR_H

#include <dslocale.h>

typedef const struct DSErrorT *DSError;
typedef struct DSErrorT DSErrorS;

struct DSErrorT {
    unsigned long err;
    const dslchar str;
    unsigned long domain;
};

int dsErrorInit();
void dsErrorDestroy();
int dsErrorNewDomain(const char *domain);
int dsErrorAddErrors(const char *domain, DSError errors, unsigned long errors_number);

const char *dsErrorName(int err, const char *locale);


#ifdef _DS_ERROR_C
static char *ds_no_error=ds_string("Ok");
static char *ds_unknown_error=ds_string("Unknown Error");

static DSErrorS ds_errors[] = {
	{0x11, ds_string("DS_ERR_MEM")},
	{0x12, ds_string("DS_ERR_STACK")},
	{0x13, ds_string("DS_ERR_NOTFOUND")},
	{0x14, ds_string("DS_ERR_INVALID_PATH")},
	{0x15, ds_string("DS_ERR_INVALID_FILE")},
	{0x16, ds_string("DS_ERR_NOTIMPLEMENTED")},
	{0x17, ds_string("DS_ERR_NOTAVAILABLE")},
	{0x18, ds_string("DS_ERR_NOTSUPPORTED")},
	{0x19, ds_string("DS_ERR_NOTAPPLICABLE")},
	{0x1A, ds_string("DS_ERR_INVALID_ARGUMENT")},
	{0x1B, ds_string("DS_ERR_INVALID_OPTIONS")},
	{0x1C, ds_string("DS_ERR_NOTINITIALIZED")},
	{0x1D, ds_string("DS_ERR_INITIALIZED")},
	{0x1E, ds_string("DS_ERR_BUSY")},
	{0x1F, ds_string("DS_ERR_TIMEOUT")},
	{0x20, ds_string("DS_ERR_EMPTY")},
	{0x21, ds_string("DS_ERR_PERMISSION")},
	{0x22, ds_string("DS_ERR_EXIST")},
	{0x23, ds_string("DS_ERR_MULTIPLE")},
	{0x24, ds_string("DS_ERR_TOOLONG")},
	{0x25, ds_string("DS_ERR_INVALID_REQUEST")},
	{0x26, ds_string("DS_ERR_INVALID_RESPONSE")},
	{0x27, ds_string("DS_ERR_INTERNAL")},
	{0x28, ds_string("DS_ERR_RETRY")},
	{0x29, ds_string("DS_ERR_DISABLED")},
	{0x30, ds_string("DS_ERR_DISABLED_ITEM")},
	{0x31, ds_string("DS_ERR_INCOMPLETE")},
	{0x32, ds_string("DS_ERR_CREATE")},
	{0x50, ds_string("DS_ERR_SYSCALL_SIGACTION")},
	{0x51, ds_string("DS_ERR_SYSCALL_SIGPROCMASK")},
	{0x52, ds_string("DS_ERR_SYSCALL_SIGPROCMASK_BLOCK")},
	{0x53, ds_string("DS_ERR_SYSCALL_SIGPROCMASK_UNBLOCK")},
	{0x54, ds_string("DS_ERR_SYSCALL_SIGTIMEDWAIT")},
	{0x55, ds_string("DS_ERR_SYSCALL_NANOSLEEP")},
	{0x60, ds_string("DS_ERR_SYSCALL_PTHREAD_CREATE")},
	{0x61, ds_string("DS_ERR_SYSCALL_PTHREAD_SIGMASK_BLOCK")},
	{0x62, ds_string("DS_ERR_SYSCALL_PTHREAD_SIGMASK_UNBLOCK")},
	{0x63, ds_string("DS_ERR_SYSCALL_PTHREAD_KEY_CREATE")},
	{0x64, ds_string("DS_ERR_SYSCALL_PTHREAD_SET_SPECIFIC")},
	{0x65, ds_string("DS_ERR_SYSCALL_PTHREAD_GET_SPECIFIC")},
	{0x66, ds_string("DS_ERR_SYSCALL_PTHREAD_MUTEX_INIT")},
	{0x67, ds_string("DS_ERR_SYSCALL_PTHREAD_MUTEX_LOCK")},
	{0x68, ds_string("DS_ERR_SYSCALL_PTHREAD_MUTEX_UNLOCK")},
	{0x69, ds_string("DS_ERR_SYSCALL_PTHREAD_COND_INIT")},
	{0x6A, ds_string("DS_ERR_SYSCALL_PTHREAD_COND_SEND")},
	{0x6B, ds_string("DS_ERR_SYSCALL_PTHREAD_COND_WAIT")},
	{0x6C, ds_string("DS_ERR_SYSCALL_PTHREAD_SETSCHEDPARAM")},
	{0x6D, ds_string("DS_ERR_SYSCALL_FORK")},
	{0x90, ds_string("DS_ERR_SYSCALL_SELECT")},
	{0x91, ds_string("DS_ERR_SYSCALL_PSELECT")},
	{0x70, ds_string("DS_ERR_SYSCALL_LOCALTIME")},
	{0x71, ds_string("DS_ERR_SYSCALL_GMTIME")},
	{0x100, ds_string("DS_ERR_SOCKET")},
	{0x101, ds_string("DS_ERR_SOCKET_CREATE")},
	{0x102, ds_string("DS_ERR_SOCKET_BIND")},
	{0x104, ds_string("DS_ERR_SOCKET_LISTEN")},
	{0x105, ds_string("DS_ERR_SOCKET_SELECT")},
	{0x106, ds_string("DS_ERR_SOCKET_IOCTL")},
	{0x107, ds_string("DS_ERR_SOCKET_FCNTL")},
	{0x108, ds_string("DS_ERR_SOCKET_READ")},
	{0x109, ds_string("DS_ERR_SOCKET_WRITE")},
	{0x10A, ds_string("DS_ERR_SOCKET_ACCEPT")},
	{0x10B, ds_string("DS_ERR_SOCKET_CONNECT")},
	{0x10C, ds_string("DS_ERR_SOCKET_GETSOCKOPT")},
	{0x10D, ds_string("DS_ERR_SOCKET_SETSOCKOPT")},
	{0x10E, ds_string("DS_ERR_SOCKET_GETADDRINFO")},
	{0x10F, ds_string("DS_ERR_SOCKET_GETHOSTBYNAME")},
	{0x110, ds_string("DS_ERR_FS")},
	{0x111, ds_string("DS_ERR_FS_CHMOD")},
	{0x150, ds_string("DS_ERR_WINAPI")},
	{0x151, ds_string("DS_ERR_WINAPI_WINSOCK_INIT")},
	{0x152, ds_string("DS_ERR_WINAPI_EVENT_CREATE")},
	{0x153, ds_string("DS_ERR_WINAPI_PIPE_CREATE")},
	{0x154, ds_string("DS_ERR_WINAPI_PIPE_CONNECT")},
	{0x155, ds_string("DS_ERR_WINAPI_MUTEX_CREATE")},
	{0x156, ds_string("DS_ERR_WINAPI_GETHOSTBYNAME")},
	{0x200, ds_string("DS_ERR_BASETHREAD")},
	{0x201, ds_string("DS_ERR_BASETHREAD_CREATE")},
	{0x202, ds_string("DS_ERR_THREAD")},
	{0x203, ds_string("DS_ERR_THREAD_CREATE")},
	{0x204, ds_string("DS_ERR_THREAD_UNREGISTERED")},
	{0x205, ds_string("DS_ERR_THREAD_CANCELED")},
	{0x206, ds_string("DS_ERR_THREAD_RETURNCODE")},
	{0x207, ds_string("DS_ERR_DOUBLETHREAD")},
	{0x208, ds_string("DS_ERR_DOUBLETHREAD_CREATE")},
	{0x210, ds_string("DS_ERR_CONNECTIONSERVER")},
	{0x211, ds_string("DS_ERR_CONNECTIONSERVER_CREATE")},
	{0x212, ds_string("DS_ERR_CONNECTIONSERVER_ADDRINUSE")},
	{0x215, ds_string("DS_ERR_CONNECTION")},
	{0x216, ds_string("DS_ERR_CONNECTION_CREATE")},
	{0x217, ds_string("DS_ERR_CONNECTION_BROKENPIPE")},
	{0x218, ds_string("DS_ERR_CONNECTION_NOTCONNECTED")},
	{0x219, ds_string("DS_ERR_CONNECTION_CONNECTED")},
	{0x21A, ds_string("DS_ERR_CONNECTION_REFUSED")},
	{0x21B, ds_string("DS_ERR_CONNECTION_TIMEOUT")},
	{0x21C, ds_string("DS_ERR_CONNECTION_NETWORK")},
	{0x21D, ds_string("DS_ERR_CONNECTION_CONNECT")},
	{0x21E, ds_string("DS_ERR_CONNECTION_INVALID_SPEED")},
	{0x21F, ds_string("DS_ERR_CONNECTION_READ")},
	{0x220, ds_string("DS_ERR_CONNECTION_WRITE")},
	{0x221, ds_string("DS_ERR_CONNECTION_UNKNOWN_HOST")},
	{0x222, ds_string("DS_ERR_CONNECTION_ADDRESS_TOO_LONG")},
	{0x223, ds_string("DS_ERR_CONNECTION_ADDRESS_INVALID")},
	{0x230, ds_string("DS_ERR_SIGNAL")},
	{0x231, ds_string("DS_ERR_SIGNAL_CREATE")},
	{0x232, ds_string("DS_ERR_SIGNAL_EVENT")},
	{0x240, ds_string("DS_ERR_LOCK")},
	{0x241, ds_string("DS_ERR_LOCK_CREATE")},
	{0x243, ds_string("DS_ERR_LOCK_SELFLOCKED")},
	{0x244, ds_string("DS_ERR_LOCK_NOTLOCKED")},
	{0x245, ds_string("DS_ERR_CONDITION")},
	{0x246, ds_string("DS_ERR_CONDITION_CREATE")},
	{0x247, ds_string("DS_ERR_LOCK_BUSY")},
	{0x250, ds_string("DS_ERR_FILE")},
	{0x251, ds_string("DS_ERR_FILE_CREATE")},
	{0x252, ds_string("DS_ERR_FILE_NOTFOUND")},
	{0x253, ds_string("DS_ERR_FILE_OPEN")},
	{0x255, ds_string("DS_ERR_DIRECTORY")},
	{0x256, ds_string("DS_ERR_DIRECTORY_CREATE")},
	{0x257, ds_string("DS_ERR_DIRECTORY_NOTFOUND")},
	{0x258, ds_string("DS_ERR_DIRECTORY_NOTADIRECTORY")},
	{0x260, ds_string("DS_ERR_PLUGIN")},
	{0x261, ds_string("DS_ERR_PLUGIN_CREATE")},
	{0x262, ds_string("DS_ERR_PLUGIN_LOAD")},
	{0x263, ds_string("DS_ERR_PLUGIN_INFOFUNCTION")},
	{0x264, ds_string("DS_ERR_PLUGIN_INVALIDINFO")},
	{0x265, ds_string("DS_ERR_PLUGINSET")},
	{0x266, ds_string("DS_ERR_PLUGINSET_CREATE")},
	{0x267, ds_string("DS_ERR_PLUGINSET_PATH")},
	{0x270, ds_string("DS_ERR_TIMECONVERT")},
	{0x300, ds_string("DS_ERR_MALLOC")},
	{0x301, ds_string("DS_ERR_MALLOC_CREATE")},
	{0x310, ds_string("DS_ERR_ALLOCATOR")},
	{0x311, ds_string("DS_ERR_ALLOCATOR_CREATE")},
	{0x312, ds_string("DS_ERR_ALLOCATOR_INCOMPLETE")},
	{0x313, ds_string("DS_ERR_ALLOCATOR_ALLOCATE")},
	{0x314, ds_string("DS_ERR_ALLOCATOR_INVALID")},
	{0x315, ds_string("DS_ERR_ALLOCATOR_NOTOWNER")},
	{0x320, ds_string("DS_ERR_BUFFERPOOL")},
	{0x321, ds_string("DS_ERR_BUFFERPOOL_CREATE")},
	{0x322, ds_string("DS_ERR_BUFFERPOOL_GET")},
	{0x323, ds_string("DS_ERR_BUFFERPOOL_SIMPLEBUFFER")},
	{0x324, ds_string("DS_ERR_BUFFERPOOL_SIMPLEBUFFER_CREATE")},
	{0x330, ds_string("DS_ERR_BUFFER")},
	{0x331, ds_string("DS_ERR_BUFFER_CREATE")},
	{0x332, ds_string("DS_ERR_DATABUFFER")},
	{0x333, ds_string("DS_ERR_DATABUFFER_CREATE")},
	{0x334, ds_string("DS_ERR_DATABUFFER_OPTIONS")},
	{0x335, ds_string("DS_ERR_DATABUFFER_OPTIONS_MISSING_BUFFERPOOL")},
	{0x336, ds_string("DS_ERR_DATABUFFER_OPTIONS_INVALID_BUFFERPOOL")},
	{0x340, ds_string("DS_ERR_HOLDER")},
	{0x341, ds_string("DS_ERR_HOLDER_CREATE")},
	{0x342, ds_string("DS_ERR_HOLDER_BUSY")},
	{0x345, ds_string("DS_ERR_HOLDERBUCKET")},
	{0x346, ds_string("DS_ERR_HOLDERBUCKET_CREATE")},
	{0x350, ds_string("DS_ERR_FORMAT")},
	{0x351, ds_string("DS_ERR_FORMAT_CREATE")},
	{0x352, ds_string("DS_ERR_FORMAT_CREATE_INSTANCE")},
	{0x353, ds_string("DS_ERR_FORMAT_NATIVE_CREATE")},
	{0x354, ds_string("DS_ERR_FORMAT_NATIVE_CREATE_INSTANCE")},
	{0x400, ds_string("DS_ERR_HASHTABLE")},
	{0x401, ds_string("DS_ERR_HASHTABLE_CREATE")},
	{0x402, ds_string("DS_ERR_STRINGLIST")},
	{0x403, ds_string("DS_ERR_STRINGLIST_CREATE")},
	{0x404, ds_string("DS_ERR_DOUBLEHASH")},
	{0x405, ds_string("DS_ERR_DOUBLEHASH_CREATE")},
	{0x410, ds_string("DS_ERR_ENUMERATION")},
	{0x411, ds_string("DS_ERR_ENUMERATION_CREATE")},
	{0x412, ds_string("DS_ERR_ENUMERATION_INVALID_DESCRIPTION")},
	{0x415, ds_string("DS_ERR_TREENODE")},
	{0x416, ds_string("DS_ERR_TREENODE_CREATE")},
	{0x420, ds_string("DS_ERR_BINARYCONFIGURATION")},
	{0x421, ds_string("DS_ERR_BINARYCONFIGURATION_CREATE")},
	{0x422, ds_string("DS_ERR_BINARYCONFIGURATION_UNKNOWN_FLOAT_FORMAT")},
	{0x423, ds_string("DS_ERR_BINARYCONFIGURATION_UNKNOWN_ALLOCATION_ALIGMENT")},
	{0x424, ds_string("DS_ERR_BINARYCONFIGURATION_UNKNOWN_STRUCT_ALIGNMENT")},
	{0x430, ds_string("DS_ERR_DIME")},
	{0x431, ds_string("DS_ERR_DIME_CREATE")},
	{0x432, ds_string("DS_ERR_DIME_INVALID_TYPE")},
	{0x440, ds_string("DS_ERR_ERROR")},
	{0x441, ds_string("DS_ERR_ERROR_INIT")},
	{0x442, ds_string("DS_ERR_DEBUG")},
	{0x450, ds_string("DS_ERR_LOCALE")},
	{0x451, ds_string("DS_ERR_LOCALE_CREATE")},
	{0x452, ds_string("DS_ERR_LOCALE_DOMAIN")},
	{0x453, ds_string("DS_ERR_LOCALE_DOMAIN_REGISTER")},
	{0x454, ds_string("DS_ERR_LOCALE_DOMAIN_SET")},
	{0x500, ds_string("DS_ERR_XML")},
	{0x510, ds_string("DS_ERR_XML_PARSER")},
	{0x511, ds_string("DS_ERR_XML_PARSER_CREATE")},
	{0x520, ds_string("DS_ERR_XML_XSD")},
	{0x521, ds_string("DS_ERR_XML_XSD_CREATE")},
	{0x530, ds_string("DS_ERR_XML_XSLT")},
	{0x531, ds_string("DS_ERR_XML_XSLT_CREATE")},
	{0x540, ds_string("DS_ERR_XML_XPATH")},
	{0x541, ds_string("DS_ERR_XML_XPATH_CREATE")},
	{0x542, ds_string("DS_ERR_XML_XPATH_REGISTERNAMESPACE")},
	{0x543, ds_string("DS_ERR_XML_WRITER")},
	{0x550, ds_string("DS_ERR_XML_SCHEMA")},
	{0x551, ds_string("DS_ERR_XML_SCHEMA_CREATE")},
	{0x600, ds_string("DS_ERR_CONFIG")},
	{0x601, ds_string("DS_ERR_CONFIG_INVALID_CONFIG")},
	{0x602, ds_string("DS_ERR_CONIFG_NOTFOUND")},
	{0x700, ds_string("DS_ERR_DATA_INVALID_TYPE_SIZE")}
};
#define ds_errors_number 204

#endif /* _DS_ERROR_C */

#endif /* _DS_ERROR_H */
