#ifndef _DS_MODULE_H
#define _DS_MODULE_H

#define DS_MODULE(ctx) ((DSModule)ctx)

#include <dslib_error.h>

#include <dsdebug.h>
#include <dsmalloc.h>

typedef struct DSModule *DSModule;
typedef struct DSModuleNameT *DSModuleName;


struct DSModuleNameT {
    int id;
    const char *name;
};
typedef struct DSModuleNameT DSModuleNameS;

int dsModuleInit();
void dsModuleDestroy();
int dsModuleAddModules(DSModuleName names, size_t modules_number);
const char *dsModuleName(int id);

enum DSModuleFlagsT {
    DS_MODULE_FLAG_INITIALIZED = 1
};
typedef enum DSModuleFlagsT DSModuleFlags;

struct DSModule {
    int module_id;
    const char *module_name;
    DSMallocContext malloc;
    DSDebugPriority debug_limit;
    DSModuleFlags flags;
    
    DSModule owner;
};
typedef struct DSModule DSModuleS;

int dsModuleSetup(DSModule ctx, int module_id, const char *name);
void dsModuleClear(DSModule ctx);

int dsModuleSetDebugLimit(DSModule ctx, DSDebugPriority limit);
int dsModuleSetName(DSModule ctx, int module_id, const char *name);
int dsModuleSetFlags(DSModule ctx, DSModuleFlags flags);
int dsModuleResetFlags(DSModule ctx, DSModuleFlags flags);


DSModule dsModuleConstructor(DSMallocContext mctx, int module_id, size_t size);
DSModule dsModuleSubConstructor(DSModule module, DSCreateContextFunction creator);
void dsModuleDestructor(DSModule ctx, DSFreeContextFunction cleaner);

DSModule dsModuleCreate(DSMallocContext mctx, int module_id, const char *name, size_t size);

#define dsModuleNewModule(module, type) \
     (DS##type##Context)dsModuleSubConstructor(DS_MODULE(module), (DSCreateContextFunction)ds##type##CreateContext)

#define dsNewModule(type, mctx) \
     (DS##type##Context)dsModuleConstructor(mctx, DS_MODULE_##type, sizeof(DS##type##ContextS))

#define dsDestroyModule(type, ctx) \
    dsModuleDestructor(DS_MODULE(ctx), (DSFreeContextFunction)ds##type##FreeContext)

#define dsZeroModule(type, ctx) \
    memset(ctx, 0, sizeof(DS##type##ContextS))

#define dsModuleDebug(module, priority, ...) \
    dsDebugModule(((module)&&(priority<DS_MODULE(module)->debug_limit))?DS_MODULE(module)->debug_limit:priority, (module)?DS_MODULE(module)->module_name:NULL, module, __VA_ARGS__)

#define dsModuleError(module, err) \
    dsErrorModule(err, (module)?DS_MODULE(module)->module_id:0)

#define dsModuleMarkInitialized(ctx) \
    dsModuleSetFlags(DS_MODULE(ctx), DS_MODULE_FLAG_INITIALIZED)

#define ds_module_assert(module) \
    ds_assert_module(module, DS_MODULE(module)->flags&DS_MODULE_FLAG_INITIALIZED)

#define ds_module_assert_return(module) \
    ds_assert_module_return(module, DS_MODULE(module)->flags&DS_MODULE_FLAG_INITIALIZED)

#define ds_module_assert_default(module, default) \
    ds_assert_module_default(module, DS_MODULE(module)->flags&DS_MODULE_FLAG_INITIALIZED, default)


#endif /* _DS_MODULE_H */
