#include <stdio.h>

#include <dsthread.h>
#include <dslock.h>
#include <dssignal.h>
#include <dsdebug.h>

static char RunFlag = 1;

static void signal_exit_handler(int signo) { 
    RunFlag = 0; 
}

DSLockContext lock;


int Reader1(DSThreadContext ctx) {
    
    while (dsThreadCheckStatusBitsOff((DSBaseThreadContext)ctx,DS_THREAD_TERMINATE_FLAG)) {
	dsDebugServer(DS_DEBUG_CRITICAL, "Reader1. Trying");
	dsLockRead(lock);
	dsDebugServer(DS_DEBUG_CRITICAL, "Reader1. Locked");
	dsWaitSeconds(2);
	dsDebugServer(DS_DEBUG_CRITICAL, "Reader1. UnLocked");
	dsLockReturn(lock);

	dsWaitSeconds(2);
    }

}

int Reader2(DSThreadContext ctx) {

    dsWaitSeconds(1);
    
    while (dsThreadCheckStatusBitsOff((DSBaseThreadContext)ctx,DS_THREAD_TERMINATE_FLAG)) {
	dsDebugServer(DS_DEBUG_CRITICAL, "Reader2. Trying");
	dsLockRead(lock);
	dsDebugServer(DS_DEBUG_CRITICAL, "Reader2. Locked");
	dsWaitSeconds(4);
	dsDebugServer(DS_DEBUG_CRITICAL, "Reader2. UnLocked");
	dsLockReturn(lock);
	dsWaitSeconds(2);
    }
}

int Writer(DSThreadContext ctx) {
    while (dsThreadCheckStatusBitsOff((DSBaseThreadContext)ctx,DS_THREAD_TERMINATE_FLAG)) {
	dsWaitSeconds(2);
	
	dsDebugServer(DS_DEBUG_CRITICAL, "Writer. Trying");
	dsLockWrite(lock);
	dsDebugServer(DS_DEBUG_CRITICAL, "Writer. Locked");
	dsWaitSeconds(4);
	dsDebugServer(DS_DEBUG_CRITICAL, "Writer. UnLocked");
	dsLockReturn(lock);
    }
}



main() {
    int err;

    DSThreadContext r1, r2, w;
    DSDuration duration;
    
    
    const char *str;
    
    dsLibInit();

    dsSignalINTSetAction(signal_exit_handler);
    dsSignalINTBlock();
    
    duration = dsDurationCreate(NULL);
    dsDurationImportSeconds(duration, 3600);

    lock = dsLockCreateContext(NULL);
    dsLockInitContext(lock);

    r1 = dsThreadCreateContext(NULL);
    r2 = dsThreadCreateContext(NULL);
    w = dsThreadCreateContext(NULL);

    dsThreadInitContext(r1);
    dsThreadInitContext(r2);
    dsThreadInitContext(w);


    dsThreadRun(r1, (DSThreadFunction)Reader1, r1);
    dsThreadRun(r2, (DSThreadFunction)Reader2, r2);
    dsThreadRun(w, (DSThreadFunction)Writer, w);

    
    dsSignalINTUnblock();

    dsWait(duration);
    
    dsSignalINTBlock();


    dsThreadDestroyContext(r1);
    dsThreadDestroyContext(r2);
    dsThreadDestroyContext(w);

    dsLockDestroyContext(lock);
    
    dsDurationDestroy(duration);    
    
    dsLibFree();
}
