#ifndef _DS_BASE_THREAD_H
#define _DS_BASE_THREAD_H

#include <dslib.h>

#if defined(HAVE_SETJMP_H)
# include <setjmp.h>
#endif /* defined(HAVE_SETJMP_H) */
#if defined(HAVE_PTHREAD_H)
# include <pthread.h>
#elif defined(HAVE_WIN32_THREADS)
# include <windows.h>
#else /* defined(HAVE_PTHREAD_H) */
# error "Missing threading library!"
#endif /* defined(HAVE_PTHREAD_H) */

typedef struct DSThreadHandleT *DSThreadHandle;
typedef struct DSThreadPropertiesT *DSThreadProperties;
typedef struct DSBaseThreadContextT *DSBaseThreadContext;
typedef int (*DSThreadFunction)(void*);

#include <dslock.h>
#include <dscondition.h>
#include <dsmalloc.h>
#include <dssignal.h>

#define DS_THREAD_TERMINATE_FLAG 1

enum DSThreadSchedulerT {
    DS_THREAD_SCHEDULER_STANDARD = 0,
    DS_THREAD_SCHEDULER_INHERIT,
    DS_THREAD_SCHEDULER_REALTIME,
    DS_THREAD_SCHEDULER_REALTIME_RR,
    DS_THREAD_SCHEDULER_IDLE,
    DS_THREAD_SCHEDULER_MAX
};
typedef enum DSThreadSchedulerT DSThreadScheduler;
typedef int DSThreadPriority;
typedef void *DSThreadSchedulerProperties;

struct DSThreadHandleT {
#if defined(HAVE_PTHREAD_H)
    pthread_t id;
#else /* defined(HAVE_WIN32_THREADS) */
    DWORD id;
#endif /* defined(HAVE_PTHREAD_H) */
};
typedef struct DSThreadHandleT DSThreadHandleS;

struct DSThreadPropertiesT {
    DSMallocContext malloc;
    void *attr;
    DSThreadFunction function;
};
typedef struct DSThreadPropertiesT DSThreadPropertiesS;

struct DSThreadRealHandleT {
    DSThreadHandleS handle;
    DSThreadPropertiesS prop;
    DSBaseThreadContext ctx;
    
    DSThreadPriority priority;
    
#ifdef _WIN32
    HANDLE thread_handle;
    HANDLE winsignals[DS_SIGNAL_NUMBER];
#else
# ifndef HAVE_PSELECT_SYSCALL
    unsigned char jmpstatus;
#  ifdef HAVE_SETJMP_H
    sigjmp_buf jmpbuf;
#  endif /* HAVE_SETJMP_H */
# endif /* HAVE_PSELECT_SYSCALL */
#endif /* _WIN32 */
};
typedef struct DSThreadRealHandleT *DSThreadRealHandle;
typedef struct DSThreadRealHandleT DSThreadRealHandleS;


struct DSBaseThreadContextT {
    struct DSConditionContextT cond;
    DSLockContext lock;

    unsigned int handles;
    DSThreadRealHandle *handle;
    
    DSThreadScheduler sched;
    DSThreadPriority priority;

    dword status;

    unsigned char initialized;
};

int dsThreadInit();
void dsThreadDestroy();
int dsThreadRegister(DSThreadRealHandle handle);
int dsThreadUnregister(DSThreadRealHandle handle);

int dsThreadGetCurrentHandle(DSThreadHandle handle);
DSBaseThreadContext dsThreadFindCurrentContext();
DSThreadRealHandle dsThreadFindCurrentHandle();
int dsThreadHandleCmp(DSThreadHandle handle1, DSThreadHandle handle2);
int dsThreadSignal(DSThreadHandle handle, DSSignal sig); 

#define DS_BASE_THREAD_ID_SIZE 64
char *dsThreadGetID(char *thread_id); /* May be removed in future */


int dsBaseThreadInitContext(DSBaseThreadContext ctx);
void dsBaseThreadFreeContext(DSBaseThreadContext ctx);
void dsBaseThreadZeroContext(DSBaseThreadContext ctx);

int dsThreadRegisterHandles(DSBaseThreadContext ctx, unsigned handles, ...);
int dsThreadUnregisterHandles(DSBaseThreadContext ctx);

int dsThreadSetScheduler(DSBaseThreadContext ctx, DSThreadScheduler sched, DSThreadSchedulerProperties props);
int dsThreadSetPriority(DSBaseThreadContext ctx, DSThreadPriority prio);
DSThreadPriority dsThreadGetMinimalPriority(DSBaseThreadContext ctx);
DSThreadPriority dsThreadGetMaximalPriority(DSBaseThreadContext ctx);
DSThreadPriority dsThreadGetKernelPriority(DSBaseThreadContext ctx);
DSThreadPriority dsThreadGetSystemPriority(DSBaseThreadContext ctx);
DSThreadPriority dsThreadGetUserPriority(DSBaseThreadContext ctx);
int dsThreadAlterPriority(DSThreadRealHandle handle, DSThreadPriority prio);
int dsThreadIncreasePriority(DSThreadRealHandle handle);
int dsThreadDecreasePriority(DSThreadRealHandle handle);
int dsThreadAlterSelfPriority(DSThreadPriority prio);
int dsThreadIncreaseSelfPriority();
int dsThreadDecreaseSelfPriority();

void dsThreadSetStatus(DSBaseThreadContext ctx, dword bits);
void dsThreadSetStatusBits(DSBaseThreadContext ctx, dword bits);
void dsThreadResetStatusBits(DSBaseThreadContext ctx, dword bits);

/* Theoretically imideatly after status checked in Plugin Check/Wait functions it can be
changed from another thread. Realy it must not be the case, becuase we have only one 
controller thread, but.... If You need to be aware of such situation it's required to 
reimplement (or better add new) functions presesting write lock after return. And only 
from controller return lock */

int dsThreadCheckStatusBitsOn(DSBaseThreadContext ctx, dword bits);
int dsThreadCheckStatusBitsOff(DSBaseThreadContext ctx, dword bits);
int dsThreadCheckStatusBits(DSBaseThreadContext ctx, dword onbits, dword offbits);
int dsThreadCheckStatusSomeBitsOn(DSBaseThreadContext ctx, dword bits);
int dsThreadCheckStatusSomeBitsOff(DSBaseThreadContext ctx, dword bits);
int dsThreadCheckStatusSomeBits(DSBaseThreadContext ctx, dword onbits, dword offbits);
void dsThreadWaitStatusBitsOn(DSBaseThreadContext ctx, dword bits);
void dsThreadWaitStatusBitsOff(DSBaseThreadContext ctx, dword bits);
void dsThreadWaitStatusBits(DSBaseThreadContext ctx, dword onbits, dword offbits);
void dsThreadWaitStatusSomeBitsOn(DSBaseThreadContext ctx, dword bits);
void dsThreadWaitStatusSomeBitsOff(DSBaseThreadContext ctx, dword bits);
void dsThreadWaitStatusSomeBits(DSBaseThreadContext ctx, dword onbits, dword offbits);

void dsThreadTerminate(DSBaseThreadContext ctx);
int dsThreadIsTerminated(DSBaseThreadContext ctx);

#ifdef _WIN32
DWORD WINAPI dsThreadGo(DSThreadRealHandle ctx);
#else /* _WIN32 */
int *dsThreadReturnInt(DSThreadRealHandle ctx, int err);
int *dsThreadGo(DSThreadRealHandle ctx);
#endif /* _WIN32 */

#define dsThreadSendCondition(ctx) dsConditionBroadcast((DSConditionContext)ctx)

#endif /* _DS_BASE_THREAD_H */
