#ifndef _DS_CONDITION_H
#define _DS_CONDITION_H

#include <dslib.h>

#if defined(HAVE_PTHREAD_H)
# include <pthread.h>
#elif defined(HAVE_WIN32_THREADS)
# include <windows.h>
#else /* defined(HAVE_PTHREAD_H) */
# error "Missing threading library!"
#endif /* defined(HAVE_PTHREAD_H) */

#include <time.h>
#include <dsmalloc.h>

#if defined(HAVE_PTHREAD_H)
#else /* defined(HAVE_WIN32_THREADS) */
#include <dslock.h>
#endif /* defined(HAVE_PTHREAD_H) */

struct DSConditionContextT {
    DSMallocContext malloc;
#if defined(HAVE_PTHREAD_H)
    pthread_cond_t cond;
    pthread_mutex_t mutex;
#else /* defined(HAVE_WIN32_THREADS) */
    HANDLE cond[2];
    DSLockContext mutex;
    HANDLE okey_cond;
    DSLockContext waiting_lock;
    unsigned long waiting;
    unsigned char prelocked;
#endif /* defined(HAVE_PTHREAD_H) */
    unsigned char initialized;
};

typedef struct DSConditionContextT *DSConditionContext;

void dsConditionZeroContext(DSConditionContext ctx);
DSConditionContext dsConditionCreateContext(DSMallocContext mctx);
int dsConditionInitContext(DSConditionContext ctx);
void dsConditionFreeContext(DSConditionContext ctx);
void dsConditionDestroyContext(DSConditionContext ctx);

void dsConditionBroadcast(DSConditionContext ctx);
void dsConditionSend(DSConditionContext ctx);
void dsConditionPreLock(DSConditionContext ctx);
void dsConditionUnLock(DSConditionContext ctx);
void dsConditionWait(DSConditionContext ctx);
void dsConditionWaitLocked(DSConditionContext ctx);

#include <dstime.h>
int dsConditionTimedWait(DSConditionContext ctx, DSDuration timeout);
int dsConditionTimedWaitLocked(DSConditionContext ctx, DSDuration timeout);
int dsConditionTimedWaitUntil(DSConditionContext ctx, DSTime time);
int dsConditionTimedWaitLockedUntil(DSConditionContext ctx, DSTime time);

#endif /* _DS_CONDITION_H */
