#ifndef _DS_CONNECTION_SERVER_H
#define _DS_CONNECTION_SERVER_H

enum DSConnectionMaskT {
    DS_CONNECTION_LOCAL = 1,
    DS_CONNECTION_NETWORK,
    DS_CONNECTION_BOTH
};
typedef enum DSConnectionMaskT DSConnectionMask;
    
typedef struct DSConnectionServerContextT *DSConnectionServerContext;


#include <dslib.h>
#include <dssoftsignal.h>

#ifdef HAVE_NETINET_IN_H
# include <netinet/in.h>
#endif /* HAVE_NETINET_IN_H */

#include <dsmalloc.h>
#include <dsconnection.h>
#include <dssignal.h>


struct DSConnectionServerOptionsT {
    char *local_name;
    char *local_path;
    char *local_extension;
    in_port_t network_port;

    DSSignalMaskS sigmask;
};
typedef struct DSConnectionServerOptionsT *DSConnectionServerOptions;
typedef struct DSConnectionServerOptionsT DSConnectionServerOptionsS;


struct DSConnectionServerContextT {
    DSMallocContext malloc;

    char *local_name;

    DSSignalMaskS sigmask;
    DSSoftSignalList softsig;

#ifdef _WIN32
    LPOVERLAPPED oOverlap;
    HANDLE hEventLocal;
    HANDLE local_handle;
    int local_ready;
# ifdef HAVE_SOCKETS
    HANDLE hEventNetwork;
    SOCKET network_handle;
# endif /* HAVE_SOCKETS */
#else /* _WIN32 */
    int local_socket;
    int network_socket;
#endif /* _WIN32 */
};
typedef struct DSConnectionServerContextT DSConnectionServerContextS;

int dsConnectionServerInit();
void dsConnectionServerDestroy();


DSConnectionServerOptions dsConnectionServerGetDefaultOptions();
int dsConnectionServerInitDefaultOptions(DSConnectionServerOptions options);
void dsConnectionServerDestroyOptions(DSConnectionServerOptions options);

int dsConnectionServerOptionsEnableLocal(DSConnectionServerOptions options, const char *local_name);
int dsConnectionServerOptionsSetLocalPath(DSConnectionServerOptions options, const char *local_path);
int dsConnectionServerOptionsSetLocalExtension(DSConnectionServerOptions options, const char *local_extension);
int dsConnectionServerOptionsEnableNetwork(DSConnectionServerOptions options, in_port_t network_port);
int dsConnectionServerOptionsSetSignalMask(DSConnectionServerOptions options, DSSignalMask sigmask);

DSConnectionServerContext dsConnectionServerCreateContext(DSMallocContext mctx);
int dsConnectionServerInitContext(DSConnectionServerContext ctx, DSConnectionServerOptions options);
void dsConnectionServerFreeContext(DSConnectionServerContext ctx);
void dsConnectionServerDestroyContext(DSConnectionServerContext ctx);

DSConnectionContext dsConnectionServerAcceptLocal(DSConnectionServerContext ctx);
DSConnectionContext dsConnectionServerAcceptNetwork(DSConnectionServerContext ctx);
DSConnectionContext dsConnectionServerAccept(DSConnectionServerContext ctx);


#endif /* _DS_CONNECTION_SERVER_H */
