#ifndef _DS_DATAGRAM_SERVER_H
#define _DS_DATAGRAM_SERVER_H

#include <dsmalloc.h>
#include <dslock.h>
#include <dsconnection.h>

typedef struct DSConnectionListItemT *DSConnectionListItem;
typedef struct DSConnectionListT *DSConnectionList;
typedef struct DSDatagramServerContextT *DSDatagramServerContext;

struct DSConnectionListItemT {
    DSTimeS expire;
    DSConnectionAddress addr;
    DSConnectionListItem next;

    size_t filter_size;
    char *filter;
};
typedef struct DSConnectionListItemT DSConnectionListItemS;

/* subscript only to messages starting from filter (starting not equal) */
int dsConnectionListItemSetFilter(DSConnectionList list, DSConnectionListItem item, size_t filter_size, const char *filter);

struct DSConnectionListT {
    DSMallocContext malloc;
    DSConnectionListItem first;
    DSLockContext lock;
};
typedef struct DSConnectionListT DSConnectionListS;

DSConnectionList dsConnectionListCreate(DSMallocContext mctx);
void dsConnectionListDestroy(DSConnectionList list);
void dsConnectionListClear(DSConnectionList list);
DSConnectionListItem dsConnectionListAdd(DSConnectionList list, DSConnectionAddress *addr, DSDuration livetime);
void dsConnectionListManage(DSConnectionList list);
int dsConnectionListIsEmpty(DSConnectionList list);

struct DSDatagramServerContextT {
    DSMallocContext malloc;
    DSConnectionContext con;
    DSConnectionList clients;
    DSDurationS expiration;
    unsigned long manager_iterations;

    DSLockContext write_lock;
    DSDatagramServerContext master;
    unsigned long refs;
    
    size_t buffer_start;
    size_t buffer_size;
    void *buffer;
};

typedef struct DSDatagramServerContextT DSDatagramServerContextS;

/*
DSDatagramServer
================
    Multiple writes can be performed simultaneously.

*/

DSDatagramServerContext dsDatagramServerCreateContext(DSMallocContext mctx);
int dsDatagramServerInitContext(DSDatagramServerContext ctx, DSSignalMask sigmask, in_port_t port);
int dsDatagramServerInitSlaveContext(DSDatagramServerContext ctx, DSDatagramServerContext master);
void dsDatagramServerFreeContext(DSDatagramServerContext ctx);
void dsDatagramServerDestroyContext(DSDatagramServerContext ctx);

int dsDatagramServerRef(DSDatagramServerContext ctx);
int dsDatagramServerUnref(DSDatagramServerContext ctx);

int dsDatagramServerSetExpirationDuration(DSDatagramServerContext ctx, DSDuration duration);
int dsDatagramServerSetMaxManagerIterations(DSDatagramServerContext ctx, unsigned long iterations);

int dsDatagramServerManage(DSDatagramServerContext ctx);
int dsDatagramServerManageClients(DSDatagramServerContext ctx, const char *data, size_t len);
int dsDatagramServerRead(DSDatagramServerContext ctx, DSConnectionAddress *addr, char *data, size_t len, DSConnectionFlags flags, size_t *wbytes);
int dsDatagramServerGetWriteLock(DSDatagramServerContext ctx);
int dsDatagramServerReturnWriteLock(DSDatagramServerContext ctx);
int dsDatagramServerWrite(DSDatagramServerContext ctx, DSConnectionAddress *addr, const char *data, size_t len, DSConnectionFlags flags, size_t *wbytes);
int dsDatagramServerBroadcast(DSDatagramServerContext ctx, DSConnectionList list, const char *title, const char *data, size_t len, DSConnectionFlags flags, size_t *wbytes);
int dsDatagramServerBroadcastIteration(DSDatagramServerContext ctx, DSConnectionList list, const char *title, const char *data, size_t len, DSConnectionFlags flags, size_t *wbytes);
int dsDatagramServerBroadcastStart(DSDatagramServerContext ctx, DSConnectionList list);
int dsDatagramServerBroadcastDone(DSDatagramServerContext ctx, DSConnectionList list);
int dsDatagramServerBroadcastCancel(DSDatagramServerContext ctx, DSConnectionList list);

int dsDatagramServerDrop(DSDatagramServerContext ctx);
int dsDatagramServerAddClient(DSDatagramServerContext ctx, DSConnectionAddress *addr);

int dsDatagramServerIsEmpty(DSDatagramServerContext ctx);

#endif /* _DS_DATAGRAM_SERVER_H */
