#ifndef _DS_DIRECTORY_H
#define _DS_DIRECTORY_H

#include <dslib.h>

#include <time.h>

#if defined(HAVE_DIRENT_H)
#include <dirent.h>
#elif defined(_WIN32)
#include <windows.h>
#else
#error "Can't find FileSystem Browsing Routines"
#endif /* HAVE_DIRENT_H */

#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif /* HAVE_SYS_TYPES_H */

#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif /* HAVE_SYS_STAT_H */


struct DSFileContextT {
    char *fullname;
    char *name;
    char *path;
#if defined(HAVE_DIRENT_H)
    struct stat filestat;
#else
    WIN32_FIND_DATA FindFileData;
    HANDLE hFind;
#endif /* HAVE_DIRENT_H */
    unsigned char initialized;
};
typedef struct DSFileContextT *DSFileContext;
typedef struct DSFileContextT DSFileContextS;


struct DSDirectoryContextT {
    DSFileContextS current;
    char *path;
    char *mask;

#if defined(HAVE_DIRENT_H)
    DIR *dir;
#else
    unsigned char status;
    HANDLE hFind;
#endif /* HAVE_DIRENT_H */

    unsigned char initialized;
};

typedef struct DSDirectoryContextT *DSDirectoryContext;
typedef struct DSDirectoryContextT DSDirectoryContextS;


int dsFileOpen(DSFileContext ctx, const char *path, const char *name);
void dsFileClose(DSFileContext ctx);
const char *dsFileGetName(const DSFileContext ctx);
const char *dsFileGetPath(const DSFileContext ctx);
const char *dsFileGetFullName(const DSFileContext ctx);
time_t dsFileGetModificationTime(const DSFileContext ctx);

int dsDirectoryOpen(DSDirectoryContext dir, const char *path, const char *mask);
const DSFileContext dsDirectoryRead(DSDirectoryContext dir);
void dsDirectoryClose(DSDirectoryContext dir);

time_t dsDirectoryGetFileModificationTime(const char *fn);

int dsDirectoryCreate(const char *path);
int dsFileCreateDirectory(const char *fn);
char *dsDirectoryConcatenate(DSMallocContext mctx, const char *dir1, const char *dir2);
int dsDirectoryParse(DSMallocContext mctx, char **dir);
int dsDirectoryParseFrom(DSMallocContext mctx, char **dir, const char *orig);


#endif /* _DS_DIRECTORY_H */
