#include <dslib.h>

#include <dsassert.h>

#include <dslib.h>
#include <dsmalloc.h>
#include <dsdebug.h>
#include <dssignal.h>
#include <dstools.h>

#include <dsdoublethread.h>

void dsDoubleThreadZeroContext(DSDoubleThreadContext ctx) {
    dsBaseThreadZeroContext((DSBaseThreadContext)ctx);
/*
    ThreadHandleZero(ctx->thread1);
    ThreadHandleZero(ctx->thread2);
*/
    ctx->thread1.prop.function = NULL;
    ctx->thread2.prop.function = NULL;
}

DSDoubleThreadContext opcDoubleThreadCreateContext(DSMallocContext mctx) {
    DSDoubleThreadContext ctx;

    ctx = (DSDoubleThreadContext)dsMalloc(mctx,sizeof(struct DSDoubleThreadContextT));
    dsDebugServer(DS_DEBUG_NOTICE, "DoubleThread %p: CreateContext", ctx);    
    if (ctx) {
	ctx->base.cond.malloc=mctx;
	dsDoubleThreadZeroContext(ctx);
    }
    return ctx;

}

int dsDoubleThreadInitContext(DSDoubleThreadContext ctx) {
    int err;
    ds_assert_param(ctx);

    dsDebugServer(DS_DEBUG_NOTICE, "DoubleThread %p: InitContext", ctx);    
    dsDoubleThreadZeroContext(ctx);

    err = dsBaseThreadInitContext((DSBaseThreadContext)ctx);
    if (!err) err = dsThreadRegisterHandles((DSBaseThreadContext)ctx, 2, &ctx->thread1, &ctx->thread2);
    if (err) {
	dsDoubleThreadFreeContext(ctx);
	return err;
    }

    ctx->base.initialized = 1;
    return 0;
}

int dsDoubleThreadFreeContext(DSDoubleThreadContext ctx) {
    int err;
    ds_assert_param(ctx);

    dsDebugServer(DS_DEBUG_NOTICE, "DoubleThread %p: FreeContext",ctx);    

    
    if (ctx->base.initialized) err=dsDoubleThreadTerminateAndWait(ctx);
    else err=0;

/*    
    ThreadHandleFree(&ctx->thread1);
    ThreadHandleFree(&ctx->thread2);
*/
    dsBaseThreadFreeContext((DSBaseThreadContext)ctx);
    
    dsDoubleThreadZeroContext(ctx);
    
    return err;
}

int dsDoubleThreadDestroyContext(DSDoubleThreadContext ctx) {
    int err;
    ds_assert_param(ctx);

    dsDebugServer(DS_DEBUG_NOTICE, "DoubleThread %p: DestroyContext",ctx);    

    err=dsDoubleThreadFreeContext(ctx);
    dsFree(ctx->base.cond.malloc,ctx);
    return err;
}


void dsDoubleThreadNotify(DSDoubleThreadContext ctx) {
    ds_assert_return(ctx);
    ds_assert_return(ctx->base.initialized);
    
    dsSignalSendTo(&ctx->thread1.handle, DS_SIGNAL_NOTIFY);
    dsSignalSendTo(&ctx->thread2.handle, DS_SIGNAL_NOTIFY);
}

int dsDoubleThreadRun(DSDoubleThreadContext ctx, DSThreadFunction thread_function1, DSThreadFunction thread_function2, void *attr) {
    int err;

#ifdef HAVE_PTHREAD_H
    int custom = 0;
    pthread_attr_t atrd;
    struct sched_param ptrd;
#endif /* HAVE_PTHREAD_H */
    
    ds_assert_param(ctx);
    ds_assert_param(thread_function1);
    ds_assert_param(thread_function2);
    ds_assert_param(ctx->base.initialized);

    ds_assert_param(ctx->thread1.prop.function == NULL);

    ctx->thread1.prop.malloc = ctx->base.cond.malloc;
    ctx->thread2.prop.malloc = ctx->base.cond.malloc;
    ctx->thread1.prop.function = thread_function1;
    ctx->thread2.prop.function = thread_function2;
    ctx->thread1.prop.attr = attr;
    ctx->thread2.prop.attr = attr;
    ctx->thread1.priority = ctx->base.priority;
    ctx->thread2.priority = ctx->base.priority;
    
#if defined(HAVE_PTHREAD_H)
    switch(ctx->base.sched) {
	    case DS_THREAD_SCHEDULER_INHERIT:
		pthread_attr_init(&atrd);
		pthread_attr_setinheritsched(&atrd, PTHREAD_INHERIT_SCHED);
		custom = 1;
	    break;
	    case DS_THREAD_SCHEDULER_REALTIME:
		pthread_attr_init(&atrd);
		pthread_attr_setschedpolicy(&atrd,SCHED_FIFO);
    		ptrd.sched_priority=ctx->base.priority;
    		pthread_attr_setschedparam(&atrd, &ptrd);
		custom = 1;
	    break;
	    case DS_THREAD_SCHEDULER_REALTIME_RR:
		pthread_attr_init(&atrd);
		pthread_attr_setschedpolicy(&atrd,SCHED_RR);
    		ptrd.sched_priority=ctx->base.priority;
    		pthread_attr_setschedparam(&atrd, &ptrd);
		custom = 1;
	    break;
    }

    err = pthread_create(&ctx->thread1.handle.id,custom?&atrd:NULL,(void*(*)(void*))dsThreadGo,(void*)&ctx->thread1);
    if (err) {
	err = DS_ERR_SYSCALL_PTHREAD_CREATE;
	if (custom) pthread_attr_destroy(&atrd);
    }
#else /* defined(HAVE_WIN32_THREADS) */
    ctx->thread1.handle.id = 0;
    ctx->thread1.thread_handle = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)dsThreadGo, (LPVOID)&ctx->thread1, 0, &err);
    if (!ctx->thread1.thread_handle) err = DS_ERR_WINAPI;
    else err = 0;
#endif /* defined(HAVE_PTHREAD_H) */
    if (err) return err;
    else {
	dsThreadSetStatusBits((DSBaseThreadContext)ctx, DS_DOUBLETHREAD_RUNNING1_FLAG);
        dsThreadSetStatusBits((DSBaseThreadContext)ctx,DS_DOUBLETHREAD_JOINREQUIRED1_FLAG);
    }

#if defined(HAVE_PTHREAD_H)
    err = pthread_create(&ctx->thread2.handle.id,custom?&atrd:NULL,(void*(*)(void*))dsThreadGo,(void*)&ctx->thread2);
    if (err) err = DS_ERR_SYSCALL_PTHREAD_CREATE;
    if (custom) pthread_attr_destroy(&atrd);
#else /* defined(HAVE_WIN32_THREADS) */
    ctx->thread2.handle.id = 0;
    ctx->thread2.thread_handle = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)dsThreadGo, (LPVOID)&ctx->thread2, 0, &err);
    if (!ctx->thread2.thread_handle) err = DS_ERR_WINAPI;
    else err = 0;
#endif /* defined(HAVE_PTHREAD_H) */
    if (err) {
	dsDoubleThreadTerminateAndWait(ctx);
	return err;
    }
    else {
	dsThreadSetStatusBits((DSBaseThreadContext)ctx, DS_DOUBLETHREAD_RUNNING2_FLAG);
	dsThreadSetStatusBits((DSBaseThreadContext)ctx,DS_DOUBLETHREAD_JOINREQUIRED2_FLAG);
    }
    
#if defined(HAVE_PTHREAD_H)
#else /* defined(HAVE_WIN32_THREADS) */
    while ((!ctx->thread1.handle.id)||(!ctx->thread2.handle.id));
#endif /* defined(HAVE_PTHREAD_H) */
    dsDebugServer(DS_DEBUG_INFO,"DoubleThread %p: Run %8.lx, %8.lx",ctx, ctx->thread1.handle.id, ctx->thread2.handle.id);


    return 0;
}

int dsDoubleThreadWaitTerminate(DSDoubleThreadContext ctx) {
    int err1 = 0, err2 = 0;
#if defined(HAVE_PTHREAD_H)
    int *retcode;
#else /* defined(HAVE_WIN32_THREADS) */
    DWORD retcode;
#endif /* defined(HAVE_PTHREAD_H) */
    DSLockContext lock;
    
    ds_assert_param(ctx);
    ds_assert_param(ctx->base.initialized);
    
    lock=ctx->base.lock;

    dsLockWrite(lock);
    if ((ctx->base.status&DS_DOUBLETHREAD_JOINREQUIRED1_FLAG)==DS_DOUBLETHREAD_JOINREQUIRED1_FLAG) {
	ctx->base.status&=~DS_DOUBLETHREAD_JOINREQUIRED1_FLAG;
	dsLockReturn(lock);

#if defined(HAVE_PTHREAD_H)
	pthread_join(ctx->thread1.handle.id,(void**)&retcode);
#else /* defined(HAVE_WIN32_THREADS) */
	for (err1 = STILL_ACTIVE;err1 == STILL_ACTIVE;) {
	    WaitForSingleObject(ctx->thread1.thread_handle, INFINITE);
	    retcode = GetExitCodeThread(ctx->thread1.thread_handle, &err1);
	    if (!retcode) err1 = 0;
	}
	CloseHandle(ctx->thread1.thread_handle);
#endif /* defined(HAVE_PTHREAD_H) */
	dsLockWrite(lock);
	ctx->base.status&=~DS_DOUBLETHREAD_RUNNING1_FLAG;
	dsLockReturn(lock);
	dsConditionBroadcast((DSConditionContext)ctx);

	if (retcode) {
#if defined(HAVE_PTHREAD_H)
	    if (*((void**)retcode)==PTHREAD_CANCELED) err1=DS_ERR_THREAD_CANCELED;
	    else err1=*retcode;
	    free(retcode);
#else /* defined(HAVE_WIN32_THREADS) */
#endif /* defined(HAVE_PTHREAD_H) */
	} else err1=DS_ERR_THREAD_RETURNCODE;

	dsLockWrite(lock);
    } 

    if ((ctx->base.status&DS_DOUBLETHREAD_JOINREQUIRED2_FLAG)==DS_DOUBLETHREAD_JOINREQUIRED2_FLAG) {
	ctx->base.status&=~DS_DOUBLETHREAD_JOINREQUIRED2_FLAG;
	dsLockReturn(lock);
#if defined(HAVE_PTHREAD_H)
	pthread_join(ctx->thread2.handle.id,(void**)&retcode);
#else /* defined(HAVE_WIN32_THREADS) */
	for (err2 = STILL_ACTIVE;err2 == STILL_ACTIVE;) {
	    WaitForSingleObject(ctx->thread2.thread_handle, INFINITE);
	    retcode = GetExitCodeThread(ctx->thread2.thread_handle, &err2);
	    if (!retcode) err2 = 0;
	}
	CloseHandle(ctx->thread2.thread_handle);
#endif /* defined(HAVE_PTHREAD_H) */
	dsLockWrite(lock);
	ctx->base.status&=~DS_DOUBLETHREAD_RUNNING2_FLAG;
	dsLockReturn(lock);
	dsConditionBroadcast((DSConditionContext)ctx);

	if (retcode) {
#if defined(HAVE_PTHREAD_H)
	    if (*((void**)retcode)==PTHREAD_CANCELED) err2=DS_ERR_THREAD_CANCELED;
	    else err2=*retcode;
	    free(retcode);
#else /* defined(HAVE_WIN32_THREADS) */
#endif /* defined(HAVE_PTHREAD_H) */
	} else err2=DS_ERR_THREAD_RETURNCODE;
    } else dsLockReturn(lock);

    dsDebugServer((err1||err2)?DS_DEBUG_WARNING:DS_DEBUG_DEBUG, "DoubleThread %p: RetCode: %i, %i",ctx, err1, err2);

    dsThreadWaitStatusBitsOff((DSBaseThreadContext)ctx,DS_DOUBLETHREAD_RUNNING1_FLAG|DS_DOUBLETHREAD_RUNNING2_FLAG);

    return (err1||err2);
}

int dsDoubleThreadTerminateAndWait(DSDoubleThreadContext ctx) {
    ds_assert_param(ctx);
    dsThreadTerminate((DSBaseThreadContext)ctx);
    dsDoubleThreadNotify(ctx);
    return dsDoubleThreadWaitTerminate(ctx);
}

int dsDoubleThreadIsRunning(DSDoubleThreadContext ctx) {
    ds_assert_param(ctx);
    return dsThreadCheckStatusSomeBitsOn((DSBaseThreadContext)ctx, DS_DOUBLETHREAD_RUNNING1_FLAG|DS_DOUBLETHREAD_RUNNING1_FLAG);
}
