#ifndef _DS_DOUBLETHREAD_H
#define _DS_DOUBLETHREAD_H

#include <dslib.h>

#if defined(HAVE_PTHREAD_H)
# include <pthread.h>
#elif defined(HAVE_WIN32_THREADS)
# include <windows.h>
#else /* defined(HAVE_PTHREAD_H) */
# error "Missing threading library!"
#endif /* defined(HAVE_PTHREAD_H) */

#include <dsbasethread.h>

#define DS_DOUBLETHREAD_JOINREQUIRED1_FLAG 2
#define DS_DOUBLETHREAD_JOINREQUIRED2_FLAG 8
#define DS_DOUBLETHREAD_RUNNING1_FLAG 4
#define DS_DOUBLETHREAD_RUNNING2_FLAG 16

struct DSDoubleThreadPropertiesT {
    DSMallocContext malloc;
    void *attr;
    DSThreadFunction function1;
    DSThreadFunction function2;
};
typedef struct DSDoubleThreadPropertiesT *DSDoubleThreadProperties;

struct DSDoubleThreadContextT {
    struct DSBaseThreadContextT base;

    DSThreadRealHandleS thread1;
    DSThreadRealHandleS thread2;
};

typedef struct DSDoubleThreadContextT *DSDoubleThreadContext;


DSDoubleThreadContext dsDoubleThreadCreateContext(DSMallocContext mctx);
int dsDoubleThreadInitContext(DSDoubleThreadContext ctx);
int dsDoubleThreadFreeContext(DSDoubleThreadContext ctx);
int dsDoubleThreadDestroyContext(DSDoubleThreadContext ctx);
void dsDoubleThreadZeroContext(DSDoubleThreadContext ctx);

void dsDoubleThreadNotify(DSDoubleThreadContext ctx);
int dsDoubleThreadRun(DSDoubleThreadContext ctx, DSThreadFunction thread_function1, DSThreadFunction thread_function2, void *attr);
void dsDoubleThreadTerminate(DSDoubleThreadContext ctx);
int dsDoubleThreadWaitTerminate(DSDoubleThreadContext ctx);
int dsDoubleThreadTerminateAndWait(DSDoubleThreadContext ctx);
int dsDoubleThreadIsRunning(DSDoubleThreadContext ctx);

#define dsDoubleThreadTerminate(ctx) dsThreadTerminate((DSBaseThreadContext)ctx)
#define dsDoubleThreadSendCondition(ctx) dsConditionBroadcast((DSConditionContext)ctx)

#endif /* _DS_DOUBLETHREAD_H */
