#ifndef _DS_MALLOC_H
#define _DS_MALLOC_H

#include <stdio.h>

#include <dslib.h>
/*
OPCMalloc is special memory allocation for OPC Server data structures and
intercommunication between OPCRequest threads.
It should use shared memory to hold all data and share it between all 
OPCRequest clients. OPCRequest local data structures should be allocated
with standard malloc or some other routines wich would use process local
memory.
*/

typedef struct DSMallocContextT *DSMallocContext;

#include <dshashtable.h>
#include <dslock.h>

struct DSMallocContextT {
    DSMallocContext parrent;
#ifdef DS_TRACK_MEMORY
    DSHashTable allocations;
    DSLockContext allocations_lock;
    unsigned char tracker_initialized;
#endif /* DS_TRACK_MEMORY */
};

int dsMallocInit();
void dsMallocDestroy();

DSMallocContext dsMallocGetDefault();
int dsMallocSetDefault(DSMallocContext ctx);

DSMallocContext dsMallocCreateContext(DSMallocContext mctx);
void dsMallocDestroyContext(DSMallocContext ctx);

#define dsNewClass(mctx, type) \
    (type)dsMallocZero(mctx, sizeof(type##S))

#define dsNewClassArray(mctx, type, n) \
    (type)dsMallocZero(mctx, n*sizeof(type##S))
    
void *dsMalloc(DSMallocContext ctx, size_t size);
void *dsMallocZero(DSMallocContext ctx, size_t size);
void *dsRealloc(DSMallocContext ctx, void *ptr, size_t size);
void *dsCalloc(DSMallocContext ctx, size_t nmemb, size_t size);
void dsFree(DSMallocContext ctx, void *ptr);

int dsExchangeString(DSMallocContext mctx, char **dst, const char *src);
int dsExchangeStringChars(DSMallocContext mctx, char **dst, const char *src, unsigned chars);
int dsSetString(DSMallocContext mctx, char **dst, const char *src);
int dsSetStringChars(DSMallocContext mctx, char **dst, const char *src, unsigned chars);
char *dsCreateFullName(DSMallocContext mctx, const char *path, const char *filename, const char *extension);
int dsSetFullName(char *fullname, size_t buffsize, const char *path, const char *filename, const char *extension);

#ifdef _WIN32 
void *dsMallocWin32Heap(size_t size);
void *dsReallocWin32Heap(void *ptr, size_t size);
void *dsCallocWin32Heap(size_t nmemb, size_t size);
void *dsStrdupWin32Heap(const char *str);
void dsFreeWin32Heap(void *ptr);
#endif /* _WIN32 */

#endif /* _DS_MALLOC_H */
