#ifndef _DS_PLUGIN_H
#define _DS_PLUGIN_H

#include <time.h>

#include <dsthread.h>
#include <dsmalloc.h>

#ifdef _WIN32
# define DS_PLUGIN_EXTENSION ".dll"
# define dsPluginInfoFunctionImplementation __declspec(dllexport) DSPluginInfoContext PluginInfo()
#else
# define DS_PLUGIN_EXTENSION ".so"
# define dsPluginInfoFunctionImplementation DSPluginInfoContext PluginInfo()
#endif 

#define DS_PLUGIN_RUN_FLAG 256
#define DS_PLUGIN_RELOAD_FLAG 512
#define DS_PLUGIN_BREAK_FLAG 1024
#define DS_PLUGIN_ERROR_FLAG 2048

typedef struct DSPluginContextT *DSPluginContext;

typedef void* (*DSPluginCreateFunction)();
typedef int (*DSPluginInitFunction)(void*,DSPluginContext,void*);
typedef void (*DSPluginFreeFunction)(void*);
typedef void (*DSPluginDestroyFunction)(void*);
typedef int (*DSPluginStepFunction)(void*);

typedef struct DSPluginInfoContextT *DSPluginInfoContext;
struct DSPluginInfoContextT {
    char title[32];
    char version[16];
    DSPluginCreateFunction create;
    DSPluginInitFunction init;
    DSPluginStepFunction step;
    DSPluginFreeFunction free;
    DSPluginDestroyFunction destroy;
};


struct DSPluginContextT {
    struct DSThreadContextT thread;
    char *fullname;
    char *filename;
    time_t modification_time;
    void *plugin_attribute;
    unsigned char mark;
    
    DSMallocContext malloc;
    DSPluginContext next;
};

void dsPluginZeroContext(DSPluginContext ctx);
DSPluginContext dsPluginCreateContext(DSMallocContext mctx);
int dsPluginInitContext(DSPluginContext ctx, const char *filename, void *attr);
int dsPluginFreeContext(DSPluginContext ctx);
int dsPluginDestroyContext(DSPluginContext ctx);

void dsPluginRun(DSPluginContext ctx);
void dsPluginStop(DSPluginContext ctx);
void dsPluginReload(DSPluginContext ctx);
void dsPluginTerminate(DSPluginContext ctx);
int dsPluginContinue(DSPluginContext ctx);


void dsPluginMark(DSPluginContext ctx);
void dsPluginUnMark(DSPluginContext ctx);
int dsPluginMarked(DSPluginContext ctx);

#endif /* _DS_PLUGIN_H */
