#ifndef _DS_SIGNAL_H
#define _DS_SIGNAL_H

#include <dslib.h>
#include <dstime.h>
#include <dssoftsignal.h>

#ifdef HAVE_SIGNAL_H
# include <signal.h>
#else /* HAVE_SIGNAL_H */
# error "singal.h is required!"
#endif /* HAVE_SIGNAL_H */

#define DS_SOFTSIG_NUMBER 8

enum DSPosixSignalT {
    DS_POSIX_SIGNAL_SIGINT = 1,	//2
    DS_POSIX_SIGNAL_SIGTERM,	//15
    DS_POSIX_SIGNAL_SIGTSTP,
    DS_POSIX_SIGNAL_SIGCONT,
    DS_POSIX_SIGNAL_NUMBER
};
typedef enum DSPosixSignalT DSPosixSignal;
typedef enum DSPosixSignalT *DSPosixSignalMask;

typedef struct DSSignalMaskT *DSSignalMask;	// 0 - terminated

/* 32 max */
enum DSSignalT {
    DS_SIGNAL_USR1 = 0,
    DS_SIGNAL_USR2,
    DS_SIGNAL_NUMBER
};
typedef enum DSSignalT DSSignal;

struct DSSignalMaskT {
#ifdef _WIN32
    unsigned long mask;
#else /* _WIN32 */
    sigset_t mask;
#endif /* _WIN32 */
};
typedef struct DSSignalMaskT DSSignalMaskS;

#include <dsbasethread.h>
#include <dslib.h>
#include <dsmalloc.h>
#include <dslock.h>
#include <dsstringlist.h>
#include <dsconnection.h>
#include <dsconnectionserver.h>

#define DS_SIGNAL_NOTIFY DS_SIGNAL_USR1


#ifdef _WIN32
# ifdef _DS_SIGNAL_C
HANDLE winsignals[DS_SIGNAL_NUMBER];
# else /* _DS_SIGNAL_C */
extern HANDLE winsignals[DS_SIGNAL_NUMBER];
# endif /* _DS_SIGNAL_C */
#else
# ifdef _DS_SIGNAL_C
int user_sigs[DS_SIGNAL_NUMBER]={SIGUSR1, SIGUSR2};
# else /* _DS_SIGNAL_C */
extern int user_sigs[DS_SIGNAL_NUMBER];
# endif /* _DS_SIGNAL_C */
#endif /* _WIN32 */

#ifndef _DS_SIGNAL_C
extern void (*DS_SIGNAL_NOP_HANDLER)(int);
#endif /* _DS_SIGNAL_C */


struct DSSignalContextT {
    DSSignalMaskS default_signal_mask;

    DSMallocContext malloc;
    DSLockContext lock;
    
    DSStringList usr[DS_SIGNAL_NUMBER];

    unsigned char ExitFlag;
};
typedef struct DSSignalContextT *DSSignalContext;
typedef struct DSSignalContextT DSSignalContextS;

int dsSignalInit();
void dsSignalDestroy();


int dsPosixSignalSetAction(DSPosixSignal dsig, DSPosixSignalMask block, void (*action)(int));
int dsPosixSignalBlock(DSPosixSignal dsig);
int dsPosixSignalUnblock(DSPosixSignal dsig);

/*
int dsSignalINTSetAction(void (*action)(int));
int dsSignalINTBlock();
int dsSignalINTUnblock();
*/

int dsSignalGetDefaultMask(DSSignalMask sigmask);
int dsSignalMaskEmpty(DSSignalMask sigmask);
int dsSignalMaskFill(DSSignalMask sigmask);
int dsSignalMaskAddSignal(DSSignalMask sigmask, DSSignal sig);
int dsSIgnalMaskRemoveSignal(DSSignalMask sigmask, DSSignal sig);


DSSignalContext dsSignalCreateContext(DSMallocContext mctx);
int dsSignalInitContext(DSSignalContext ctx);
void dsSignalFreeContext(DSSignalContext ctx);
void dsSignalDestroyContext(DSSignalContext ctx);

int dsSignalRegister(DSSignalContext ctx, DSSignal sig);
int dsSignalCancel(DSSignalContext ctx, DSSignal sig);
int dsSignalSend(DSSignalContext ctx, DSSignal sig);

int dsSignalSendTo(DSThreadHandle thread, DSSignal sig);

int dsSignalTerminate(DSSignalContext ctx);
int dsSignalIsTerminated(DSSignalContext ctx);


#ifdef _WIN32
int dsSignalWinWait(DSSignalMask sigmask, DSSoftSignalList softsig, HANDLE hEvent, const DSDuration timeout);
int dsSignalWinWaitConnection(DSConnectionServerContext con, DSSignalMask sigmask, DSSoftSignalList softsig, DSConnectionMask *conmask, const DSDuration timeout);
#else /* _WIN32 */
int dsSignalCheckFDRead(int s);
int dsSignalCheckFDWrite(int s);
#endif /* _WIN32 */

#ifdef HAVE_SOCKETS
int dsSignalWaitRead(DSConnectionContext con, DSSignalMask sigmask, DSSoftSignalList softsig, const DSDuration timeout);
int dsSignalWaitWrite(DSConnectionContext con, DSSignalMask sigmask, DSSoftSignalList softsig, const DSDuration timeout);
int dsSignalCheckRead(DSConnectionContext con);
int dsSignalCheckWrite(DSConnectionContext con);

int dsSignalWaitConnection(DSConnectionServerContext con, DSSignalMask sigmask, DSSoftSignalList softsig, DSConnectionMask *conmask, const DSDuration timeout);
int dsSignalCheckConnection(DSConnectionServerContext con, DSConnectionMask *conmask);
#endif /* HAVE_SOCKETS */

int dsSignalWait(DSSignalMask sigmask, DSSoftSignalList softsig, DSDurationConst duration);
int dsSignalWaitRT(DSSignalMask sigmask, DSSoftSignalList softsig, DSDurationConst duration);
int dsSignalWaitUntil(DSSignalMask sigmask, DSSoftSignalList softsig, DSTimeConst time);
int dsSignalWaitRTUntil(DSSignalMask sigmask, DSSoftSignalList softsig, DSTimeConst time);

#endif /* _DS_SIGNAL_H */
