#include <stdio.h>

#include <dslib.h>

#ifdef HAVE_FCNTL_H
# include <fcntl.h>
#endif /* HAVE_FCNTL_H */
#ifdef HAVE_SYS_FILE_H
# include <sys/file.h>
#endif /* HAVE_SYS_FILE_H */
#ifdef HAVE_SYS_TYPES_H
# include <sys/types.h>
#endif /* HAVE_SYS_TYPES_H */
#ifdef HAVE_SYS_STAT_H
# include <sys/stat.h>
#endif /* HAVE_SYS_STAT_H */


#include <dsassert.h>

#include <dssystemlock.h>

void dsSystemLockZeroContext(DSSystemLockContext ctx) {
    ds_assert_return(ctx);
    ctx->name = NULL;
#ifdef _WIN32
    ctx->lockfd = INVALID_HANDLE_VALUE;
#else
    ctx->lockfd = -1;
#endif /* _WIN32 */
    
}

DSSystemLockContext dsSystemLockCreateContext(DSMallocContext mctx) {
    DSSystemLockContext ctx;
    
    ctx = (DSSystemLockContext)dsMalloc(mctx, sizeof(DSSystemLockContextS));
    if (ctx) {
	ctx->malloc = mctx;
	dsSystemLockZeroContext(ctx);
    }
    return ctx;
}

int dsSystemLockInitContext(DSSystemLockContext ctx, const char *name) {
    ds_assert_param(ctx);
    ds_assert_param(name);
    
    dsSystemLockZeroContext(ctx);
    
    ctx->name = dsCreateFullName(ctx->malloc, DS_LOCKDIR, name, ".lock");
    if (!ctx->name) return DS_ERR_MEM;
    return 0;
}


void dsSystemLockFreeContext(DSSystemLockContext ctx) {
    ds_assert_return(ctx);

    dsSystemLockReturn(ctx);    
    if (ctx->name) dsExchangeString(ctx->malloc, &ctx->name, NULL);
    dsSystemLockZeroContext(ctx);
}

void dsSystemLockDestroyContext(DSSystemLockContext ctx) {
    dsSystemLockFreeContext(ctx);
    dsFree(ctx->malloc, ctx);
}


int dsSystemLockAcquire(DSSystemLockContext ctx) {
#ifdef _WIN32
//    HANDLE lockfd;
#else
    int lockfd;
#endif /* _WIN32 */

    ds_assert_param(ctx);
    ds_assert_param(ctx->name);
    
#ifdef _WIN32
    if (ctx->lockfd != INVALID_HANDLE_VALUE) return 0;

/* DS: Try    
    ctx->lockfd = CreateFile(ctx->name, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, FILE_FLAG_DELETE_ON_CLOSE, NULL);
	// I don;t need this?
    LockFile(ctx->lockfd,0,0,0,0);
*/
#else
    if (ctx->lockfd >= 0) return 0;
    
    lockfd = open(ctx->name, O_RDWR|O_CREAT, 0644);
    if (lockfd<0) return DS_ERR_FILE_CREATE;
    if (flock(lockfd, LOCK_EX|LOCK_NB)) {
	close(lockfd);
	return DS_ERR_BUSY;
    }
    fcntl(lockfd, F_SETFD, 1);
    ctx->lockfd = lockfd;
#endif /* _WIN32 */

    return 0;
}

int dsSystemLockWait(DSSystemLockContext ctx) {
    return DS_ERR_NOTSUPPORTED;
}

int dsSystemLockReturn(DSSystemLockContext ctx) {
    ds_assert_param(ctx);
    ds_assert_param(ctx->name);

#ifdef _WIN32 
    if (ctx->lockfd == INVALID_HANDLE_VALUE) return 0;

    CloseHandle(ctx->lockfd);    
    ctx->lockfd = INVALID_HANDLE_VALUE;
#else
    if (ctx->lockfd < 0) return 0;

    flock(ctx->lockfd, LOCK_UN);
    close(ctx->lockfd);
    ctx->lockfd = -1;
#endif /* _WIN32 */

    return 0;
}
