#include <stdio.h>

#include <dslib.h>
#ifdef HAVE_ERRNO_H
# include <errno.h>
#endif /* HAVE_ERRNO_H */
#ifdef HAVE_UNISTD_H
# include <unistd.h>
#endif /* HAVE_UNISTD_H */
#ifdef HAVE_SYS_TYPES_H
# include <sys/types.h>
#endif /* HAVE_SYS_TYPES_H */

#include <dsassert.h>

#include <dssystemservice.h>

void dsSystemServiceZeroContext(DSSystemServiceContext ctx) {
    ds_assert_return(ctx);
    ctx->name = NULL;
}

DSSystemServiceContext dsSystemServiceCreateContext(DSMallocContext mctx) {
    DSSystemServiceContext ctx;
    
    ctx = (DSSystemServiceContext)dsMalloc(mctx, sizeof(DSSystemServiceContextS));
    if (ctx) {
	ctx->malloc = mctx;
	dsSystemServiceZeroContext(ctx);
    }
    return ctx;
}

int dsSystemServiceInitContext(DSSystemServiceContext ctx, const char *name) {
    ds_assert_param(ctx);
    ds_assert_param(name);
    
    dsSystemServiceZeroContext(ctx);
    dsExchangeString(ctx->malloc, &ctx->name, name);
    return 0;
}


void dsSystemServiceFreeContext(DSSystemServiceContext ctx) {
    ds_assert_return(ctx);

    if (ctx->name) dsExchangeString(ctx->malloc, &ctx->name, NULL);
    dsSystemServiceZeroContext(ctx);
}

void dsSystemServiceDestroyContext(DSSystemServiceContext ctx) {
    dsSystemServiceFreeContext(ctx);
    dsFree(ctx->malloc, ctx);
}


int dsSystemServiceDaemonize() {
#ifdef _WIN32
    return DS_ERR_NOTSUPPORTED;
#else /* _WIN32 */
    pid_t newpid;
    
    newpid = fork();
    if (newpid==(pid_t)-1) {
	if ((errno == EAGAIN)||(errno == ENOMEM)) return DS_ERR_MEM;
	return DS_ERR_SYSCALL_FORK;
    }

    if (newpid) exit(0);
    return 0;
#endif /* _WIN32 */
}
