#ifndef _DS_DISPERSION_H
#define _DS_DISPERSION_H

#include <dsmalloc.h>

struct DSCorrelationContextT {
    DSMallocContext malloc;
    
    unsigned int n;
    double *corr;
    double *s;
    unsigned long events;
};

typedef struct DSCorrelationContextT *DSCorrelationContext;
typedef struct DSCorrelationContextT DSCorrelationContextS;

struct DSDispersionContextT {
    DSMallocContext malloc;    

    double m;
    double d;
    unsigned long events;
};
typedef struct DSDispersionContextT *DSDispersionContext;
typedef struct DSDispersionContextT DSDispersionContextS;

enum DSCorrelationDataTypeT {
    DS_CORRELATION_INT8,
    DS_CORRELATION_UINT8,
    DS_CORRELATION_INT16,
    DS_CORRELATION_UINT16,
    DS_CORRELATION_INT32,
    DS_CORRELATION_UINT32,
    DS_CORRELATION_INT64,
    DS_CORRELATION_UINT64,
    DS_CORRELATION_FLOAT,
    DS_CORRELATION_DOUBLE,
    DS_CORRELATION_LONGDOUBLE
};
typedef enum DSCorrelationDataTypeT DSCorrelationDataType;

typedef double (*DSCorrelationGetValueFunction)(void *data, unsigned int i);

DSDispersionContext dsDispersionCreateContext(DSMallocContext mctx);
int dsDispersionInitContext(DSDispersionContext disp);
void dsDispersionFreeContext(DSDispersionContext disp);
void dsDispersionDestroyContext(DSDispersionContext disp);

int dsDispersionEvent(DSDispersionContext disp, double value);
int dsDispersionFinalyze(DSDispersionContext disp);
int dsDispersionClear(DSDispersionContext disp);

double dsDispersionGetMedian(DSDispersionContext disp);
double dsDispersionGetDispersion(DSDispersionContext disp);

DSCorrelationContext dsCorrelationCreateContext(DSMallocContext mctx);
int dsCorrelationInitContext(DSCorrelationContext corr, unsigned int elements);
void dsCorrelationFreeContext(DSCorrelationContext corr);
void dsCorrelationDestroyContext(DSCorrelationContext corr);

int dsCorrelationEvent(DSCorrelationContext corr, DSCorrelationDataType type, void *data);
int dsCorrelationComplexEvent(DSCorrelationContext corr, DSCorrelationGetValueFunction get, void *data);
int dsCorrelationFinalyze(DSCorrelationContext corr);
int dsCorrelationClear(DSCorrelationContext corr);

double dsCorrelationGetMedian(DSCorrelationContext corr, unsigned int i);
double dsCorrelationGetDispersion(DSCorrelationContext corr, unsigned int i);
double dsCorrelationGetCorrelation(DSCorrelationContext corr, unsigned int i, unsigned int j);

double *dsCorrelationGetMedianArray(DSCorrelationContext corr);
double *dsCorrelationGetDispersionArray(DSCorrelationContext corr);
double *dsCorrelationGetCorrelationArray(DSCorrelationContext corr);

#endif /* _DS_DISPERSION_H */
