#ifndef _DS_DOUBLEHASH_H
#define _DS_DOUBLEHASH_H

#include <dsmalloc.h>


/* 
 DoubleHashCleanFunction:
    arg1: Pointer to current data
    arg2: Supplied attribute
    return:
	0: Clean isn't required
	1: Clean from DoubleHash table
	2: Clean from DoubleHash and Destroy (if required)
*/
typedef int (*DSDoubleHashCleanFunction)(void *, void *);

typedef struct DSDoubleHashContextT *DSDoubleHashContext;

struct DSDoubleHashContextT {
    DSMallocContext mctx;
    
    void **buffer;
        
    unsigned long size;		/* should be prime number */
    unsigned long step_size;
    unsigned long max_size;

    unsigned char gift;
    void (*cfunc)(void *data);
};

DSDoubleHashContext dsDoubleHashCreateContext(DSMallocContext mctx);
int dsDoubleHashInitContext(DSDoubleHashContext ctx, unsigned long size, unsigned long step_size, unsigned char free_on_clear, void (*cfunc)(void *data));
void dsDoubleHashFreeContext(DSDoubleHashContext ctx);
void dsDoubleHashDestroyContext(DSDoubleHashContext ctx);

unsigned long dsDoubleHashInsert(DSDoubleHashContext ctx, void *data, unsigned long hash);
unsigned long dsDoubleHashInsertS(DSDoubleHashContext ctx, void *data, const char *str);
void dsDoubleHashRemove(DSDoubleHashContext ctx, unsigned long pos);
void dsDoubleHashUnlink(DSDoubleHashContext ctx, unsigned long pos);

const void *dsDoubleHashGetBuffer(DSDoubleHashContext ctx);
const void *dsDoubleHashGetLink(DSDoubleHashContext ctx, unsigned long pos);
void *dsDoubleHashGet(DSDoubleHashContext ctx, unsigned long pos);

int dsDoubleHashSetMaximumSize(DSDoubleHashContext ctx, unsigned long max_size);

unsigned long dsDoubleHashClean(DSDoubleHashContext ctx, DSDoubleHashCleanFunction clean_func, void *attr);

#endif /* _DS_DOUBLEHASH_H */
