#ifndef _DS_HASHTABLE_H
#define _DS_HASHTABLE_H

#include <dsmalloc.h>

#define hash_table_coef 0.618

enum DSHashTypeT {
    DS_HASH_TYPE_POINTER = 0,		/* First element of data is pointer on search string */
    DS_HASH_TYPE_STATICPOINTER,		/* Same, but this string shouldn't be deallocated on free */
    DS_HASH_TYPE_ARRAY,			/* First element of data is string (array of characters) */
    DS_HASH_TYPE_FUNCTION,		/* Search string should be got from data with supported function */
    DS_HASH_TYPE_HASH,			/* First element of data is hash (no string at all) */
    DS_HASH_TYPE_HASH_FUNCTION,		/* Hash acquired from data with supported function (no string) */
    DS_HASH_TYPE_LAST
};
typedef enum DSHashTypeT DSHashType;

typedef struct DSHashStringPtrT *DSHashStringPtr;
typedef struct DSHashStringArrayT *DSHashStringArray;
typedef struct DSHashStringHashT *DSHashStringHash;
typedef struct DSHashTableBucketT *DSHashTableBucket;
typedef struct DSHashTableT *DSHashTable;
typedef int (*DSHashTableWalkFunction)(const void *, void*);
typedef int (*DSHashTableCleanFunction)(const void *);
typedef char * (*DSHashTableStringFunction)(const void *);
typedef unsigned long (*DSHashTableHashFunction)(const void *);
typedef void (*DSHashTableDestructorFunction)(void *data);
typedef DSHashTableDestructorFunction HashTableDestructorFunction;

char *dsHashTableFake(const void *ctx);

enum DSHashTableDestructionModeT { 
    DS_HASHTABLE_DESTRUCTIONMODE_SINGLE = 0,
    DS_HASHTABLE_DESTRUCTIONMODE_MULTI,
    DS_HASHTABLE_DESTRUCTIONMODE_ALL
};
typedef enum DSHashTableDestructionModeT DSHashTableDestructionMode;

typedef void (*DSHashTablePPDFunction)(DSHashTable, unsigned char, void *);

struct DSHashStringPtrT {
    char *str;
};

struct DSHashStringArrayT {
    char str[1];
};

struct DSHashStringHashT {
    unsigned long hash;
};


struct DSHashTableBucketT {
    unsigned long hash;
    void *data;
    char *string;
    DSHashTableBucket next;
};

struct DSHashTableT {
    DSMallocContext malloc;
    unsigned char gift;
    unsigned char mode;
    unsigned long size;
    unsigned long string_size;
    void (*cfunc)(void *data);
    DSHashTableStringFunction string_func;
    DSHashTablePPDFunction prefunc, postfunc;
    void *ppfuncattr;
    DSHashTableBucket *bucket;
};

DSHashTable dsHashTableInit(DSMallocContext mctx, unsigned long size, DSHashType hash_type, DSHashTableStringFunction string_func);
DSHashTable dsHashTableHashInit(DSMallocContext mctx, unsigned long size, DSHashType hash_type, DSHashTableHashFunction hash_func);
int dsHashTableGrow(DSHashTable ctx, unsigned long size);
int dsHashTableSetDestructor(DSHashTable ctx, DSHashTableDestructorFunction cfunc);
int dsHashTableSetPPDFunctions(DSHashTable ctx, DSHashTablePPDFunction prefunc, DSHashTablePPDFunction postfunc, void *attr);
int dsHashTableSetStringSize(DSHashTable ctx, unsigned long size);
void dsHashTableFree(DSHashTable ctx);
void dsHashTableDestroy(DSHashTable ctx);
int dsHashTableInsert(DSHashTable ctx, void *data);
int dsHashTableHashInsert(DSHashTable ctx, unsigned long hash, void *data);
int dsHashTableHashStringInsert(DSHashTable ctx, unsigned long hash, char *str, void *data);
int dsHashTableAdd(DSHashTable ctx, void *data, unsigned char report);
int dsHashTableHashAdd(DSHashTable ctx, unsigned long hash, void *data, unsigned char report);
int dsHashTableHashStringAdd(DSHashTable ctx, unsigned long hash, char *str, void *data, unsigned char report);
int dsHashTableDeleteMulti(DSHashTable ctx, const char *string);
int dsHashTableHashDeleteMulti(DSHashTable ctx, unsigned long hash);
int dsHashTableDelete(DSHashTable ctx, const char *string);
int dsHashTableHashDelete(DSHashTable ctx, unsigned long hash);
int dsHashTableRemove(DSHashTable ctx, const void *data);
int dsHashTableUnlink(DSHashTable ctx, const void *data);
void *dsHashTableFindNext(DSHashTable ctx, const char *string, const void *start);
void *dsHashTableFind(DSHashTable ctx, const char *string);
void *dsHashTableHashFind(DSHashTable ctx, unsigned long hash);
void *dsHashTableHashFindNext(DSHashTable ctx, unsigned long hash, const void *start);
void *dsHashTableHashStringFind(DSHashTable ctx, unsigned long hash, const char *string);
void *dsHashTableHashStringFindNext(DSHashTable ctx, unsigned long hash, const char *string, const void *start);
int dsHashTableWalk(DSHashTable ctx, DSHashTableWalkFunction func, void *attr);
int dsHashTableClean(DSHashTable ctx, DSHashTableCleanFunction func);
int dsHashTableIsEmpty(DSHashTable ctx);

unsigned long dsHashTableGetHash(DSHashTable ctx, const char *string);
unsigned long dsHashTableGetPos(DSHashTable ctx, const char *string);
unsigned long dsHashTableHashGetPos(DSHashTable ctx, unsigned long hash);

#endif /* _DS_HASHTABLE_H */
