#include <stdio.h>
#include <string.h>
#include <libintl.h>

#include <dslib.h>
#include <dsassert.h>
#include <dsmalloc.h>

#include <dslocale.h>


#define CURRENT_DOMAIN_SIZE 255
static char current_domain[CURRENT_DOMAIN_SIZE];


int dsLocaleRegisterDomain(const char *domain, const char *dir) {
#ifndef PHARLAP
    if (!bindtextdomain(domain,dir?dir:DS_LOCALEDIR)) return DS_ERR_LOCALE_DOMAIN_REGISTER;
#endif /* ! PHARLAP */
    return 0;
}

int dsLocaleSetDomain(const char *domain) {
#ifndef PHARLAP
    if (!textdomain(domain)) return DS_ERR_LOCALE_DOMAIN_SET;
#endif /* ! PHARLAP */
    return 0;
}

int dsLocaleInit() {
    int err;
    err = dsLocaleRegisterDomain(DS_GETTEXT_DOMAIN,DS_LOCALEDIR);
    if (!err) dsLocaleSetDomain(DS_GETTEXT_DOMAIN);
    return err;
}


void dsLocaleDestroy() {
}


dschar dsLocaleGetText(const dslchar str, const char *locale) {
#ifdef PHARLAP
    return str;
#else
    return gettext(str);
#endif /* PHARLAP */
}

dschar dsLocaleGetDomainText(const char *domain, const dslchar str, const char *locale) {
#ifdef PHARLAP
    return str;
#else
    return dgettext(domain, str);
#endif /* PHARLAP */
}



int dsLocaleExchangeString(DSMallocContext mctx, dslchar *dst, const char *src) {
    dslchar tmp;
    if (!src) {
	if (*dst) dsFree(mctx, *dst);
	return 0;
    }
    tmp = (dslchar)dsMalloc(mctx, (strlen(src)+1)*sizeof(char));
    if (!tmp) return DS_ERR_MEM;
    strcpy(tmp,src);
    if (*dst) dsFree(mctx, *dst);
    *dst=tmp;
    return 0;
}
