#include <stdio.h>
#include <stdarg.h>

#define _DS_INCLUDE_XML_HEADERS

#include <dsassert.h>
#include <dsstring.h>
#include <dsxml.h>


DSString dsStringCreate(DSMallocContext mctx, const char *s, ...) {
    DSString res;
    va_list ap;
    
    ds_assert_null(s);

    va_start(ap, s);
    res = dsStringVCreate(mctx, s, ap);
    va_end(ap);
    
    return res;
}

DSString dsStringVCreate(DSMallocContext mctx, const char *s, va_list ap) {
    int size;
    DSString res;
    
#ifndef _WIN32
    int rsize;
    DSString newres;
    va_list apc;
#endif /* ! _WIN32 */

    ds_assert_null(s);

#ifdef _WIN32
    size = _vscprintf(s, ap) + 1;
    res = dsMalloc(mctx, size);
    if (res) vsprintf(res, s, ap);
#else /* _WIN32 */
    size = strlen(s) + 1;

    for (res = dsMalloc(mctx, size); res; res = newres) {
	va_copy(apc, ap);
	rsize = vsnprintf (res, size, s, apc);
	va_end(apc);
	
	if ((rsize > -1) && (rsize < size)) return res;
	if (rsize > -1) size = rsize + 1; /* GLibc 2.1 beheviour */
	else size*=2;
	
	newres = dsRealloc(mctx, res, size);
	if (!newres) dsFree(mctx, res);
    }
#endif /* _WIN32 */

    return res;
}

void dsStringDestroy(DSMallocContext mctx, DSString str) {
    dsExchangeString(mctx, &str, NULL);
}

int dsStringExchange(DSMallocContext mctx, DSString *str, DSStringConstant s) {
    return dsExchangeString(mctx, str, s);
}

int dsStringExchangeBytes(DSMallocContext mctx, DSString *str, DSStringConstant s, unsigned int bytes) {
    return dsExchangeStringChars(mctx, str, s, bytes);
}

DSString dsStringCopy(DSMallocContext mctx, DSStringConstant str) {
    DSString res = NULL;
    
    if (dsStringExchange(mctx, &res, str)) return NULL;
    return res;
}

DSStringConstant dsStringStrchr(DSStringConstant str, char ch) {
    return xmlStrchr(str, ch);
}

DSStringConstant dsStringStrrchr(DSStringConstant str, char ch) {
    DSStringConstant res = NULL, newres;
    
    for (res = dsStringStrchr(str, ch); res; res = newres) {
	newres = dsStringStrchr(res + 1, ch);
	if (!newres) break;
    }

    return res;
}

DSStringConstant dsStringStrstr(DSStringConstant str, DSStringConstant sub) {
    return xmlStrstr(str, sub);
}
