#include <stdio.h>
#include <string.h>

#include <dsassert.h>
#include <dsdebug.h>

#include <dsmalloc.h>
#include <dsstringlist.h>

DSStringList dsStringListCreate(DSMallocContext mctx) {
    DSStringList sl;
    sl=(DSStringList)dsMalloc(mctx,sizeof(struct DSStringListT));
    if (sl) {
	sl->malloc=mctx;
	sl->item=NULL;
	sl->next=NULL;
	sl->free=NULL;
    }
    return sl;
}

void dsStringListDestroy(DSStringList list) {
    DSStringList cur;


    ds_assert_return(list);

    while (list) {
	cur=list;
	if (cur->item) dsFree(cur->malloc, cur->item);
	list=list->next;
	dsFree(cur->malloc,cur);
    }
}

void dsStringListFree(DSStringList list) {

    ds_assert_return(list);

    if (list->next) {
	dsStringListDestroy(list->next);
	list->next=NULL;
    }
    if (list->item) {
	dsFree(list->malloc, list->item);
	list->item=NULL;
    }
    list->free=list;
}

int dsStringListIsEmpty(DSStringList list) {
    DSStringList cur;
    
    ds_assert_param(list);
    
    for (cur=list;cur;cur=cur->next)
	if (cur->item) return 0;
    return 1;
}

int dsStringListIsIn(DSStringList list, const void *item, unsigned long size) {
    DSStringList cur;
    
    ds_assert_param(list);
    ds_assert_param(item);
    ds_assert_param(size>0);
    
    for (cur=list;cur;cur=cur->next) {
	if ((cur->item)&&(cur->size==size)&&(memcmp(cur->item,item,size)==0))
	    return 1;
    }
    return 0;
}

static int opcStringListRemoveSM(DSStringList list, const void *item, unsigned long size, unsigned char multi) {
    int res=0;
    DSStringList cur;
    DSStringList pre=NULL;
    DSStringList end=NULL;
    
    ds_assert_param(list);
    ds_assert_param(item);
    ds_assert_param(size>0);
    
    list->free=NULL;
    for (cur=list;cur;cur=cur->next) {
	if (cur->item) {
	    if ((cur->size==size)&&(memcmp(cur->item,item,size)==0)) {
		res++;
		if (pre) {
		    if (!end) for (end=cur;end->next;end=end->next);
		    pre->next=cur->next;
		    cur->next=NULL;
		    end->next=cur;
		    end=end->next;
		    if (!list->free) list->free=end;
		} else {
		    dsFree(cur->malloc, cur->item);
		    cur->item=NULL;
		    list->free=cur;
		}
		if (!multi) break;
	    }
	} else {
	    if (!list->free) list->free=cur;
	}
	pre=cur;
    }
    if (!list->free) list->free=pre;
    return res;
}

int dsStringListRemove(DSStringList list, const void *item, unsigned long size) {
    return opcStringListRemoveSM(list,item,size,0);
}

int dsStringListRemoveMulti(DSStringList list, const void *item, unsigned long size) {
    return opcStringListRemoveSM(list,item,size,1);
}

int dsStringListAdd(DSStringList list, const void *item,  unsigned long size) {
    DSStringList cur;

    ds_assert_param(list);
    ds_assert_param(item);
    ds_assert_param(size>0);
    
    if (list->free) cur=list->free;
    else cur=list;
    
    for (;cur;cur=cur->next) {
    	if (!cur->item) {
	    list->free=cur;
	    break;
	}
	else if (!cur->next) {
	    cur->next=(DSStringList)dsMalloc(list->malloc,sizeof(struct DSStringListT));
	    cur = cur->next;
	    if (!cur) return DS_ERR_MEM;
	    cur->next = NULL;
	    cur->malloc = list->malloc;
	    list->free = cur;
	    break;
	}
    }

    cur->item=(char*)dsMalloc(cur->malloc,size);
    if (!cur->item) return DS_ERR_MEM;
    memcpy(cur->item,item,size);
    cur->size=size;
    return 0;
}
