#include <dsassert.h>

#include <dstreenode.h>


void dsTreeNodeZeroContext(DSTreeNodeContext ctx) {
    ds_assert_return(ctx);
    ctx->destructor=NULL;
    ctx->parrent=NULL;
    ctx->next=NULL; ctx->prev=NULL;
}

DSTreeNodeContext dsTreeNodeCreateContext(DSMallocContext mctx) {
    DSTreeNodeContext ctx;

    ctx = (DSTreeNodeContext)dsMalloc(mctx,sizeof(struct DSTreeNodeContextT));
    if (ctx) {
	ctx->malloc=mctx;
	dsTreeNodeZeroContext(ctx);
    }
    return ctx;
}

int dsTreeNodeInitContext(DSTreeNodeContext ctx, DSDestroyContextFunction destructor) {
    ds_assert_param(ctx);
    dsTreeNodeZeroContext(ctx);    
    return 0;
}

void dsTreeNodeFreeContext(DSTreeNodeContext ctx) {
    ds_assert_return(ctx);
    dsTreeNodeUnlink(ctx);
    dsTreeNodeZeroContext(ctx);
}

void dsTreeNodeDestroyContext(DSTreeNodeContext ctx) {
    dsTreeNodeFreeContext(ctx);
    dsFree(ctx->malloc,ctx);
}

int dsTreeNodeUnlink(DSTreeNodeContext ctx) {
    if (ctx->parrent) {
	if (ctx->prev) ctx->prev->next = ctx->next;
	else ctx->parrent->first_child = ctx->next;
	
	if (ctx->next) ctx->next->prev = ctx->prev;
	else ctx->parrent->last_child = ctx->prev;
	
	ctx->parrent=NULL;
	ctx->prev=NULL; ctx->next=NULL;
    }
    return 0;
}


int dsTreeNodeInsertAfter(DSTreeNodeContext ctx, DSTreeNodeContext sibling) {
    ds_assert_param(ctx);
    ds_assert_param(ctx->parrent);
    ds_assert_param(!sibling->parrent);
    
    sibling->next = ctx->next;
    ctx->next = sibling;
    sibling->prev = ctx;
    
    if (!sibling->next) ctx->parrent->last_child=sibling;
    sibling->parrent = ctx->parrent;

    return 0;
}

int dsTreeNodeInsertBefore(DSTreeNodeContext ctx, DSTreeNodeContext sibling) {
    ds_assert_param(ctx);
    ds_assert_param(ctx->parrent);
    ds_assert_param(!sibling->parrent);
    
    sibling->prev = ctx->prev;
    ctx->prev = sibling;
    sibling->next = ctx;
    
    if (!sibling->prev) ctx->parrent->first_child=sibling;
    sibling->parrent = ctx->parrent;
    
    return 0;
}

void dsTreeComplexNodeZeroContext(DSTreeComplexNodeContext ctx) {
    dsTreeNodeZeroContext((DSTreeNodeContext)ctx);
    ctx->first_child=NULL; ctx->last_child=NULL;
}

DSTreeComplexNodeContext dsTreeComplexNodeCreateContext(DSMallocContext mctx) {
    DSTreeComplexNodeContext ctx;

    ctx = (DSTreeComplexNodeContext)dsMalloc(mctx,sizeof(struct DSTreeComplexNodeContextT));
    if (ctx) {
	((DSTreeNodeContext)ctx)->malloc=mctx;
	dsTreeComplexNodeZeroContext(ctx);
    }
    return ctx;
}

int dsTreeComplexNodeInitContext(DSTreeComplexNodeContext ctx, DSDestroyContextFunction destructor) {
    ds_assert_param(ctx);
    dsTreeComplexNodeZeroContext(ctx);
    return 0;
}

void dsTreeComplexNodeFreeContext(DSTreeComplexNodeContext ctx) {
    DSTreeNodeContext node,next_node;
    ds_assert_return(ctx);
    
    for (node=ctx->first_child;node;node=next_node) {
	node->parrent=NULL;
	next_node = node->next;
	if (node->destructor) node->destructor(node);
	else dsTreeNodeFreeContext(node);
    }
    dsTreeNodeFreeContext((DSTreeNodeContext)ctx);
}

void dsTreeComplexNodeDestroyContext(DSTreeComplexNodeContext ctx) {
    dsTreeComplexNodeFreeContext(ctx);
    dsFree(((DSTreeNodeContext)ctx)->malloc,ctx);
}


int dsTreeComplexNodeAddChild(DSTreeComplexNodeContext ctx, DSTreeNodeContext child) {
    ds_assert_param(ctx);
    ds_assert_param(child);
    ds_assert_param(!child->parrent);

    if (ctx->last_child) ctx->last_child->next = child;
    else ctx->first_child = child;
    child->prev = ctx->last_child;
    child->parrent = ctx;
    ctx->last_child = child;
    return 0;
}
