#define _DS_INCLUDE_XML_HEADERS

#include <stdio.h>
#include <dsassert.h>

#include <dsxml.h>
#include <dsmalloc.h>

int dsXmlInit() {
    xmlInitParser();
    xmlInitCharEncodingHandlers();
    xmlKeepBlanksDefault(0);
    xmlSubstituteEntitiesDefault(0);
#ifdef _WIN32
//    xmlMemSetup	(dsFreeWin32Heap, dsMallocWin32Heap, dsReallocWin32Heap, dsStrdupWin32Heap);
#endif /* _WIN32 */

    return 0;
}

void dsXmlDestroy() {
    xmlCleanupCharEncodingHandlers();
    xmlCleanupParser();
}

/* DS: Rewrite me! Stack instead of recursion, FindFirst, FindNext */
xmlNodePtr dsXmlFindNode(xmlNodePtr parent, const char *name, const xmlNsPtr ns) {
    xmlNodePtr cur;
    xmlNodePtr ret;
    
    cur = parent;
    while(cur != NULL) {

/*	
	if (cur->ns) {
	    printf("%s %p(%s)\n",cur->name, cur->ns, cur->ns->prefix);
	    printf("%s\n",cur->ns->href);
	} else printf("%s %p\n",cur->name, cur->ns);
	Hm.. Problem in SearchNS
	

	printf("%s %p(%s) %p(%s)\n",cur->name, cur->ns, cur->ns->prefix, ns, ns->prefix);
	printf("%s\n%s\n",cur->ns->href,ns->href);
	Problem is following:
	<SOAP-ENV:Envelope xmlns:ns1="http://opcfoundation.org/webservices/XMLDA/1.0/">
	    ....
		<GetStatus  xmlns="http://opcfoundation.org/webservices/XMLDA/1.0/">
	
	SearchNS founds ns1 and think what it isn't equal to namespace in GetStatus
*/

        if ((cur->type == XML_ELEMENT_NODE)&&(ns==cur->ns)&&
	   ((name==NULL)||(xmlStrEqual(cur->name, (xmlChar*)name)))) return(cur);
	if(cur->children != NULL) {
	    ret = dsXmlFindNode(cur->children, name, ns);
	    if(ret != NULL) return(ret);	    
	}
	cur = cur->next;
    }
    
    if (!ns) return NULL;

    cur = parent;
    while(cur != NULL) {
        if ((cur->type == XML_ELEMENT_NODE)&&((ns==cur->ns)||(!xmlStrcmp(ns->href,cur->ns->href)))&&
	   ((name==NULL)||(xmlStrEqual(cur->name, (xmlChar*)name)))) return(cur);
	if(cur->children != NULL) {
	    ret = dsXmlFindNode(cur->children, name, ns);
	    if(ret != NULL) return(ret);	    
	}
	cur = cur->next;
    }

    return(NULL);
}

xmlNodePtr dsXmlFindNodeByHref(xmlNodePtr parent, const char *name, const char *ns) {
    xmlNodePtr cur;
    xmlNodePtr ret;

    cur = parent;
    while(cur != NULL) {
        if ((cur->type == XML_ELEMENT_NODE)&&
	   ((ns==NULL)||(!xmlStrcmp((xmlChar*)ns,cur->ns->href)))&&
	   ((name==NULL)||(xmlStrEqual(cur->name, (xmlChar*)name)))) return(cur);
	if ((cur->children != NULL)) {
	    ret = dsXmlFindNodeByHref(cur->children, name, ns);
	    if(ret != NULL) return(ret);	    
	}
	cur = cur->next;
    }

    return(NULL);
}

xmlNodePtr dsXmlFindSubNode(xmlNodePtr node, const char *name) {
    xmlNodePtr cur;

    ds_assert_null(node);
    ds_assert_null(name);
        
    for (cur = node->children; cur; cur = cur->next) {
        if ((cur->type == XML_ELEMENT_NODE)&&((name==NULL)||(xmlStrEqual(cur->name, (xmlChar*)name)))) return(cur);
    }
    return NULL;
}

xmlNodeSetPtr dsXmlFindSubNodes(xmlNodePtr node, const char *pname, const char *name) {
    unsigned int nodes = 0;
    xmlNodeSetPtr nodeset;
    xmlNodePtr cur;

    ds_assert_null(node);
    ds_assert_null(name);
    
    if (pname) {
	node = dsXmlFindSubNode(node, pname);
	if (!node) return NULL;
    }

    for (cur = node->children; cur; cur = cur->next) {
        if ((cur->type == XML_ELEMENT_NODE)&&((name==NULL)||(xmlStrEqual(cur->name, (xmlChar*)name)))) nodes++;
    }
    
    nodeset = (xmlNodeSetPtr)malloc(sizeof(xmlNodeSet));
    if (!nodeset) return nodeset;
    
    nodeset->nodeTab = (xmlNodePtr*)malloc(nodes*sizeof(xmlNodePtr));
    if (!nodeset->nodeTab) {
	free(nodeset);
	return NULL;
    }
    nodeset->nodeNr = nodes;
    nodeset->nodeMax = nodes;

    for (nodes = 0, cur = node->children; cur; cur = cur->next) {
        if ((cur->type == XML_ELEMENT_NODE)&&((name==NULL)||(xmlStrEqual(cur->name, (xmlChar*)name)))) nodeset->nodeTab[nodes++] = cur;
    }
    
    return nodeset;
}

void dsXmlFreeSubNodes(xmlNodeSetPtr nodes) {
    ds_assert_return(nodes);
    
    if (nodes->nodeTab) free(nodes->nodeTab);
    free(nodes);
}


const char *dsXmlGetPropString(xmlNodePtr node, const char *name, const char *default_value) {
    const char *str;
    xmlAttrPtr attr;
    
    ds_assert_default(name, default_value);
    
    if (!node) return default_value;
    
    attr=xmlHasProp(node, (xmlChar*)name);
    if (attr) {
	str=(char*)attr->children->content;
	if (str) return str;
    }
    return default_value;
}


void dsXmlSetPropString(xmlNodePtr node, const char *name, const char *value) {
    ds_assert_return(node);
    ds_assert_return(name);
    
    if (value) xmlSetProp(node, (xmlChar*)name, (xmlChar*)value);
    else xmlUnsetProp(node, (xmlChar*)name);
}

const char *dsXmlGetNodeString(xmlNodePtr node, const char *name, const char *default_value) {
    xmlNodePtr subnode;

    if (!node) return default_value;

    if (name) subnode = dsXmlFindSubNode(node, name);
    else subnode = node;
    if ((subnode)&&(subnode->children)&&(subnode->children->type == XML_TEXT_NODE)&&(subnode->children->content)) return (const char*)subnode->children->content;
    return default_value;
}

xmlNodePtr dsXmlSetNodeString(xmlNodePtr node, const char *name, const char *value) {
    xmlNodePtr child;

    ds_assert_null(node);
    ds_assert_null(name);

    if (value) {
	child = dsXmlFindSubNode(node, name);
        if (!child) {
    	    child = xmlNewNode(NULL, name);
	    if (child) xmlAddChild(node, child);
	    else return NULL;
	}
	xmlNodeSetContent(child, value);	    

	return child;
    } else {
	child = dsXmlFindSubNode(node, name);
	if (child) {
	    xmlUnlinkNode(child);
	    xmlFreeNode(child);
	}
	
	return NULL;
    }
}

int dsXmlGetPropInt(xmlNodePtr node, const char *name, int default_value) {
    char *str;
    xmlAttrPtr attr;

    ds_assert_default(name, default_value);

    if (!node) return default_value;
    
    attr=xmlHasProp(node, (xmlChar*)name);
    if (attr) {
	 str=(char*)attr->children->content;
	 if (str) return strtol(str, NULL, 0);
    }
    return default_value;
}


int dsXmlGetNodeInt(xmlNodePtr node, const char *name, int default_value) {
    const char *str;

    if (!node) return default_value;

    str = dsXmlGetNodeString(node, name, NULL);
    if (str) return strtol(str, NULL, 0);
    
    return default_value;    
}

void dsXmlSetPropInt(xmlNodePtr node, const char *name, int value) {
    char str[16];

    ds_assert_return(node);
    ds_assert_return(name);
    
    sprintf(str, "%i", value);
    xmlSetProp(node, (xmlChar*)name, (xmlChar*)str);
}

xmlNodePtr dsXmlSetNodeInt(xmlNodePtr node, const char *name, int value) {
    char str[16];

    ds_assert_null(node);
    ds_assert_null(name);
    
    sprintf(str, "%i", value);
    return dsXmlSetNodeString(node, name, str);
}


long dsXmlGetPropLong(xmlNodePtr node, const char *name, long default_value) {
    const char *str;
    xmlAttrPtr attr;

    ds_assert_default(name, default_value);

    if (!node) return default_value;
    
    attr=xmlHasProp(node, (xmlChar*)name);
    if (attr) {
	 str=(const char*)attr->children->content;
	 if (str) return strtol(str, NULL, 0);
    }
    return default_value;
}

long dsXmlGetNodeLong(xmlNodePtr node, const char *name, long default_value) {
    const char *str;

    if (!node) return default_value;

    str = dsXmlGetNodeString(node, name, NULL);
    if (str) return strtol(str, NULL, 0);
    
    return default_value;    
}

void dsXmlSetPropLong(xmlNodePtr node, const char *name, long value) {
    char str[16];

    ds_assert_return(node);
    ds_assert_return(name);
    
    sprintf(str, "%li", value);
    xmlSetProp(node, (xmlChar*)name, (xmlChar*)str);
}

xmlNodePtr dsXmlSetNodeLong(xmlNodePtr node, const char *name, long value) {
    char str[16];

    ds_assert_null(node);
    ds_assert_null(name);
    
    sprintf(str, "%li", value);
    return dsXmlSetNodeString(node, name, str);
}

double dsXmlGetPropDouble(xmlNodePtr node, const char *name, double default_value) {
    const char *str;
    xmlAttrPtr attr;
    double value;

    ds_assert_default(name, default_value);

    if (!node) return default_value;
    
    attr=xmlHasProp(node, (xmlChar*)name);
    if (attr) {
	 str=(const char*)attr->children->content;
	 if ((str)&&(sscanf(str,"%lg",&value)==1)) return value;
    }
    return default_value;
}

double dsXmlGetNodeDouble(xmlNodePtr node, const char *name, double default_value) {
    const char *str;
    double value;

    if (!node) return default_value;

    str = dsXmlGetNodeString(node, name, NULL);

    if ((str)&&(sscanf(str,"%lg",&value)==1)) return value;
    return default_value;    
}

void dsXmlSetPropDouble(xmlNodePtr node, const char *name, double value) {
    char str[16];

    ds_assert_return(node);
    ds_assert_return(name);
    
    sprintf(str, "%g", value);
    xmlSetProp(node, (xmlChar*)name, (xmlChar*)str);
}

xmlNodePtr dsXmlSetNodeDouble(xmlNodePtr node, const char *name, double value) {
    char str[16];

    ds_assert_null(node);
    ds_assert_null(name);
    
    sprintf(str, "%g", value);
    return dsXmlSetNodeString(node, name, str);
}

unsigned char dsXmlGetPropBoolean(xmlNodePtr node, const char *name, unsigned char default_value) {
    const char *str;
    xmlAttrPtr attr;

    ds_assert_default(name, default_value);

    if (!node) return default_value;
    
    attr=xmlHasProp(node, (xmlChar*)name);
    if (attr) {
	str=(const char*)attr->children->content;
	if (str) {
	    if ((str[0]=='T')||(str[0]=='t')||(str[0]=='1')) return 1;
	    else return 0;
	}
    }
    return default_value;
}

unsigned char dsXmlGetNodeBoolean(xmlNodePtr node, const char *name, unsigned char default_value) {
    const char *str;

    if (!node) return default_value;

    str = dsXmlGetNodeString(node, name, NULL);
    if (str) {
	if ((str[0]=='T')||(str[0]=='t')||(str[0]=='1')) return 1;
	else return 0;
    }
    return default_value;    
}

void dsXmlSetPropBoolean(xmlNodePtr node, const char *name, int value) {
    ds_assert_return(node);
    ds_assert_return(name);

    if (value) xmlSetProp(node, (xmlChar*)name, (xmlChar*)"true");
    else xmlSetProp(node, (xmlChar*)name, (xmlChar*)"false");
}

xmlNodePtr dsXmlSetNodeBoolean(xmlNodePtr node, const char *name, int value) {
    ds_assert_null(node);
    ds_assert_null(name);

    if (value) return dsXmlSetNodeString(node, name, (xmlChar*)"true");
    else return dsXmlSetNodeString(node, name, (xmlChar*)"false");
}


DSTime dsXmlGetPropDateTime(xmlNodePtr node, const char *name, DSTime time) {
    const char *str;

    ds_assert_default(name, time);
    ds_assert_default(time, time);

    if (!node) return time;

    str = dsXmlGetPropString(node, name, NULL);
    if ((str)&&(dsTimeSetDateTime(time, str))) return NULL;

    return time;
}

DSTime dsXmlGetNodeDateTime(xmlNodePtr node, const char *name, DSTime time) {
    const char *str;

    ds_assert_default(time, time);

    if (!node) return time;

    str = dsXmlGetNodeString(node, name, NULL);
    if ((str)&&(dsTimeSetDateTime(time, str))) return NULL;

    return time;
}

void dsXmlSetPropDateTime(xmlNodePtr node, const char *name, DSTime time) {
    char str[64];

    ds_assert_return(node);
    ds_assert_return(name);
    ds_assert_return(time);
    
    dsTimeGetDateTime(time, str);
    xmlSetProp(node, (xmlChar*)name, (xmlChar*)str);
}

xmlNodePtr dsXmlSetNodeDateTime(xmlNodePtr node, const char *name, DSTime time) {
    char str[64];

    ds_assert_null(node);
    ds_assert_null(name);
    ds_assert_null(time);
    
    dsTimeGetDateTime(time, str);
    return dsXmlSetNodeString(node, name, str);
}


DSDuration dsXmlGetPropDuration(xmlNodePtr node, const char *name, DSDuration duration) {
    const char *str;

    ds_assert_default(name, duration);
    ds_assert_default(duration, duration);

    if (!node) return duration;

    str = dsXmlGetPropString(node, name, NULL);
    if ((str)&&(dsDurationSet(duration, str))) return NULL;

    return duration;
}

DSDuration dsXmlGetNodeDuration(xmlNodePtr node, const char *name, DSDuration duration) {
    const char *str;

    ds_assert_default(duration, duration);

    if (!node) return duration;

    str = dsXmlGetNodeString(node, name, NULL);
    if ((str)&&(dsDurationSet(duration, str))) return NULL;

    return duration;
}

void dsXmlSetPropDuration(xmlNodePtr node, const char *name, DSDuration duration) {
    char str[64];

    ds_assert_return(node);
    ds_assert_return(name);
    ds_assert_return(duration);
    
    dsDurationGet(duration, str);
    xmlSetProp(node, (xmlChar*)name, (xmlChar*)str);
}

xmlNodePtr dsXmlSetNodeDuration(xmlNodePtr node, const char *name, DSDuration duration) {
    char str[64];

    ds_assert_null(node);
    ds_assert_null(name);
    ds_assert_null(duration);
    
    dsDurationGet(duration, str);
    return dsXmlSetNodeString(node, name, str);
}


DSDebugPriority dsXmlGetPropDebugLevel(xmlNodePtr node, const char *name, DSDebugPriority default_value) {
    DSDebugPriority level;
    const char *str;
    
    ds_assert_default(name, default_value);

    if (!node) return default_value;

    str = dsXmlGetPropString(node, name, NULL);
    if (str) {
	level = dsDebugGetLevelByName(str);
	if (level != (DSDebugPriority)-1) return level;
    }
    
    return default_value;
}

DSDebugPriority dsXmlGetNodeDebugLevel(xmlNodePtr node, const char *name, DSDebugPriority default_value) {
    DSDebugPriority level;
    const char *str;
    
    if (!node) return default_value;

    str = dsXmlGetNodeString(node, name, NULL);
    if (str) {
	level = dsDebugGetLevelByName(str);
	if (level != (DSDebugPriority)-1) return level;
    }
    
    return default_value;
}

void dsXmlSetPropDebugLevel(xmlNodePtr node, const char *name, DSDebugPriority value) {
    const char *str;

    ds_assert_return(node);
    ds_assert_return(name);
    
    str = dsDebugGetName(value);
    if (str) xmlSetProp(node, (xmlChar*)name, (xmlChar*)str);
}

xmlNodePtr dsXmlSetNodeDebugLevel(xmlNodePtr node, const char *name, DSDebugPriority value) {
    const char *str;

    ds_assert_null(node);
    ds_assert_null(name);
    
    str = dsDebugGetName(value);
    if (str) return dsXmlSetNodeString(node, name, str);
    return NULL;
}


int dsXmlDumpMemory(xmlNodePtr node, char **buf, size_t *size) {
    int bufsize;
    xmlDocPtr doc;
    xmlNodePtr copy;
    
    ds_assert_param(node);
    ds_assert_param(buf);

    doc = xmlNewDoc("1.0");    
    if (!doc) return DS_ERR_XML;

    copy = xmlCopyNode(node, 1);
    if (!copy) {
	xmlFreeDoc(doc);
	return DS_ERR_XML;
    }
    
    xmlDocSetRootElement(doc, copy);
    xmlDocDumpFormatMemoryEnc(doc, (xmlChar**)buf, &bufsize, "UTF-8", 1);
    xmlFreeDoc(doc);
    if (!*buf) return DS_ERR_XML;
    
    if (size) *size = bufsize;
    return 0;
}
