#ifndef _DS_XML_H
#define _DS_XML_H

#ifdef _DS_INCLUDE_XML_HEADERS
#include <libxml/tree.h>
#include <libxml/xpath.h>
#include <libxml/xpathInternals.h>
#include <libxml/schemasInternals.h>
#include <libxml/xmlschemas.h>
#include <libxml/parser.h>
#include <libxml/encoding.h>
#include <libxml/xmlsave.h>
#include <libxml/xmlstring.h>
#else /*  _DS_INCLUDE_XML_HEADERS */
typedef char xmlChar;
typedef void *xmlParserCtxtPtr;
typedef void *xmlDocPtr;
typedef void *xmlNodePtr;
typedef void *xmlNodeSetPtr;
typedef void *xmlNsPtr;
typedef void *xmlSchemaParserCtxtPtr;
typedef void *xmlSchemaPtr;
typedef void *xmlSchemaValidCtxtPtr;
typedef void *xmlXPathContextPtr;    

typedef void *xmlOutputWriteCallback;
typedef void *xmlOutputCloseCallback;
#endif /* _DS_INCLUDE_XML_HEADERS */

#define DS_XML_XSI_NAMESPACE "http://www.w3.org/2001/XMLSchema-instance"
#define DS_XML_XSD_NAMESPACE "http://www.w3.org/2001/XMLSchema"

#include <dsdebug.h>
#include <dstime.h>

/*
Formatted Output Problem
========================
    xmlBlanksDefault(0) should be called at initialization stage in order
    to allow indented output. However, this function should be called for
    each application thread. Since, XML libraries are not mandatory intialized
    the question is it Okey to put initialization in dsThread.
*/

int dsXmlInit();
void dsXmlDestroy();

const xmlChar *dsXmlStrrchr(const xmlChar *str, xmlChar ch);

xmlNodePtr dsXmlFindNode(xmlNodePtr parent, const char *name, const xmlNsPtr ns);
xmlNodePtr dsXmlFindNodeByHref(xmlNodePtr parent, const char *name, const char *ns);
xmlNodePtr dsXmlFindSubNode(xmlNodePtr node, const char *name);
xmlNodeSetPtr dsXmlFindSubNodes(xmlNodePtr node, const char *pname, const char *name);
void dsXmlFreeSubNodes(xmlNodeSetPtr nodes);

const char *dsXmlGetPropString(xmlNodePtr node, const char *name, const char *default_value);
const char *dsXmlGetNodeString(xmlNodePtr node, const char *name, const char *default_value);

int dsXmlGetPropInt(xmlNodePtr node, const char *name, int default_value);
int dsXmlGetNodeInt(xmlNodePtr node, const char *name, int default_value);
long dsXmlGetPropLong(xmlNodePtr node, const char *name, long default_value);
long dsXmlGetNodeLong(xmlNodePtr node, const char *name, long default_value);
double dsXmlGetPropDouble(xmlNodePtr node, const char *name, double default_value);
double dsXmlGetNodeDouble(xmlNodePtr node, const char *name, double default_value);
unsigned char dsXmlGetPropBoolean(xmlNodePtr node, const char *name, unsigned char default_value);
unsigned char dsXmlGetNodeBoolean(xmlNodePtr node, const char *name, unsigned char default_value);
DSTime dsXmlGetPropDateTime(xmlNodePtr node, const char *name, DSTime time);
DSTime dsXmlGetNodeDateTime(xmlNodePtr node, const char *name, DSTime time);
DSDuration dsXmlGetPropDuration(xmlNodePtr node, const char *name, DSDuration duration);
DSDuration dsXmlGetNodeDuration(xmlNodePtr node, const char *name, DSDuration duration);
DSDebugPriority dsXmlGetPropDebugLevel(xmlNodePtr node, const char *name, DSDebugPriority default_value);
DSDebugPriority dsXmlGetNodeDebugLevel(xmlNodePtr node, const char *name, DSDebugPriority default_value);

void dsXmlSetPropString(xmlNodePtr node, const char *name, const char *value);
xmlNodePtr dsXmlSetNodeString(xmlNodePtr node, const char *name, const char *value);
void dsXmlSetPropInt(xmlNodePtr node, const char *name, int value);
xmlNodePtr dsXmlSetNodeInt(xmlNodePtr node, const char *name, int value);
void dsXmlSetPropLong(xmlNodePtr node, const char *name, long value);
xmlNodePtr dsXmlSetNodeLong(xmlNodePtr node, const char *name, long value);
void dsXmlSetPropDouble(xmlNodePtr node, const char *name, double value);
xmlNodePtr dsXmlSetNodeDouble(xmlNodePtr node, const char *name, double value);
void dsXmlSetPropBoolean(xmlNodePtr node, const char *name, int value);
xmlNodePtr dsXmlSetNodeBoolean(xmlNodePtr node, const char *name, int value);
xmlNodePtr dsXmlSetNodeBoolean(xmlNodePtr node, const char *name, int value);
void dsXmlSetPropDateTime(xmlNodePtr node, const char *name, DSTime time);
xmlNodePtr dsXmlSetNodeDateTime(xmlNodePtr node, const char *name, DSTime time);
void dsXmlSetPropDuration(xmlNodePtr node, const char *name, DSDuration duration);
xmlNodePtr dsXmlSetNodeDuration(xmlNodePtr node, const char *name, DSDuration duration);
void dsXmlSetPropDebugLevel(xmlNodePtr node, const char *name, DSDebugPriority value);
xmlNodePtr dsXmlSetNodeDebugLevel(xmlNodePtr node, const char *name, DSDebugPriority value);

int dsXmlDumpMemory(xmlNodePtr node, char **buf, size_t *size);

#endif /* _DS_XML_H */
