#include <opclib.h>

#include <dsassert.h>
#include <dsdebug.h>
#include <dserror.h>

#include <dsdata.h>
#include <dsdatahandler.h>

#include <opcdata.h>
#include <opcdatainfo.h>

OPCDataInfo opcDataInfoCreate(DSMallocContext mctx) {
    OPCDataInfo info;
    
    info = (OPCDataInfo)dsMalloc(mctx, sizeof(OPCDataInfoS));
    if (info) opcDataInfoSetup(info);
    return info;
}

void opcDataInfoDestroy(DSMallocContext mctx, OPCDataInfo info) {
    ds_assert_return(info);
    dsFree(mctx, info);
}

int opcDataInfoSetup(OPCDataInfo info) {
    ds_assert_param(info);
    dsTimeCurrentTime(&info->time);
    opcQualityCopy(&info->quality, OPCQualityGood);
    return 0;
}

int opcDataInfoSetTime(OPCDataInfo info, DSTime time) {
    ds_assert_param(info);
    ds_assert_param(time);
    return dsTimeCopy(&info->time, time);
}

int opcDataInfoSetQuality(OPCDataInfo info, OPCQuality quality) {
    ds_assert_param(info);
    ds_assert_param(quality);
    return opcQualityCopy(&info->quality, quality);
}

DSTimeConst opcDataInfoGetTime(OPCDataInfo info) {
    ds_assert_default(info, NULL);
    return &info->time;
}

OPCQualityConst opcDataInfoGetQuality(OPCDataInfo info) {
    ds_assert_default(info, NULL);
    return &info->quality;
}

DSData opcDataInfoCreateEnclosed(DSMallocContext mctx) {
    OPCDataInfo info;
    DSData data;
    
    info = opcDataInfoCreate(mctx);
    if (!info) return NULL;
    
    data = opcDataInfoEnclosureCreate(info, mctx);
    if (!data) {
	opcDataInfoDestroy(mctx, info);
	return NULL;
    }
    
    opcDataInfoEnclosurePassOwnership(data, mctx);
    
    return data;
}

DSData opcDataInfoEnclosureCreate(OPCDataInfo info, DSMallocContext mctx) {
    DSData data;
    
    data = dsDataCreate(mctx);
    if (data) opcDataInfoEnclosureSetup(data, info);
    return data;
}

int opcDataInfoEnclosureSetup(DSData data, OPCDataInfo info) {
    ds_assert_param(data);
    ds_assert_param(info);
    
    dsDataSetPointer(data, NULL, info, sizeof(OPCDataInfoS));
    dsDataSetHandler(data, DS_DATA_HANDLER_STATIC_DATA);
    dsDataSetUserType(data, OPC_DATA_TYPE_DATAINFO);

    return 0;
}

int opcDataInfoEnclosurePassOwnership(DSData data, DSMallocContext owner) {
    ds_assert_param(data);
    ds_assert_param(dsDataCheckUserType(data, OPC_DATA_TYPE_DATAINFO));

    dsDataSetOwner(data, owner);
    dsDataSetHandler(data, NULL);
    return 0;
}

OPCDataInfo opcDataInfoEnclosureExtract(DSData data) {
    ds_assert_default(dsDataCheckUserType(data, OPC_DATA_TYPE_DATAINFO), NULL);

    return (OPCDataInfo)data->data;
}

int opcDataInfoCopy(OPCDataInfo dst, OPCDataInfo src) {
    ds_assert_param(src);
    ds_assert_param(dst);
    memcpy(dst, src, sizeof(OPCDataInfoS));

}

