#include <stdio.h>

#include <opclib.h>
#include <dsassert.h>
#include <dshashtable.h>
#include <dsstringlist.h>

#include <dstime.h>
#include <dstools.h>
#include <opctree.h>
#include <opcitem.h>

#include <opcinterface.h>


OPCInterface opcCreateInterface(OPCContext ctx) {
    OPCInterface server_interface;
    ds_assert_null(ctx);
    
    server_interface = (OPCInterface)malloc(sizeof(struct OPCInterfaceT));
    if (!server_interface) return NULL;
    server_interface->sctx = ctx;

    server_interface->start_time = NULL;
    server_interface->locale = NULL;
    server_interface->current_locale = NULL;
    server_interface->last_error = 0;

    return server_interface;
}

void opcDestroyInterface(OPCInterface ctx) {
    ds_assert_return(ctx);

    if (ctx->start_time) free(ctx->start_time);
    if (ctx->locale) opcLocaleDestroy(ctx->locale);
    free(ctx);
}


const char *opcInterfaceGetVersion(OPCInterface ctx) {
    ds_assert_null(ctx);
    ds_assert_null(ctx->sctx);
    
    return opcVersionGet(ctx->sctx->version);
}

const char *opcInterfaceGetStartTime(OPCInterface ctx) {
    ds_assert_null(ctx);
    ds_assert_null(ctx->sctx);
    
    ctx->start_time = dsTimeGetDateTime(ctx->sctx->start_time,NULL);
    return ctx->start_time;
}

const char *opcInterfaceSetLocale(OPCInterface ctx, char *attr) {
    ds_assert_null(ctx);
    ds_assert_null(ctx->sctx);
    
    ctx->locale=opcLocalesGetLocale(ctx->sctx->locales, attr);
    if (ctx->locale) {
	ctx->current_locale = ctx->locale->locale;
	return ctx->current_locale;
    } else {
	ctx->current_locale = opc_default_locale;
	return opc_default_locale;
    }
}

const char *opcInterfaceGetCharset(OPCInterface ctx) {
    ds_assert_null(ctx);
    if (ctx->locale) return ctx->locale->charset;
    return NULL;
}

dschar opcInterfaceGetStatusInfo(OPCInterface ctx) {
    ds_assert_null(ctx);
    ds_assert_null(ctx->current_locale);
    return opcStateGetStatusInfo(ctx->sctx->state,ctx->current_locale);
}

dschar opcInterfaceGetVendorInfo(OPCInterface ctx) {
    ds_assert_null(ctx);
    ds_assert_null(ctx->current_locale);
    return opcStateGetVendorInfo(ctx->sctx->state,ctx->current_locale);
}

dschar opcInterfaceGetStatus(OPCInterface ctx) {
    ds_assert_null(ctx);
    ds_assert_null(ctx->sctx);
    ds_assert_null(ctx->current_locale);
    return opcStateGetStatus(ctx->sctx->state,ctx->current_locale);
}

int opcInterfaceListSupportedLocales(OPCInterface ctx, OPCLocalesListFunctionT list_function, void *attr) {
    ds_assert_param(ctx);
    ds_assert_param(ctx->sctx);
    return opcLocalesList(ctx->sctx->locales, list_function, attr);
}

int opcInterfaceListSupportedInterfaces(OPCInterface ctx, OPCVersionListFunctionT list_function, void *attr) {
    ds_assert_param(ctx);
    ds_assert_param(ctx->sctx);
    return opcVersionList(ctx->sctx->version, list_function, attr);
}

OPCConfigurationContext opcInterfaceGetConfiguration(OPCInterface ctx) {
    ds_assert_null(ctx);
    ds_assert_null(ctx->sctx);
    return ctx->sctx->config;
}

OPCTreeContext opcInterfaceGetServerTree(OPCInterface ctx) {
    ds_assert_null(ctx);
    ds_assert_null(ctx->sctx);
    return ctx->sctx->tree;
}

OPCItemContext opcInterfaceGetItem(OPCInterface ctx, const char *path, const char *name) {
    OPCTreeContext tree;
    OPCItemContext ictx;
    size_t size = 2;
    char *tmp;
    
    ds_assert_null(ctx);
    tree = opcInterfaceGetServerTree(ctx);
    if (!tree) return NULL;
    

    if (name) size+=strlen(name);
    if (path) size+=strlen(path);

    if ((name)&&(path)) {
	tmp = (xmlChar*)malloc((size)*sizeof(char));
	if (!tmp) return NULL;

	if (name[0]=='/') sprintf(tmp,"%s%s",path,name);
	else sprintf(tmp,"%s/%s",path,name);
    } else if (name) {
	if ((name[0]=='/')||(name[0]==0)) tmp=(char*)name;
	else {
	    tmp = (xmlChar*)malloc((size)*sizeof(char));
	    if (!tmp) return NULL;

	    sprintf(tmp,"/%s",name);
	}
    } else if (path) tmp=(char*)path;
    else return NULL;

    ictx = opcTreeGetItem(tree,tmp);
    if ((tmp!=name)&&(tmp!=path)) free(tmp);

    return ictx;
}

void opcInterfaceReturnItem(OPCInterface ctx, OPCItemContext ictx) {
    ds_assert_return(ictx);
    opcItemReturn(ictx);
}

OPCPropertiesContext opcInterfaceGetProperties(OPCInterface ctx) {
    ds_assert_null(ctx);
    ds_assert_null(ctx->sctx);
    return ctx->sctx->properties;
}

/*
OPCRequestContext opcInterfaceCreateReadRequest(OPCInterface ctx);
int opcInterfaceStoreRequest(OPCInterface ctx, const char *key, OPCRequestContext request);
OPCRequestContext opcInterfaceGetRequest(OPCInterface ctx, const char *key);
*/
