#ifndef _OPC_ITEM_H
#define _OPC_ITEM_H

typedef enum OPCItemBrowseFilterT OPCItemBrowseFilter;
enum OPCItemBrowseFilterT {
    ALL = 0,
    BRANCH,
    ITEM
};

typedef struct OPCItemContextT *OPCItemContext;

#include <dslock.h>
#include <dstime.h>
#include <dslocale.h>
#include <dscondition.h>
#include <dsstringlist.h>

#include <dsbinarytype.h>
#include <dsdatamanager.h>
#include <dsdatamanagerrep.h>
#include <dsdatamanagerfast.h>
#include <dsdatachunk.h>
#include <dsdataconverter.h>

#include <opcmodule.h>

#include <opcitemproperties.h>
#include <opctree.h>
#include <opcquality.h>
#include <opcdatainfo.h>

#define OPC_ITEM_LINKED_FLAG 1
#define OPC_ITEM_CONNECTED_FLAG 2

typedef int (*OPCVendorFilter)(OPCItemContext, const char *);



struct OPCItemContextT {
    DSModuleS module;
    
    char *fullname; /* hashed, should be first */
    char *name;
    dslchar title;
    
    DSLockContext lock;
    DSConditionContext client_cond;
    
    DSStringList subitems;
    
    OPCTreeContext tree;

    OPCVendorFilter vendor_filter;

    OPCDataInfoS info;
    DSDataManagerContext dm;
    OPCItemPropertiesContext prop;
    
    unsigned long clients;
    
    DSDataChunkForm native_representation;
    DSDataChunkForm opcxml_representation;
    DSDataChunkForm representation;

    unsigned char status;	/*  1 Bit: Linked in tree,
				    2 Bit: Real Item */
			

    unsigned char manual;	/* By default all items in automatic mode, this means
				if parrent item are unregistered???? */
    
    unsigned char mark;		/* Marked from deletion. If in FreeContext this flag set then
				Item is locked and alredy unlinked from tree */
};
typedef struct OPCItemContextT OPCItemContextS;

// Locking functions
OPCItemContext opcItemCreateContext(DSMallocContext mctx);
int opcItemInitContext(OPCItemContext ctx, const char *fullname, const char *title);
int opcItemFreeContext(OPCItemContext ctx);
void opcItemDestroyContext(OPCItemContext ctx);

int opcItemLock(OPCItemContext ctx);
int opcItemUnlock(OPCItemContext ctx);
int opcItemGet(OPCItemContext ctx); /* Should be called only inderect truth opcTreeGetItem */
int opcItemReturn(OPCItemContext ctx);

int opcItemSetAssociation(OPCItemContext ctx);
int opcItemSetTitle(OPCItemContext ctx, const char *title);
int opcItemSetManualMode(OPCItemContext ctx);
int opcItemResetManualMode(OPCItemContext ctx);
int opcItemSetVendorFilter(OPCItemContext ctx, OPCVendorFilter vendor_filter);

char *opcItemGetFullName(OPCItemContext ctx); /* Should be free'd */
const char *opcItemGetFullNamePtr(OPCItemContext ctx); /* For internal use only */
int opcItemBrowse(OPCItemContext ctx, OPCItemBrowseFilter browse_filter, const char *name_filter, const char *vendor_filter, DSStringList *result);

// Non-Locking functions. Should be locked from caller.

int opcItemHaveChilds(OPCItemContext ctx);
int opcItemIsConnected(OPCItemContext ctx);
dschar opcItemGetTitle(OPCItemContext ctx, const char *locale);

// Internal functions called from opcTree & opcItem and synchronised from caller functions 
int opcItemBrowseFilter(OPCItemContext ctx, OPCItemBrowseFilter filter);
int opcItemNameFilter(OPCItemContext ctx, const char *filter);

int opcItemLink(OPCItemContext ctx, OPCItemContext ictx);
int opcItemUnlink(OPCItemContext ctx, OPCItemContext ictx);

int opcItemSetTree(OPCItemContext ctx, OPCTreeContext tree);
int opcItemHaveManualSubChilds(OPCItemContext ctx);

void opcItemMark(OPCItemContext ctx);
void opcItemMarkSubTree(OPCItemContext ctx);
void opcItemUnMark(OPCItemContext ctx);
void opcItemUnMarkSubTree(OPCItemContext ctx);
int opcItemIsMarked(OPCItemContext ctx);

int opcItemMarkChildsCallback(OPCItemContext ctx, OPCItemContext ictx);
int opcItemMarkParrentsCallback(OPCItemContext ctx, OPCItemContext ictx);

int opcItemCheckConnection(OPCItemContext ctx);
int opcItemCheckAge(OPCItemContext ctx, const DSTime current_time, const DSDuration maxage);


DSDataManagerFastContext opcItemRegisterFastDataManager(OPCItemContext ctx, DSBinaryType type, DSDataManagerFastOptions options, unsigned long flags);
DSDataManagerRepContext opcItemRegisterLegacyDataManager(OPCItemContext ctx, DSBinaryType type, DSDataManagerRepOptions options, unsigned long flags);
DSDataManagerContext opcItemRegisterDataManager(OPCItemContext ctx, DSDataManagerContext dm);
void opcItemRemoveDataManager(OPCItemContext ctx);

int opcItemEnableDataManager(OPCItemContext ctx);
int opcItemDisableDataManager(OPCItemContext ctx);

int opcItemPushData(OPCItemContext ctx, DSDataChunkForm repid, DSData info, DSData data);
int opcItemPushNativeData(OPCItemContext ctx, DSData info, DSData data);
int opcItemPushNativePointer(OPCItemContext ctx, DSData info, void *ptr, size_t size);
int opcItemPushNativeAllocation(OPCItemContext ctx, DSData info, DSAllocatorContext allocator, DSAllocation allocation);
int opcItemPushOPCXMLData(OPCItemContext ctx, DSData info, DSData data);
int opcItemPushOPCXMLNode(OPCItemContext ctx, DSData info, xmlNodePtr node);
int opcItemPushOPCXMLString(OPCItemContext ctx, DSData info, xmlChar *string);

DSTimeConst opcItemGetTimestamp(OPCItemContext ctx);

int opcItemGetData(OPCItemContext ctx, DSData *info, DSData *data);
void opcItemReturnData(OPCItemContext ctx, DSData info);


// Properties
//int opcItemRegisterProperty(OPCItemContext ctx, "propertyname");
//int opcItemRegisterNewProperty(OPCItemContext ctx, 
// int opcItemRegisterStandardFormats(OPCItemContext ctx)
//int opcItemRegisterFormat(OPCItemContext ctx, 
//


#endif /* _OPC_ITEM_H */
