#include <dslib.h>
#include <dsassert.h>
#include <dsdebug.h>

#include <dsallocatorsimple.h>
#include <opcitemproperties.h>

OPCItemPropertiesContext opcItemPropertiesCreateContext(DSMallocContext mctx) {
    OPCItemPropertiesContext ctx;

    ctx = dsNewClass(mctx, OPCItemPropertiesContext);
    if (ctx) ctx->malloc=mctx;
    return ctx;
}

int opcItemPropertiesInitContext(OPCItemPropertiesContext ctx, OPCPropertiesContext properties) {
    int err;
    OPCPropertyContext prop;
    OPCPropertyID i, maxid;
    
    ds_assert_param(ctx);

    ctx->lock = dsLockCreateContext(ctx->malloc);
    if (!ctx->lock) {
	opcItemPropertiesFreeContext(ctx);
	return DS_ERR_LOCK_CREATE;
    }

    err = dsLockInitContext(ctx->lock);
    if (err) {
	opcItemPropertiesFreeContext(ctx);
	return err;
    }

    ctx->converter = dsDataConverterCreateOPCXML(DS_MODULE(ctx)->malloc, NULL);
    if (!ctx->converter) {
	opcItemPropertiesFreeContext(ctx);
	return dsErrorModule(DS_ERR_CREATE, DS_MODULE_DataConverter);
    }

    ctx->allocator = (DSAllocatorContext)dsAllocatorSimpleCreateContext(DS_MODULE(ctx)->malloc);
    if (!ctx->allocator) {
	opcItemPropertiesFreeContext(ctx);
	return dsErrorModule(DS_ERR_CREATE, DS_MODULE_AllocatorSimple);
    }
	    
    err = dsAllocatorSimpleInitContext((DSAllocatorSimpleContext)ctx->allocator, NULL);
    if (err) {
	opcItemPropertiesFreeContext(ctx);
	return err;
    }

    maxid = properties->curid;
    ctx->maxid = maxid;
    ctx->opcproperties = properties;

    ctx->properties = (OPCItemPropertyContext)dsMalloc(ctx->malloc, maxid*sizeof(OPCItemPropertyContextS));
    if (!ctx->properties) return DS_ERR_MEM;
    
    for (i=0;i<maxid;i++) {
	prop = opcPropertiesGetProperty(properties, i);
	if (prop) {
	    err = opcItemPropertyInit(ctx->properties+i, prop);
	    if (err) dsDebugServerError(DS_DEBUG_ERROR, err, "opcItemProperties(%p): CreateProperty \"%s\" is failed, ", ctx, prop->name);
	} else memset(ctx->properties+i, 0, sizeof(OPCItemPropertyContextS));
    }

    return 0;
}

void opcItemPropertiesFreeContext(OPCItemPropertiesContext ctx) {
    OPCPropertyID i, maxid;
    ds_assert_return(ctx);

    if (ctx->properties) {
	maxid = ctx->maxid;
	for (i=0;i<maxid;i++) opcItemPropertyFree(ctx->properties+i);
	dsFree(ctx->malloc, ctx->properties);    
    }

    if (ctx->lock) dsLockDestroyContext(ctx->lock);
    if (ctx->allocator) dsAllocatorDestroyContext(ctx->allocator);
    if (ctx->converter) dsDataConverterDestroyContext(ctx->converter);

    memset(ctx, 0, sizeof(OPCItemPropertiesContextS));
    
}

void opcItemPropertiesDestroyContext(OPCItemPropertiesContext ctx) {
    ds_assert_return(ctx);
    
    opcItemPropertiesFreeContext(ctx);
    dsFree(ctx->malloc,ctx);
}
