#include <opclib.h>
#include <dsassert.h>
#include <dsdebug.h>
#include <dserror.h>

#include <opcproperty.h>


void opcPropertyZeroContext(OPCPropertyContext ctx) {
    ds_assert_return(ctx);
    ctx->id=0;    /* 0 - property by default (unconnected) */
    ctx->name=NULL;
    ctx->description=NULL;
    ctx->xmlnamespace=NULL;
    ctx->type=NULL;
}

OPCPropertyContext opcPropertyCreateContext(DSMallocContext mctx) {
    OPCPropertyContext ctx;

    ctx = (OPCPropertyContext)dsMalloc(mctx,sizeof(struct OPCPropertyContextT));
    if (ctx) {
	ctx->malloc=mctx;
	opcPropertyZeroContext(ctx);
    }

    dsDebugServer(DS_DEBUG_DEBUG, "opcPropertyCreateContext(%p)", ctx);    

    return ctx;
}

int opcPropertyInitContext(OPCPropertyContext ctx, const char *name, OPCPropertyType ptype, DSBinaryType type) {
    ds_assert_param(ctx);
    ds_assert_param(name);
    
    opcPropertyZeroContext(ctx);

    if (dsExchangeString(ctx->malloc,&ctx->name,name)) return DS_ERR_MEM;
    ctx->type = type;
    ctx->proptype = ptype;
    
    dsDebugServer(DS_DEBUG_DEBUG, "opcPropertyInitContext(%p): %s Ok", ctx, name);    
    return 0;
}

int opcPropertyInitStandardContext(OPCPropertyContext ctx, const char *name,  OPCPropertyType ptype, DSBinaryTypeStandard datatype, const void *attr) {
    int err;
    DSBinaryType type;
    
    ds_assert_param(ctx);
    ds_assert_param(name);

    dsDebugServer(DS_DEBUG_NOTICE, "opcPropertyInitContext(%p): %s", ctx, name);

    type = dsBinaryTypeStandardCreate(ctx->malloc, datatype, attr);

/*  
    Some parameters are not typed
    if (!type) return dsErrorModule(DS_ERR_CREATE, DS_MODULE_BinaryType);
*/
    err = opcPropertyInitContext(ctx, name, ptype, type);
    if (err) dsBinaryTypeDestroy(type);
    
    return err;
}

void opcPropertyFreeContext(OPCPropertyContext ctx) {
    dsDebugServer(DS_DEBUG_DEBUG, "opcPropertyFreeContext(%p)", ctx);    
    ds_assert_return(ctx);
    if (ctx->type) dsBinaryTypeDestroy(ctx->type);
    if (ctx->name) dsExchangeString(ctx->malloc,&ctx->name,NULL);
    if (ctx->xmlnamespace) dsExchangeString(ctx->malloc,&ctx->xmlnamespace,NULL);
}

void opcPropertyDestroyContext(OPCPropertyContext ctx) {
    dsDebugServer(DS_DEBUG_DEBUG, "opcPropertyDestroyContext(%p)", ctx);    
    opcPropertyFreeContext(ctx);
    dsFree(ctx->malloc,ctx);
}

int opcPropertySetNamespace(OPCPropertyContext ctx, const char *xmlnamespace) {
    ds_assert_param(ctx);
    if (dsExchangeString(ctx->malloc,&ctx->xmlnamespace,xmlnamespace)) return DS_ERR_MEM;
    return 0;
}

int opcPropertySetId(OPCPropertyContext ctx, OPCPropertyID id) {
    ds_assert_param(ctx);
    ctx->id = id;    
    return 0;
}

int opcPropertySetOPCId(OPCPropertyContext ctx, unsigned short int id) {
    ds_assert_param(ctx);
    ctx->opcid = id;    
    return 0;
}

int opcPropertySetDescription(OPCPropertyContext ctx, const char *description) {
    ds_assert_param(ctx);
    if (dsLocaleExchangeString(ctx->malloc, &ctx->description, description)) return DS_ERR_MEM;
    return 0;
}

OPCPropertyType opcPropertyGetType(OPCPropertyContext ctx) {
    ds_assert_default(ctx, OPC_PROPERTY_TYPE_NOTSUPPORTED);
    return ctx->proptype;
}

const char *opcPropertyGetName(OPCPropertyContext ctx) {
    ds_assert_default(ctx, NULL);
    return ctx->name;
}

const char *opcPropertyGetDescription(OPCPropertyContext ctx) {
    ds_assert_default(ctx, NULL);
    return ctx->description?ctx->description:"Property";
}


//opcPropertyCreateInstance()

//opcPropertyDestroyInstance()
