#ifndef _OPC_PROPERTY_H
#define _OPC_PROPERTY_H

#include <dsmalloc.h>
#include <dslocale.h>
#include <dsbinarytype.h>
#include <dsbinarytypestandard.h>

typedef int OPCPropertyID;
#define OPC_PROPERTY_ID_INVALID ((OPCPropertyID)-1)

typedef struct OPCPropertyContextT *OPCPropertyContext;


enum OPCPropertyTypeT {
    OPC_PROPERTY_TYPE_NOTSUPPORTED = 0,
    OPC_PROPERTY_TYPE_STANDARD,
    OPC_PROPERTY_TYPE_VALUE,
    OPC_PROPERTY_TYPE_QUALITY,
    OPC_PROPERTY_TYPE_TIMESTAMP,
    OPC_PROPERTY_TYPE_TIMEZONE,
    OPC_PROPERTY_TYPE_DATATYPE
};
typedef enum OPCPropertyTypeT OPCPropertyType;

struct OPCPropertyContextT {
    char *name;
    dslchar description;
    char *xmlnamespace;
    OPCPropertyType proptype;
    unsigned short int opcid;
    DSBinaryType type;
    DSMallocContext malloc;
    OPCPropertyID id;
};

OPCPropertyContext opcPropertyCreateContext(DSMallocContext mctx);
int opcPropertyInitContext(OPCPropertyContext ctx, const char *name, OPCPropertyType ptype, DSBinaryType type); /* type is eaten */
int opcPropertyInitStandardContext(OPCPropertyContext ctx, const char *name, OPCPropertyType ptype, DSBinaryTypeStandard datatype, const void *attr);
void opcPropertyFreeContext(OPCPropertyContext ctx);
void opcPropertyDestroyContext(OPCPropertyContext ctx);

int opcPropertySetNamespace(OPCPropertyContext ctx, const char *xmlnamespace);
int opcPropertySetOPCId(OPCPropertyContext ctx, unsigned short int id);
int opcPropertySetId(OPCPropertyContext ctx, OPCPropertyID id);
int opcPropertySetDescription(OPCPropertyContext ctx, const char *description);

OPCPropertyType opcPropertyGetType(OPCPropertyContext ctx);
const char *opcPropertyGetName(OPCPropertyContext ctx);
const char *opcPropertyGetDescription(OPCPropertyContext ctx);

#endif /* _OPC_PROPERTY_H */
