#include <stdio.h>

#include <opclib.h>

#include <dsassert.h>
#include <dshashtable.h>
#include <dsstringlist.h>
#include <dslocale.h>
#include <dstime.h>

#include <opcstate.h>
#include <opclocale.h>
#include <opctree.h>

#include <opcserverinterface.h>

int opcServerStartTimeSet(OPCContext ctx) {
    ds_assert_param(ctx);
    
    if (!ctx->start_time) {
	ctx->start_time=dsTimeCreate(ctx->malloc);
	if (!ctx->start_time) return DS_ERR_MEM;
    }
    dsTimeCurrentTime(ctx->start_time);
    return 0;
}

int opcServerStateSet(OPCContext ctx, unsigned int state, const char *status_str) {
    int err;
    ds_assert_param(ctx);
    
    if (!ctx->state) {
	ctx->state=opcStateCreateContext(ctx->malloc);
	if (!ctx->state) return DS_ERR_MEM;
	err=opcStateInitContext(ctx->state,ctx);
	if (err) return err;
    }
    return opcStateSetStatus(ctx->state,state,status_str);
}

int opcServerVendorInfoSet(OPCContext ctx, const char *vi) {
    int err;
    ds_assert_param(ctx);

    if (!ctx->state) {
	ctx->state=opcStateCreateContext(ctx->malloc);
	if (!ctx->state) return DS_ERR_MEM;
	err=opcStateInitContext(ctx->state,ctx);
	if (err) return err;
    }
    
    return opcStateSetVendor(ctx->state,vi);
}

int opcServerVersionSet(OPCContext ctx, const char *major, const char *minor, unsigned long build) {
    int err=0;
    int i=0;
    
    ds_assert_param(ctx);
    
    if (!ctx->version) {
        ctx->version=opcVersionInit(ctx->malloc,ctx);
	if (!ctx->version) return DS_ERR_MEM;
    } else {
	opcVersionInterfacesClear(ctx->version);
    }
    return opcVersionSet(ctx->version,major,minor,build);
}

int opcServerInterfaceAdd(OPCContext ctx, const char *interface_version, const char **extensions) {
    return opcVersionInterfaceAdd(ctx->version,interface_version,extensions);
}


int opcServerLocalesAdd(OPCContext ctx, const char **locales) {
    int err;
    
    ds_assert_param(ctx);
    
    if (!ctx->locales) {
	ctx->locales=opcLocalesCreateContext(ctx->malloc);
	if (!ctx->locales) return DS_ERR_MEM;
	err=opcLocalesInitContext(ctx->locales);
	if (err) return err;
    }
    
    return opcLocalesAddLocales(ctx->locales,locales);
}

int opcServerInitTree(OPCContext ctx) {
    int err;
    OPCTreeContext tree;
    ds_assert_param(ctx);
    
    tree = opcTreeCreateContext(ctx->malloc);
    if (!tree) return DS_ERR_MEM;
    
    err = opcTreeInitContext(tree,ctx);
    if (err) opcTreeDestroyContext(tree);
    else ctx->tree=tree;
    return err;
}


int opcServerRegisterItem(OPCContext ctx, OPCItemContext *ictx, const char *path, const char *name) {
    int err;
    OPCItemContext item;
    char *fullname, *title;
    size_t i;
    unsigned char flag = 0;
    char *fp;
        
    ds_assert_param(ctx);
    ds_assert_param(ctx->tree);
    ds_assert_param(name);
    
    if (path) {
	for (i=strlen(path);((i>0)&&(path[i-1]=='/'));i--) flag=1;
    } else {
	flag=1;
	i=0;
    }
    
    if (flag) {
	fp = (char*)dsMalloca((i+1)*sizeof(char));
	if (!fp) return DS_ERR_MEM;
	strncpy(fp,path,i);
	fp[i]=0;
    }  else fp = (char*)path;
     
    item=opcItemCreateContext(ctx->malloc);
    fullname=(char*)dsMalloca((i+strlen(name)+2)*sizeof(char));
    if ((!fullname)||(!item)) {
	if (fullname) free(fullname);
	if (item) opcItemDestroyContext(item);
	if (flag) free((char*)path);
	return OPC_ERR_ITEM_CREATE;
    }

    sprintf(fullname,"%s/%s",fp,name);
    title = fullname + i + 1;

    err=opcItemInitContext(item, fullname, title);
    dsFreea(fullname);
    
    if (!err) err=opcTreeRegisterItem(ctx->tree,item,fp);
    if (flag) dsFreea(fp);
    
    if (!err) *ictx=item;
    else {
	opcItemDestroyContext(item);	
	*ictx=NULL;
    }
    return err;
}

int opcServerDestroyItem(OPCContext ctx, OPCItemContext ictx) {
    ds_assert_param(ctx);
    ds_assert_param(ictx);
    
    opcItemDestroyContext(ictx);
    return 0;
}

