#define _OPC_STANDARD_PROPERTIES_C

#include <opclib.h>

#include <dsassert.h>
#include <dsbinarytypestandard.h>

#include <opcstandardproperties.h>

static const char *opc_property_enumeration_quality[] = { 
    "bad", "badConfigurationError", "badNotConnected", 
    "badDeviceFailure", "badSensorFailure", "badLastKnownValue", 
    "badCommFailure", "badOutOfService", "uncertain",
    "uncertainLastUsableValue", "uncertainSensorNotAccurate", "uncertainEUExceeded",
    "uncertainSubNormal", "good", "goodLocalOverride", NULL
};


static const char *opc_property_enumeration_limitstatus[] = {
    "none","low","high","constant", NULL
};


static const char * opc_property_enumeration_accessrights[] = {
    "unknown", "readable", "writable", "readWritable", NULL
};

static const char * opc_property_enumeration_eutype[] = {
    "noEnum","analog","enumerated",NULL
};

static struct OPCStandardPropertyT opc_property[] = {
    {"dataType",NULL,OPC_PROPERTY_TYPE_DATATYPE,1,DS_BINARY_TYPE_STANDARD_QNAME,NULL,""},
    {"value",NULL,OPC_PROPERTY_TYPE_VALUE,2,0,NULL,""},
    {"quality",NULL,OPC_PROPERTY_TYPE_QUALITY,3,DS_BINARY_TYPE_STANDARD_ENUMERATION,opc_property_enumeration_quality,""},
    {"timestamp",NULL,OPC_PROPERTY_TYPE_TIMESTAMP,4,DS_BINARY_TYPE_STANDARD_DATETIME,NULL,""},
    {"accessRights",NULL,OPC_PROPERTY_TYPE_STANDARD,5,DS_BINARY_TYPE_STANDARD_ENUMERATION,opc_property_enumeration_accessrights,""},
    {"scanRate",NULL,OPC_PROPERTY_TYPE_STANDARD,6,DS_BINARY_TYPE_STANDARD_FLOAT,NULL,""},
    {"euType",NULL,OPC_PROPERTY_TYPE_NOTSUPPORTED,7,DS_BINARY_TYPE_STANDARD_ENUMERATION,opc_property_enumeration_eutype,""},
    {"euInfo",NULL,OPC_PROPERTY_TYPE_NOTSUPPORTED,8,DS_BINARY_TYPE_STANDARD_STRINGLIST,NULL,""},
    {"engineeringUnits",NULL,OPC_PROPERTY_TYPE_STANDARD,100,DS_BINARY_TYPE_STANDARD_STRING,NULL,""},
    {"description",NULL,OPC_PROPERTY_TYPE_STANDARD,101,DS_BINARY_TYPE_STANDARD_STRING,NULL,""},
    {"highEU",NULL,OPC_PROPERTY_TYPE_STANDARD,102,DS_BINARY_TYPE_STANDARD_DOUBLE,NULL,""},
    {"lowEU",NULL,OPC_PROPERTY_TYPE_STANDARD,103,DS_BINARY_TYPE_STANDARD_DOUBLE,NULL,""},
    {"highIR",NULL,OPC_PROPERTY_TYPE_STANDARD,104,DS_BINARY_TYPE_STANDARD_DOUBLE,NULL,""},
    {"lowIR",NULL,OPC_PROPERTY_TYPE_STANDARD,105,DS_BINARY_TYPE_STANDARD_DOUBLE,NULL,""},
    {"closeLabel",NULL,OPC_PROPERTY_TYPE_STANDARD,106,DS_BINARY_TYPE_STANDARD_STRING,NULL,""},
    {"openLabel",NULL,OPC_PROPERTY_TYPE_STANDARD,107,DS_BINARY_TYPE_STANDARD_STRING,NULL,""},
    {"timeZone",NULL,OPC_PROPERTY_TYPE_TIMEZONE,108,DS_BINARY_TYPE_STANDARD_UNSIGNED,NULL,""},
    {NULL,NULL,OPC_PROPERTY_TYPE_NOTSUPPORTED,0,0,NULL,NULL}
};

int opcPropertiesCreateTypes() { return 0; }
void opcPropertiesDestroyTypes() {}

/*
DSBinaryTypeInteger opc_property_type_quality = NULL;
DSBinaryTypeInteger opc_property_type_limitstatus = NULL;
DSBinaryTypeInteger opc_property_type_accessrights = NULL;
DSBinaryTypeInteger opc_property_type_eutype = NULL;

int opcPropertiesCreateTypes() {
    opc_property_type_quality = dsBinaryTypeStandardCreateEnumerationFromArray(NULL, opc_property_enumeration_quality, 0);
    opc_property_type_limitstatus = dsBinaryTypeStandardCreateEnumerationFromArray(NULL, opc_property_enumeration_limitstatus, 0);
    opc_property_type_accessrights = dsBinaryTypeStandardCreateEnumerationFromArray(NULL, opc_property_enumeration_accessrights, 0);
    opc_property_type_eutype = dsBinaryTypeStandardCreateEnumerationFromArray(NULL, opc_property_enumeration_eutype, 0);
    
    if ((!opc_property_type_quality)||(!opc_property_type_limitstatus)||
	(!opc_property_type_accessrights)||(!opc_property_type_eutype)) {
	    opcPropertiesDestroyTypes();
	    return DS_ERR_INTERNAL;
    }
    return 0;
}

void opcPropertiesDestroyTypes() {
    if (opc_property_type_quality) dsBinaryTypeIntegerDestroy(opc_property_type_quality);
    if (opc_property_type_limitstatus) dsBinaryTypeIntegerDestroy(opc_property_type_limitstatus);
    if (opc_property_type_accessrights) dsBinaryTypeIntegerDestroy(opc_property_type_accessrights);
    if (opc_property_type_eutype) dsBinaryTypeIntegerDestroy(opc_property_type_eutype);

    opc_property_type_quality = NULL;
    opc_property_type_limitstatus = NULL;
    opc_property_type_accessrights = NULL;
    opc_property_type_eutype = NULL;
}
*/

int opcPropertiesRegisterStandard(OPCPropertiesContext ctx) {
    int err;
    OPCPropertyContext prop;
    OPCStandardProperty pctx;
    
    ds_assert_param(ctx);
    
    for (pctx=opc_property;pctx->name;pctx++) {
	prop = opcPropertyCreateContext(ctx->malloc);
	if (!prop) return OPC_ERR_PROPERTY_CREATE;



	err = opcPropertyInitStandardContext(prop, pctx->name, pctx->proptype, pctx->type, pctx->attr);
	if (err) {
	    opcPropertyDestroyContext(prop);
	    return err;
	}
	opcPropertySetOPCId(prop,pctx->id);
	if (pctx->xmlnamespace) opcPropertySetNamespace(prop,pctx->xmlnamespace);
	if (pctx->description) opcPropertySetDescription(prop,pctx->description);
	
	err = opcPropertiesRegisterProperty(ctx, prop);
	if (err) {
	    opcPropertyDestroyContext(prop);
	    return err;
	}
    }
    return 0;
}
