#define _OPC_STATE_C

#include <stdio.h>

#include <opclib.h>
#include <dsassert.h>
#include <dsdebug.h>
#include <dslocale.h>

#include <dsmalloc.h>

#include <opc.h>
#include <opcstate.h>

static void opcStateZeroContext(OPCStateContext ctx) {
    ds_assert_return(ctx);
    ctx->lock = NULL;
    ctx->status_info = NULL;
    ctx->vendor_info = NULL;
}

OPCStateContext opcStateCreateContext(DSMallocContext mctx) {
    OPCStateContext state;

    dsDebugServer(DS_DEBUG_NOTICE, "opcStateCreateContext");    

    state = (OPCStateContext)dsMalloc(mctx,sizeof(struct OPCStateContextT));
    if (state) {
	state->malloc=mctx;
	opcStateZeroContext(state);
    }
    return state;
}

int opcStateInitContext(OPCStateContext ctx, OPCContext sctx) {
    int err = 0;

    dsDebugServer(DS_DEBUG_NOTICE, "opcStateInitContext");    

    ds_assert_param(ctx);
    
    ctx->lock = dsLockCreateContext(ctx->malloc);
    if (!ctx->lock) return DS_ERR_MEM;
    err = dsLockInitContext(ctx->lock);
    if (!err) err=dsLocaleExchangeString(ctx->malloc,&ctx->vendor_info,"");
    if (!err) err=dsLocaleExchangeString(ctx->malloc,&ctx->status_info,"");
    if (err) opcStateFreeContext(ctx);
    return err;
}

void opcStateFreeContext(OPCStateContext ctx) {
    dsDebugServer(DS_DEBUG_NOTICE, "opcStateFreeContext");    

    ds_assert_return(ctx);
    if (ctx->vendor_info) dsLocaleExchangeString(ctx->malloc,&ctx->vendor_info,NULL);
    if (ctx->status_info) dsLocaleExchangeString(ctx->malloc,&ctx->status_info,NULL);
    if (ctx->lock) dsLockDestroyContext(ctx->lock);
    opcStateZeroContext(ctx);
}

void opcStateDestroyContext(OPCStateContext ctx) {
    dsDebugServer(DS_DEBUG_NOTICE, "opcStateDestroyContext");    

    ds_assert_return(ctx);
    opcStateFreeContext(ctx);
    dsFree(ctx->malloc,ctx);
}

int opcStateSetStatus(OPCStateContext ctx, OPCState state, const char *status) {
    int err = 0;
    
    ds_assert_param(ctx);
    ds_assert_param(ctx->lock);
    
    dsLockWrite(ctx->lock);
    if (state>=OPC_STATE_LASTSTATE) err=OPC_ERR_STATE_INVALID_STATE;
    if (!err) {
	ctx->state=state;
	err=dsLocaleExchangeString(ctx->malloc,&ctx->status_info,status);
    }
    dsLockReturn(ctx->lock);
    return err;
}

int opcStateSetVendor(OPCStateContext ctx, const char *vendor) {
    int err;

    ds_assert_param(ctx);
    ds_assert_param(ctx->lock);
    
    dsLockWrite(ctx->lock);
    err=dsLocaleExchangeString(ctx->malloc,&ctx->vendor_info,vendor);
    dsLockReturn(ctx->lock);
    return err;
}


dschar opcStateGetStatusInfo(OPCStateContext ctx, const char *locale) {
    dschar tmp;
    
    ds_assert_null(ctx);
    ds_assert_null(ctx->lock);
    ds_assert_null(ctx->status_info);
    ds_assert_null(locale);
    
    dsLockRead(ctx->lock);
    tmp=dsLocaleGetText(ctx->status_info,locale);
    dsLockReturn(ctx->lock);
    
    return tmp;
}

dschar opcStateGetVendorInfo(OPCStateContext ctx, const char *locale) {
    dschar tmp;

    ds_assert_null(ctx);
    ds_assert_null(ctx->lock);
    ds_assert_null(ctx->vendor_info);
    ds_assert_null(locale);

    dsLockRead(ctx->lock);
    tmp=dsLocaleGetText(ctx->vendor_info,locale);
    dsLockReturn(ctx->lock);
    return tmp;
}

dschar opcStateGetStatus(OPCStateContext ctx, const char *locale) {
    ds_assert_null(ctx);

    return dsLocaleGetText(opc_states[ctx->state],locale);
}
