#include <stdio.h>

#include <opclib.h>
#include <dsassert.h>
#include <dsdebug.h>

#include <dsmalloc.h>

#include <opc.h>
#include <dsstringlist.h>
#include <dstools.h>
#include <opcversion.h>

OPCVersion opcVersionInit(DSMallocContext mctx, OPCContext ctx) {
    OPCVersion ver;

    dsDebugServer(DS_DEBUG_NOTICE, "opcVersionInit");    

    ver = (OPCVersion)dsMalloc(mctx,sizeof(struct OPCVersionT));
    if (ver) {
	ver->malloc=mctx;
	ver->major=NULL;
	ver->minor=NULL;
	ver->version=NULL;
	ver->SupportedInterfaceVersions=NULL;
    }
    return ver;
}

void opcVersionInterfacesClear(OPCVersion ctx) {
    ds_assert_return(ctx);
    
    if (ctx->SupportedInterfaceVersions) {
	dsStringListDestroy(ctx->SupportedInterfaceVersions);
	ctx->SupportedInterfaceVersions=NULL;
    }
}

void opcVersionDestroy(OPCVersion ctx) {
    dsDebugServer(DS_DEBUG_NOTICE, "opcVersionDestroy");    

    ds_assert_return(ctx);
    
    opcVersionInterfacesClear(ctx);
    if (ctx->major) dsExchangeString(ctx->malloc,&ctx->major,NULL);
    if (ctx->minor) dsExchangeString(ctx->malloc,&ctx->minor,NULL);
    if (ctx->version) dsExchangeString(ctx->malloc,&ctx->version,NULL);
    dsFree(ctx->malloc,ctx);
}

int opcVersionSet(OPCVersion ctx, const char *major, const char *minor, unsigned long build) {
    ds_assert_param(ctx);
    ds_assert_param(major);
    ds_assert_param(minor);
    
    if (dsExchangeString(ctx->malloc,&ctx->major,major)) return DS_ERR_MEM;
    if (dsExchangeString(ctx->malloc,&ctx->minor,minor)) return DS_ERR_MEM;
    ctx->build=build;

    if (ctx->version) dsExchangeString(ctx->malloc,&ctx->version,NULL);
    ctx->version=(char*)dsMalloc(ctx->malloc,(strlen(major)+strlen(minor)+16)*sizeof(char));
    if (!ctx->version) return DS_ERR_MEM;
    
    if (ctx->build)
	sprintf(ctx->version,"%s.%s.%u",ctx->major,ctx->minor,ctx->build);
    else
	sprintf(ctx->version,"%s.%s",ctx->major,ctx->minor);
    
    return 0;
}

int opcVersionInterfaceAdd(OPCVersion ctx, const char *interface_version, const char **extensions) {
    struct OPCInterfaceVersionT tmp;
    
    ds_assert_param(ctx);
    ds_assert_param(interface_version);
    
    if (!ctx->SupportedInterfaceVersions) {
	ctx->SupportedInterfaceVersions=dsStringListCreate(ctx->malloc);
	if (!ctx->SupportedInterfaceVersions) return DS_ERR_MEM;
    }

    tmp.interface_version=interface_version;
    tmp.extensions=extensions;

    if (dsStringListAdd(ctx->SupportedInterfaceVersions,(void*)&tmp,sizeof(struct OPCInterfaceVersionT)))
	return DS_ERR_MEM;
    return 0;    
}

const char *opcVersionGet(OPCVersion ctx) {
    ds_assert_null(ctx);
    ds_assert_null(ctx->version);
    
    return ctx->version;
}

int opcVersionList(OPCVersion ctx, OPCVersionListFunctionT list_function, void *attr) {
    DSStringList siv;
    ds_assert_param(ctx);
    ds_assert_param(list_function);
    
    for (siv=ctx->SupportedInterfaceVersions;siv;siv=siv->next) {
	list_function(((OPCInterfaceVersion)siv->item)->interface_version,attr);
    }
    return 0;
}
