#define _OPC_XMLDA_REQUEST_C
#define _DS_INCLUDE_XML_HEADERS

#include <stdio.h>

#include <libxml/parser.h>
#include <libxml/encoding.h>
#include <libxml/xmlsave.h>


#include <opclib.h>
#include <dsassert.h>
#include <dshashtable.h>
#include <dsxml.h>

#include <opc.h>
#include <opcconfiguration.h>
#include <opcxmldarequest.h>
#include <opcxmldarequestsecurity.h>
#include <opcxmldarequestprocess.h>

OPCXmldaRequestContext opcXmldaRequestCreateContext() {
    OPCXmldaRequestContext ctx;
    ctx = (OPCXmldaRequestContext)malloc(sizeof(struct OPCXmldaRequestContextT));
    if (ctx) {
	ctx->parser_ctx=NULL;
	ctx->doc=NULL;
	ctx->adoc=NULL;
	
	ctx->opc_commands=NULL;
	ctx->sctx=NULL;
	
//	ctx->start_time=NULL;
//	ctx->end_time=NULL;
	ctx->client_request_handle=NULL;

	ctx->extensions=0;
	ctx->locale=NULL;
    }
    return ctx;
}

int opcXmldaRequestInitContext(OPCXmldaRequestContext ctx, OPCContext sctx, DSHashTable opc_commands) {
    ds_assert_param(ctx);
    
    
    ctx->sctx=sctx;
    ctx->opc_commands=opc_commands;
    
    return 0;
}

void opcXmldaRequestFreeContext(OPCXmldaRequestContext ctx) {
    ds_assert_return(ctx);
    if (ctx->parser_ctx) {
	xmlFreeParserCtxt(ctx->parser_ctx);
	ctx->parser_ctx=NULL;
    }
    if (ctx->doc) {
	xmlFreeDoc(ctx->doc);
	ctx->doc=NULL;
    }
    if (ctx->adoc) {
	xmlFreeDoc(ctx->adoc);
	ctx->adoc=NULL;
    }
/*    if (ctx->start_time) {
	dsTimeDestroy(ctx->start_time);
	ctx->start_time=NULL;
    }
    if (ctx->end_time) {
	dsTimeDestroy(ctx->end_time);
	ctx->end_time=NULL;
    }*/
    if (ctx->client_request_handle) {
	xmlFree(ctx->client_request_handle);
	ctx->client_request_handle=NULL;
    }

    ctx->extensions=0;
    ctx->locale=NULL;
}

void opcXmldaRequestDestroyContext(OPCXmldaRequestContext ctx) {
    ds_assert_return(ctx);
    opcXmldaRequestFreeContext(ctx);
    free(ctx);
}

int opcXmldaRequestReturnParserError(int err) {
    switch (err) {
	case 0:
	    return 0;
	default:
	    return OPC_ERR_XLIB_REQUEST_PARSE;
    }
}

int opcXmldaRequestXMLData(OPCXmldaRequestContext ctx, char *data, unsigned long length) {
    int err;
    int preprocessed=0;
    if (length<4) return OPC_ERR_XLIB_REQUEST_PARSE_CHUNK_TO_SMALL;
    if (!ctx->parser_ctx) {
	ctx->parser_ctx=xmlCreatePushParserCtxt(NULL,NULL,data,4,"OPCXmldaRequest");
        if (!ctx->parser_ctx) return OPC_ERR_XLIB_PARSER_CREATE;
#ifndef _DEBUG_XML
//	xmlSetFeature(ctx->parser_ctx,"SAX function error",&func);
	ctx->parser_ctx->sax->error=NULL;
#endif /* _DEBUG_XML */
	preprocessed=4;
    }
    err=xmlParseChunk(ctx->parser_ctx,data+preprocessed,length-preprocessed,0);
    
    if (err) return opcXmldaRequestReturnParserError(err);

    return 0;
}

/*
int opcXmldaRequestDebug(OPCXmldaRequestContext ctx) {
    int size;
    char *mem;
    
    xmlDocDumpMemory(ctx->doc,&mem,&size);
    puts("++++++++++++++++++++++++++++++++++++++++++");
    puts(mem);
    puts("++++++++++++++++++++++++++++++++++++++++++");
    free(mem);
}
*/

int opcXmldaRequestProcess(OPCXmldaRequestContext ctx, OPCXmldaRequestOpenAttachmentFunction open_attachment, DSSendFunction send_attachment, OPCXmldaRequestCloseAttachmentFunction close_attachment, void *cctx) {
    int err;
    xmlNodePtr root, tnode;
    
	// Parse Request
    err=xmlParseChunk(ctx->parser_ctx,NULL,0,1);
    if (err) return opcXmldaRequestReturnParserError(err);
    ctx->doc=ctx->parser_ctx->myDoc;
    xmlFreeParserCtxt(ctx->parser_ctx);
    ctx->parser_ctx=NULL;
    
	// Attachment Send Function
    ctx->open_attachment = open_attachment;
    ctx->send_attachment = send_attachment;
    ctx->close_attachment = close_attachment;
    ctx->cctx = cctx;
    
//    opcXmldaRequestDebug(ctx);

	// Rip SOAP headers (if any)
    root = xmlDocGetRootElement(ctx->doc);
    if (!root) return OPC_ERR_XLIB_REQUEST_VALIDATE_EMPTY;

/*  DS: Temporary some problems with finding namespaces    
    ns = xmlSearchNsByHref(ctx->doc,root,(const xmlChar*)opchdr_namespace);
    if (ns) ctx->extensions|=OPC_EXTENSION_HDR;
    else ns = xmlSearchNsByHref(ctx->doc,root,(const xmlChar*)opc_namespace);
    if (!ns) return OPC_ERR_XLIB_REQUEST_VALIDATE_NSNOTFOUND;
    
    ctx->req=dsXmlFindNode(root,NULL,ns);
*/

    ctx->req=dsXmlFindNodeByHref(root,NULL,OPC_XMLDA_NAMESPACE);
    if (!ctx->req) return OPC_ERR_XLIB_REQUEST_VALIDATE_NOTFOUND_NS;
/*
    {
	 FILE *f;
	 xmlChar *mem;
	 int memsize;
	 xmlDocDumpMemory(ctx->doc,&mem,&memsize);
	 
	 f = fopen("/tmp/debug.info", "a+");
	 fprintf(f, "----------------- Request -------------------------\n");
	 fprintf(f, mem);
	 fprintf(f, "---------------------------------------------------\n");
	 free(mem);
	 fclose(f);
    }    
*/

    if (ctx->req==root) ctx->soap=0;
    else {
	tnode=dsXmlFindNodeByHref(root,"Envelope",NULL);
	if (tnode) ctx->soapns=tnode->nsDef;
	else ctx->soapns=NULL;
	ctx->soap=1;
    }
    
	// Security Parse
    err=opcXmldaRequestSecurityRequest(ctx);
    
	// Validate

	// OPC Parse / Create
    err=opcXmldaRequestProcessOPC(ctx);
	    
	// Security Create
    err=opcXmldaRequestSecurityResponse(ctx);


	// Build Answer
//    xmlDocDumpMemory(ctx->adoc,&answer,&ctx->answer_size);
//    xmlDocDumpMemoryEnc(ctx->adoc,&answer,&ctx->answer_size,ctx->charset);
//    ctx->answer=(char*)answer;
//    puts("--------------------------- Finished -----------------------------------");
//    puts(ctx->answer);
/*

    {
	 FILE *f;
	 xmlChar *mem;
	 int memsize;
	 xmlDocDumpMemory(ctx->adoc,&mem,&memsize);
	 f = fopen("/tmp/debug.info", "a+");
	 fprintf(f, "----------------- Response ------------------------");
	 fprintf(f, mem);
	 fprintf(f, "---------------------------------------------------\n");
	 free(mem);
	 fclose(f);
    }    
*/
    return 0;
}

int opcXmldaRequestSend(OPCXmldaRequestContext ctx, xmlOutputWriteCallback func_write, xmlOutputCloseCallback func_close, void *cctx) {
    int err = 0;

#ifdef opc_supported_xmlsave
    xmlSaveCtxtPtr save;
#else /* opc_supported_xmlsave */
    xmlChar *answer;
    int answer_size;
#endif /* opc_supported_xmlsave */


    ds_assert_param(ctx);
    if (!ctx) return DS_ERR_INVALID_ARGUMENT;
    if (!func_write) return 0;

#ifdef opc_supported_xmlsave
    save = xmlSaveToIO(func_write, NULL, cctx, ctx->charset, 0);
    if (!save) return OPC_ERR_XLIB_REQUEST_SERIALIZE_START;
    if (xmlSaveDoc(save, ctx->adoc)<0) err = OPC_ERR_XLIB_REQUEST_SERIALIZE;
    if (xmlSaveClose(save)<0) 
	if (!err) err = OPC_ERR_XLIB_REQUEST_SERIALIZE_FINISH;
#else /* opc_supported_xmlsave */
    xmlDocDumpMemoryEnc(ctx->adoc,&answer,&answer_size,ctx->charset);
    if (!answer) return OPC_ERR_XLIB_REQUEST_SERIALIZE_START;
    if (func_write(cctx, (void*)answer, answer_size)<0) err = OPC_ERR_XLIB_REQUEST_SERIALIZE;
    if ((func_close)&&(func_close(cctx)<0))
	if (!err) err = OPC_ERR_XLIB_REQUEST_SERIALIZE_FINISH;
    xmlFree(answer);
#endif /* opc_supported_xmlsave */

    return err;
}
