#ifndef _OPC_XMLDA_REQUEST_H
#define _OPC_XMLDA_REQUEST_H

#include <dsxml.h>
#include <dshashtable.h>

typedef struct OPCXmldaRequestContextT *OPCXmldaRequestContext;

#include <opc.h>
#include <opcinterface.h>
#include <opcxmldarequestprocess.h>

#define OPC_XMLDA_NAMESPACE "http://opcfoundation.org/webservices/XMLDA/1.0/"

typedef void *(*OPCXmldaRequestOpenAttachmentFunction)(void *ctx, const char *item, DSDataConst info, DSDataConst data);
typedef void (*OPCXmldaRequestCloseAttachmentFunction)(void *ctx);

struct OPCXmldaRequestContextT {
    OPCContext sctx;
    OPCInterface server_interface;
    DSHashTable opc_commands;
    
    xmlParserCtxtPtr parser_ctx;
    xmlDocPtr doc;
    xmlNodePtr req;
    xmlDocPtr adoc;
    xmlNodePtr resp;

    const char* locale;
    const char *charset;
    
    unsigned char soap;
    xmlNsPtr soapns;
    
    unsigned long extensions;
    
    xmlChar *client_request_handle;
    
    DSTimeS start_time, end_time;
    
    OPCXmldaRequestOpenAttachmentFunction open_attachment;
    DSSendFunction send_attachment;
    OPCXmldaRequestCloseAttachmentFunction close_attachment;
    void *cctx;
};

OPCXmldaRequestContext opcXmldaRequestCreateContext();
int opcXmldaRequestInitContext(OPCXmldaRequestContext ctx, OPCContext sctx, DSHashTable opc_commands);
void opcXmldaRequestFreeContext(OPCXmldaRequestContext ctx);
void opcXmldaRequestDestroyContext(OPCXmldaRequestContext ctx);
int opcXmldaRequestXMLData(OPCXmldaRequestContext ctx, char *data, unsigned long length);
int opcXmldaRequestProcess(OPCXmldaRequestContext ctx, OPCXmldaRequestOpenAttachmentFunction open_attachment, DSSendFunction send_attachment, OPCXmldaRequestCloseAttachmentFunction close_attachment, void *cctx);
int opcXmldaRequestSend(OPCXmldaRequestContext ctx, xmlOutputWriteCallback write, xmlOutputCloseCallback close, void *cctx);


#endif /* _OPC_XMLDA_REQUEST_H */

