#define _DS_INCLUDE_XML_HEADERS

#include <stdio.h>
#include <dsassert.h>

#include <libxml/tree.h>

#include <opclib.h>

#include <opctree.h>
#include <opcitem.h>

#include <opcxmldarequesttools.h>

#include <opcxmldarequestbrowse.h>

void opcXmldaRequestBrowseZeroContext(OPCXmldaRequestBrowseContext ctx) {
    ds_assert_return(ctx);
    
    ctx->item=NULL;
    ctx->tree=NULL;
    ctx->subitems=NULL;
    ctx->position=NULL;
    ctx->browse_filter = 0;
    ctx->name_filter = NULL;
    ctx->vendor_filter = NULL;
}

OPCXmldaRequestBrowseContext opcXmldaRequestBrowseCreateContext() {
    OPCXmldaRequestBrowseContext ctx;
    ctx = (OPCXmldaRequestBrowseContext)malloc(sizeof(struct OPCXmldaRequestBrowseContextT));
    if (ctx) opcXmldaRequestBrowseZeroContext(ctx);
    return ctx;
}

int opcXmldaRequestBrowseInitContext(OPCXmldaRequestBrowseContext ctx, OPCXmldaRequestContext rctx, xmlNodePtr node) {
    int err = 0;
    OPCInterface server_interface;
    OPCTreeContext tree;
    xmlChar *attr;
    xmlChar *browse_filter;

    ds_assert_param(ctx);
    ds_assert_param(node);
    ds_assert_param(rctx);
    server_interface = rctx->server_interface;
    ds_assert_param(server_interface);

    opcXmldaRequestBrowseZeroContext(ctx);
    
    ctx->locale = rctx->locale;
    
    tree = opcInterfaceGetServerTree(server_interface);
    if (!tree) return OPC_ERR_INTERFACE_GET_SERVERTREE;
    else ctx->tree = tree;

    ctx->item=opcXmldaRequestGetItemName(node);
    if (!ctx->item) return OPC_ERR_XLIB_REQUEST_VALIDATE_NOTFOUND_ITEMNAME;
    
    attr = xmlGetProp(node, "MaxElementsReturned");
    if (attr) {
	if ((sscanf(attr,"%d\n", &ctx->max_elements_returned)!=1)||(ctx->max_elements_returned<0)) {
	    opcXmldaRequestBrowseFreeContext(ctx);
	    return OPC_ERR_XLIB_REQUEST_PROCESS_BROWSE_INVALID_RETURNED_ELEMENTS_NUMBER;
	}
	xmlFree(attr);
	if (!ctx->max_elements_returned) ctx->max_elements_returned=OPC_SERVER_CONFIG_MAX_ELEMENTS_RETURNED;
    } else ctx->max_elements_returned = OPC_SERVER_CONFIG_MAX_ELEMENTS_RETURNED;

    browse_filter = xmlGetProp(node, "BrowseFilter");
    if (browse_filter) {
	switch (browse_filter[0]) {
	    case 'a':
		if (!xmlStrcmp(browse_filter,"all")) ctx->browse_filter=ALL;
		else err = OPC_ERR_XLIB_REQUEST_PROCESS_BROWSE_INVALID_BROWSE_FILTER;
		break;
	    case 'b':
		if (!xmlStrcmp(browse_filter,"branch")) ctx->browse_filter=BRANCH;
		else err = OPC_ERR_XLIB_REQUEST_PROCESS_BROWSE_INVALID_BROWSE_FILTER;
		break;
	    case 'i':
		if (!xmlStrcmp(browse_filter,"item")) ctx->browse_filter=ITEM;
		else err = OPC_ERR_XLIB_REQUEST_PROCESS_BROWSE_INVALID_BROWSE_FILTER;
		break;
	    default:
		err = OPC_ERR_XLIB_REQUEST_PROCESS_BROWSE_INVALID_BROWSE_FILTER;
	}
	xmlFree(browse_filter);
	if (err) {
	    opcXmldaRequestBrowseFreeContext(ctx);
	    return err;
	}
    }

    ctx->name_filter = xmlGetProp(node, "ElementNameFilter");
    ctx->vendor_filter = xmlGetProp(node, "VendorFilter");
    
    printf("Browse: \"%s\".\n",ctx->item);
    err = opcTreeBrowse(tree,ctx->item,ctx->browse_filter,ctx->name_filter,ctx->vendor_filter,&ctx->subitems);
    if (!ctx->subitems) {
	opcXmldaRequestBrowseFreeContext(ctx);
	return err;
    }
    
    ctx->position = ctx->subitems;
    
    return 0;
}

int opcXmldaRequestBrowseCheckContext(OPCXmldaRequestBrowseContext ctx, OPCXmldaRequestContext rctx, xmlNodePtr node) {
    int err = 0;
    OPCInterface server_interface;
    xmlChar *attr;
    xmlChar *browse_filter;

    ds_assert_param(ctx);
    ds_assert_param(node);
    ds_assert_param(rctx);
    server_interface = rctx->server_interface;
    ds_assert_param(server_interface);
    
    ctx->locale = rctx->locale;

    ctx->tree = opcInterfaceGetServerTree(server_interface);
    if (!ctx->tree) return OPC_ERR_INTERFACE_GET_SERVERTREE;
    
    attr=opcXmldaRequestGetItemName(node);
    if (!attr) err = OPC_ERR_XLIB_REQUEST_VALIDATE_NOTFOUND_ITEMNAME;
    else {
	if (xmlStrcmp(attr,ctx->item)) err = OPC_ERR_XLIB_REQUEST_PROCESS_CONTINUATION_POINT_INVALID_ITEMNAME;
	xmlFree(attr);
    }
    if (err) return err;

    attr = xmlGetProp(node, "MaxElementsReturned");
    if (attr) {
	if ((sscanf(attr,"%d\n", &ctx->max_elements_returned)!=1)||(ctx->max_elements_returned<0)) {
	    return OPC_ERR_XLIB_REQUEST_PROCESS_BROWSE_INVALID_RETURNED_ELEMENTS_NUMBER;
	}
	xmlFree(attr);
	if (!ctx->max_elements_returned) ctx->max_elements_returned=OPC_SERVER_CONFIG_MAX_ELEMENTS_RETURNED;
    } else ctx->max_elements_returned = OPC_SERVER_CONFIG_MAX_ELEMENTS_RETURNED;
    

    browse_filter = xmlGetProp(node, "BrowseFilter");
    if (browse_filter) {
	switch (browse_filter[0]) {
	    case 'a':
		if (!xmlStrcmp(browse_filter,"all")) {
		    if (ctx->browse_filter!=ALL) err = OPC_ERR_XLIB_REQUEST_PROCESS_CONTINUATION_POINT_INVALID_BROWSE_FILTER;
		} else err = OPC_ERR_XLIB_REQUEST_PROCESS_CONTINUATION_POINT_INVALID_BROWSE_FILTER;
		break;
	    case 'b':
		if (!xmlStrcmp(browse_filter,"branch")) {
		    if (ctx->browse_filter!=BRANCH) err = OPC_ERR_XLIB_REQUEST_PROCESS_CONTINUATION_POINT_INVALID_BROWSE_FILTER;
		} else err = OPC_ERR_XLIB_REQUEST_PROCESS_CONTINUATION_POINT_INVALID_BROWSE_FILTER;
		break;
	    case 'i':
		if (!xmlStrcmp(browse_filter,"item")) {
		    if (ctx->browse_filter!=ITEM) err = OPC_ERR_XLIB_REQUEST_PROCESS_CONTINUATION_POINT_INVALID_BROWSE_FILTER;
		} else err = OPC_ERR_XLIB_REQUEST_PROCESS_CONTINUATION_POINT_INVALID_BROWSE_FILTER;
		break;
	    default:
		err = OPC_ERR_XLIB_REQUEST_PROCESS_CONTINUATION_POINT_INVALID_BROWSE_FILTER;
	}
	xmlFree(browse_filter);
	if (err) return err;
    }

    attr = xmlGetProp(node, "ElementNameFilter");
    if (attr) {
	if (xmlStrcmp(attr, ctx->name_filter)) err = OPC_ERR_XLIB_REQUEST_PROCESS_CONTINUATION_POINT_INVALID_NAME_FILTER;
	xmlFree(attr);
    } else if (ctx->name_filter) err = OPC_ERR_XLIB_REQUEST_PROCESS_CONTINUATION_POINT_INVALID_NAME_FILTER;
    
    attr = xmlGetProp(node, "VendorFilter");
    if (attr) {
	if (xmlStrcmp(attr, ctx->vendor_filter)) err = OPC_ERR_XLIB_REQUEST_PROCESS_CONTINUATION_POINT_INVALID_VENDOR_FILTER;
	xmlFree(attr);
    } else if (ctx->vendor_filter) err = OPC_ERR_XLIB_REQUEST_PROCESS_CONTINUATION_POINT_INVALID_VENDOR_FILTER;

    if (err) return err;


    return 0;
}



void opcXmldaRequestBrowseFreeContext(OPCXmldaRequestBrowseContext ctx) {
    ds_assert_return(ctx);
    if (ctx->item) xmlFree(ctx->item);
    if (ctx->subitems) dsStringListDestroy(ctx->subitems);
    if (ctx->name_filter) xmlFree(ctx->name_filter);
    if (ctx->vendor_filter) xmlFree(ctx->vendor_filter);
    opcXmldaRequestBrowseZeroContext(ctx);
}

void opcXmldaRequestBrowseDestroyContext(OPCXmldaRequestBrowseContext ctx) {
    ds_assert_return(ctx);
    opcXmldaRequestBrowseFreeContext(ctx);
    free(ctx);
}


int opcXmldaRequestBrowseNextElement(OPCXmldaRequestBrowseContext ctx, xmlNodePtr pnode) {
    OPCTreeContext tree;
    DSStringList sl;
    OPCItemContext ictx = NULL;
    xmlNodePtr node;
    
    ds_assert_param(ctx);
    ds_assert_param(pnode);
    tree = ctx->tree;
    ds_assert_param(tree);

    for (sl = ctx->position;!ictx;sl=sl->next) {
	while ((sl)&&(!sl->item)) sl = sl->next;
	if (!sl) return OPC_ERR_XLIB_REQUEST_PROCESS_NOMORE_ITEMS;
	ictx=opcTreeGetItem(tree,sl->item);
    }
    
    node = xmlNewChild(pnode, NULL, "Elements", NULL);
    xmlSetProp(node,"Name", opcItemGetTitle(ictx,ctx->locale));
    xmlSetProp(node,"ItemName", ictx->fullname);
    xmlSetProp(node,"ItemPath", "");

    if (opcItemHaveChilds(ictx)) xmlSetProp(node,"HasChildren", "true");
    else xmlSetProp(node,"HasChildren", "false");
    if (opcItemIsConnected(ictx)) xmlSetProp(node, "IsItem", "true");
    else xmlSetProp(node, "IsItem", "false");
    
    // Properties are here

    ctx->position = sl;
    opcItemReturn(ictx);
    return 0;
}

int opcXmldaRequestBrowseAnswer(OPCXmldaRequestBrowseContext ctx, xmlNodePtr pnode) {
    DSStringList sl;
    int err;
    int i;
    
    ds_assert_param(ctx);
    ds_assert_param(pnode);

    printf("Items: %i\n",ctx->max_elements_returned);
    for (i=ctx->max_elements_returned;i;i--) {
	err=opcXmldaRequestBrowseNextElement(ctx,pnode);
	if (err) {
	    if (err==OPC_ERR_XLIB_REQUEST_PROCESS_NOMORE_ITEMS) break;
	    else {
		// Add errors about items
	    }
	}
    }

    if (err==OPC_ERR_XLIB_REQUEST_PROCESS_NOMORE_ITEMS) return 0;
    if (ctx->position) {
	for (sl=ctx->position;sl;sl=sl->next)
	    if (sl->item) {
		ctx->position = sl;
		return OPC_ERR_XLIB_REQUEST_PROCESS_MORE_ITEMS;
	    }
    }
    return 0;
}

