#define _DS_INCLUDE_XML_HEADERS
#define _OPC_XMLDA_REQUESTPOOL_C

#include <stdio.h>

#include <opclib.h>
#include <dsassert.h>
#include <dshashtable.h>
#include <dsdebug.h>
#include <dsxml.h>

#include <opc.h>
#include <opcserverinterface.h>
#include <opcconfiguration.h>
#include <opcxmldarequest.h>
#include <opcxmldarequestpool.h>

void opcXmldaRequestPoolZeroContext(OPCXmldaRequestPoolContext ctx) {
    ds_assert_return(ctx);
    
    ctx->opc_commands=NULL;
    ctx->schema_parser = NULL;
    ctx->schema_rep = NULL;
    ctx->schema_ctx = NULL;
    
}

OPCXmldaRequestPoolContext opcXmldaRequestPoolCreateContext(DSMallocContext mctx) {
    OPCXmldaRequestPoolContext ctx;

    dsDebugServer(DS_DEBUG_NOTICE, "opcXmldaRequestPoolCreateContext");    

    ctx = (OPCXmldaRequestPoolContext)dsMalloc(mctx,sizeof(struct OPCXmldaRequestPoolContextT));
    if (ctx) {
	ctx->malloc = mctx;
	opcXmldaRequestPoolZeroContext(ctx);
    }
    return ctx;
}

int opcXmldaRequestPoolInitContext(OPCXmldaRequestPoolContext ctx, OPCContext sctx) {
    unsigned long i;
    int err=0;

    dsDebugServer(DS_DEBUG_NOTICE, "opcXmldaRequestPoolInitContext");    

    ds_assert_param(ctx);
    ds_assert_param(opc_interfaces);

    opcXmldaRequestPoolZeroContext(ctx);

	// Setting link on OPC Server Context
    ctx->sctx=sctx;

	// Setting up command hash
    ctx->opc_commands = dsHashTableInit(ctx->malloc,OPC_COMMANDS,DS_HASH_TYPE_POINTER,NULL);
    if (!ctx->opc_commands) {
	xmlCleanupCharEncodingHandlers();
	xmlCleanupParser();
	return OPC_ERR_XLIB_COMMANDS_INIT;
    }

    for (i=0;opc_commands_ctx[i].name;i++) {
	err+=dsHashTableInsert(ctx->opc_commands,(void*)&opc_commands_ctx[i]);
    }

    if (err) {
	opcXmldaRequestPoolFreeContext(ctx);
	return DS_ERR_MEM;
    }

	// Setting supported interface versions
    for (i=0;opc_interfaces[i];i++) {
	err=opcServerInterfaceAdd(sctx, opc_interfaces[i], opc_extensions[i]);
	if (err) {
	    opcXmldaRequestPoolFreeContext(ctx);
	    return err;
	}
    }

	// Loading XSD schema
    ctx->schema_parser=xmlSchemaNewParserCtxt(sctx->config->opc_schema);
    if (!ctx->schema_parser) {
	opcXmldaRequestPoolFreeContext(ctx);
	return OPC_ERR_XLIB_INVALID_SCHEMA;
    }
    ctx->schema_rep=xmlSchemaParse(ctx->schema_parser);
    if (!ctx->schema_rep) {
	opcXmldaRequestPoolFreeContext(ctx);
	return OPC_ERR_XLIB_INVALID_SCHEMA;
    }
    ctx->schema_ctx=xmlSchemaNewValidCtxt(ctx->schema_rep);
    if (!ctx->schema_ctx) {
	opcXmldaRequestPoolFreeContext(ctx);
	return OPC_ERR_XLIB_INVALID_SCHEMA;
    }

/*
    err = opcNamespacesSetOPCNamespace(&sctx->config->namespaces, ctx->schema_rep->targetNamespace);
    if (err) {
	opcXmldaRequestPoolFreeContext(ctx);
	return err;
    }
*/

    return 0;
}

void opcXmldaRequestPoolFreeContext(OPCXmldaRequestPoolContext ctx) {
    dsDebugServer(DS_DEBUG_NOTICE, "opcXmldaRequestPoolFreeContext");    

    ds_assert_return(ctx);

    if (ctx->schema_ctx) xmlSchemaFreeValidCtxt(ctx->schema_ctx);
    if (ctx->schema_rep) xmlSchemaFree(ctx->schema_rep);
    if (ctx->schema_parser) xmlSchemaFreeParserCtxt(ctx->schema_parser);

    if (ctx->opc_commands) dsHashTableDestroy(ctx->opc_commands);

    opcXmldaRequestPoolZeroContext(ctx);
}


void opcXmldaRequestPoolDestroyContext(OPCXmldaRequestPoolContext ctx) {
    dsDebugServer(DS_DEBUG_NOTICE, "opcXmldaRequestPoolDestroyContext");    

    opcXmldaRequestPoolFreeContext(ctx);
    dsFree(ctx->malloc,ctx);
}

OPCXmldaRequestContext opcXmldaRequestPoolGetRequest(OPCXmldaRequestPoolContext ctx) {
    int err;
    OPCXmldaRequestContext rctx;
    
    ds_assert_null(ctx);
    
    rctx = opcXmldaRequestCreateContext();
    if (rctx) {
	err=opcXmldaRequestInitContext(rctx,ctx->sctx,ctx->opc_commands);
	if (err) {
	    free(rctx);
	    return NULL;
	}
    }
    return rctx;
}

int opcXmldaRequestPoolReturnRequest(OPCXmldaRequestPoolContext ctx, OPCXmldaRequestContext rctx) {
    ds_assert_param(ctx);
    ds_assert_param(rctx);
    opcXmldaRequestDestroyContext(rctx);
    return 0;
}
