#define _DS_INCLUDE_XML_HEADERS
#define _OPC_XMLDA_REQUESTPROCESS_C
#include <stdio.h>

#include <opclib.h>
#include <dsassert.h>
#include <dsdebug.h>
#include <dserror.h>
#include <dshashtable.h>
#include <dsxml.h>

#include <opcitem.h>

#include <opcxmldarequest.h>
#include <opcxmldarequestprocess.h>
#include <opcxmldarequestbrowse.h>


int opcXmldaRequestProcessOPC(OPCXmldaRequestContext ctx) {
    xmlNodePtr node;
    xmlNsPtr soapns,opcns;
    OPCCommand cmd;
    char *stmp;
    
    ds_assert_param(ctx);

    ctx->adoc = xmlNewDoc("1.0");
    
    if (ctx->soap) {
	node=xmlNewChild((xmlNodePtr)ctx->adoc,NULL,"Envelope",NULL);
	if (ctx->soapns) soapns=xmlNewNs(node,ctx->soapns->href,ctx->soapns->prefix);
	else soapns=xmlNewNs(node,"http://www.w3.org/2001/12/soap-envelope",NULL);
	xmlSetNs(node,soapns);
	node=xmlNewChild(node,NULL,"Body",NULL);
    } else {
	node = (xmlNodePtr)ctx->adoc;
    }
    
    dsTimeCurrentTime(&ctx->start_time);
    
    cmd=dsHashTableFind(ctx->opc_commands,ctx->req->name);
    if (!cmd) return OPC_ERR_XLIB_REQUEST_PROCESS_UNKNOWN_REQUEST;

    stmp=(char*)dsMalloca(strlen(cmd->name)+9);
    sprintf(stmp,"%sResponse",cmd->name);
    ctx->resp=xmlNewChild(node,NULL,stmp,NULL);
    dsFreea(stmp);
    if (ctx->extensions&OPC_EXTENSION_HDR)
	opcns=xmlNewNs(ctx->resp,opchdr_namespace,NULL);
    else 
	opcns=xmlNewNs(ctx->resp,OPC_XMLDA_NAMESPACE,NULL);
    xmlSetNs(ctx->resp,opcns);

    ctx->server_interface = opcCreateInterface(ctx->sctx);
    if (!ctx->server_interface) return DS_ERR_MEM;

    cmd->func(ctx);
    
    opcDestroyInterface(ctx->server_interface);

    return 0;
}

const char *opcXmldaRequestGetClientRequestHandle(OPCXmldaRequestContext ctx, xmlNodePtr node) {
    ds_assert_null(ctx);
    
    if (ctx->client_request_handle) xmlFree(ctx->client_request_handle);
    if (node) ctx->client_request_handle = xmlGetProp(node,"ClientRequestHandle");
    else ctx->client_request_handle = NULL;
    return ctx->client_request_handle;
}

const char *opcXmldaRequestGetLocale(OPCXmldaRequestContext ctx, xmlNodePtr node) {
    xmlChar *attr;
    const char *charset;
    
    ds_assert_null(ctx);

    if (node) attr = xmlGetProp(node,"LocaleID");
    else attr = NULL;
    ctx->locale=opcInterfaceSetLocale(ctx->server_interface,attr);
    charset=opcInterfaceGetCharset(ctx->server_interface);
    if ((charset)&&(xmlParseCharEncoding(charset)!=XML_CHAR_ENCODING_UTF8)&&(xmlFindCharEncodingHandler(charset))) {
	ctx->charset = charset;
    } else {
	ctx->charset = opc_default_charset;
    }
    if (attr) xmlFree(attr);
    return ctx->locale;
}


int opcXmldaRequestSupportedLocales(const char *locale, xmlNodePtr pnode) {
    xmlNewTextChild(pnode,NULL,"SupportedLocaleIDs",locale);
    return 0;
}

int opcXmldaRequestSupportedInterfaces(const char *server_interface, xmlNodePtr pnode) {
/*    const char **sie; */

    xmlNewChild(pnode,NULL,"SupportedInterfaceVersions",server_interface);

/*
    if (ctx->extensions&OPC_EXTENSION_HDR) {
	for (siv=sctx->version->SupportedInterfaceVersions;siv;siv=siv->next) {
	    sie=((OPCInterface)siv->item)->extensions;
	    if ((!sie)||(!sie[0])) continue;
	    node=xmlNewChild(pnode,NULL,"SupportedInterfaceExtensions",NULL);
	    xmlSetProp(node,"Version", ((OPCInterface)siv->item)->server_interface);
	    for (i=0;sie[i];i++) xmlNewChild(node,NULL,sie[i],NULL);
	}
    }
*/

    return 0;
}

/*
opcXmldaRequestSetDateTimeProp(xmlNodePtr node, DSTime time) {
    char *str;
    ds_assert_null(node);
    ds_assert_null(time);
    
    str=dsTimeGetDateTime(sctx->start_time,NULL);
    
}
*/

int opcXmldaRequestReplyBase(OPCXmldaRequestContext ctx, xmlNodePtr pnode) {
    char *time;
    
    dsTimeCurrentTime(&ctx->end_time);

    time=dsTimeGetDateTime(&ctx->start_time,NULL);
    if (!time) return DS_ERR_MEM;
    xmlSetProp(pnode,"RcvTime",time);
    dsTimeGetDateTime(&ctx->end_time,time);
    xmlSetProp(pnode,"ReplyTime",time);
    free(time);

    if ((ctx->client_request_handle)&&(xmlStrlen(ctx->client_request_handle)))
	xmlSetProp(pnode,"ClientRequestHandle",ctx->client_request_handle);

    xmlSetProp(pnode,"RevisedLocaleID",ctx->locale);
    xmlSetProp(pnode,"ServerState", opcInterfaceGetStatus(ctx->server_interface));

    return 0;
}

int opcXmldaRequestGetStatus(OPCXmldaRequestContext ctx) {
    OPCInterface server_interface;
    xmlNodePtr rnode, pnode;
    
    ds_assert_param(ctx);
    server_interface = ctx->server_interface;
    ds_assert_param(server_interface);
    
    opcXmldaRequestGetClientRequestHandle(ctx, ctx->req);
    opcXmldaRequestGetLocale(ctx,ctx->req);
    rnode=xmlNewChild(ctx->resp,NULL,"GetStatusResult",NULL);    
    pnode=xmlNewChild(ctx->resp,NULL,"Status",NULL);
    xmlSetProp(pnode,"StartTime",opcInterfaceGetStartTime(server_interface));
    xmlSetProp(pnode,"ProductVersion",opcInterfaceGetVersion(server_interface));
    
    xmlNewTextChild(pnode,NULL,"StatusInfo",opcInterfaceGetStatusInfo(server_interface));
    xmlNewTextChild(pnode,NULL,"VendorInfo",opcInterfaceGetVendorInfo(server_interface));
    opcInterfaceListSupportedLocales(server_interface,(OPCLocalesListFunctionT)opcXmldaRequestSupportedLocales, (void*)pnode);
    opcInterfaceListSupportedInterfaces(server_interface,(OPCVersionListFunctionT)opcXmldaRequestSupportedInterfaces, (void*)pnode);
    return opcXmldaRequestReplyBase(ctx, rnode);
}

int opcXmldaRequestBrowse(OPCXmldaRequestContext ctx) {
    OPCContext sctx;
    OPCConfigurationContext config;
    OPCXmldaRequestBrowseContext bctx;
    DSHolderBucketContext bucket;
    xmlChar *atmp;
    xmlNodePtr rnode;
    unsigned long pos, secret;
    char *ch;
    char stmp[32];
    int err;

    ds_assert_param(ctx);
    sctx = ctx->sctx;
    ds_assert_param(sctx);
    config = sctx->config;
    ds_assert_param(config);

    opcXmldaRequestGetClientRequestHandle(ctx, ctx->req);
    opcXmldaRequestGetLocale(ctx,ctx->req);

    rnode=xmlNewChild(ctx->resp,NULL,"BrowseResult",NULL);

    atmp = xmlGetProp(ctx->req, "ContinuationPoint");
    if (atmp) {
	ch = strchr(atmp,':');

	*ch=0;
	pos = atol(ch+1);
	secret = atol(atmp);
	xmlFree(atmp);
	
	bucket=dsHolderGet(sctx->holder, pos, DS_HOLDER_BUCKET_MAGIC_REQUEST_BROWSE,secret);
	if (!bucket) err = OPC_ERR_XLIB_REQUEST_PROCESS_CONTINUATION_POINT_INVALID;
	else {
	    bctx = (OPCXmldaRequestBrowseContext)bucket->data;

	    err=opcXmldaRequestBrowseCheckContext(bctx,ctx,ctx->req);
	    if (!err) {
		err=opcXmldaRequestBrowseAnswer(bctx,ctx->resp);
		if (err==OPC_ERR_XLIB_REQUEST_PROCESS_MORE_ITEMS) {
		    pos = dsHolderPut(sctx->holder, bucket);
		    if (pos==(unsigned long)-1) err = DS_ERR_HOLDER_BUSY;
		    else sprintf(stmp,"%lu:%lu",secret,pos);
		} else dsHolderBucketDestroyContext(bucket);
	    } else dsHolderBucketDestroyContext(bucket);
	}
    } else {
	bctx=opcXmldaRequestBrowseCreateContext();
	if (!bctx) return DS_ERR_MEM;
	err=opcXmldaRequestBrowseInitContext(bctx,ctx,ctx->req);
	if (!err) {
	    err=opcXmldaRequestBrowseAnswer(bctx,ctx->resp);
	    if (err==OPC_ERR_XLIB_REQUEST_PROCESS_MORE_ITEMS) {
		bucket = dsHolderBucketInitContext(sctx->malloc, bctx, &config->browse_timeout, &config->browse_timeout, (DSDestroyContextFunction)opcXmldaRequestBrowseDestroyContext);
	        if (!bucket) err = DS_ERR_MEM;
		else {
		    dsHolderBucketSetNumbers(bucket, DS_HOLDER_BUCKET_MAGIC_REQUEST_BROWSE);
		    secret = bucket->secret;
		    pos = dsHolderPut(sctx->holder, bucket);
		    if (pos==(unsigned long)-1) err = DS_ERR_HOLDER_BUSY;
		    else sprintf(stmp,"%lu:%lu",secret,pos);
		}	    
	    } 
	}
    }

    if (err==OPC_ERR_XLIB_REQUEST_PROCESS_MORE_ITEMS) {
	xmlSetProp(ctx->resp,"ContinuationPoint", stmp);
	xmlSetProp(ctx->resp,"MoreElements","true");
    } else {
	xmlSetProp(ctx->resp,"MoreElements","false");
	opcXmldaRequestBrowseDestroyContext(bctx);
    }
    if ((err)&&(err!=OPC_ERR_XLIB_REQUEST_PROCESS_MORE_ITEMS)) return err;
    
//    pnode=xmlNewChild(ctx->resp,NULL,"Properties",NULL);
//    pnode=xmlNewChild(ctx->resp,NULL,"Errors",NULL);

    return opcXmldaRequestReplyBase(ctx,rnode);
}

int opcXmldaRequestGetProperties(OPCXmldaRequestContext ctx) {
    int err;
    OPCInterface server_interface;
    DSDataManagerUserizeOptionsS req_options;
    OPCConfigurationContext config;
    OPCTreeContext tree;
    xmlNodePtr node, inode, rnode, pnode;
    xmlNodePtr options = NULL, itemlist = NULL;
    xmlNodePtr aitem, anode,qnode = NULL;
    OPCContext sctx;
    OPCItemContext ictx;
    DSData data;
    DSData info;
    DSDataType type;
    DSDataS userdata;
    DSTimeS deadline, current_time;
    const char *ItemPathG, *ItemPath, *ItemName;
    unsigned long MaxAgeG, MaxAge;
    void *attachment;
    char timestamp[64];
    xmlNsPtr xsins, xsdns;
    const char *datatype = NULL;
    unsigned char ReturnAllProperties, ReturnPropertyValues, ReturnErrorText;

    OPCPropertiesContext properties;
    
    OPCPropertyID id, nprops;
    OPCPropertyID *props;
    OPCPropertyContext property;
    OPCPropertyType proptype;
    const char *propname;
    
    
    ds_assert_param(ctx);
    
    server_interface = ctx->server_interface;
    ds_assert_param(server_interface);

    dsDataManagerUserizeOptionsInitOPCXML(&req_options);
    
/*    
    sctx = ctx->sctx;
    ds_assert_param(sctx);

    tree = opcInterfaceGetServerTree(server_interface);
    ds_assert_param(tree);

    config = opcInterfaceGetConfiguration(server_interface);
    ds_assert_param(config);
*/    
/*
    {
	xmlChar *debug;
	int debug_size;
	xmlDocDumpMemory(ctx->req->doc, &debug,&debug_size);
	puts(debug);
    }
*/


    options = ctx->req;

    opcXmldaRequestGetClientRequestHandle(ctx, options);
    opcXmldaRequestGetLocale(ctx, options);
    
    ReturnErrorText = dsXmlGetPropBoolean(options, "ReturnErrorText", 0);
    ReturnPropertyValues = dsXmlGetPropBoolean(options, "ReturnPropertyValues", 0);
    ReturnAllProperties = dsXmlGetPropBoolean(options, "ReturnAllProperties", 0);
    ItemPathG = dsXmlGetPropString(itemlist, "ItemPath", NULL);


    rnode = xmlNewChild(ctx->resp,NULL,"GetPropertiesResult",NULL);    
    xsins = xmlNewNs(ctx->resp,"http://www.w3.org/2001/XMLSchema-instance","xsi");
    xsdns = xmlNewNs(ctx->resp,"http://www.w3.org/2001/XMLSchema","xsd");
    

    properties = opcInterfaceGetProperties(server_interface);
    if (!properties) return DS_ERR_INTERNAL;
    
    nprops = opcPropertiesGetLastID(properties);
    
    props = (OPCPropertyID*)dsMalloca(nprops*sizeof(OPCPropertyID));
    if (!props) return DS_ERR_MEM;
    
    if (ReturnAllProperties) {
	for (id=0;id<nprops;id++) props[id] = 1;
    } else {
	memset(props, 0, nprops*sizeof(OPCPropertyID));
	
	for (node=ctx->req->children; node; node=node->next) {
	    if ((node->type != XML_ELEMENT_NODE)||(!node->children)||(node->children->type != XML_TEXT_NODE)) continue;
	    if ((xmlStrcmp(node->name, "PropertyNames"))) continue;
	    
	    propname = node->children->content;
	    if (!propname) continue; /* Add error */

	    id = opcPropertiesFindPropertyId(properties, propname);
	    if (id == OPC_PROPERTY_ID_INVALID) continue; /* Add error */
	    props[id] = 1;
	}
    }
    
    for (inode=ctx->req->children; inode; inode=inode->next) {
	if (inode->type != XML_ELEMENT_NODE) continue;
	if ((xmlStrcmp(inode->name, "ItemIDs"))) continue;

	ItemName = dsXmlGetPropString(inode, "ItemName", NULL);
	ItemPath = dsXmlGetPropString(inode, "ItemPath", NULL);
	if (!ItemPath) ItemPath = ItemPathG;

	ictx = opcInterfaceGetItem(server_interface, ItemPath, ItemName);
	if (!ictx) continue; /* Add Error */
	
	pnode = xmlNewChild(ctx->resp, NULL, "PropertyLists", NULL);
	if (ItemName) xmlSetProp(pnode, "ItemName", ItemName);
	if (ItemPath) xmlSetProp(pnode, "ItemPath", ItemPath);

	for (id = 0; id < nprops; id++) {
	    if (!props[id]) continue;
	    
	    property = opcPropertiesGetProperty(properties, id);
	    if (!property) continue; /* Add Error */
	    
	    proptype = opcPropertyGetType(property);
	    propname = opcPropertyGetName(property);
	    puts(propname);

	    anode = xmlNewChild(pnode, NULL, "Properties", NULL);
	    xmlSetProp(anode, "Description", opcPropertyGetDescription(property));
	    xmlSetProp(anode, "Name", propname);
	    
	    if (ReturnPropertyValues) {
	     switch (proptype) {
	      case OPC_PROPERTY_TYPE_DATATYPE:
		err=opcItemGetData(ictx, &info, &data);
		if (!err) {
		    err=dsDataManagerUserize(&userdata,data,&req_options);
		    if (!err) {
			type = dsDataGetType(&userdata);
			if (type == DS_DATA_TYPE_XMLNODEPTR) {
		    	    datatype = dsXmlGetPropString((xmlNodePtr)dsDataGet(&userdata), "type", NULL);
		    	    if (datatype) {
				qnode = xmlNewChild(anode, NULL, "Value", datatype);
			    }
			}
			dsDataDestroyContent(&userdata);
		    }
		    opcItemReturnData(ictx, info);
		}
		if (!datatype) {
		    if (strstr(ictx->fullname, "nteger")) {
			qnode = xmlNewChild(anode, NULL, "Value", "xsd:unsignedInt");
		    } else if (strstr(ictx->fullname, "ouble")) {
			qnode = xmlNewChild(anode, NULL, "Value", "xsd:double");
		    } else {
			qnode = xmlNewChild(anode, NULL, "Value", "ArrayOfAnyType");
		    }
		}
		xmlSetNsProp(qnode, xsins, "type", "xsd:QName");
	      break;
	      case OPC_PROPERTY_TYPE_QUALITY:
	      case OPC_PROPERTY_TYPE_TIMESTAMP:
	      case OPC_PROPERTY_TYPE_TIMEZONE:
	      case OPC_PROPERTY_TYPE_VALUE:
	      
	      break;
	      case OPC_PROPERTY_TYPE_STANDARD:
	        if (!xmlStrcmp(propname, "accessRights")) {
		    qnode = xmlNewChild(anode, NULL, "Value", "readable");
		    xmlSetNsProp(qnode, xsins, "type", "xsd:string");
		}

		else if (!xmlStrcmp(propname, "scanRate")) {
		    qnode = xmlNewChild(anode, NULL, "Value", "100");
		    xmlSetNsProp(qnode, xsins, "type", "xsd:double");
		}
	      break;
	      case OPC_PROPERTY_TYPE_NOTSUPPORTED:
	        xmlFreeNode(anode);
	        /* Error */
	      break;
	      default:
	        xmlFreeNode(anode);
	        /* Error */
		;
	     }
	    }
	}
	opcItemReturn(ictx);
    }
    
    dsFreea(props);

/*
    {
	xmlChar *debug;
	int debug_size;
	xmlDocDumpMemory(ctx->adoc, &debug,&debug_size);
	puts(debug);
    }
*/

    return opcXmldaRequestReplyBase(ctx, rnode);
}

static int opcXmldaRequestReadInternal(OPCXmldaRequestContext ctx, unsigned char subscr, const char *handle) {
    int err;
    OPCInterface server_interface;
    DSDataManagerUserizeOptionsS req_options;
    OPCConfigurationContext config;
    OPCTreeContext tree;
    xmlNodePtr node, rnode, pnode;
    xmlNodePtr options = NULL, itemlist = NULL;
    xmlNodePtr aitem, anode,qnode;
    OPCContext sctx;
    OPCItemContext ictx;
    DSData data;
    DSData info;

    DSDataS userdata;
    DSDataType type;

    DSTimeS deadline, current_time;
    unsigned char ReturnErrorText = 1, ReturnDiagnosticInfo = 0;
    unsigned char ReturnItemTime = 0, ReturnItemName = 0, ReturnItemPath = 0;
    const char *ItemPathG, *ItemPath, *ItemName;
    const char *ItemHandle, *SamplingRate;
    unsigned long MaxAgeG, MaxAge;
    void *attachment;
    char timestamp[64];

    ds_assert_param(ctx);

    dsDataManagerUserizeOptionsInitOPCXML(&req_options);
    
    server_interface = ctx->server_interface;
    ds_assert_param(server_interface);
    
    sctx = ctx->sctx;
    ds_assert_param(sctx);

    tree = opcInterfaceGetServerTree(server_interface);
    ds_assert_param(tree);

    config = opcInterfaceGetConfiguration(server_interface);
    ds_assert_param(config);

    switch (subscr) {    
	case 0:
	    rnode=xmlNewChild(ctx->resp,NULL,"ReadResult",NULL);    
	break;
	case 1:
	    rnode=xmlNewChild(ctx->resp,NULL,"SubscribeResult",NULL);    
	break;
	default:
	    rnode=xmlNewChild(ctx->resp,NULL,"SubscriptionPolledRefreshResult",NULL);    
    }    
    
    for (node=ctx->req->children; node; node=node->next) {
	if (node->type != XML_ELEMENT_NODE) continue;
	if ((!options)&&(!xmlStrcmp(node->name, "Options"))) {
	    options=node;
	    if (itemlist) break;
	} else if (!xmlStrcmp(node->name, "ItemList")) { 
	    itemlist=node;
	    if (options) break;
	}
    }

    opcXmldaRequestGetClientRequestHandle(ctx, options);
    opcXmldaRequestGetLocale(ctx, options);


    // ToDo: analyse options
    dsTimeCurrentTime(&deadline);
    dsTimeAdd(&deadline, &config->request_timeout);
    if (options) {
	//ReturnErrorText, ReturnDiagnosticInfo, ReturnItemTime, ReturnItemPath, ReturnItemName
	//_RequestDeadline_

	if (!dsXmlGetPropDateTime(options, "RequestDeadline", &deadline)) {
	    // ToDo: Invalid Deadline specified!
	}
	
	ReturnErrorText = dsXmlGetPropBoolean(options, "ReturnErrorText", 1);
	ReturnDiagnosticInfo = dsXmlGetPropBoolean(options, "ReturnDiagnosticInfo", 0);
	ReturnItemTime = dsXmlGetPropBoolean(options, "ReturnItemTime", 0);
	ReturnItemName = dsXmlGetPropBoolean(options, "ReturnItemName", 0);
	ReturnItemPath = dsXmlGetPropBoolean(options, "ReturnItemPath", 0);
	
	
    }    
    
    if (itemlist) {

	    /* Seting request global options */
	req_options.flags|=DS_DATA_MANAGER_USERIZE_FLAG_XMLNODE_BEAR;
	
	pnode = xmlNewChild(ctx->resp,NULL,"RItemList",NULL);
	// ReqType, ItemPath, MaxAge
	if (subscr == 2) {
	    xmlSetProp(pnode, "SubscriptionHandle", handle);
	}

	ItemPathG = dsXmlGetPropString(itemlist, "ItemPath", NULL);
	MaxAgeG = dsXmlGetPropLong(itemlist, "MaxAge", 0);
	
	xmlNewNs(pnode,"http://www.w3.org/2001/XMLSchema-instance","xsi");
	xmlNewNs(pnode,"http://www.w3.org/2001/XMLSchema","xsd");
		
	    // Deadline, timeout
	for (node=itemlist->children;((node)&&(dsTimeCurrentCmp(&deadline)>0));node=node->next) {
	    if (node->type != XML_ELEMENT_NODE) continue;
	        // ItemPath, _ReqType_?, ItemName, ClientItemHandle, _MaxAge_
	    
	    dsDebugServer(DS_DEBUG_INFO,"PPPPPPPPPPPOST1");
	    
	    ItemName = dsXmlGetPropString(node, "ItemName", NULL);
	    ItemPath = dsXmlGetPropString(node, "ItemPath", NULL);
	    if (!ItemPath) ItemPath=ItemPathG;
	    ItemHandle = dsXmlGetPropString(node, "ClientItemHandle", NULL);
	    SamplingRate = dsXmlGetPropString(node, "RequestedSamplingRate", NULL);
	    MaxAge = dsXmlGetPropLong(node, "MaxAge", -1);
	    if (MaxAge==(unsigned long)-1) MaxAge=MaxAgeG;
//	    if (ItemPath) puts(ItemPath);
//	    if (ItemName) puts(ItemName);
	    ictx=opcInterfaceGetItem(server_interface, ItemPath, ItemName);

	    if (!ictx) err = OPC_ERR_XLIB_REQUEST_PROCESS_ITEM_NOTFOUND;
	    else err=opcItemCheckConnection(ictx);
	    

	    if (!err) {
		err=opcItemGetData(ictx, &info, &data);
	    }
	    
	    if (!err) {
		err=dsDataManagerUserize(&userdata,data,&req_options);
		if (!err) {
		    DSTimeS hacktime;
	    	    aitem=xmlNewChild(pnode, NULL, "Items", NULL);
		    if (subscr==1) {
			if (SamplingRate) xmlSetProp(aitem, "RevisedSamplingRate", SamplingRate);
			aitem=xmlNewChild(aitem, NULL, "ItemValue", NULL); /*DS hack */
		    }
		    dsTimeCurrentTime(&hacktime); /*DS hack*/
		    dsTimeGetDateTime(&hacktime, timestamp);
		    if (ReturnItemTime) xmlSetProp(aitem, "Timestamp", timestamp);
		    if (ReturnItemName) xmlSetProp(aitem, "ItemName", ictx->fullname);
		    if (ReturnItemPath) xmlSetProp(aitem, "ItemPath", "");
		    if (ItemHandle) xmlSetProp(aitem, "ClientItemHandle", ItemHandle);
		    if (ReturnDiagnosticInfo) xmlNewChild(aitem, NULL, "DiagnosticInfo", "No Info");
		    if (subscr) {
			qnode = xmlNewChild(aitem, NULL, "Quality", NULL);
			xmlSetProp(qnode, "QualityField", "good");
		    }
		    
		    type = dsDataGetType(&userdata);
		    switch (type) {
			case DS_DATA_TYPE_INCOMPLETE:
			    dsDataDestroyContent(&userdata);
			    err = DS_ERR_EMPTY;
			break;
			case DS_DATA_TYPE_XMLNODEPTR:
			    if (userdata.type == DS_DATA_TYPE_REFERENCE) {
				anode=xmlDocCopyNode((xmlNodePtr)dsDataGet(&userdata), ctx->adoc, 1);
			        if (anode) xmlAddChild(aitem, anode);
				else err = OPC_ERR; /* ToDo: error */

				dsDataDestroyContent(&userdata);
			    } else {
				xmlAddChild(aitem, (xmlNodePtr)dsDataGet(&userdata));
			    }
			break;
			default:
			  if ((ctx->send_attachment)||(ctx->open_attachment)) {
			    if (ctx->open_attachment) {
				attachment = ctx->open_attachment(ctx->cctx, ictx->fullname, info, &userdata);
				if (!attachment) err = OPC_ERR;
			    }
			    if ((!err)&&(ctx->send_attachment))
			    	err = dsDataPipeContent(&userdata,ctx->send_attachment,attachment);
			    if ((!err)&&(ctx->close_attachment))
				ctx->close_attachment(attachment);
			    if (!err) {
				anode = xmlNewChild(aitem, NULL, "Value", NULL);
				anode = xmlNewChild(anode, NULL, "Attachment", NULL);
				xmlSetProp(anode, "href", ictx->fullname);
			    }
			  } else {
			    /* ToDo: Implement storing of attachments for server post processing */
			    err = DS_ERR_NOTIMPLEMENTED;
			  }
			  dsDataDestroyContent(&userdata);
		    }
		}
		opcItemReturnData(ictx, info);
	    }
	    dsDebugServer(DS_DEBUG_INFO,"PPPPPPPPPPPOST2");
/*	    
	    if (!err) {
		puts("Checking Age");
		opcDurationImportMS(&maxage_duration, MaxAge);
		err=opcItemCheckAge(ictx, &maxage_duration);
		if (err==OPC_ERR_ITEM_VALUE_OUTOFDATE_ACQUIRING) {
		    puts("reaquire");
		    // ToDo: wait
		    // err=0
		}
		printf("Age: %lu\n",err);
	    }
*/

/*	    
	    if (!err) err=opcItemGetOPCXMLData(ictx, &info, &data);
	    
	    if (!err) {
	        aitem=xmlNewChild(pnode, NULL, "Items", NULL);
		// ItemPath, ItemName, ClientItemHandle, Timestamp, ResultID
		// ValueTypeQualifier
		// [ DiagnosticInfo, Value, Quality ]
	        if ((data)&&(data->ptr)) {
		    puts("+++++ GotInformation ++++++++++");
		// Localize strings

		    anode=xmlDocCopyNode((xmlNodePtr)data->ptr, ctx->adoc, 1);
//		    anode=xmlCopyNode((xmlNodePtr)data->ptr,1);
//    attr=xmlHasProp((xmlNodePtr)data->ptr,"type");
//    printf("ANODE: NS: %p %s\n",attr->ns, attr->ns->href);
//    attr=xmlHasProp(anode,"type");
//    printf("ANODE: NS: %p\n",attr->ns);
//    xmlReconciliateNs(ctx->answer, anode);
//    attr=xmlHasProp(anode,"type");
//    printf("ANODE: NS: %p\n",attr->ns);
		    
		    if (node) xmlAddChild(aitem, anode);
		    else {
			// ToDO: error
			opcItemReturn(ictx);
			continue;
		    }
    attr=xmlHasProp(anode,"type");
    printf("ANODE: NS: %p\n",attr->ns);
//
//    xmlStaticCopyNode((xmlNodePtr)data->ptr, ctx->adoc, aitem, 1);

		    anode=xmlDocCopyNode((xmlNodePtr)data->ptr, ctx->adoc, 1);
		    xmlAddChild(aitem, anode);
    attr=xmlHasProp(anode,"type");
    printf("ANODE: NS: %p\n",attr->ns);

		}
		puts(ictx->fullname);
	    }
*/	    
	    dsDebugServer(DS_DEBUG_INFO,"PPPPPPPPPPPOST3");

	    // ToDo: Error message
	    if (err) {
		if (ItemPath&&ItemName)
		    dsDebugRequest(DS_DEBUG_ERROR, "Error processing item \"%s/%s\": %s", ItemPath, ItemName, dsError(err, NULL));
		else if (ItemPath||ItemName)		    
		    dsDebugRequest(DS_DEBUG_ERROR, "Error processing item \"%s\": %s", ItemName?ItemName:ItemPath, dsError(err, NULL));
		else 
		    dsDebugRequest(DS_DEBUG_ERROR, "Error processing item \"/\": %s", dsError(err, NULL));
	    }
	    dsDebugServer(DS_DEBUG_INFO,"PPPPPPPPPPPOST5");
	    if (ictx) opcItemReturn(ictx);
	    dsDebugServer(DS_DEBUG_INFO,"PPPPPPPPPPPOST6");
	}

	dsDebugServer(DS_DEBUG_INFO,"PPPPPPPPPPPOST4");
	
//	xmlReconciliateNs(ctx->answer, aitem);
    }


		dsDebugServer(DS_DEBUG_INFO,"PPPPPPPPPPPOST");

    return opcXmldaRequestReplyBase(ctx, rnode);
}

int opcXmldaRequestRead(OPCXmldaRequestContext ctx) {
    return opcXmldaRequestReadInternal(ctx, 0, NULL);
}


int opcXmldaRequestSubscribe(OPCXmldaRequestContext ctx) {
    xmlDocPtr doc;
    char stmp[32];
        
    doc = xmlNewDoc("1.0");
    xmlDocSetRootElement(doc, xmlDocCopyNode(ctx->req, doc, 1));
    

    sprintf(stmp,"%lu", (unsigned long)doc);
    xmlSetProp(ctx->resp, "ServerSubHandle", stmp);

    return opcXmldaRequestReadInternal(ctx, 1, NULL);
}

int opcXmldaRequestSubscriptionPolledRefresh(OPCXmldaRequestContext ctx) {
    int err;
    xmlDocPtr doc, sdoc;
    xmlNodePtr node, snode;
    DSTimeS time, curtime;
    DSDurationS duration;
    unsigned long wait;
    const char *handle;
    char tmp1[128],tmp2[128];

    dsTimeCurrentTime(&time);
    dsXmlGetPropDateTime(ctx->req, "HoldTime", &time);
    wait = dsXmlGetPropLong(ctx->req, "WaitTime", 1000) / 2;
    duration.s = wait / 1000;
    duration.us = (wait - (wait/1000)*1000)*1000;
    dsTimeAdd(&time, &duration);

    for (node=ctx->req->children; node; node=node->next) {
	if (node->type != XML_ELEMENT_NODE) continue;
	if (!xmlStrcmp(node->name, "ServerSubHandles")) break;
    }

    if ((node)&&(node->children)&&(node->children->type==XML_TEXT_NODE))  {
	handle = node->children->content;
	sscanf(handle, "%lu", (unsigned long*)&doc); /* ptr = long int */
	
	sdoc = ctx->doc;
	snode = ctx->req;
	
	ctx->doc = doc;
	ctx->req = xmlDocGetRootElement(doc);

	for (node=ctx->req->children; node; node=node->next) {
	    if (node->type != XML_ELEMENT_NODE) continue;
	    if (!xmlStrcmp(node->name, "Options")) {
		xmlUnlinkNode(node);
		xmlFreeNode(node);
		break;
	    }
	}

	for (node=snode->children; node; node=node->next) {
	    if (node->type != XML_ELEMENT_NODE) continue;
	    if (!xmlStrcmp(node->name, "Options")) {
		xmlAddChild(ctx->req, xmlDocCopyNode(node, ctx->doc, 1));
		break;
	    }
	}

        dsTimeCurrentTime(&curtime);
	dsTimeCalcDuration(&duration, &curtime, &time);
	dsTimeGetDateTime(&curtime, tmp1);
	dsTimeGetDateTime(&time, tmp2);
	
	
	if ((duration.s > 0)||((duration.s == 0)&&(duration.us > 0))) {
	    if (duration.s > 5) duration.s = 5;
#ifdef _WIN32
	    Sleep(duration.s*1000 + duration.us/1000);
#else
	    duration.us*=1000;
	    nanosleep((struct timespec*)&duration, NULL);
	
#endif
	}

	err=opcXmldaRequestReadInternal(ctx, 2, handle);
	ctx->req = snode;
	ctx->doc = sdoc;
	
	return err;
    }

    return 0;
}

int opcXmldaRequestSubscriptionCancel(OPCXmldaRequestContext ctx) {
    xmlDocPtr doc;

    doc = (xmlDocPtr)dsXmlGetPropLong(ctx->req, "ServerSubHandle", 0);
    xmlFreeDoc(doc);
    return 0;
}

int opcXmldaRequestWrite(OPCXmldaRequestContext ctx) {
    return 0;
}
