#define _BACKEND_C

#include <stdio.h>

#include <opclib.h>
#include <dsassert.h>
#include <dsdebug.h>
#include <dspluginset.h>
#include <dssignal.h>

#include <opcdefines.h>

#include <backend.h>

void opcBackendZeroContext(OPCBackendContext ctx) {
    ds_assert_return(ctx);
    
    ctx->sctx=NULL;
    ctx->plugins=NULL;
    ctx->sigctx=NULL;
    ctx->config=NULL;
}

OPCBackendContext opcBackendCreateContext(DSMallocContext mctx) {
    OPCBackendContext ctx;

    dsDebugServer(DS_DEBUG_NOTICE, "opcBackendCreateContext");    

    ctx = (OPCBackendContext)dsMalloc(mctx, sizeof(struct OPCBackendContextT));
    if (ctx) {
	ctx->malloc=mctx;
	opcBackendZeroContext(ctx);
    }
    return ctx;    
}

void opcBackendFreeContext(OPCBackendContext ctx) {
    ds_assert_return(ctx);

    dsDebugServer(DS_DEBUG_NOTICE, "opcBackendFreeContext");    
    if (ctx->plugins) dsPluginSetDestroyContext(ctx->plugins);
    
    opcBackendZeroContext(ctx);
}

void opcBackendDestroyContext(OPCBackendContext ctx) {
    ds_assert_return(ctx);

    dsDebugServer(DS_DEBUG_NOTICE, "opcBackendDestroyContext");

    opcBackendFreeContext(ctx);
    dsFree(ctx->malloc,ctx);
}

int opcBackendInitContext(OPCBackendContext ctx, OPCConfigContext config, OPCContext sctx, DSSignalContext sigctx) {
    int err;

    ds_assert_param(ctx);    
    ds_assert_param(config);
    ds_assert_param(sctx);

    dsDebugServer(DS_DEBUG_NOTICE, "opcBackendInitContext");    

    opcBackendZeroContext(ctx);

        // Load all server plugins
    ctx->plugins = dsPluginSetCreateContext(ctx->malloc);
    if (!ctx->plugins) return DS_ERR_PLUGINSET_CREATE;
    err = dsPluginSetInitContext(ctx->plugins,config->srvpath?config->srvpath:OPC_SERVERS_DIR,(void*)sctx);
    if (err) {
	opcBackendFreeContext(ctx);
	return err;
    }
    
    ctx->sctx=sctx;
    ctx->sigctx=sigctx;
    ctx->config=config;

    return 0;
}



int opcBackendListen(OPCBackendContext ctx) {
    ds_assert_complexctx(ctx,ctx->plugins);
    dsPluginSetStart(ctx->plugins);
    return 0;
}
