#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys\stat.h>


#include <winsock.h>
#include <time.h>
#include <microweb\httpserv.h>


#include <dserror.h>
#include <dsdime.h>
#include <opcxmldarequest.h>
#include <opcerrors.h>

#include <microweb.h>

#define OPC_HTTP_PORT 81

static OPCConfigContext opcconfig = NULL;
static OPCXmldaRequestPoolContext request_pool = NULL;
static DSBufferPoolContext buffer_pool = NULL;
static DSSignalContext sigctx = NULL;

int MicrowebConfig(OPCConfigContext config, OPCXmldaRequestPoolContext rpctx, DSBufferPoolContext bp, DSSignalContext sctx) {
    opcconfig = config;
    request_pool = rpctx;
    buffer_pool = bp;
    sigctx = sctx;
    return 0;
}

int MicrowebStart() {
    if (!StartWebServer(OPC_HTTP_PORT)) return OPC_ERR_SERVER;
    return 0;
}

void MicrowebStop() {
    StopWebServer();
}

#define DEFAULT_ROOT "c:\\OPC\\www"// default root directory for serving files
#define DEFAULT_FILE "index.htm" // default file name
#define EOS '\0'// end-of-string character
#define COUNTOF(a) (sizeof(a) / sizeof((a)[0]))

char *pErr404 = "404 Not Found";
char *pErr500 = "500 Internal Server Error";
char *pErr501 = "501 Not Implemented";

struct {
	char	*pExt;		// file extension
	char	*pType;		// media type
} MediaType[] = {
	{".avi", "video/avi"},
	{".doc", "application/msword"},
	{".gif", "image/gif"},
	{".htm", "text/html"},
	{".html", "text/html"},
	{".jpg", "image/jpeg"},
	{".ps", "application/postscript"},
	{".rtf", "application/rtf"},
	{".tif", "image/tiff"},
	{".txt", "text/plain"},
	{".wav", "audio/wav"},
	{".zip", "application/zip"}
};

static void BuildErrorPage(REQ_OUTPUTS *pOutp, const char *pFormat, ...)
{
	char	ErrPage[1024];
	va_list	ArgList;

	// construct the error page in our buffer
	strcpy(ErrPage, "<HTML>\n<BODY>\n");
	va_start(ArgList, pFormat);
	vsprintf(&ErrPage[strlen(ErrPage)], pFormat, ArgList);
	va_end(ArgList);
	strcat(ErrPage, "\n</BODY>\n</HTML>\n");

	// return it in the output structure
	pOutp->ContentLength = strlen(ErrPage);
	pOutp->pEntityBody = malloc(pOutp->ContentLength);
	if (pOutp->pEntityBody == NULL)
	{
		pOutp->ContentLength = 0;
		return;
	}
	memcpy(pOutp->pEntityBody, ErrPage, pOutp->ContentLength);
	pOutp->LastModTime = time(NULL);
	strcpy(pOutp->ContentType, "text/html");
	return;
}


struct MicrowebOutputBufferT {
    char *buffer;
    unsigned long used_size;
    unsigned long allocated_size;
    unsigned char format;
    unsigned char err;
};
typedef struct MicrowebOutputBufferT *MicrowebOutputBuffer;
typedef struct MicrowebOutputBufferT MicrowebOutputBufferS;

static void MicrowebOutputBufferInit(MicrowebOutputBuffer outbuf) {
    outbuf->buffer = NULL;
    outbuf->format = 0;
    outbuf->err = 0;
    outbuf->used_size = 0;
    outbuf->allocated_size = 0;
}

static void MicrowebOutputBufferFree(MicrowebOutputBuffer outbuf) {
    if (outbuf->buffer) {
	free(outbuf->buffer);
	outbuf->buffer = NULL;
    }
}

static int SendDataCallback(MicrowebOutputBuffer outbuf, const char *buffer, unsigned long len) {
    char *newbuf;
    unsigned long newsize;
    
    if (outbuf->used_size + len > outbuf->allocated_size) {
	newsize = outbuf->allocated_size * 2;
	if (newsize < outbuf->used_size + len) newsize = outbuf->used_size + len;
	if (newsize&3) newsize = (newsize&~3)+4; /* Buffer size should be multiple of 4 */
	
	newbuf = realloc(outbuf->buffer, newsize);
	if (!newbuf) return -1;
	outbuf->buffer = newbuf;
	outbuf->allocated_size = newsize;
    }
    memcpy(outbuf->buffer + outbuf->used_size, buffer, len);
    outbuf->used_size += len;
    return 0;
}


static int SendCallback(MicrowebOutputBuffer outbuf, const char *buffer, int len) {
    int err;
    int size;
    
    if ((!outbuf)||(!buffer)||(len<0)) return -1;
    if (!len) return 0;
    

    if (outbuf->format) {
	if (outbuf->format == 1) {
	    err = dsDimeSendHeader(opc_soap_message_id, len, DS_DIME_TYPE_XML, NULL, 0, 0, 1, (DSSendFunction)SendDataCallback, outbuf);
	    outbuf->format = 2;	    
	} else {
	    err = dsDimeSendChunk(len, 0, (DSSendFunction)SendDataCallback, outbuf);
	}
	if (err) {
	    outbuf->err = err;
	    return -1;
	}

	if (len&3) size = (len&~3)+4; /* Buffer size should be multiple of 4 */
	else size = len;
    } else size = len;

    err = SendDataCallback(outbuf, buffer, size);
    if (err) {
	outbuf->err = err;
	return -1;
    }
    
    return len;
}


static void *OpenDIMEAttachmentCallback(MicrowebOutputBuffer outbuf, const char *item, OPCDataInfo info, OPCUserData data) {
    int err;

    if (!data->size) return NULL;
    
    if (outbuf->format) {
	err = dsDimeSendHeader(item,data->size,DS_DIME_TYPE_BINARY,NULL,0,0,0,(DSSendFunction)SendDataCallback, outbuf);
	if (err) {
	    outbuf->err = err;
	    return NULL;
	}
    } else {
	err = dsDimeSendHeader(item,data->size,DS_DIME_TYPE_BINARY,NULL,1,0,0,(DSSendFunction)SendDataCallback, outbuf);
	if (err) {
	    outbuf->err = err;
	    return NULL;
	}
	outbuf->format = 1;
    }

    return (void*)outbuf;
}



//
// This function is called by HTTPSERV.LIB for a GET or HEAD method, and reads 
// the content ("entity-body") for a requested URI from disk into memory.
//
// NOTE: This function MUST be reentrant.  In the interests of good response 
//	times, HTTPSERV.LIB does NOT serialize calls to this function.  If
//	multiple HTTP requests arrive close together, two (or more) calls
//	to this function may execute concurrently.
//
void __cdecl GetURI(REQ_INPUTS *pInp, REQ_OUTPUTS *pOutp)
{
	char	*pPath;		// ptr to full file path and name
	FILE	*hDoc;		// document file handle
	char	*pExt;		// file extension
	struct _stat Stats;	// file status
	size_t	nBytes;		// number of bytes read
	int	i;

//
// Allocate a buffer to hold an absolute path to the target file, and
// prepend our default root directory to the front, and add a default file
// name on the end of the path if the path ends in '\'
//
	pPath = malloc(strlen(pInp->pPath) + strlen(DEFAULT_ROOT) +
						strlen(DEFAULT_FILE) + 1);
	if (pPath == NULL)
	{
		printf("Can't malloc memory for path in GetURI\n");
		strcpy(pOutp->StatusCode, pErr500);
		BuildErrorPage(pOutp, "Server error, can't allocate memory");
		return;
	}
	strcpy(pPath, DEFAULT_ROOT);
	strcat(pPath, pInp->pPath);
	if (*(pPath + strlen(pPath) - 1) == '\\')
		strcat(pPath, DEFAULT_FILE);

	if (_stat(pPath, &Stats) != 0)
	{
		// file not found
		strcpy(pOutp->StatusCode, pErr404);
		BuildErrorPage(pOutp, "Requested URI %s not found",pInp->pURI);
		goto FREE_PATH;
	}

//
// Get file size, last modified time, and content type based on the
// file extension.  We don't return a content encoding or an expiration time.
//
	pOutp->ContentLength = (int) Stats.st_size;
	pOutp->LastModTime = Stats.st_mtime;
	pOutp->ContentType[0] = EOS;
	pOutp->ContentEncoding[0] = EOS;
	pOutp->HeaderFields[0] = EOS;
	pOutp->Expires = 0;
	pExt = strrchr(pPath, '.');

	if (pExt != NULL)
	{
		for (i = 0; i < COUNTOF(MediaType); ++ i)
			if (strcmp(MediaType[i].pExt, pExt) == 0)
			{
				strcpy(pOutp->ContentType, MediaType[i].pType);
				break;
			}
	}

//
// Read the file into memory.  We skip this step for the HEAD method, which
// only wants information about the file, not the file itself.
//
	if (!pInp->fHeadMethod)
	{
		hDoc = fopen(pPath, "rb");
		if (hDoc == NULL)
		{
			// file not found
			strcpy(pOutp->StatusCode, pErr404);
			BuildErrorPage(pOutp, "Requested URI %s not found",
								pInp->pURI);
			goto FREE_PATH;
		}
		pOutp->pEntityBody = malloc(pOutp->ContentLength);
		if (pOutp->pEntityBody == NULL)
		{
			printf("Can't malloc memory for file %s\n", pPath);
			strcpy(pOutp->StatusCode, pErr500);
			BuildErrorPage(pOutp, "Server error, can't allocate "
						"memory");
			goto CLOSE_FILE;
		}
		nBytes = fread(pOutp->pEntityBody,1,pOutp->ContentLength,hDoc);
		if (nBytes != pOutp->ContentLength)
		{
			if (feof(hDoc) != 0)
				// end of file -- adjust length
				pOutp->ContentLength = nBytes;
			else
			{
				printf("Error reading data from file %s\n", 
									pPath);
				strcpy(pOutp->StatusCode, pErr500);
				BuildErrorPage(pOutp, "Server error reading "
						"file for URI %s", pInp->pURI);
				goto FREE_ENTITY;
			}
		}
		fclose(hDoc);
	}
	free(pPath);
	return;

FREE_ENTITY:
	free(pOutp->pEntityBody);
CLOSE_FILE:
	fclose(hDoc);
FREE_PATH:
	free(pPath);
	return;
}

//
// This function is called by HTTPSERV.LIB for a POST method.  Since we don't
// handle POSTs in this simple server, we just return a "not implemented"
// error.
//
// NOTE: This function MUST be reentrant.  In the interests of good response 
//	times, HTTPSERV.LIB does NOT serialize calls to this function.  If
//	multiple HTTP requests arrive close together, two (or more) calls
//	to this function may execute concurrently.
//
void __cdecl PostURI(REQ_INPUTS *pInp, REQ_OUTPUTS *pOutp)
{
    int err;
    MicrowebOutputBufferS outbuf;
    OPCXmldaRequestContext rctx;
    
    if ((!opcconfig)||(!request_pool)||(!pInp->pEntityBody)) {
	strcpy(pOutp->StatusCode, pErr501);
	BuildErrorPage(pOutp, "Microweb server is not configured or invalid request (%p %p %p)", opcconfig, request_pool, pInp->pEntityBody);
	return;
    }
    
    if ((pInp->pContentType)&&(strncmp(pInp->pContentType, "text", 4))) {
	strcpy(pOutp->StatusCode, pErr501);
	BuildErrorPage(pOutp, "MIME type is not supported");
	return;
    }

    MicrowebOutputBufferInit(&outbuf);

    rctx = opcXmldaRequestPoolGetRequest(request_pool);
    
    if (rctx) err = opcXmldaRequestXMLData(rctx, pInp->pEntityBody, pInp->ContentLength?pInp->ContentLength:strlen(pInp->pEntityBody));
    else err = OPC_ERR_XLIB_REQUEST_CREATE;
    
    if (!err) err = opcXmldaRequestProcess(rctx, (OPCXmldaRequestOpenAttachmentFunction)&OpenDIMEAttachmentCallback, (DSSendFunction)&SendDataCallback, NULL, &outbuf);
    if (!err) err = opcXmldaRequestSend(rctx, (xmlOutputWriteCallback)SendCallback, NULL, &outbuf);
    
    if (rctx) opcXmldaRequestPoolReturnRequest(request_pool, rctx);

    if (err) {
	MicrowebOutputBufferFree(&outbuf);
	strcpy(pOutp->StatusCode, pErr500);
	BuildErrorPage(pOutp, "Error proccessing OPC XML-DA request (Error: %lx)", err);
	return;
    }

    pOutp->ContentLength = outbuf.used_size;
    pOutp->pEntityBody = outbuf.buffer;
    pOutp->LastModTime = 0;
    strcpy(pOutp->ContentType, "text/xml");
    strcpy(pOutp->ContentEncoding, "UTF-8");
    pOutp->HeaderFields[0] = EOS;
    pOutp->Expires = 0;
}

//
// This function is called by HTTPSERV.LIB with a record of each processed 
// HTTP request.  We use this to keep a log file of server activity on disk.
//
// NOTE: This function does not need to be reentrant.  Because it is called
//	after the HTTP request has been serviced, HTTPSERV.LIB is able to
//	serialize calls to this function without impacting response times.
//	However, we still need to use csLogFile to serialize access to the
//	log file with the main program thread, which reads the file.
//
void __cdecl LogRequest(REQUEST_RECORD *pReq)
{
    return;
}

//
// This function is called if a server error occurs.
//	fFatal		T ==> fatal server error, the server LIB has done
//				the equivalent of a call to StopWebServer()
//				and is no longer responding to HTTP requests
//			F ==> non-fatal error (usually an error servicing
//				a specific HTTP request), the server is
//				still running and responding to HTTP requests
//	pErrMsg		a (possibly multi-line) error message
//
void __cdecl WebServerError(BOOL fFatal, char *pErrMsg)
{
	return;
}
