/* RCS_Id: httputil.c 10.2 1999/12/10 16:44:34 clark Exp $ */
//
// httputil.c - This module contains utility functions used by the other
// modules in this directory.
//

  //////////////
 // Includes //
//////////////

#define WIN32_LEAN_AND_MEAN
#define WIN32_EXTRA_LEAN
#include <winsock.h>

#include <stdlib.h>
#include <time.h>
#include <httpserv.h>
#include <embkern.h>

#include "httputil.h"


  ///////////////////////////////////
 // Private Function Declarations //
///////////////////////////////////

static BOOL QueryIsComplete (char *pStr);


  /////////////////////////////////
 // Public Function Definitions //
/////////////////////////////////

//
// SetNewString() - Allocates a string and then sets the string to the
//	contents of the string passed in.
//
// Input:
//	pSourceStr	The string to copy.
//	SourceLen	The length of the string to copy.
//
// Results:
//	Success		A pointer to a new string.
//	Failure		NULL
//
char *SetNewString (char *pSourceStr, DWORD SourceLen)
{
    char *pDestStr;


    pDestStr = (char *) malloc(sizeof(*pDestStr) * (SourceLen + 1));

    if ( pDestStr != NULL ) {

	strncpy(pDestStr, pSourceStr, SourceLen);

	pDestStr[SourceLen] = '\0';
    }


    return pDestStr;
}

//
// BuildNewQueryString() - This function is used to build the first part of a
//	new query string.  This function is used when there is query related
//	information in REQ_INPUTS' pQuery field and pEntityBody field.
//
// Inputs:
//	pQuery		The original query string.
//	fMoreToAppend	Is there more to append?
//
// Input/Output:
//	QueryLen	The length of the new query string to build.  This value
//			may be changed if characters have to be added.
//
// Return:
//	Success		A pointer to the new query string.
//	Failure		NULL.
//
char *BuildNewQueryString (char *pQuery, DWORD *QueryLen, BOOL fMoreToAppend)
{
    BOOL   fComplete;
    char  *pNewQuery;


    // Find the last boundary and make sure that it is followed by an equal
    // sign.  If it is not, we have to append an equal sign or
    // GetFormVariableList() will break.
    fComplete = QueryIsComplete(pQuery);

    if ( !fComplete ) (*QueryLen)++;

    pNewQuery = (char *) malloc(sizeof(*pNewQuery) * (*QueryLen));

    if ( pNewQuery != NULL ) {

	strcpy(pNewQuery, pQuery);

	if ( !fComplete )

	    strcat(pNewQuery, "=");

	if ( !fMoreToAppend )

	    pNewQuery[strlen(pNewQuery)] = '\0';

	else {

	    if ( pNewQuery[strlen(pNewQuery) - 1] != '&' )

		strcat(pNewQuery, "&");
	}
    }


    return (pNewQuery);
}


  //////////////////////////////////
 // Private Function Definitions //
//////////////////////////////////

//
// QueryisComplete() - This function makes sure that the query string
//	passed into it is complete.  That is it makes sure that the last
//	ampersand is followed by an equal sign.  In short, this function makes
//	sure that each variable has a setting.  If the last character of the
//	query string is an ampersand, the string is considered complete.
//
// Input:
//	pStr	The string to check (must be null-terminated).
//
// Return:
//	TRUE	The string is complete
//	FALSE	The string is not complete.
//
static BOOL QueryIsComplete (char *pStr)
{
    char       *pTemp;
    char       *pLastBound;
    BOOL        fComplete;


    pLastBound = pStr;

    while ( (pTemp = strchr(pLastBound, '&')) != NULL ) pLastBound = pTemp + 1;

    // If this condition is true, the last character in the query is an
    // ampersand.  Pass back true.
    if ( (pLastBound != pStr) &&  (*pLastBound == '\0') )

	fComplete = TRUE;

    else

	fComplete = strchr(pLastBound, '=') != NULL;

    return (fComplete);
}
