/* RCS_Id: urlpost.c 10.2 1999/12/10 16:45:01 clark Exp $ */
//
// urlpost.c - This module contains function(s) that can be used to extract a
//	URL-encoded style POST from the pEntityBody field of a REQ_INPUTS
//	structure and insert it into the pQuery field.
//

  //////////////
 // Includes //
//////////////

#define WIN32_LEAN_AND_MEAN
#define WIN32_EXTRA_LEAN
#include <winsock.h>

#include <stdlib.h>
#include <time.h>
#include <httpserv.h>

#include "httputil.h"


  /////////////////////////////////
 // Public Function Definitions //
/////////////////////////////////

//
// GetURLEncodedPostInfo() - This function is used when setting pInp to a POST
//	of the form url-encoded.  This function takes the query string out of
//	the entity body and moves it into the query string (pQuery).
//
// Outputs:
//	pInp	If this function is successful, this structure will have a query
//		string containing the fields past in with the POST method, and
//		the entity body will be freed and set to NULL.  The content
//		length field will be set to zero.
//
// Return:
//	Success			FORM_NO_ERR	(0)
//	Fail (malloc)		FORM_ALLOC_ERR	(-2)
//
short GetURLEncodedPostInfo (REQ_INPUTS *pInp)
{
    char  *pNewQuery;
    DWORD  NewLen;


    //
    // If the query string already has something in it, we have to append to
    // the query string.  If the query string is NULL, we have to allocate
    // and build the query string.
    //
    if ( pInp->pQuery == NULL ) {

	NewLen = pInp->ContentLength + 1;

	pInp->pQuery = (char *) malloc(sizeof(*pInp->pEntityBody) * NewLen);

	if ( pInp->pQuery != NULL )   {
	    // If there was no entity body the form was empty (contained no
	    // fields).
	    if ( pInp->pEntityBody == NULL )

		pInp->pQuery[0] = '\0';

	    else {

		strncpy(pInp->pQuery, pInp->pEntityBody, pInp->ContentLength);

		pInp->pQuery[pInp->ContentLength] = '\0';
	    }
	}
    }

    else {

	// We add two to account for a NULL character and an ampersand.
	NewLen = pInp->ContentLength + strlen(pInp->pQuery) + 2;

	pNewQuery = BuildNewQueryString(pInp->pQuery,
					&NewLen,
					pInp->pEntityBody != NULL);

	if (pInp->pQuery)
	    free(pInp->pQuery);

	pInp->pQuery = pNewQuery;

	if (pInp->pQuery != NULL)
	    strncat(pInp->pQuery, pInp->pEntityBody, pInp->ContentLength);

	pInp->pQuery[NewLen - 1] = '\0';
    }

    if ( pInp->pQuery == NULL )

	return FORM_ALLOC_ERR;


    // If there was no entity body the form was empty (contained no fields).
    if (pInp->pEntityBody != NULL)
    {
	free(pInp->pEntityBody);
	pInp->pEntityBody = NULL;
	pInp->ContentLength = 0;
    }
    return FORM_NO_ERR;
}
