#define _OPC_CONFIG_C
#include <opclib.h>
#include <dsassert.h>

#include <dsdirectory.h>

#include <opcconfiguration.h>
#include <opcconfig.h>

void opcConfigZeroContext(OPCConfigContext ctx) {
    opcConfigurationZeroContext((OPCConfigurationContext)ctx);
    ctx->srvpath = NULL;
}

OPCConfigContext opcConfigCreateContext(DSMallocContext mctx) {
    OPCConfigContext ctx;
    ctx = (OPCConfigContext)dsMalloc(mctx, sizeof(struct OPCConfigContextT));
    if (ctx) {
	ctx->opc.malloc = mctx;
	ctx->opc.destructor = (DSDestroyContextFunction)opcConfigDestroyContext;
	opcConfigZeroContext(ctx);
    }
    return ctx;
}

int opcConfigInitContext(OPCConfigContext ctx, const char *data_directory, const char *config) {
    int err;
    
    ds_assert_param(ctx);
    
    opcConfigZeroContext(ctx);

    err = dsDirectoryCreate(OPC_TEMP_DIR);
    if (err) return err;

//    ctx->debug_level = DS_DEBUG_DEBUG;
//    dsDebugSetLevel(ctx->debug_level);
    ctx->srvport = 0;
    ctx->max_frontend_threads = OPC_SERVER_CONFIG_SCALE_FACTOR;
    strncpy(ctx->srvname, OPC_SERVER_CONFIG_DEFAULT_NAME, MAX_SERVER_NAME);
    
    err = opcConfigurationInitContext((OPCConfigurationContext)ctx, OPC_SERVER_CONFIG_SCALE_FACTOR);
    if (err) {
	opcConfigFreeContext(ctx);
	return err;
    }
    
    if (data_directory) {
	err = opcConfigurationSetDataDirectory((OPCConfigurationContext)ctx, data_directory);
	if (err) {
	    opcConfigFreeContext(ctx);
	    return err;
	}

	ctx->srvpath = dsDirectoryConcatenate(ctx->opc.malloc, data_directory, OPC_SERVER_CONFIG_SERVERS_DIRECTORY);
	
	dsExchangeString(ctx->opc.malloc, &ctx->srvpath, data_directory);
    }

/* Config used: 
1. Use specified as parameter, exit if failed (if specified)
2. Try config in current directory
3. Try config in data directory (if specified)
4. Try config in system directory
*/
    return 0;
}

void opcConfigFreeContext(OPCConfigContext ctx) {
    ds_assert_return(ctx);

    if (ctx->srvpath) dsExchangeString(ctx->opc.malloc, &ctx->srvpath, NULL);
    
    opcConfigurationFreeContext((OPCConfigurationContext)ctx);
    opcConfigZeroContext(ctx);
}

void opcConfigDestroyContext(OPCConfigContext ctx) {
    ds_assert_return(ctx);
    opcConfigFreeContext(ctx);
    dsFree(ctx->opc.malloc, ctx);
}
